// dynload.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "dynload.h"
#include "paramset.h"
#include "shapes.h"
#include "materials.h"
#ifndef WIN32
#include <dlfcn.h>
#endif
// Runtime Loading Forward Declarations
template <class D> D*LoadDSO(const string &name,
		StringHashTable &hashTable, const string &subdir,
		const string &searchPath) {
	D *dso = (D *)hashTable.Search(name);
	if (!dso) {
		string filename = subdir + name;
#ifdef WIN32
		filename += ".dll";
#else
		filename += ".so";
#endif
		string path = SearchPath(searchPath, filename);
		if (path != "") {
			dso = new D(path.c_str());
			hashTable.Add(name, dso);
		}
		else
			Error("Unable to find DSO/DLL for \"%s\"",
				name.c_str());
	}
	return dso;
}
static string SearchPath(const string &searchpath, const string &filename);
// Runtime Loading Static Data
static StringHashTable shape_dsos, filter_dsos;
static StringHashTable material_dsos, bump_dsos;
static StringHashTable light_dsos, arealight_dsos, volume_dsos;
static StringHashTable surf_integrator_dsos, vol_integrator_dsos, tonemap_dsos;
static StringHashTable accelerator_dsos, camera_dsos, sampler_dsos;
// Runtime Loading Local Classes
class ShapeDSO : public DSO {
public:
	// ShapeDSO Constructor
	ShapeDSO(const string &name)
		: DSO(name) {
		CreateShape = (CreateShapeFunc)(GetSymbol("CreateShape"));
	}
	typedef Reference<Shape> (*CreateShapeFunc)(const Transform &o2w,
		const ParamSet &params);
	CreateShapeFunc CreateShape;
};
class MaterialDSO : public DSO {
	typedef Reference<Material> (*CreateMaterialFunc)(const Transform &shader2world, const ParamSet &, const ParamSet &, const Spectrum &Cs, Texture<Float> *disp);
public:
	MaterialDSO( const string &name ): DSO( name )
	{
		CreateMaterial =
			(CreateMaterialFunc)(GetSymbol("CreateMaterial"));
	}
	CreateMaterialFunc CreateMaterial;
};
class BumpDSO : public DSO {
	typedef Texture<Float> *(*CreateBumpFunc)(const Transform &shader2world, const ParamSet &, const ParamSet &);
public:
	BumpDSO( const string &name ): DSO( name )
	{
		CreateBump =
			(CreateBumpFunc)(GetSymbol("CreateBump"));
	}
	CreateBumpFunc CreateBump;
};
class LightDSO : public DSO {
	typedef Light *(*CreateLightFunc)(bool castsShadows,
		const Transform &light2world, const ParamSet &params);
public:
	LightDSO( const string &name ): DSO( name )
	{
		CreateLight =
			(CreateLightFunc)(GetSymbol("CreateLight"));
	}
	CreateLightFunc CreateLight;
};
class VolumeRegionDSO : public DSO {
	typedef VolumeRegion *(*CreateVolumeFunc)(const Transform &volume2world,
		const ParamSet &params);
public:
	VolumeRegionDSO( const string &name ): DSO( name )
	{
		CreateVolumeRegion =
			(CreateVolumeFunc)(GetSymbol("CreateVolumeRegion"));
	}
	CreateVolumeFunc CreateVolumeRegion;
};
class AreaLightDSO : public DSO {
	typedef AreaLight *(*CreateAreaLightFunc)(bool castsShadows,
		const Transform &light2world, const ParamSet &params,
		const Reference<Shape> &shape);
public:
	AreaLightDSO( const string &name ): DSO( name )
	{
		CreateAreaLight =
			(CreateAreaLightFunc)(GetSymbol("CreateAreaLight"));
	}
	CreateAreaLightFunc CreateAreaLight;
};
class SurfaceIntegratorDSO : public DSO {
	typedef SurfaceIntegrator *(*CreateSurfaceIntegratorFunc)(const ParamSet &params);
public:
	SurfaceIntegratorDSO( const string &name ): DSO( name )
	{
		CreateSurfaceIntegrator =
			(CreateSurfaceIntegratorFunc)(GetSymbol("CreateSurfaceIntegrator"));
	}
	CreateSurfaceIntegratorFunc CreateSurfaceIntegrator;
};
class VolumeIntegratorDSO : public DSO {
	typedef VolumeIntegrator *(*CreateVolumeIntegratorFunc)(const ParamSet &params);
public:
	VolumeIntegratorDSO( const string &name ): DSO( name )
	{
		CreateVolumeIntegrator =
			(CreateVolumeIntegratorFunc)(GetSymbol("CreateVolumeIntegrator"));
	}
	CreateVolumeIntegratorFunc CreateVolumeIntegrator;
};
class AcceleratorDSO : public DSO {
	typedef Primitive *(*CreateAcceleratorFunc)(const vector<Reference<Primitive> > &prims, const ParamSet &params);
public:
	AcceleratorDSO( const string &name ): DSO( name )
	{
		CreateAccelerator =
			(CreateAcceleratorFunc)(GetSymbol("CreateAccelerator"));
	}
	CreateAcceleratorFunc CreateAccelerator;
};
class SamplerDSO : public DSO {
	typedef Sampler *(*CreateSamplerFunc)(const ParamSet &params, int
xres, int yres, int xsamp, int ysamp, const Extent2D &crop, Filter *filter);
public:
	SamplerDSO( const string &name ): DSO( name )
	{
		CreateSampler =
			(CreateSamplerFunc)(GetSymbol("CreateSampler"));
	}
	CreateSamplerFunc CreateSampler;
};
class CameraDSO : public DSO {
	typedef Camera *(*CreateCameraFunc)(const ParamSet &params, const Transform &world2cam, Film *film);
public:
	CameraDSO( const string &name ): DSO( name )
	{
		CreateCamera =
			(CreateCameraFunc)(GetSymbol("CreateCamera"));
	}
	CreateCameraFunc CreateCamera;
};
class FilterDSO : public DSO {
	typedef Filter *(*CreateFilterFunc)(const ParamSet &params,
		Float xwidth, Float ywidth);
public:
	FilterDSO( const string &name ): DSO( name ) {
		CreateFilter =
			(CreateFilterFunc)(GetSymbol("CreateFilter"));
	}
	CreateFilterFunc CreateFilter;
};
class ToneMapDSO : public DSO {
	typedef ToneMap *(*CreateToneMapFunc)(const ParamSet &params);
public:
	ToneMapDSO( const string &name ): DSO( name ) {
		CreateToneMap =
			(CreateToneMapFunc)(GetSymbol("CreateToneMap"));
	}
	CreateToneMapFunc CreateToneMap;
};
// Runtime Loading Methods
Reference<Shape> CreateShape(const string &name, const string &searchpath,
		const Transform &object2world, const ParamSet &paramSet) {
	ShapeDSO *dso = LoadDSO<ShapeDSO>(name, shape_dsos, "shapes/",
		searchpath);
	if (dso)
		return dso->CreateShape(object2world, paramSet);
	return NULL;
}
static string SearchPath(const string &searchpath,
		const string &filename) {
	const char *start = searchpath.c_str();
	const char *end = start;
	while (*start) {
		while (*end && *end != ':')
			++end;
		string component(start, end);

		string fn = component + "/" + filename;
		FILE *f = fopen(fn.c_str(), "r");
		if (f) {
			fclose(f);
			return fn;
		}
		if (*end == ':') ++end;
		start = end;
	}
	return "";
}
Reference<Material> CreateMaterial(const string &name, const string &searchPath,
		const Transform &mtl2world,
		const ParamSet &pgeom, const ParamSet &pshader,
		const Spectrum &Cs, Texture<Float> *disp) {
	// synonym action
	MaterialDSO *dso = NULL;
	if (name == "paintedplastic")
		dso = LoadDSO<MaterialDSO>("plastic", material_dsos,
			"materials/", searchPath);
	else
		dso = LoadDSO<MaterialDSO>(name, material_dsos,
			"materials/", searchPath);
	if (dso) {
		Reference<Material> ret =
			dso->CreateMaterial(mtl2world, pgeom, pshader,
				Cs, disp);
		pshader.ReportUnused();
		return ret;
	}
	return NULL;
}
Texture<Float> *CreateBump(const string &name, const string &searchPath,
		const Transform &bump2world,
		const ParamSet &pgeom, const ParamSet &pdisp) {
	if (name == "") return NULL;
	BumpDSO *dso = LoadDSO<BumpDSO>(name, bump_dsos, "materials/",
		searchPath);
	if (dso) {
		Texture<Float> *ret = dso->CreateBump(bump2world, pgeom, pdisp);
		pdisp.ReportUnused();
		return ret;
	}
	return NULL;
}
Light *CreateLight(const string &name, const string &searchPath, bool castsShadows,
		const Transform &light2world, const ParamSet &paramSet) {
	LightDSO *dso = LoadDSO<LightDSO>(name, light_dsos, "lights/",
		searchPath);
	if (dso) {
		Light *ret = dso->CreateLight(castsShadows, light2world,
			paramSet);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
AreaLight *CreateAreaLight(const string &name, const string &searchPath,
		bool castsShadows,
		const Transform &light2world, const ParamSet &paramSet,
		const Reference<Shape> &shape) {
	AreaLightDSO *dso = LoadDSO<AreaLightDSO>(name, arealight_dsos,
		"lights/", searchPath);
	if (dso) {
		AreaLight *ret = dso->CreateAreaLight(castsShadows,
			light2world, paramSet, shape);
		// hack to silence warning since we know API will be
		// looking for this shortly...
		(void)paramSet.FindOneInt("nsamples", 1);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
VolumeRegion *CreateVolumeRegion(const string &name, const string &searchPath,
		const Transform &volume2world, const ParamSet &paramSet) {
	VolumeRegionDSO *dso = LoadDSO<VolumeRegionDSO>(name, volume_dsos, "volumes/",
		searchPath);
	if (dso) {
		VolumeRegion *ret = dso->CreateVolumeRegion(volume2world, paramSet);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
SurfaceIntegrator *CreateSurfaceIntegrator(const string &name, const string &searchPath,
		const ParamSet &paramSet) {
	SurfaceIntegratorDSO *dso = LoadDSO<SurfaceIntegratorDSO>(name, surf_integrator_dsos,
		"integrators/", searchPath);
	if (dso) {
		SurfaceIntegrator *ret = dso->CreateSurfaceIntegrator(paramSet);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
VolumeIntegrator *CreateVolumeIntegrator(const string &name, const string &searchPath,
		const ParamSet &paramSet) {
	VolumeIntegratorDSO *dso = LoadDSO<VolumeIntegratorDSO>(name, vol_integrator_dsos,
		"integrators/", searchPath);
	if (dso) {
		VolumeIntegrator *ret = dso->CreateVolumeIntegrator(paramSet);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
Primitive *CreateAccelerator(const string &name, const string &searchPath,
		const vector<Reference<Primitive> > &prims,
		const ParamSet &paramSet) {
	AcceleratorDSO *dso = LoadDSO<AcceleratorDSO>(name, accelerator_dsos,
		"accelerators/", searchPath);
	if (dso) {
		Primitive *ret = dso->CreateAccelerator(prims, paramSet);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
Camera *CreateCamera(const string &name, const string &searchPath,
		const ParamSet &paramSet,
		const Transform &world2cam, Film *film) {
	CameraDSO *dso = LoadDSO<CameraDSO>(name, camera_dsos, "cameras/",
		searchPath);
	if (dso) {
		Camera *ret = dso->CreateCamera(paramSet, world2cam, film);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
Sampler *CreateSampler(const string &name, const string &searchPath,
		const ParamSet &paramSet,
		int xres, int yres, int xsamp, int ysamp,
		const Extent2D &crop, Filter *filter) {
	SamplerDSO *dso = LoadDSO<SamplerDSO>(name, sampler_dsos, "samplers/",
		searchPath);
	if (dso) {
		Sampler *ret = dso->CreateSampler(paramSet, xres, yres,
			xsamp, ysamp, crop, filter);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
Filter *CreateFilter(const string &name, const string &searchPath,
	const ParamSet &paramSet, Float xWidth, Float yWidth) {
	FilterDSO *dso = LoadDSO<FilterDSO>(name, filter_dsos, "filters/",
		searchPath);
	if (dso) {
		Filter *ret = dso->CreateFilter(paramSet, xWidth,
			yWidth);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
ToneMap *CreateToneMap(const string &name, const string &searchPath,
	const ParamSet &paramSet) {
	ToneMapDSO *dso = LoadDSO<ToneMapDSO>(name, tonemap_dsos, "tonemaps/",
		searchPath);
	if (dso) {
		ToneMap *ret = dso->CreateToneMap(paramSet);
		paramSet.ReportUnused();
		return ret;
	}
	return NULL;
}
// DSO Method Definitions
DSO::DSO(const string &fname) {
#ifdef WIN32
	hinstLib = LoadLibrary(fname.c_str());
	if (!hinstLib)
		Error("DSO Loader can't open DLL %s", fname.c_str());
#else
	hinstLib = dlopen(fname.c_str(), RTLD_LAZY);
	if (!hinstLib)
		Error("DSO Loader can't open DLL %s (%s)", fname.c_str(),
			dlerror());
#endif
}
DSO::~DSO() {
#ifdef WIN32
	FreeLibrary(hinstLib);
#else
	dlclose(hinstLib);
#endif
}
void *DSO::GetSymbol(const string &symname) {
	void *data;
#ifdef WIN32
	data = GetProcAddress(hinstLib, symname.c_str());
#else
	data = dlsym(hinstLib, symname.c_str());
#endif
	if (!data)
		Error("Couldn't get symbol \"%s\" in DSO.", symname.c_str());
	return data;
}
