// film.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "film.h"
#include "tonemap.h"
// Film Method Definitions
Film::Film(int xres, int yres, const Extent2D &cropWindow) {
	xResolution = xres;
	yResolution = yres;
	// Compute film image extent
	xPixelStart = Ceil2Int(xResolution * cropWindow.x0);
	xPixelWidth = Ceil2Int(xResolution * cropWindow.x1) - xPixelStart;
	yPixelStart = Ceil2Int(yResolution * cropWindow.y0);
	yPixelWidth = Ceil2Int(yResolution * cropWindow.y1) - yPixelStart;
	// Allocate film image storage
	pixels = new Pixel[xPixelWidth * yPixelWidth];
}
Film::~Film() {
	delete[] pixels;
}
void Film::WriteCoefficients(const string &filename) const {
	// Compute floating-point pixel SPD coefficients
	// Allocate working imaging memory and copy data
	int nPix = xPixelWidth * yPixelWidth;
	Spectrum *Lout = new Spectrum[nPix];
	Float *AlphaOut = new Float[nPix];
	for (int i = 0; i < nPix; ++i) {
		Lout[i] = pixels[i].L;
		AlphaOut[i] = pixels[i].alpha;
	}
	// Apply filter weights
	for (int o = 0; o < xPixelWidth * yPixelWidth; ++o) {
		if (pixels[o].weightSum == 0.)
			continue;
		Float invWt = 1.f / pixels[o].weightSum;
		Lout[o] *= invWt;
		AlphaOut[o] *= invWt;
	}
	// Compute premultiplied alpha color values
	for (int i = 0; i < xPixelWidth * yPixelWidth; ++i)
		Lout[i] *= AlphaOut[i];
	TIFFWriteFloat(filename, (Float *)Lout, AlphaOut,
		xPixelWidth, yPixelWidth, COLOR_SAMPLES,
		xResolution, yResolution);
	// Release temporary image memory
	delete[] Lout;
	delete[] AlphaOut;
}
void Film::WriteDisplayImage(const DisplayInfo &dinfo) const {
	// Compute floating-point pixel SPD coefficients
	// Allocate working imaging memory and copy data
	int nPix = xPixelWidth * yPixelWidth;
	Spectrum *Lout = new Spectrum[nPix];
	Float *AlphaOut = new Float[nPix];
	for (int i = 0; i < nPix; ++i) {
		Lout[i] = pixels[i].L;
		AlphaOut[i] = pixels[i].alpha;
	}
	// Apply filter weights
	for (int o = 0; o < xPixelWidth * yPixelWidth; ++o) {
		if (pixels[o].weightSum == 0.)
			continue;
		Float invWt = 1.f / pixels[o].weightSum;
		Lout[o] *= invWt;
		AlphaOut[o] *= invWt;
	}
	// Compute premultiplied alpha color values
	for (int i = 0; i < xPixelWidth * yPixelWidth; ++i)
		Lout[i] *= AlphaOut[i];
	// Apply display imaging pipeline
	// Convert image to XYZ
	Float *xyz = new Float[3*nPix];
	for (int i = 0; i < nPix; ++i)
		Lout[i].XYZ(&xyz[3*i]);
	// Apply tone reproduction to image
	if (dinfo.toneMap) {
		// Scale XYZ pixels so that \yxyz is luminance
		for (int i = 0; i < xPixelWidth * yPixelWidth; ++i) {
			X(i) *= 683.f;
			Y(i) *= 683.f;
			Z(i) *= 683.f;
		}
		dinfo.toneMap->Map(xyz, xPixelWidth, yPixelWidth, dinfo);
		// Scale tone mapped pixels to $[0,1]$ range
		Float invMaxDisplayY = 1.f / dinfo.maxDisplayY;
		for (int i = 0; i < xPixelWidth * yPixelWidth; ++i) {
			X(i) *= invMaxDisplayY;
			Y(i) *= invMaxDisplayY;
			Z(i) *= invMaxDisplayY;
		}
	}
	// Convert image to display RGB
	Float *rgb = new Float[3*nPix];
	// Define RGB access macros
	#define R(i) (rgb[3*(i)])
	#define G(i) (rgb[3*(i)+1])
	#define B(i) (rgb[3*(i)+2])
	for (int i = 0; i < nPix; ++i) {
		R(i) = dinfo.rWeight[0]*X(i) + dinfo.rWeight[1]*Y(i) +
			dinfo.rWeight[2]*Z(i);
		G(i) = dinfo.gWeight[0]*X(i) + dinfo.gWeight[1]*Y(i) +
			dinfo.gWeight[2]*Z(i);
		B(i) = dinfo.bWeight[0]*X(i) + dinfo.bWeight[1]*Y(i) +
			dinfo.bWeight[2]*Z(i);
	}
	delete[] xyz;
	xyz = NULL;
	// Scale and handle out-of-gamut RGB values
	for (int i = 0; i < nPix; ++i) {
		R(i) *= dinfo.gain;
		G(i) *= dinfo.gain;
		B(i) *= dinfo.gain;
	}
	if (dinfo.clampToGamut) {
		for (int i = 0; i < nPix; ++i) {
			R(i) = Clamp(R(i), 0., 1.);
			G(i) = Clamp(G(i), 0., 1.);
			B(i) = Clamp(B(i), 0., 1.);
		}
	}
	// Apply gamma correction
	if (dinfo.invGamma != 1.f) {
		for (int i = 0; i < nPix; ++i) {
			R(i) = powf(R(i), dinfo.invGamma);
			G(i) = powf(G(i), dinfo.invGamma);
			B(i) = powf(B(i), dinfo.invGamma);
		}
	}
	// Map image to display range
	if (dinfo.integerFormat) {
		for (int i = 0; i < nPix; ++i) {
			R(i) *= dinfo.maxDisplayValue;
			G(i) *= dinfo.maxDisplayValue;
			B(i) *= dinfo.maxDisplayValue;
			AlphaOut[i] *= dinfo.maxDisplayValue;
		}
	}
	// Dither image
	if (dinfo.ditherAmount > 0) {
		for (int i = 0; i < nPix; ++i) {
			R(i) += RandomFloat(-dinfo.ditherAmount,
				dinfo.ditherAmount);
			G(i) += RandomFloat(-dinfo.ditherAmount,
				dinfo.ditherAmount);
			B(i) += RandomFloat(-dinfo.ditherAmount,
				dinfo.ditherAmount);
		}
	}
	// Save display image to disk
	if (dinfo.integerFormat)
		TIFFWrite8Bit(dinfo.filename, rgb, AlphaOut, xPixelWidth,
			yPixelWidth, 3, xResolution, yResolution);
	else
		TIFFWriteFloat(dinfo.filename, rgb, AlphaOut, xPixelWidth,
			yPixelWidth, 3, xResolution, yResolution);
	// Release temporary image memory
	delete[] Lout;
	delete[] AlphaOut;
}
