// kdtree2.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef KDTREE2_H
#define KDTREE2_H
#include "lrt.h"
#include "geometry.h"
// Kd Tree Declarations
enum SplitAxis { SPLIT_X, SPLIT_Y, SPLIT_Z, NOTHING };

template <class NodeData> struct KdNode2 {
	NodeData data;
	SplitAxis split;
};

template <class NodeData, class LookupProc> class KdTree2 {
public:
	// KdTree Method Declarations
	KdTree2(const vector<NodeData> &data);
	~KdTree2() {}
	void recursiveBuild(int index,
											vector<NodeData> &data, int start, int end);
	void Lookup(const Point &P, const LookupProc &process,
							Float &maxDist) const;
	void recursiveLookup(int index, const Point &P,
											 const LookupProc &process, Float &maxDist) const;
private:
	// KdTree Private Data
	vector< KdNode2<NodeData> > tab;
	int limit;
};

template<class NodeData> struct CompareNodeX {
	bool operator()(const NodeData &d1,
			const NodeData &d2) const {
		return d1.P.x < d2.P.x;
	}
};
template<class NodeData> struct CompareNodeY {
	bool operator()(const NodeData &d1,
			const NodeData &d2) const {
		return d1.P.y < d2.P.y;
	}
};
template<class NodeData> struct CompareNodeZ {
	bool operator()(const NodeData &d1,
			const NodeData &d2) const {
		return d1.P.z < d2.P.z;
	}
};

// Kd Tree Method Definitions
template <class NodeData, class LookupProc>
KdTree2<NodeData, LookupProc>::KdTree2(const vector<NodeData> &d) {
	vector<NodeData> data = d;
	tab.reserve(data.size()*2);
	for(unsigned int i=0;i<data.size()*2;i++)
		tab[i].split=NOTHING;
	recursiveBuild(1,data, 0, int(data.size()));
	limit=data.size()-1;
}

template <class NodeData, class LookupProc>
void
KdTree2<NodeData, LookupProc>::recursiveBuild(int index,
		vector<NodeData> &data, int start, int end) {
	// Create leaf node of kd tree if we've reached the bottom
	if (start >= end) {
		tab[index-1].split=NOTHING;
		return;
	}
	if (start + 1 == end) {
		tab[index-1].data=data[start];
		tab[index-1].split=SPLIT_X;
		return;
	}
	// Choose split direction and partition data
	// Compute bounds of data from [[start]] to [[end]]
	BBox bound;
	for (int i = start; i < end; ++i)
		bound = Union(bound, data[i].P);
	Vector diag = bound.pMax - bound.pMin;
	SplitAxis split;
	int splitPos = (start+end)/2;
	if (diag.x > diag.y && diag.x > diag.z) {
		split = SPLIT_X;
		std::nth_element(&data[start], &data[splitPos], &data[end],
			CompareNodeX<NodeData>());
	}
	else if (diag.y > diag.z) {
		split = SPLIT_Y;
		std::nth_element(&data[start], &data[splitPos], &data[end],
			CompareNodeY<NodeData>());
	}
	else {
		split = SPLIT_Z;
		std::nth_element(&data[start], &data[splitPos], &data[end],
			CompareNodeZ<NodeData>());
	}

	tab[index-1].data=data[splitPos];
	tab[index-1].split=split;
	recursiveBuild(index*2, data, start, splitPos);
	recursiveBuild(index*2+1, data, splitPos+1, end);
}

template <class NodeData, class LookupProc> void
KdTree2<NodeData, LookupProc>::Lookup(const Point &P,
		const LookupProc &process, Float &maxDist) const {
	recursiveLookup(1, P, process, maxDist);
}

template <class NodeData, class LookupProc> void
KdTree2<NodeData, LookupProc>::recursiveLookup(int index,
		const Point &P,	const LookupProc &process,
		Float &maxDist) const {
	const int indexbis=index-1;
	if (indexbis>limit || tab[indexbis].split == NOTHING) return;
	if (tab[indexbis].split == SPLIT_X) {
		// Process Kd node's children based on $x$ split
		if (P.x <= tab[indexbis].data.P.x) {
			recursiveLookup(2*index, P, process, maxDist);
			if (P.x + maxDist >= tab[indexbis].data.P.x)
				recursiveLookup(2*index+1, P, process, maxDist);
		}
		else {
			recursiveLookup(2*index+1, P, process, maxDist);
			if (P.x - maxDist <= tab[indexbis].data.P.x)
				recursiveLookup(2*index, P, process, maxDist);
		}
	}
	else if (tab[indexbis].split == SPLIT_Y) {
		// Process Kd node's children based on $y$ split
		if (P.y <= tab[indexbis].data.P.y) {
			recursiveLookup(2*index, P, process, maxDist);
			if (P.y + maxDist >= tab[indexbis].data.P.y)
				recursiveLookup(2*index+1, P, process, maxDist);
		}
		else {
			recursiveLookup(2*index+1, P, process, maxDist);
			if (P.y - maxDist <= tab[indexbis].data.P.y)
				recursiveLookup(2*index, P, process, maxDist);
		}
	}
	else {
		// Process Kd node's children based on $z$ split
		if (P.z <= tab[indexbis].data.P.z) {
			recursiveLookup(2*index, P, process, maxDist);
			if (P.z + maxDist >= tab[indexbis].data.P.z)
				recursiveLookup(2*index+1, P, process, maxDist);
		}
		else {
			recursiveLookup(2*index+1, P, process, maxDist);
			if (P.z - maxDist <= tab[indexbis].data.P.z)
				recursiveLookup(2*index, P, process, maxDist);
		}
	}
	// Hand Kd tree node to processing function
	if (DistanceSquared(tab[indexbis].data.P, P) < maxDist*maxDist)
		process(tab[indexbis].data, maxDist);
}
#endif // KDTREE2_H
