// kdtree.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef KDTREE_H
#define KDTREE_H
#include "lrt.h"
#include "geometry.h"
// Kd Tree Declarations
enum SplitAxis { SPLIT_X, SPLIT_Y, SPLIT_Z };
template <class NodeData> struct KdNode {
	KdNode(const NodeData &d, SplitAxis a)
		: data(d) {
		children[0] = children[1] = NULL;
		split = a;
	}
	~KdNode() { delete children[0]; delete children[1]; }
	// KdNode Data
	NodeData data;
	SplitAxis split;
	KdNode *children[2];
};
template <class NodeData, class LookupProc> class KdTree {
public:
	// KdTree Method Declarations
	KdTree(const vector<NodeData> &data);
	~KdTree() { delete root; }
	void recursiveBuild(KdNode<NodeData> **node,
			vector<NodeData> &data, int start, int end);
	void Lookup(const Point &P, const LookupProc &process,
							Float &maxDist) const;
	NodeData *LookupOne(const Point &P, Float &maxDist) const;
	void recursiveLookup(KdNode<NodeData> *node, const Point &P,
		const LookupProc &process, Float &maxDist) const;

	// KdTree Private Data
	KdNode<NodeData> *root;
};
template<class NodeData> struct CompareNodeX {
	bool operator()(const NodeData &d1,
			const NodeData &d2) const {
		return d1.P.x < d2.P.x;
	}
};
template<class NodeData> struct CompareNodeY {
	bool operator()(const NodeData &d1,
			const NodeData &d2) const {
		return d1.P.y < d2.P.y;
	}
};
template<class NodeData> struct CompareNodeZ {
	bool operator()(const NodeData &d1,
			const NodeData &d2) const {
		return d1.P.z < d2.P.z;
	}
};
// Kd Tree Method Definitions
template <class NodeData, class LookupProc>
KdTree<NodeData, LookupProc>::KdTree(const vector<NodeData> &d) {
	vector<NodeData> data = d;
	recursiveBuild(&root, data, 0, int(data.size()));
}
template <class NodeData, class LookupProc>
void
KdTree<NodeData, LookupProc>::recursiveBuild(KdNode<NodeData> **node,
		vector<NodeData> &data, int start, int end) {
	// Create leaf node of kd tree if we've reached the bottom
	if (start >= end) {
		*node = NULL;
		return;
	}
	if (start + 1 == end) {
		*node = new KdNode<NodeData>(data[start], SPLIT_X);
		return;
	}
	// Choose split direction and partition data
	// Compute bounds of data from [[start]] to [[end]]
	BBox bound;
	for (int i = start; i < end; ++i)
		bound = Union(bound, data[i].P);
	Vector diag = bound.pMax - bound.pMin;
	SplitAxis split;
	int splitPos = (start+end)/2;
	if (diag.x > diag.y && diag.x > diag.z) {
		split = SPLIT_X;
		std::nth_element(&data[start], &data[splitPos], &data[end],
			CompareNodeX<NodeData>());
	}
	else if (diag.y > diag.z) {
		split = SPLIT_Y;
		std::nth_element(&data[start], &data[splitPos], &data[end],
			CompareNodeY<NodeData>());
	}
	else {
		split = SPLIT_Z;
		std::nth_element(&data[start], &data[splitPos], &data[end],
			CompareNodeZ<NodeData>());
	}
	// Allocate Kd tree node and continue recursively
	*node = new KdNode<NodeData>(data[splitPos], split);
	recursiveBuild(&((*node)->children[0]), data, start, splitPos);
	recursiveBuild(&((*node)->children[1]), data, splitPos+1, end);
}


template <class NodeData, class LookupProc> void
KdTree<NodeData, LookupProc>::Lookup(const Point &P,
		const LookupProc &process, Float &maxDist) const {
	recursiveLookup(root, P, process, maxDist);
}
template <class NodeData, class LookupProc> void
KdTree<NodeData, LookupProc>::recursiveLookup(KdNode<NodeData> *node,
		const Point &P,	const LookupProc &process,
		Float &maxDist) const {
	if (!node) return;
	if (node->split == SPLIT_X) {
		// Process Kd node's children based on $x$ split
		if (P.x <= node->data.P.x) {
			recursiveLookup(node->children[0], P, process, maxDist);
			if (P.x + maxDist >= node->data.P.x)
				recursiveLookup(node->children[1], P, process, maxDist);
		}
		else {
			recursiveLookup(node->children[1], P, process, maxDist);
			if (P.x - maxDist <= node->data.P.x)
				recursiveLookup(node->children[0], P, process, maxDist);
		}
	}
	else if (node->split == SPLIT_Y) {
		// Process Kd node's children based on $y$ split
		if (P.y <= node->data.P.y) {
			recursiveLookup(node->children[0], P, process, maxDist);
			if (P.y + maxDist >= node->data.P.y)
				recursiveLookup(node->children[1], P, process, maxDist);
		}
		else {
			recursiveLookup(node->children[1], P, process, maxDist);
			if (P.y - maxDist <= node->data.P.y)
				recursiveLookup(node->children[0], P, process, maxDist);
		}
	}
	else {
		// Process Kd node's children based on $z$ split
		if (P.z <= node->data.P.z) {
			recursiveLookup(node->children[0], P, process, maxDist);
			if (P.z + maxDist >= node->data.P.z)
				recursiveLookup(node->children[1], P, process, maxDist);
		}
		else {
			recursiveLookup(node->children[1], P, process, maxDist);
			if (P.z - maxDist <= node->data.P.z)
				recursiveLookup(node->children[0], P, process, maxDist);
		}
	}
	// Hand Kd tree node to processing function
	if (DistanceSquared(node->data.P, P) < maxDist*maxDist)
		process(node->data, maxDist);
}
#endif // KDTREE_H
