// light.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef LIGHT_H
#define LIGHT_H
#include "lrt.h"
#include "geometry.h"
#include "transform.h"
#include "color.h"
#include "paramset.h"
#include "mc.h"
// Light Classes
class Light {
public:
	// Light Interface
	virtual ~Light();
	Light(bool shadows, const Transform &l2w,
			const Spectrum &power) {
		CastsShadows = shadows;
		LightToWorld = l2w;
		WorldToLight = LightToWorld.GetInverse();
		Power = power;
		doGlow = false;
	}
	virtual Spectrum L(const Point &Ps, const Point &Pl,
		VisibilityTester *) const;
	virtual Spectrum dE(const Point &P, const Normal &N,
		Vector *w, VisibilityTester *vis) const = 0;
	#if 0
	virtual Spectrum E(const DifferentialGeometry &dg, Vector *w) const = 0;
	#endif
	virtual Spectrum Le(const RayDifferential &r) const;
	virtual Spectrum Sample_L(const Point &P, Float u1, Float u2, Vector *wo, Float *weight,
			bool *deltaLight, VisibilityTester *) const = 0;
	virtual Spectrum Sample_L(const Point &P, const Normal &N, Float u1, Float u2,
			Vector *wo, Float *weight, bool *deltaLight,
			VisibilityTester *) const;
	virtual Spectrum Sample_L(const Scene *scene, Float u1, Float u2,
			Float u3, Float u4, Ray *ray, Float *weight,
			bool *deltaLight) const = 0;
	virtual Float Weight(const Point &P, const Normal &,
			const Vector &w) const {
		return Weight(P, w);
	}
	virtual Float Weight(const Point &,
			const Vector &) const = 0;

	virtual Spectrum Glow(const RayDifferential &r) const {
	  return Spectrum(0.);
	}
	Point copyPos;
	bool doGlow;
protected:
	// Light Protected Data
	bool CastsShadows;
	Transform WorldToLight, LightToWorld;
	Spectrum Power;
};
struct VisibilityTester {
	void SetSegment(const Point &p1, const Point &p2, bool castsShadows) {
		r = Ray(p1, p2-p1, RAY_EPSILON, 1.f - RAY_EPSILON);
		traceRay = castsShadows;
	}
	void SetRay(const Point &p, const Vector &d, bool castsShadows) {
		r = Ray(p, d, RAY_EPSILON);
		traceRay = castsShadows;
	}
	bool Unoccluded(const Scene *scene) const;
	Spectrum Transmittance(const Scene *scene) const;

	Ray r;
	bool traceRay;
};
// AreaLight Classes
class AreaLight : public Light {
public:
	// AreaLight Interface
	AreaLight(bool shadows, const Transform &light2world,
		const Spectrum &power, const Reference<Shape> &shape);
	virtual Spectrum L(const Point &x, const Vector &w) const {
		return Power/(area * M_PI);
	}
	virtual Spectrum B(const Point &x) const {
		return Power/area;
	}
	Spectrum L(const Point &Ps, const Point &Pl,
		VisibilityTester *visibility) const;
	virtual Spectrum Sample_L(const Point &P, const Normal &N,
		Float u1, Float u2, Vector *wo, Float *weight, bool *deltaLight,
		VisibilityTester *visibility) const;
	virtual Spectrum Sample_L(const Point &P, Float u1, Float u2, Vector *wo,
		Float *weight, bool *deltaLight, VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, Float u1, Float u2,
			Float u3, Float u4, Ray *ray, Float *weight,
			bool *deltaLight) const;
	Float Weight(const Point &, const Vector &) const;
	Float Weight(const Point &, const Normal &, const Vector &) const;
	Spectrum dE(const Point &P, const Normal &N, Vector *w, VisibilityTester *visibility) const;
protected:
	// AreaLight Protected Data
	Reference<Shape> shape;
	Float area;
	friend class MultiAreaLight;
};
class MultiAreaLight : public Light {
public:
	// MultiAreaLight Methods
	MultiAreaLight(AreaLight *p, int nSamples)
		: Light(p->CastsShadows, p->LightToWorld, p->Power) {
		parent = p;
		scale = 1.f / (Float)nSamples;
	}
	Spectrum Le(const RayDifferential &r) const;
	Spectrum L(const Point &x, const Vector &w) const;
	Spectrum B(const Point &x) const;
	Spectrum L(const Point &Ps, const Point &Pl,
		VisibilityTester *visibility) const;
	Spectrum dE(const Point &P, const Normal &N,
		Vector *w, VisibilityTester *vis) const;
	Spectrum Sample_L(const Point &P, Float u1, Float u2, Vector *wo, Float *weight,
			bool *deltaLight, VisibilityTester *) const;
	Spectrum Sample_L(const Point &P, const Normal &N, Float u1, Float u2,
			Vector *wo, Float *weight, bool *deltaLight,
			VisibilityTester *) const;
	Spectrum Sample_L(const Scene *scene, Float u1, Float u2,
			Float u3, Float u4, Ray *ray, Float *weight,
			bool *deltaLight) const;
	Float Weight(const Point &P, const Normal &N, const Vector &w) const;
	Float Weight(const Point &P, const Vector &w) const;
private:
	// MultiAreaLight Private Data
	AreaLight *parent;
	Float scale;
};
#endif // LIGHT_H
