// mc.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "geometry.h"
#include "shapes.h"
#include "mc.h"
// MC Function Definitions
void ComputeStep1dCDF(Float *f, int nSteps, Float *c, Float *cdf) {
	// Compute integral of step function at $x_i$
	int i;
	cdf[0] = 0.;
	for (i = 1; i < nSteps+1; ++i)
		cdf[i] = cdf[i-1] + f[i-1] / nSteps;
	// Transform step function integral into cdf
	 *c = cdf[nSteps];
	for (i = 1; i < nSteps+1; ++i)
		cdf[i] /= *c;
}
Float SampleStep1d(Float *f, Float *cdf, Float c,
		int nSteps, Float u, Float *weight) {
	// Find surrounding cdf segments
	Float *ptr = std::lower_bound(cdf, cdf+nSteps+1, u);
	int offset = (int) (ptr-cdf-1);
	// Return offset along current cdf segment
	u = (u - cdf[offset]) / (cdf[offset+1] - cdf[offset]);
	*weight = f[offset] / (c * nSteps);
	return (offset + u) / nSteps;
}
Constant2DSampler::Constant2DSampler(Float *_f, int _nx, int _ny) {
#if 0
	f = f;
	nx = _nx;
	ny = _ny;
	// Compute $f_x(x)$ cdf
	fx = new Float[nx];
	for (int x = 0; x < nx; ++x) {
		fx[x] = 0.;
		for (int y = 0; y < ny; ++y)
			fx[x] += f[y*nx + x];
	}
	xcdf = new Float[nx];
	ComputeStep1dCDF(fx, nx, &xc, xcdf);
	// Compute $y$ column cdfs
	ycdfs = new Float *[nx];
	ycs = new Float[nx];
	Float *ftmp = new Float[ny];
	for (int x = 0; x < nx; ++x) {
		for (int y = 0; y < ny; ++y)
			ftmp[y] = f[y*nx + x];
		ycdfs[x] = new Float[ny];
		ComputeStep1dCDF(ftmp, ny, &ycs[x], ycdfs[x]);
	}
	delete[] ftmp;
#endif
}
Constant2DSampler::~Constant2DSampler() {
	delete[] f;
	delete[] fx;
	delete[] xcdf;
	delete[] ycs;
	for (int x = 0; x < nx; ++x)
		delete[] ycdfs[x];
	delete[] ycdfs;
}
void Constant2DSampler::Sample(Float u1, Float u2, Float *x, Float *y, Float *weight) const {
#if 0
	Float xweight, yweight;
	*x = SampleStep1d(fx, xcdf, nx, u1, &xweight);
	int xoffset = int(*x * nx);
	*y = SampleStep1d(ycdfs[xoffset], ny, u2, &yweight);
	*weight = *xweight * *yweight;
#endif
}
void ComputeLinear1dCDF(Float *values, int nValues, Float delta,
		Float *cdf) {
	// Compute integral of piecewise linear function
	int i;
	cdf[0] = 0;
	for (i = 1; i < nValues; ++i)
		cdf[i] = cdf[i-1] +
			(0.5f * values[i-1] + 0.5f * values[i]) * delta;
	// Compute piecewise linear function's cdf
	Float c = 1.f / cdf[nValues-1];
	for (i = 1; i < nValues; ++i)
		cdf[i] *= c;
	cdf[nValues-1] = 1.;
}
Float SampleLinear1dCDF(Float *values, Float *cdf, int nValues,
		Float dx, Float u) {
	Float *ptr = std::lower_bound(cdf, cdf+nValues, u);
	int o = (int) (ptr-cdf-1);
	// Compute offset [[delta]] along segment
	Float regionInt = 0.5f * (values[o] + values[o+1]);
	Float uPrime = (u - cdf[o]) / (cdf[o+1] - cdf[o]);
	Float A = .5f * (values[o+1] - values[o]) / regionInt;
	Float B = values[o] / regionInt;
	Float C = -uPrime;
	Float t0, delta;
	bool ok = Quadratic(A, B, C, &t0, &delta);
	Assert(ok && (t0 < 0 || t0 > 1) && (delta >= 0.f && delta <= 1.f));
	return (o+delta) * dx;
}
Vector RejectionSampleHemisphere() {
	Vector wo;
	// Sample BxDF hemisphere uniformly to compute [[wo]]
	while (1) {
		wo = Vector(RandomFloat(-1, 1), RandomFloat(-1, 1),
			RandomFloat(-1, 1));
		if (wo.LengthSquared() < 1.f) break;
	}
	wo = wo.Hat();
	if (wo.z < 0.) wo.z *= -1;
	return wo;
}
Float UniformHemisphereWeight() {
	return 1.f / (2.f * M_PI);
}
Vector UniformSampleHemisphere(Float u1, Float u2) {
	Float z = 1 - u1;
	Float r = sqrtf(max(0.f, 1.f - z*z));
	Float phi = 2 * M_PI * u2;
	Float x = r * sinf(phi);
	Float y = r * cosf(phi);
	return Vector(x, y, z);
}
Vector UniformSampleSphere(Float u1, Float u2) {
	Float z = 1.f - 2.f * u1;
	Float r = sqrtf(max(0.f, 1.f - z*z));
	Float phi = 2.f * M_PI * u2;
	Float x = r * sinf(phi);
	Float y = r * cosf(phi);
	return Vector(x, y, z);
}
Vector UniformSampleCone(Float u1, Float u2, Float costhetamax) {
	// Uniformly sample $\theta$ and $\phi$ in cone
	Float costheta = Lerp(u1, costhetamax, 1.f);
	Float sintheta = sqrtf(1.f - costheta*costheta);
	Float phi = u2 * 2.f * M_PI;
	return Vector(cosf(phi) * sintheta, sinf(phi) * sintheta, costheta);
}
Vector UniformSampleCone(Float u1, Float u2, Float costhetamax,
		const Vector &x, const Vector &y, const Vector &z) {
	// Uniformly sample $\theta$ and $\phi$ in cone
	Float costheta = Lerp(u1, costhetamax, 1.f);
	Float sintheta = sqrtf(1.f - costheta*costheta);
	Float phi = u2 * 2.f * M_PI;
	return cosf(phi) * sintheta * x + sinf(phi) * sintheta * y +
		costheta * z;
}
Float UniformConeWeight(Float cosThetaMax) {
	return 1.f / (2.f * M_PI * (1.f - cosThetaMax));
}
void UniformSampleDisk(Float u1, Float u2, Float *x, Float *y) {
	Float r = sqrtf(u2);
	Float theta = 2.0f * M_PI * u1;
	*x = r * cosf(theta);
	*y = r * sinf(theta);
}
void ConcentricSampleDisk(Float u1, Float u2,
		Float *dx, Float *dy) {
	Float r, theta;
	// Map uniform random numbers to $[-1,1]^2$
	Float sx = 2 * u1 - 1;
	Float sy = 2 * u2 - 1;
	// Map square to $(r,\theta)$
	// Handle degeneracy at the origin
	if (sx == 0.0 && sy == 0.0) {
		*dx = 0.0;
		*dy = 0.0;
		return;
	}
	if (sx >= -sy) {
		if (sx > sy) {
			// Handle first region
			r = sx;
			if (sy > 0.0)
				theta = sy/r;
			else
				theta = 8.0f + sy/r;
		}
		else {
			// Handle second region
			r = sy;
			theta = 2.0f - sx/r;
		}
	}
	else {
		if (sx <= sy) {
			// Handle third region
			r = -sx;
			theta = 4.0f - sy/r;
		}
		else {
			// Handle fourth region
			r = -sy;
			theta = 6.0f + sx/r;
		}
	}
	theta *= M_PI / 4;
	*dx = r*cosf(theta);
	*dy = r*sinf(theta);
}
