// mipmap.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef MIPMAP_H
#define MIPMAP_H 1
#include "lrt.h"
// MIPMap Declarations
template <class T> class MIPMap {
public:
	// MIPMap Public Interface
	MIPMap(int xres, int yres, T *data);
	~MIPMap();
	T Lookup(Float s, Float t, Float width = 0.f) const;
	T Lookup(Float s, Float t, Float ds0, Float dt0,
		Float ds1, Float dt1) const;
private:
	// MIPMap Private Methods
	const T &texel(int level, int s, int t) const;
	T &texel(int level, int s, int t);
	T bilerp(int level, Float s, Float t) const;
	T EWA(Float s, Float t, Float ds0, Float dt0,
			Float ds1, Float dt1, int level) const;
	// MIPMap Private Data
	T **data;
	int nLevels;
	int *sRes, *tRes, *sBlocks;
	#define LOG_BLOCK_SIZE 2
	#define BLOCK_SIZE (1 << LOG_BLOCK_SIZE)
	#define UP(x) (((x)+BLOCK_SIZE-1) & ~(BLOCK_SIZE-1))
	#define WEIGHT_LUT_SIZE 128
	static Float *weightLut;
};
// MIPMap Method Definitions
template <class T>
MIPMap<T>::MIPMap(int sr, int tr, T *d) {
	nLevels = Log2Int(max(sr, tr));
	sRes = new int[3*nLevels];
	tRes = sRes + nLevels;
	sBlocks = tRes + nLevels;
	data = new T *[nLevels];
	// Initialize most detailed level of MIPMap
	sRes[0] = sr;
	tRes[0] = tr;
	sBlocks[0] = UP(sr) >> LOG_BLOCK_SIZE;
	data[0] = (T *)AllocL2CacheAligned(UP(sr) * UP(tr) * sizeof(T));
	for (int t = 0; t < tr; ++t)
		for (int s = 0; s < sr; ++s)
			texel(0, s, t) = d[s + t * sr];
	for (int i = 1; i < nLevels; ++i) {
		// Initialize $i$th MIPMap level from $i-1$st level
		sRes[i] = max(1, sRes[i-1]/2);
		tRes[i] = max(1, tRes[i-1]/2);
		sBlocks[i] = UP(sRes[i]) >> LOG_BLOCK_SIZE;
		Assert(sBlocks[i] > 0);
		data[i] = (T *)AllocL2CacheAligned(UP(sRes[i]) * UP(tRes[i]) * sizeof(T));
		// Filter four texels from finer level of pyramid
		for (int t = 0; t < tRes[i]; ++t)
			for (int s = 0; s < sRes[i]; ++s)
				texel(i, s, t) = .25f * (texel(i-1, 2*s, 2*t) +
					texel(i-1, 2*s+1, 2*t) + texel(i-1, 2*s, 2*t+1) +
					texel(i-1, 2*s+1, 2*t));
	}
	// Initialize MIPMap filter weights if needed
	if (!weightLut) {
		weightLut = (Float *)AllocL2CacheAligned(WEIGHT_LUT_SIZE *
			sizeof(Float));
		for (int i = 0; i < WEIGHT_LUT_SIZE; ++i) {
			Float alpha = 2;
			Float r2 = float(i) / float(WEIGHT_LUT_SIZE - 1);
			weightLut[i] = expf(-alpha * r2);
		}
	}
}
template <class T>
const T &MIPMap<T>::texel(int level, int s, int t) const {
	// Compute texel block and offsets for $(s,t)$
	s = Clamp(s, 0, sRes[level]-1);
	t = Clamp(t, 0, tRes[level]-1);
	#define BLOCK(a) ((a) >> LOG_BLOCK_SIZE)
	#define OFFSET(a) ((a) & (BLOCK_SIZE-1))
	int bs = BLOCK(s), bt = BLOCK(t);
	int os = OFFSET(s), ot = OFFSET(t);
	// Compute overall offset to texel and return value
	int offset = BLOCK_SIZE * BLOCK_SIZE *
		((UP(sRes[level]) >> LOG_BLOCK_SIZE) * bt + bs);
	offset += BLOCK_SIZE * ot + os;
	return data[level][offset];
}
template <class T>
T &MIPMap<T>::texel(int level, int s, int t) {
	// Compute texel block and offsets for $(s,t)$
	s = Clamp(s, 0, sRes[level]-1);
	t = Clamp(t, 0, tRes[level]-1);
	#define BLOCK(a) ((a) >> LOG_BLOCK_SIZE)
	#define OFFSET(a) ((a) & (BLOCK_SIZE-1))
	int bs = BLOCK(s), bt = BLOCK(t);
	int os = OFFSET(s), ot = OFFSET(t);
	// Compute overall offset to texel and return value
	int offset = BLOCK_SIZE * BLOCK_SIZE *
		((UP(sRes[level]) >> LOG_BLOCK_SIZE) * bt + bs);
	offset += BLOCK_SIZE * ot + os;
	return data[level][offset];
}
template <class T>
MIPMap<T>::~MIPMap() {
	for (int i = 0; i < nLevels; ++i)
		FreeCacheAligned(data[i]);
	delete[] data;
	delete[] sRes;
}
template <class T>
T MIPMap<T>::bilerp(int level, Float s, Float t) const {
	level = Clamp(level, 0, nLevels-1);
	s = s * sRes[level] - 0.5f;
	t = t * tRes[level] - 0.5f;
	int s0 = Floor2Int(s), t0 = Floor2Int(t);
	Float ds = s - s0, dt = t - t0;
	return (1.-ds)*(1.-dt) * texel(level, s0, t0) +
		(1.-ds)*dt * texel(level, s0, t0+1) +
		ds*(1.-dt) * texel(level, s0+1, t0) +
		ds*dt * texel(level, s0+1, t0+1);
}
template <class T>
T MIPMap<T>::Lookup(Float s, Float t, Float width) const {
	static StatsCounter mipTrilerps("Texture",
		"Trilinear MIPMap lookups");
	++mipTrilerps;
	// Compute MIPMap level for trilinear filtering
	Float level = nLevels - 1 + Log2(max(width, 1e-8f));
	// Perform trilinear interpolation at appropriate MIPMap level
	if (level < 0)
		return bilerp(0, s, t);
	else if (level >= nLevels - 1)
		return texel(nLevels-1, 0, 0);
	else {
		int iLevel = Floor2Int(level);
		Float delta = level - iLevel;
		return (1.f-delta) * bilerp(iLevel, s, t) +
			delta * bilerp(iLevel+1, s, t);
	}
}
template <class T>
T MIPMap<T>::Lookup(Float s, Float t, Float ds0, Float dt0,
		Float ds1, Float dt1) const {
	static StatsCounter ewaLookups("Texture",
		"EWA filter lookups");
	++ewaLookups;
	// Compute ellipse minor and major axes
	if (ds0*ds0 + dt0*dt0 < ds1*ds1 + dt1*dt1) {
		swap(ds0, ds1);
		swap(dt0, dt1);
	}
	Float majorLength = sqrtf(ds0*ds0 + dt0*dt0);
	Float minorLength = sqrtf(ds1*ds1 + dt1*dt1);
	// Clamp ellipse eccentricity if too large
	const Float maxEccentricity = 10.f;
	Float e = majorLength / minorLength;
	if (e > maxEccentricity) {
		Float scale = e / maxEccentricity;
		ds1 *= scale;
		dt1 *= scale;
		minorLength *= scale;
	}
	// Choose level of detail for EWA lookup
	int lod = max(0, nLevels + Log2Int(minorLength / 2.5f));
	if (lod >= nLevels-1)
		return texel(nLevels-1, 0, 0);
	else
		return EWA(s, t, ds0, dt0, ds1, dt1, lod);
}
template <class T>
T MIPMap<T>::EWA(Float s, Float t, Float ds0, Float dt0,
		Float ds1, Float dt1, int level) const {
	// Convert EWA coordinates to appropriate scale for level
	s *= sRes[level] - 1; // - 0.5f;
	t *= tRes[level] - 1; // - 0.5f;
	ds0 *= sRes[level] - 1;
	dt0 *= tRes[level] - 1;
	ds1 *= sRes[level] - 1;
	dt1 *= tRes[level] - 1;
	// Compute ellipse coefficients to bound EWA filter region
	Float A = dt0*dt0 + dt1*dt1 + 1;
	Float B = -2.f * (ds0*dt0 + ds1*dt1);
	Float C = ds0*ds0 + ds1*ds1 + 1;
	Float invF = 1.f / (A*C - B*B*0.25f);
	A *= invF;
	B *= invF;
	C *= invF;
	// Compute the ellipse's $(s,t)$ bounding box in texture space
	Float det = -B*B + 4.f*A*C;
	Float invDet = 1.f / det;
	Float uSqrt = sqrtf(det * C), vSqrt = sqrtf(A * det);
	int s0 = Ceil2Int (s - 2.f * invDet * uSqrt);
	int s1 = Floor2Int(s + 2.f * invDet * uSqrt);
	int t0 = Ceil2Int (t - 2.f * invDet * vSqrt);
	int t1 = Floor2Int(t + 2.f * invDet * vSqrt);
	static StatsRatio ewaTexels("Texture",
		"Texels per EWA lookup", false);
	ewaTexels.add((1+s1-s0) * (1+t1-t0), 1);
	// Scan over ellipse bound and compute quadratic equation
	T num(0.);
	Float den = 0;
	for (int it = t0; it <= t1; ++it) {
		Float tt = it - t;
		for (int is = s0; is <= s1; ++is) {
			Float ss = is - s;
			Float r2 = A*ss*ss + B*ss*tt + C*tt*tt;
			if (r2 < 1.) {
				// Add EWA sample for current texel
				Float weight = weightLut[min(Float2Int(r2 * WEIGHT_LUT_SIZE),
					WEIGHT_LUT_SIZE-1)];
				num += texel(level, is, it) * weight;
				den += weight;
			}
		}
	}
	return num / den;
}
template <class T> Float *MIPMap<T>::weightLut = NULL;
#endif // MIPMAP_H
