// octree.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef OCTREE_H
#define OCTREE_H
#include "lrt.h"
#include "geometry.h"
// Octree Declarations
template <class NodeData> struct OctNode {
	// OctNode Method Declarations
	OctNode() {
		for (int i = 0; i < 8; ++i)
			children[i] = NULL;
	}
	~OctNode() {
		for (int i = 0; i < 8; ++i)
			delete children[i];
	}
	OctNode *children[8];
	vector<NodeData> data;
};
template <class NodeData, class LookupProc> class Octree {
public:
	// Octree Method Declarations
	Octree(const BBox &b)
		: bound(b) {
	}
	void Add(const NodeData &dataItem, const BBox &dataBound) {
		add(&root, bound, dataItem, dataBound,
			DistanceSquared(dataBound.pMin, dataBound.pMax));
	}
	void add(OctNode<NodeData> *node, const BBox &nodeBound,
		const NodeData &dataItem, const BBox &dataBound, Float diag2);
	void Lookup(const Point &P, const LookupProc &process) {
		if (!bound.Inside(P)) return;
		lookup(&root, bound, P, process);
	}
	void lookup(OctNode<NodeData> *node, const BBox &nodeBound, const Point &P,
			const LookupProc &process);
private:
	// Octree Private Data
	BBox bound;
	OctNode<NodeData> root;
};
// Octree Method Definitions
template <class NodeData, class LookupProc>
void Octree<NodeData, LookupProc>::add(OctNode<NodeData> *node,
		const BBox &nodeBound, const NodeData &dataItem,
		const BBox &dataBound, Float diag2) {
	// Possibly add data item to current octree node
	if (DistanceSquared(nodeBound.pMin, nodeBound.pMax) < diag2) {
		node->data.push_back(dataItem);
		return;
	}
	// Otherwise add data item to octree children
	Point pMid = .5 * nodeBound.pMin + .5 * nodeBound.pMax;
	// Determine which children the item overlaps
	bool over[8];
	over[0] = over[1] = over[2] = over[3] =  (dataBound.pMin.x <= pMid.x);
	over[4] = over[5] = over[6] = over[7] =  (dataBound.pMax.x  > pMid.x);
	over[0] = over[1] = over[4] = over[5] &= (dataBound.pMin.y <= pMid.y);
	over[2] = over[3] = over[6] = over[7] &= (dataBound.pMax.y  > pMid.y);
	over[0] = over[2] = over[4] = over[6] &= (dataBound.pMin.z <= pMid.z);
	over[1] = over[3] = over[5] = over[7] &= (dataBound.pMax.z  > pMid.z);
	for (int child = 0; child < 8; ++child) {
		if (!over[child]) continue;
		// Hand data item down to child number [[child]]
		if (!node->children[child])
			node->children[child] = new OctNode<NodeData>;
		// Compute [[childBound]] for [[child]]
		BBox childBound;
		childBound.pMin.x = (child & 4) ? pMid.x : nodeBound.pMin.x;
		childBound.pMax.x = (child & 4) ? nodeBound.pMax.x : pMid.x;
		childBound.pMin.y = (child & 2) ? pMid.y : nodeBound.pMin.y;
		childBound.pMax.y = (child & 2) ? nodeBound.pMax.y : pMid.y;
		childBound.pMin.z = (child & 1) ? pMid.z : nodeBound.pMin.z;
		childBound.pMax.z = (child & 1) ? nodeBound.pMax.z : pMid.z;
		add(node->children[child], childBound, dataItem, dataBound, diag2);
	}
}
template <class NodeData, class LookupProc>
void Octree<NodeData, LookupProc>::lookup(OctNode<NodeData> *node,
		const BBox &nodeBound, const Point &P,
		const LookupProc &process) {
	for (u_int i = 0; i < node->data.size(); ++i)
		process(P, node->data[i]);
	// Determine which octree child node [[P]] is inside
	Point pMid = .5 * nodeBound.pMin + .5 * nodeBound.pMax;
	int child = (P.x > pMid.x ? 4 : 0) +
		(P.y > pMid.y ? 2 : 0) + (P.z > pMid.z ? 1 : 0);
	if (node->children[child]) {
		// Compute [[childBound]] for [[child]]
		BBox childBound;
		childBound.pMin.x = (child & 4) ? pMid.x : nodeBound.pMin.x;
		childBound.pMax.x = (child & 4) ? nodeBound.pMax.x : pMid.x;
		childBound.pMin.y = (child & 2) ? pMid.y : nodeBound.pMin.y;
		childBound.pMax.y = (child & 2) ? nodeBound.pMax.y : pMid.y;
		childBound.pMin.z = (child & 1) ? pMid.z : nodeBound.pMin.z;
		childBound.pMax.z = (child & 1) ? nodeBound.pMax.z : pMid.z;
		lookup(node->children[child], childBound, P, process);
	}
}
#endif // OCTREE_H
