// paramset.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef PARAMSET_H
#define PARAMSET_H
#include "lrt.h"
#include "geometry.h"
#include "color.h"
#include "ri.h"
#include <set>
using std::set;
// ParamSet Types
#define PARAM_TYPE_INT     (1<<0)
#define PARAM_TYPE_FLOAT   (1<<1)
#define PARAM_TYPE_POINT   (1<<2)
#define PARAM_TYPE_VECTOR  (1<<3)
#define PARAM_TYPE_NORMAL  (1<<4)
#define PARAM_TYPE_STRING  (1<<5)
#define PARAM_TYPE_COLOR   (1<<6)
#define PARAM_TYPE_VOID    (1<<7)
#define PARAM_TYPE_HPOINT  (1<<8)
#define PARAM_TYPE_ERROR       -1
#define PARAM_TYPE_UNIFORM (1<<9)
#define PARAM_TYPE_VARYING (1<<10)
#define PARAM_TYPE_VERTEX  (1<<11)
// ParamSet Declarations
class ParamSet {
public:
	// ParamSet Constructors
	ParamSet(int nv = 0) { nVertex = nv; }
	ParamSet(const ParamSet &p2) {
		*this = p2;
	}
	ParamSet &operator=(const ParamSet &p2);
	ParamSet(int n, const char **tokens, void **params, int nv = 0) {
		init(n, tokens, params, nv);
	}
	~ParamSet();
	void init(int n, const char **tokens, void **params, int nv = 0);
	void clear();

	// ParamSet Interface
	void AddFloat(const string &, const Float *, int type = PARAM_TYPE_UNIFORM, int nArray = 1);
	void AddInt(const string &, const int *,
		int type = PARAM_TYPE_UNIFORM, int nArray = 1);
	void AddPoint(const string &, const Point *,
		int type = PARAM_TYPE_UNIFORM, int nArray = 1);
	void AddVector(const string &, const Vector *,
		int type = PARAM_TYPE_UNIFORM, int nArray = 1);
	void AddNormal(const string &, const Normal *,
		int type = PARAM_TYPE_UNIFORM, int nArray = 1);
	void addSpectrum(const string &, const Spectrum *,
		int type = PARAM_TYPE_UNIFORM, int nArray = 1);
	void AddString(const string &, const string *, int nArray = 1);
	Float FindOneFloat(const string &, Float d) const;
	int FindOneInt(const string &, int d) const;
	Point FindOnePoint(const string &, const Point &d) const;
	Vector FindOneVector(const string &, const Vector &d) const;
	Normal FindOneNormal(const string &, const Normal &d) const;
	Spectrum FindOneSpectrum(const string &,
		const Spectrum &d) const;
	string FindOneString(const string &, const string &d) const;
	const Float *FindFloat(const string &, int *type, int *nArray) const;
	const int *findInt(const string &, int *type,
		int *nArray) const;
	const Point *FindPoint(const string &, int *type,
		int *nArray) const;
	const Vector *FindVector(const string &, int *type,
		int *nArray) const;
	const Normal *FindNormal(const string &, int *type,
		int *nArray) const;
	const Spectrum *FindSpectrum(const string &, int *type,
		int *nArray) const;
	const string *FindString(const string &, int *type,
		int *nArray) const;
	int TypeToNum(int type);
	void ReportUnused() const;
	// ParamSet Public Data
	int nVertex;
private:
	// ParamSet Data
	vector<ParamSetItem<int> *> ints;
	vector<ParamSetItem<Float> *> floats;
	vector<ParamSetItem<Point> *> points;
	vector<ParamSetItem<Vector> *> vectors;
	vector<ParamSetItem<Normal> *> normals;
	vector<ParamSetItem<Spectrum> *> spectra;
	vector<ParamSetItem<string> *> strings;
};
template <class T> struct ParamSetItem {
	ParamSetItem(const string &name, const T *val, int type, int count,
		int nVertex = 0);
	~ParamSetItem();
	ParamSetItem<T> *Clone(int nVertex) const;

	string name;
	int type, arraySize;
	T *data;
	bool lookedUp;
};
// ParamSetItem Methods
template <class T>
ParamSetItem<T>::ParamSetItem(const string &n, const T *v, int t,
		int c, int nVertex) {
	name = n;
	type = t;
	arraySize = c;
	lookedUp = false;
	// Determine number of data items for [[ParamSetItem]]
	int nAlloc = arraySize;
	if (type & PARAM_TYPE_VARYING) nAlloc *= 4;
	if (type & PARAM_TYPE_VERTEX) nAlloc *= nVertex;
	data = new T[nAlloc];
	for (int i = 0; i < nAlloc; ++i)
		data[i] = v[i];
}
template <class T>
ParamSetItem<T>::~ParamSetItem() {
	delete[] data;
}
template <class T>
ParamSetItem<T> *ParamSetItem<T>::Clone(int nVertex) const {
	return new ParamSetItem<T>(name, data, type,
		arraySize, nVertex);
}
#endif /* PARAMSET_H */
