// primitives.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "primitives.h"
#include "light.h"
// Primitive Methods
Primitive::~Primitive() { }

bool Primitive::CanIntersect() const {
	return true;
}

void Primitive::Refine(vector<Reference<Primitive> > &refined) const {
	Severe("Unimplemented Primitive::Refine method called!");
}
// GeometricPrimitive Methods
GeometricPrimitive::~GeometricPrimitive() {
//	delete Ns;
//	delete Ss;
}
BBox GeometricPrimitive::WorldBound() const {
	return shape->WorldBound();
}
bool GeometricPrimitive::IntersectP(const Ray &r) const {
	return shape->IntersectP(r);
}
bool GeometricPrimitive::CanIntersect() const {
	return shape->CanIntersect();
}
void GeometricPrimitive::Refine(vector<Reference<Primitive> > &refined) const {
	vector<RefinedShape> r;
	shape->Refine(r);
	for (u_int i = 0; i < r.size(); ++i) {
		GeometricPrimitive *gp = new GeometricPrimitive(r[i].shape,
			material, areaLight, r[i].Ns ? r[i].Ns : Ns,
			r[i].Ss ? r[i].Ss : Ss);
		refined.push_back(gp);
	}
}
GeometricPrimitive::GeometricPrimitive(const Reference<Shape> &s,
		const Reference<Material> &m, AreaLight *a,
		Texture<Normal> *ns, Texture<Vector> *ss) {
	shape = s;
	material = m;
	areaLight = a;
	Ns = ns;
	Ss = ss;
}
bool GeometricPrimitive::Intersect(const Ray &r,
		Intersection *isect) const {
	Float thit;
	if (shape->Intersect(r, &thit, &isect->dgGeom)) {
		isect->primitive = this;
		r.maxt = thit;
		return true;
	}
	return false;
}
void GeometricPrimitive::Bump(const DifferentialGeometry &dgg,
		DifferentialGeometry *dgs) const {
	if (Ns || Ss) {
		// Evaluate texture for shading normal and tangent
		Normal Nvert = Ns ? Ns->Evaluate(dgg).Hat() : dgg.Nn;
		Vector Svert, Tvert;
		Svert = Ss ? Ss->Evaluate(dgg).Hat() : dgg.S;
		if (Ns || Ss) {
			Tvert = Cross(Svert, Nvert);
			Svert = Cross(Tvert, Nvert);
		}
		else
			Tvert = dgg.T;
		*dgs = DifferentialGeometry(dgg.P, Svert, Tvert,
			Vector(0,0,0), Vector(0,0,0), dgg.u, dgg.v,
			dgg.shape, dgg.dudx, dgg.dvdx, dgg.dudy,
			dgg.dvdy);
	}
	else *dgs = dgg;
	material->Bump(dgg, dgs);
}
// PrimitiveList Methods
PrimitiveList::PrimitiveList(const vector<Reference<Primitive> > &p)
	: prims(p) {
	for (u_int i = 0; i < prims.size(); ++i)
		bounds = Union(bounds, prims[i]->WorldBound());
}
bool PrimitiveList::CanIntersect() const {
	for (u_int i = 0; i < prims.size(); ++i)
		if (!prims[i]->CanIntersect())
			return false;
	return true;
}
bool PrimitiveList::Intersect(const Ray &r, Intersection *isect) const {
	bool hit = false;
	for (u_int i = 0; i < prims.size(); ++i)
		if (prims[i]->Intersect(r, isect))
			hit = true;
	return hit;
}
bool PrimitiveList::IntersectP(const Ray &r) const {
	for (u_int i = 0; i < prims.size(); ++i)
		if (prims[i]->IntersectP(r))
			return true;
	return false;
}
void PrimitiveList::Refine(vector<Reference<Primitive> > &refined) const {
	for (u_int i = 0; i < prims.size(); ++i)
		if (prims[i]->CanIntersect())
			refined.push_back(prims[i]);
		else
			prims[i]->Refine(refined);
}
// Intersection Method Definitions
BSDF *Intersection::GetBSDF(const RayDifferential &ray) const {
	// Update statistics for number of points shaded
	static StatsCounter pointsShaded("Shading", "Number of points shaded");
	++pointsShaded;
	// Compute filter region for anti-aliasing
	dgGeom.ComputeDifferentials(ray);
	primitive->Bump(dgGeom, &dgShading);
	return primitive->material->GetBSDF(this);
}
Spectrum Intersection::Le(const Vector &wo) const {
	if (primitive->areaLight)
		return primitive->areaLight->L(dgGeom.P, wo);
	else return Spectrum(0.);
}
