// reflection.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "color.h"
#include "reflection.h"
#include "shapes.h"
#include "mc.h"
#include "sampling.h"
#include <stdarg.h>
// BxDF Utility Functions
Spectrum FrDiel(Float cosi, Float cost, const Spectrum &etai,
		const Spectrum &etat) {
	Spectrum Rparl = ((etat * cosi) - (etai * cost)) /
		((etat * cosi) + (etai * cost));
	Spectrum Rperp = ((etai * cosi) - (etat * cost)) /
		((etai * cosi) + (etat * cost));
	return (Rparl*Rparl + Rperp*Rperp) / 2.f;
}
Spectrum FrCond(Float cosi, const Spectrum &eta, const Spectrum &k) {
    Spectrum tmp = (eta*eta + k*k) * cosi*cosi;
    Spectrum Rparl2 = (tmp - (2.f * eta * cosi) + 1) /
    	(tmp + (2.f * eta * cosi) + 1);
    Spectrum tmp_f = eta*eta + k*k;
    Spectrum Rperp2 = (tmp_f - (2.f * eta * cosi) + cosi*cosi) /
	   (tmp_f + (2.f * eta * cosi) + cosi*cosi);
    return (Rparl2 + Rperp2) / 2.f;
}
Spectrum FresnelApproxEta(const Spectrum &intensity) {
	return (Spectrum(1.) + intensity.Sqrt()) /
		(Spectrum(1.) - intensity.Sqrt());
}
Spectrum FresnelApproxK(const Spectrum &intensity) {
    return 2.f * (intensity / (Spectrum(1.) - intensity)).Sqrt();
}
// BxDF Method Definitions
Spectrum BRDFToBTDF::f(const Vector &wo, const Vector &wi) const {
	return brdf->f(wo, otherHemisphere(wi));
}
Spectrum BRDFToBTDF::f_delta(const Vector &wo, Vector *wi) const {
	Spectrum f = brdf->f_delta(wo, wi);
	*wi = otherHemisphere(*wi);
	return f;
}
Fresnel::~Fresnel() { }
Spectrum FresnelConductor::evaluate(Float cosi) const {
	return FrCond(fabsf(cosi), eta, k);
}
Spectrum FresnelDielectric::evaluate(Float cosi) const {
	// Compute Fresnel reflectance for dielectric
	cosi = Clamp(cosi, -1.f, 1.f);
	// Compute indices of refraction for dielectric
	bool entering = cosi > 0.;
	Float ei = eta_i, et = eta_t;
	if (!entering)
		swap(ei, et);
	Float sint = ei/et * sqrtf(max(0.f, 1.f - cosi*cosi));
	if (sint > 1.) {
		// Handle total internal reflection
		return 1.;
	}
	else {
		Float cost = sqrtf(max(0.f, 1.f - sint*sint));
		return FrDiel(fabsf(cosi), cost, ei, et);
	}
}
Spectrum FresnelNoOp::evaluate(Float cosi) const {
	return 1.;
}
SpecularReflection::~SpecularReflection() {
	delete fresnel;
}
Spectrum SpecularReflection::f_delta(const Vector &wo,
		Vector *wi) const {
	// Compute perfect specular reflection direction
	*wi = Vector(-wo.x, -wo.y, wo.z);
	return fresnel->evaluate(wo.z) * R /
		fabsf(wi->z);
}
Spectrum SpecularTransmission::f_delta(const Vector &wo,
		Vector *wi) const {
	// Figure out which $\eta$ is for incident and which transmitted
	bool entering = wo.z > 0.;
	Float ei = etai, et = etat;
	if (!entering)
		swap(ei, et);
	// Compute transmitted ray direction
	Float sini2 = wo.x*wo.x + wo.y*wo.y;
	Float eta = ei / et;
	Float sint2 = eta * eta * sini2;
	// Handle total internal reflection for transmission
	if (sint2 > 1.) return 0.;
	Float cost = sqrtf(max(0.f, 1.f - sint2));
	if (entering) cost = -cost;
	Float sintOverSini = eta;
	*wi = Vector(sintOverSini * -wo.x, sintOverSini * -wo.y, cost);
	Float cosi = wo.z;
	Spectrum F = fresnel.evaluate(cosi);
	return (et*et)/(ei*ei) * (Spectrum(1.)-F) * T / fabsf(wi->z);
}
Spectrum Lambertian::f(const Vector &wo,
		const Vector &wi) const {
	return RoverPI;
}
Spectrum OrenNayar::f(const Vector &wo,
		const Vector &wi) const {
	Float sinthetai = sqrtf(wi.x*wi.x + wi.y*wi.y);
	Float sinthetao = sqrtf(wo.x*wo.x + wo.y*wo.y);
	// Compute cosine term of Oren--Nayar model
	Float sinphii = wi.y / sinthetai;
	Float cosphii = wi.x / sinthetai;
	Float sinphio = wo.y / sinthetao;
	Float cosphio = wo.x / sinthetao;
	Float dcos = cosphii * cosphio + sinphii * sinphio;
	Float maxcos = max(0.f, dcos);
	// Compute sine and tangent terms of Oren--Nayar model
	Float sinalpha, tanbeta;
	if (fabsf(wi.z) > fabsf(wo.z)) {
		sinalpha = sinthetao;
		tanbeta = sinthetai / fabsf(wi.z);
	}
	else {
		sinalpha = sinthetai;
		tanbeta = sinthetao / fabsf(wo.z);
	}
	return R * INV_PI * (A + B * maxcos * sinalpha * tanbeta);
}
Microfacet::Microfacet(const Spectrum &reflectance, Fresnel *f,
		MicrofacetDistribution *d)
		: R(reflectance) {
	fresnel = f;
	distribution = d;
}
Microfacet::~Microfacet() {
	delete distribution;
	delete fresnel;
}
Spectrum Microfacet::f(const Vector &wo, const Vector &wi) const {
	Float cosThetaO = fabsf(wo.z);
	Float cosThetaI = fabsf(wi.z);
	Vector wh = (wi + wo).Hat();
	Spectrum F = fresnel->evaluate(Dot(wi, wh));
	return R * distribution->D(wh) * G(wi, wo, wh) * F /
		 (4.f * cosThetaI * cosThetaO);
}
Lafortune::Lafortune(const Spectrum &r, int nl, const Spectrum *xx,
	const Spectrum *yy, const Spectrum *zz, const Spectrum *e)
	: R(r) {
	nLobes = nl;
	Assert(nLobes <= MAX_LOBES);
	for (int i = 0; i < nLobes; ++i) {
		x[i] = xx[i];
		y[i] = yy[i];
		z[i] = zz[i];
		exponent[i] = e[i];
	}
}
Spectrum Lafortune::f(const Vector &wo, const Vector &wi) const {
	Spectrum ret = R / M_PI;
	for (int i = 0; i < nLobes; ++i) {
		// Evaluate Lafortune model for $i$th lobe
		Spectrum v = x[i] * wo.x * wi.x + y[i] * wo.y * wi.y +
			z[i] * wo.z * wi.z;
		ret += v.Pow(exponent[i]);
	}
	return ret;
}
FresnelBlend::FresnelBlend(const Spectrum &d, const Spectrum &s,
		MicrofacetDistribution *dist)
	: Rd(d), Rs(s) {
	distribution = dist;
}
FresnelBlend::~FresnelBlend() {
	delete distribution;
}
Spectrum FresnelBlend::f(const Vector &wo, const Vector &wi) const {
	Spectrum diffuse = (28.f/(23.f*M_PI)) * Rd *
		(Spectrum(1.) - Rs) *
		(1 - powf(1 - .5f * fabsf(wi.z), 5)) *
		(1 - powf(1 - .5f * fabsf(wo.z), 5));
	Vector H = (wi + wo).Hat();
	Spectrum specular = distribution->D(H) /
		(8.f * M_PI * fabsf(Dot(wi, H)) * max(fabsf(wi.z), fabsf(wo.z))) *
		SchlickFresnel(Dot(wi, H));
	return diffuse + specular;
}
Spectrum BxDF::sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2, Float *wt) const {
	Float x, y;
	ConcentricSampleDisk(u1, u2, &x, &y);
	// Compute final direction with Malley's method
	Float z = sqrtf(max(0.f, 1.f - x*x - y*y));
	*wo = Vector(x, y, z);
	if (wi.z * wo->z < 0.) *wo = -*wo;
	*wt = Weight(wi, *wo);
	return f(wi, *wo);
}
Float BxDF::Weight(const Vector &wi, const Vector &wo) const {
	return fabsf(wo.Hat().z) * INV_PI;
}
Spectrum BRDFToBTDF::sample_f(const Vector &wi, Vector *wo, Float u1, Float u2,
		Float *wt) const {
	Spectrum f = brdf->sample_f(wi, wo, u1, u2, wt);
	*wo = -*wo;
	return f;
}
Float BRDFToBTDF::Weight(const Vector &wi, const Vector &wo) const {
	return brdf->Weight(wi, -wo);
}
Spectrum Microfacet::sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2, Float *wt) const {
	distribution->sample_f(wi, wo, u1, u2);
	*wt = distribution->Weight(wi, *wo);
	return f(wi, *wo);
}
Float Microfacet::Weight(const Vector &wi,
		const Vector &wo) const {
	return distribution->Weight(wi, wo);
}
void Blinn::sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2) const {
	Float costheta = powf(u1, 1.f / (exponent+1));
	Float sintheta = sqrtf(max(0.f, 1.f - costheta*costheta));
	Float phi = u2 * 2.f * M_PI;
	Vector H = SphericalDirection(sintheta, costheta, phi);
	// Compute incident direction by reflecting about $H$
	*wo = -wi + 2.f * Dot(wi, H) * H;
}
Float Blinn::Weight(const Vector &wi, const Vector &wo) const {
	if (wi.z * wo.z < 0.) return 0;
	Vector H = (wi + wo).Hat();
	Float costheta = fabsf(H.z);
	return ((exponent + 1.f) * powf(costheta, exponent)) /
		(4.f * Dot(wo, H));
}
void Anisotropic::sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2) const {
	Float phi, costheta;
	// Sample from first quadrant and remap to hemisphere
	if (u1 < .25f)
		sampleFirstQuadrant(4.f * u1, u2, &phi, &costheta);
	else if (u1 < .5f) {
		u1 = 4.f * (.5f - u1);
		sampleFirstQuadrant(u1, u2, &phi, &costheta);
		phi = M_PI - phi;
	}
	else if (u1 < .75f) {
		u1 = 4.f * (u1 - .5f);
		sampleFirstQuadrant(u1, u2, &phi, &costheta);
		phi += M_PI;
	}
	else {
		u1 = 4.f * (1.f - u1);
		sampleFirstQuadrant(u1, u2, &phi, &costheta);
		phi = 2.f * M_PI - phi;
	}
	Float sintheta = sqrtf(max(0.f, 1.f - costheta*costheta));
	Vector H = SphericalDirection(sintheta, costheta, phi);
	// Compute incident direction by reflecting about $H$
	*wo = -wi + 2.f * Dot(wi, H) * H;
}
void Anisotropic::sampleFirstQuadrant(Float u1, Float u2,
		Float *phi, Float *costheta) const {
	*phi = atanf(sqrtf((ex+1)*(ey+1)) * tanf(M_PI * u1 * 0.5f));
	Float cosphi = cosf(*phi), sinphi = sinf(*phi);
	*costheta = powf(u2, 1.f/(ex * cosphi * cosphi +
		ey * sinphi * sinphi + 1));
}
Float Anisotropic::Weight(const Vector &wi, const Vector &wo) const {
	if (wi.z * wo.z < 0.) return 0;
	Vector H = (wi + wo).Hat();
	return D(H) / (4.f * Dot(wo, H));
}
Spectrum Lafortune::sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2, Float *wt) const {
	int comp = RandomInt() % (nLobes+1);
	if (comp == nLobes) {
		Float x, y;
		ConcentricSampleDisk(u1, u2, &x, &y);
		// Compute final direction with Malley's method
		Float z = sqrtf(max(0.f, 1.f - x*x - y*y));
		*wo = Vector(x, y, z);
		if (wi.z * wo->z < 0.) *wo = -*wo;
	}
	else {
		// Sample lobe [[comp]] for Lafortune BRDF
		Float xlum = x[comp].y();
		Float ylum = y[comp].y();
		Float zlum = z[comp].y();
		Float costheta = powf(u1, 1.f / (exponent[comp].y() + 1));
		Float sintheta = sqrtf(max(0.f, 1.f - costheta*costheta));
		Float phi = u2 * 2.f * M_PI;
		Vector lobeCenter = Vector(xlum * wi.x, ylum * wi.y, zlum * wi.z).Hat();
		Vector lobeX, lobeY;
		CoordinateSystem(lobeCenter, &lobeX, &lobeY);
		*wo = SphericalDirection(sintheta, costheta, phi, lobeX, lobeY,
			lobeCenter);
	}
	*wt = Weight(wi, *wo);
	return f(wi, *wo);
}
Float Lafortune::Weight(const Vector &wi, const Vector &wo) const {
	Float pdfSum = fabsf(wo.z) / M_PI;
	for (int i = 0; i < nLobes; ++i) {
		Float xlum = x[i].y();
		Float ylum = y[i].y();
		Float zlum = z[i].y();
		Vector lobeCenter =
			Vector(wi.x * xlum, wi.y * ylum, wi.z * zlum).Hat();
		Float e = exponent[i].y();
		pdfSum += (e + 1.f) * powf(max(0.f, Dot(wo,lobeCenter)), e);
	}
	// balance heuristic
	return pdfSum / (1.f + nLobes);
}
Spectrum FresnelBlend::sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2, Float *wt) const {
	if (u1 < .5) {
		u1 = 2.f * u1;
		Float x, y;
		ConcentricSampleDisk(u1, u2, &x, &y);
		// Compute final direction with Malley's method
		Float z = sqrtf(max(0.f, 1.f - x*x - y*y));
		*wo = Vector(x, y, z);
		if (wi.z * wo->z < 0.) *wo = -*wo;
	}
	else {
		u1 = 2.f * (u1 - .5f);
		distribution->sample_f(wi, wo, u1, u2);
	}
	*wt = Weight(wi, *wo);
	return f(wi, *wo);
}
Float FresnelBlend::Weight(const Vector &wi, const Vector &wo) const {
	return .5f * fabsf(wo.z) / M_PI +
		.5f * distribution->Weight(wi, wo);
}
Spectrum BxDF::rho(const Vector &w) const {
	Spectrum r = 0.;
	const int nSamples = 4;
	Float samples[2*nSamples*nSamples];
	StratifiedSample2D(samples, nSamples, nSamples);
	for (int i = 0; i < nSamples*nSamples; ++i) {
		// Estimate one term of $\rho_{dh}$
		Vector wi;
		Float wt;
		Spectrum f = sample_f(w, &wi, samples[2*i], samples[2*i+1], &wt);
		if (wt > 0.) r += f * fabsf(wi.z) / wt;
	}
	return r / (nSamples*nSamples);
}
Spectrum BxDF::rho() const {
	Spectrum r = 0.;
	const int nSamples = 10;
	Float samples[4*nSamples];
	LatinHypercube(samples, nSamples, 4);
	for (int i = 0; i < nSamples; ++i) {
		// Estimate one term of $\rho_{hh}$
		Vector wo, wi;
		wo = UniformSampleHemisphere(samples[4*i], samples[4*i+1]);
		Float weight_o = 2.f * M_PI, weight_i;
		Spectrum f = sample_f(wo, &wi, samples[4*i+2], samples[4*i+3], &weight_i);
		if (weight_i > 0.)
			r += f * fabsf(wi.z * wo.z) / (weight_o * weight_i);
	}
	return r / (M_PI*nSamples);
}
// BSDF Method Definitions
BSDF::BSDF(const DifferentialGeometry &dgS,
		const DifferentialGeometry &dgG) {
	Ng = dgG.Nn;
	nSpecular = 0;
	// Orient shading normal to match geometric normal
	if (Dot(Ng, dgS.Nn) < 0) {
		Ns = -dgS.Nn;
		Ss = -dgS.S;
		Ts = -dgS.T;
	}
	else {
		Ns = dgS.Nn;
		Ss = dgS.S;
		Ts = dgS.T;
	}
}
Spectrum BSDF::rho_r() const {
	Spectrum ret(0.);
	for (u_int i = 0; i < brdfs.size(); ++i)
		ret += brdfs[i]->rho();
	return ret;
}
Spectrum BSDF::rho_r(const Vector &wo) const {
	Spectrum ret(0.);
	for (u_int i = 0; i < brdfs.size(); ++i)
		ret += brdfs[i]->rho(wo);
	return ret;
}
Spectrum BSDF::rho_t() const {
	Spectrum ret(0.);
	for (u_int i = 0; i < btdfs.size(); ++i)
		ret += btdfs[i]->rho();
	return ret;
}
Spectrum BSDF::rho_t(const Vector &wo) const {
	Spectrum ret(0.);
	for (u_int i = 0; i < btdfs.size(); ++i)
		ret += btdfs[i]->rho(wo);
	return ret;
}
// BSDF MC Methods
Spectrum BSDF::sample_f(const Vector &wiW, Vector *woW,
		Float u1, Float u2, Float *wt,
		bool *specularBounce) const {
	Vector wi = WorldToLocal(wiW);
	u_int which = RandomInt() % (brdfs.size() + btdfs.size());
	BxDF *bxdf = NULL;
	if (which < brdfs.size())
		bxdf = brdfs[which];
	else
		bxdf = btdfs[which - brdfs.size()];

	Vector wo;
	Spectrum f = bxdf->sample_f(wi, &wo, u1, u2, wt);
	*specularBounce = bxdf->IsSpecular();
	*woW = LocalToWorld(wo);

	//	*wt = 0.f;
	if (!*specularBounce) {
		if (which < brdfs.size()) {
			for (u_int i = 0; i < brdfs.size(); ++i)
				if (brdfs[i] != bxdf) {
					*wt += brdfs[i]->Weight(wi, wo);
					f += brdfs[i]->f(wi, wo);
				}
		}
		else {
			for (u_int i = 0; i < btdfs.size(); ++i)
				if (btdfs[i] != bxdf) {
					*wt += btdfs[i]->Weight(wi, wo);
					f += btdfs[i]->f(wi, wo);
				}
		}
	}
	*wt /= (brdfs.size() + btdfs.size());

	return f;
}
Float BSDF::Weight(const Vector &wiW, const Vector &woW) const {
	Vector wi = WorldToLocal(wiW), wo = WorldToLocal(woW);
	Float wt = 0;
	u_int i;
	for (i = 0; i < brdfs.size(); ++i)
		wt += brdfs[i]->Weight(wi, wo);
	for (i = 0; i < btdfs.size(); ++i)
		wt += btdfs[i]->Weight(wi, wo);
	return wt / (brdfs.size() + btdfs.size());
}
