// reflection.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef REFLECTION_H
#define REFLECTION_H
#include "lrt.h"
#include "geometry.h"
// BSDF Declarations
class BSDF {
public:
	// BSDF Method Declarations
	BSDF(const DifferentialGeometry &dg, const DifferentialGeometry &);
	inline void Add(BRDF *brdf);
	inline void Add(BTDF *btdf);
	inline ~BSDF();
	int NumComponents() const { return (int)(brdfs.size() + btdfs.size()); }
	inline int NumSpecular() const { return nSpecular; }
	Vector WorldToLocal(const Vector &v) const {
		return Vector(Dot(v, Ss), Dot(v, Ts), Dot(v, Ns)).Hat();
	}
	Vector LocalToWorld(const Vector &v) const {
		return Vector(Ss.x * v.x + Ts.x * v.y + Ns.x * v.z,
			      Ss.y * v.x + Ts.y * v.y + Ns.y * v.z,
			      Ss.z * v.x + Ts.z * v.y + Ns.z * v.z);
	}
	inline Spectrum f(const Vector &woW, const Vector &wiW) const;
	inline Spectrum f_delta(int component, const Vector &wo,
		Vector *wi) const;
	Spectrum rho_r() const;
	Spectrum rho_r(const Vector &wo) const;
	Spectrum rho_t() const;
	Spectrum rho_t(const Vector &wo) const;
	Spectrum sample_f(const Vector &wi, Vector *wo, Float u1, Float u2,
		Float *wt, bool *specularBounce) const;
	Float Weight(const Vector &wi, const Vector &wo) const;
private:
	// BSDF Member Variables
	Normal Ns, Ng;
	Vector Ss, Ts;
	int nSpecular;
	vector<BRDF *> brdfs;
	vector<BTDF *> btdfs;
};
// BxDF Declarations
class BxDF {
public:
	// BxDF Interface Declarations
	virtual ~BxDF() { }
	virtual Spectrum f(const Vector &wo, const Vector &wi) const = 0;
	virtual bool IsSpecular() const { return false; }
	virtual Spectrum f_delta(const Vector &wi, Vector *wo) const {
		return 0.;
	}
	virtual Spectrum rho(const Vector &w) const;
	virtual Spectrum rho() const;
	virtual Spectrum sample_f(const Vector &wi, Vector *wo, Float u1, Float u2,
		Float *wt) const;
	virtual Float Weight(const Vector &wi, const Vector &wo) const;
};
class BRDF : public BxDF {
};
class BTDF : public BxDF {
};
class BRDFToBTDF : public BTDF {
public:
	BRDFToBTDF(BRDF *b) { brdf = b; }
	~BRDFToBTDF() { delete brdf; }
	// BRDFToBTDF Method Declarations
	static Vector otherHemisphere(const Vector &w) {
		return Vector(w.x, w.y, -w.z);
	}
	bool IsSpecular() { return brdf->IsSpecular(); }
	Spectrum rho(const Vector &w) const {
		return brdf->rho(otherHemisphere(w));
	}
	Spectrum rho() const { return brdf->rho(); }
	Spectrum f(const Vector &wo, const Vector &wi) const;
	Spectrum f_delta(const Vector &wo, Vector *wi) const;
	Spectrum sample_f(const Vector &wi, Vector *wo, Float u1, Float u2,
		Float *wt) const;
	Float Weight(const Vector &wi, const Vector &wo) const;
private:
	BRDF *brdf;
};
class Fresnel {
public:
	// Fresnel Interface
	virtual ~Fresnel();
	virtual Spectrum evaluate(Float cosi) const = 0;
};
class FresnelConductor : public Fresnel {
public:
	// FresnelConductor Interface
	Spectrum evaluate(Float cosi) const;
	FresnelConductor(const Spectrum &e, const Spectrum &kk)
		: eta(e), k(kk) {
	}
private:
	// FresnelConductor Private Data
	Spectrum eta, k;
};
class FresnelDielectric : public Fresnel {
public:
	// FresnelDielectric Interface
	Spectrum evaluate(Float cosi) const;
	FresnelDielectric(Float ei, Float et) {
		eta_i = ei;
		eta_t = et;
	}
private:
	// FresnelDielectric Private Data
	Float eta_i, eta_t;
};
class FresnelNoOp : public Fresnel {
public:
	Spectrum evaluate(Float cosi) const;
};
class SpecularReflection : public BRDF {
public:
	// SpecularReflection Methods
	~SpecularReflection();
	SpecularReflection(const Spectrum &r, Fresnel *f)
		: R(r), fresnel(f) {
	}
	Spectrum f(const Vector &, const Vector &) const {
		return Spectrum(0.);
	}
	bool IsSpecular() const { return true; }
	Spectrum f_delta(const Vector &wi, Vector *wo) const;
	Spectrum sample_f(const Vector &wi, Vector *wo, Float u1, Float u2,
			Float *wt) const {
		*wt = 1.;
		return f_delta(wi, wo);
	}
	Float Weight(const Vector &wi, const Vector &wo) const {
		return 0.;
	}
private:
	// SpecularReflection Private Data
	Spectrum R;
	Fresnel *fresnel;
};
class SpecularTransmission : public BTDF {
public:
	// SpecularTransmission Methods
	SpecularTransmission(const Spectrum &t, Float ei, Float et)
			: fresnel(ei, et) {
		T = t;
		etai = ei;
		etat = et;
	}
	Spectrum f(const Vector &, const Vector &) const {
		return Spectrum(0.);
	}
	bool IsSpecular() const { return true; }
	Spectrum f_delta(const Vector &wi, Vector *wo) const;
	Spectrum sample_f(const Vector &wi, Vector *wo, Float u1, Float u2,
			Float *wt) const {
		*wt = 1.;
		return f_delta(wi, wo);
	}
	Float Weight(const Vector &wi, const Vector &wo) const {
		return 0.;
	}
private:
	// SpecularTransmission Private Data
	Spectrum T;
	Float etai, etat;
	FresnelDielectric fresnel;
};
class Lambertian : public BRDF {
public:
	// Lambertian Methods
	Lambertian(const Spectrum &reflectance)
		: R(reflectance), RoverPI(reflectance * INV_PI) {
	}
	Spectrum f(const Vector &wo, const Vector &wi) const;
	Spectrum rho(const Vector &w) const { return R; }
	Spectrum rho() const { return R; }
private:
	// Lambertian Private Data
	Spectrum R, RoverPI;
};
class OrenNayar : public BRDF {
public:
	// OrenNayar Methods
	Spectrum f(const Vector &wo, const Vector &wi) const;
	OrenNayar(const Spectrum &reflectance, Float sig)
		: R(reflectance) {
		Float sigma2 = Radians(sig*sig);
		A = 1.f - (sigma2 / (2.f * (sigma2 + 0.33f)));
		B = 0.45f * sigma2 / (sigma2 + 0.09f);
	}
private:
	// OrenNayar Private Data
	Spectrum R;
	Float A, B;
};
class MicrofacetDistribution {
public:
	// MicrofacetDistribution Interface
	virtual ~MicrofacetDistribution() { }
	virtual Float D(const Vector &wh) const = 0;
	virtual void sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2) const = 0;
	virtual Float Weight(const Vector &wi, const Vector &wo) const = 0;
};

class Microfacet : public BRDF {
public:
	// Microfacet Methods
	Microfacet(const Spectrum &reflectance, Fresnel *f,
		MicrofacetDistribution *d);
	~Microfacet();
	Spectrum f(const Vector &wo, const Vector &wi) const;
	Float G(const Vector &wo, const Vector &wi,
			const Vector &wh) const {
		Float NdotH = fabsf(wh.z);
		Float NdotWO = fabsf(wo.z), NdotWI = fabsf(wi.z);
		Float WOdotH = fabsf(Dot(wo, wh));
		return min(1.f, min((2.f * NdotH * NdotWO / WOdotH),
					(2.f * NdotH * NdotWI / WOdotH)));
	}
	Spectrum sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2, Float *wt) const;
	Float Weight(const Vector &wi, const Vector &wo) const;
private:
	// Microfacet Private Data
	Spectrum R;
	MicrofacetDistribution *distribution;
	Fresnel *fresnel;
};
class Blinn : public MicrofacetDistribution {
public:
	Blinn(Float e) { exponent = e; }
	// Blinn Method Declarations
	Float D(const Vector &wh) const {
		Float costhetah = fabsf(wh.z);
		return (exponent+1) * powf(max(0.f, costhetah), exponent);
	}
	virtual void sample_f(const Vector &wi, Vector *wo, Float u1, Float u2) const;
	virtual Float Weight(const Vector &wi, const Vector &wo) const;
private:
	Float exponent;
};
class Anisotropic : public MicrofacetDistribution {
public:
	Anisotropic(Float x, Float y) { ex = x; ey = y; }
	// Anisotropic Method Declarations
	Float D(const Vector &wh) const {
		Float costhetah = fabsf(wh.z);
		Float e = (ex * wh.x * wh.x + ey * wh.y * wh.y) /
			(1.f - costhetah * costhetah);
		return sqrtf((ex+1)*(ey+1)) * powf(costhetah, e);
	}
	void sample_f(const Vector &wi, Vector *wo, Float u1, Float u2) const;
	Float Weight(const Vector &wi, const Vector &wo) const;
	void sampleFirstQuadrant(Float u1, Float u2, Float *phi, Float *costheta) const;
private:
	Float ex, ey;
};
class Lafortune : public BRDF {
public:
	// Lafortune Methods
	Lafortune(const Spectrum &r, int nl, const Spectrum *x,
		const Spectrum *y, const Spectrum *z, const Spectrum *e);
	Spectrum f(const Vector &wo, const Vector &wi) const;
	Spectrum sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2, Float *wt) const;
	Float Weight(const Vector &wi, const Vector &wo) const;
private:
	// Lafortune Private Data
	Spectrum R;
	#define MAX_LOBES 3
	Spectrum x[MAX_LOBES], y[MAX_LOBES], z[MAX_LOBES];
	Spectrum exponent[MAX_LOBES];
	int nLobes;
};
class FresnelBlend : public BRDF {
public:
	// FresnelBlend Methods
	FresnelBlend(const Spectrum &Rd, const Spectrum &Rs, MicrofacetDistribution *dist);
	~FresnelBlend();
	Spectrum f(const Vector &wo, const Vector &wi) const;
	Spectrum SchlickFresnel(Float costheta) const {
		return Rs + powf(1 - costheta, 5.f) * (Spectrum(1.) - Rs);
	}
	Spectrum sample_f(const Vector &wi, Vector *wo,
		Float u1, Float u2, Float *wt) const;
	Float Weight(const Vector &wi, const Vector &wo) const;
private:
	// FresnelBlend Private Data
	Spectrum Rd, Rs;
	MicrofacetDistribution *distribution;
};
// BSDF Inline Methods
inline void BSDF::Add(BRDF *b) {
	brdfs.push_back(b);
	if (b->IsSpecular()) ++nSpecular;
}
inline void BSDF::Add(BTDF *b) {
	btdfs.push_back(b);
	if (b->IsSpecular()) ++nSpecular;
}
inline BSDF::~BSDF() {
	u_int i;
	for (i = 0; i < brdfs.size(); ++i)
		delete brdfs[i];
	for (i = 0; i < btdfs.size(); ++i)
		delete btdfs[i];
}
inline Spectrum BSDF::f(const Vector &woW,
		const Vector &wiW) const {
	Vector wi = WorldToLocal(wiW), wo = WorldToLocal(woW);
	Spectrum f = 0.;
	if (Dot(wiW, Ng) * Dot(woW, Ng) > 0)
		for (u_int i = 0; i < brdfs.size(); ++i)
			f += brdfs[i]->f(wo, wi);
	else
		for (u_int i = 0; i < btdfs.size(); ++i)
			f += btdfs[i]->f(wo, wi);
	return f;
}
inline Spectrum BSDF::f_delta(int component, const Vector &w,
		Vector *wi) const {
	BxDF *spec = NULL;
	// Find the [[component]]th specular component
	for (u_int i = 0; i < brdfs.size(); ++i) {
		if (brdfs[i]->IsSpecular()) {
			if (component-- == 0) {
				spec = brdfs[i];
				break;
			}
		}
	}
	if (spec == NULL) {
		// Look for specular reflection in BTDFs
		for (u_int i = 0; i < btdfs.size(); ++i) {
			if (btdfs[i]->IsSpecular()) {
				if (component-- == 0) {
					spec = btdfs[i];
					break;
				}
			}
		}
	}
	if (spec == NULL) return 0.;
	Vector wo = WorldToLocal(w);
	Spectrum f = spec->f_delta(wo, wi);
	*wi = LocalToWorld(*wi);
	return f;
}
#endif // REFLECTION_H
