// sampling.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "sampling.h"
#include "film.h"
#include "transport.h"
#include "volume.h"
// Sampler Method Definitions
Sampler::~Sampler() {
	delete filter;
}
Sampler::Sampler(int xres, int yres, int xsamp, int ysamp,
		const Extent2D &Crop, Filter *f) {
	xResolution = xres;
	yResolution = yres;
	xPixelSamples = xsamp;
	yPixelSamples = ysamp;
	filter = f;
	// Initialize pixel extents from crop window
	int RasterCropLeft   = Ceil2Int(xResolution * Crop.x0);
	int RasterCropRight  = Ceil2Int(xResolution * Crop.x1);
	int RasterCropTop    = Ceil2Int(yResolution * Crop.y0);
	int RasterCropBottom = Ceil2Int(yResolution * Crop.y1);
	xPixelStart = Floor2Int(RasterCropLeft   - filter->halfXWidth);
	xPixelEnd   = Ceil2Int (RasterCropRight  + filter->halfXWidth) + 1;
	yPixelStart = Floor2Int(RasterCropTop    - filter->halfYWidth);
	yPixelEnd   = Ceil2Int (RasterCropBottom + filter->halfYWidth) + 1;
}
void Sampler::AddSample(Film *film, const Point &Praster,
		const Spectrum &radiance, Float alpha) {
	// Check sample's raster depth against far clipping plane
	if (Praster.z > 1.f) return;
	// Compute sample's raster extent
	int x0 = Ceil2Int (Praster.x - filter->halfXWidth);
	int x1 = Floor2Int(Praster.x + filter->halfXWidth);
	int y0 = Ceil2Int (Praster.y - filter->halfYWidth);
	int y1 = Floor2Int(Praster.y + filter->halfYWidth);
	// Loop over filter support and add sample to pixel arrays
	for (int y = y0; y <= y1; ++y)
		for (int x = x0; x <= x1; ++x) {
			// Evaluate filter value at $(x,y)$
			Float fx = x - Praster.x;
			Float fy = y - Praster.y;
			Float filterWt = filter->Evaluate(fx, fy);
			// Update pixel values with filtered sample
			if (filterWt > 0.)
				film->UpdatePixel(x, y, radiance, alpha, Praster.z,
					filterWt);
		}
}
// Sample Method Definitions
Sample::Sample(SurfaceIntegrator *surf,
		VolumeIntegrator *vol, const Scene *scene) {
	surf->RequestSamples(this, scene);
	vol->RequestSamples(this, scene);
	// Allocate storage for sample pointers
	int nPtrs = n1D.size() + n2D.size();
	if (!nPtrs)
		oneD = twoD = NULL;
	else {
		oneD = (Float **)AllocL1CacheAligned(nPtrs * sizeof(Float *));
		twoD = oneD + n1D.size();
	}
	// Compute total number of sample values needed
	int totSamples = 0;
	for (u_int i = 0; i < n1D.size(); ++i)
		totSamples += n1D[i];
	for (u_int i = 0; i < n2D.size(); ++i)
		totSamples += 2 * n2D[i];
	// Allocate storage for sample values
	if (oneD) {
		Float *mem = (Float *)AllocL1CacheAligned(totSamples *
			sizeof(Float));
		for (u_int i = 0; i < n1D.size(); ++i) {
			oneD[i] = mem;
			mem += n1D[i];
		}
		for (u_int i = 0; i < n2D.size(); ++i) {
			twoD[i] = mem;
			mem += 2 * n2D[i];
		}
	}
}
// Sampling Function Definitions
void StratifiedSample1D(Float *samp, int nSamples, bool jitter) {
	Float invTot = 1.f / nSamples;
	for (int i = 0;  i < nSamples; ++i) {
		Float delta = jitter ? RandomFloat() : 0.5f;
		*samp++ = (i + delta) * invTot;
	}
}
void StratifiedSample2D(Float *samp, int nx, int ny, bool jitter) {
	Float dx = 1.f / nx, dy = 1.f / ny;
	for (int y = 0; y < ny; ++y)
		for (int x = 0; x < nx; ++x) {
			Float jx = jitter ? RandomFloat() : 0.5f;
			Float jy = jitter ? RandomFloat() : 0.5f;
			*samp++ = (x + jx) * dx;
			*samp++ = (y + jy) * dy;
		}
}
void Shuffle(Float *samp, int count, int dims) {
	for (int i = 0; i < count; ++i) {
		int other = RandomInt() % count;
		for (int j = 0; i < dims; ++i)
			swap(samp[dims*i + j], samp[dims*other + j]);
	}
}
void LatinHypercube(Float *samples, int nSamples, int nDim) {
	int i, j;
	// Generate samples along diagonal
	Float delta = 1.f / nSamples;
	for (i = 0; i < nSamples; ++i)
		for (j = 0; j < nDim; ++j)
			samples[nDim * i + j] = (i + RandomFloat()) * delta;
	// Permute in each dimension
	for (i = 0; i < nDim; ++i) {
		for (j = 0; j < nSamples; ++j) {
			int other = RandomInt() % nSamples;
			swap(samples[nDim * j + i], samples[nDim * other + i]);
		}
	}
}
Float VanDerCorput(u_int n, u_int scramble) {
	n = (n << 16) | (n >> 16);
	n = ((n & 0x00ff00ff) << 8) | ((n & 0xff00ff00) >> 8);
	n = ((n & 0x0f0f0f0f) << 4) | ((n & 0xf0f0f0f0) >> 4);
	n = ((n & 0x33333333) << 2) | ((n & 0xcccccccc) >> 2);
	n = ((n & 0x55555555) << 1) | ((n & 0xaaaaaaaa) >> 1);
	n ^= scramble;
	return (Float)n / (Float)0x100000000LL;
}
Float Sobol2(u_int n, u_int scramble) {
	for (u_int v = 1 << 31; n != 0; n >>= 1, v ^= v >> 1)
		if (n & 0x1) scramble ^= v;
	return (Float)scramble / (Float)0x100000000LL;
}
Float LarcherPillichshammer2(u_int n, u_int scramble) {
	for (u_int v = 1 << 31; n != 0; n >>= 1, v |= v >> 1)
		if (n & 0x1) scramble ^= v;
	return (Float)scramble / (Float)0x100000000LL;
}
