// sampling.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef SAMPLING_H
#define SAMPLING_H
#include "lrt.h"
// Sampling Constants
#define SQRT_SAMPLE_TABLE_SIZE 64
#define SAMPLE_TABLE_SIZE (SQRT_SAMPLE_TABLE_SIZE * \
                           SQRT_SAMPLE_TABLE_SIZE)
// Sampling Declarations
class Sampler {
public:
	// Sampler Interface
	virtual ~Sampler();
	Sampler(int xres, int yres, int xSamp, int ySamp,
		const Extent2D &crop, Filter *f);
	virtual bool GetNextSample(Sample *sample) = 0;
	int TotalSamples() const {
		return xPixelSamples * yPixelSamples *
			(xPixelEnd - xPixelStart) * (yPixelEnd - yPixelStart);
	}
	void AddSample(Film *film, const Point &PRaster, const Spectrum &radiance, Float alpha);
	// Sampler Options
	int xResolution, yResolution;
	int xPixelSamples, yPixelSamples;
	Filter *filter;
	int xPixelStart, xPixelEnd, yPixelStart, yPixelEnd;
};
struct Sample {
public:
	// Sample Method Declarations
	Sample(SurfaceIntegrator *surf, VolumeIntegrator *vol,
		const Scene *scene);
	u_int Add1D(u_int num) {
		n1D.push_back(num);
		return n1D.size()-1;
	}
	u_int Add2D(u_int num) {
		n2D.push_back(num);
		return n2D.size()-1;
	}
	~Sample() {
		if (oneD != NULL) {
			FreeCacheAligned(oneD[0]);
			FreeCacheAligned(oneD);
		}
	}
	// Camera sample data
	Float imageX, imageY;
	Float lensX, lensY;
	Float time;
	// Integrator sample data
	vector<u_int> n1D, n2D;
	Float **oneD, **twoD;
};
extern void StratifiedSample1D(Float *samples, int nsamples, bool jitter = true);
extern void StratifiedSample2D(Float *samples, int nx, int ny, bool jitter = true);
extern void Shuffle(Float *samp, int count, int dims);
extern void LatinHypercube(Float *samples, int nSamples, int nDim);
inline Float RadicalInverse(int n, int base) {
	Float val = 0;
	Float invBase = 1.f / base, invBi = invBase;
	while (n > 0) {
		// Compute next digit of radical inverse
		int ai = (n % base);
		val += ai * invBi;
		n /= base;
		invBi *= invBase;
	}
	return val;
}
inline Float FoldedRadicalInverse(int n, int base) {
	Float val = 0;
	Float invBase = 1.f/base, invBi = invBase;
	int modOffset = 0;
	while (val + base * invBi != val) {
		// Compute next digit of folded radical inverse
		int digit = ((n+modOffset) % base);
		val += digit * invBi;
		n /= base;
		invBi *= invBase;
		++modOffset;
	}
	return val;
}
extern Float VanDerCorput(u_int n, u_int scramble = 0);
extern Float Sobol2(u_int n, u_int scramble = 0);
extern Float LarcherPillichshammer2(u_int n, u_int scramble = 0);
class Filter {
public:
	// Filter Interface
	virtual ~Filter() { }
	Filter(Float xw, Float yw)
		: xWidth(xw), yWidth(yw), halfXWidth(0.5f*xw),
			halfYWidth(0.5f*yw), invHalfXWidth(2.f/xw),
			invHalfYWidth(2.f/yw) {
	}
	virtual Float Evaluate(Float x, Float y) const = 0;
	// Filter Data
	const Float xWidth, yWidth;
	const Float halfXWidth, halfYWidth;
	const Float invHalfXWidth, invHalfYWidth;
};
#endif // SAMPLING_H
