// scene.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "scene.h"
#include "camera.h"
#include "film.h"
#include "sampling.h"
#include "dynload.h"
#include "transport.h"
#include "volume.h"
#include "materials.h"

#include "translucentmap.h"

int debugflag=0;
bool isCreateBlob;
void createBlob(Scene* scene);

// Scene Methods
Scene::~Scene() {
	delete camera;
	delete sampler;
	delete surfaceIntegrator;
	delete volumeIntegrator;
	delete prims;
	delete displayInfo;
	for (u_int i = 0; i < lights.size(); ++i)
		delete lights[i];
	for (u_int i = 0; i < volumeRegions.size(); ++i)
		delete volumeRegions[i];
}
Scene::Scene(Camera *cam, SurfaceIntegrator *si, VolumeIntegrator *vi,
		Sampler *s, Primitive *accel, const vector<Light *> &lts,
		const vector<VolumeRegion *> &vrs, DisplayInfo *di) {
	lights = lts;
	prims = accel;
	camera = cam;
	sampler = s;
	surfaceIntegrator = si;
	volumeIntegrator = vi;
	volumeRegions = vrs;
	displayInfo = di;
	if (lts.size() == 0)
		Warning("No light sources defined in scene; "
			"possibly rendering a black image.");
	bound = prims->WorldBound();

}
void Scene::Render() {
  //create translucent map
  createTranslucentmap(this);
	// Prepare to render image
	Sample *sample = new Sample(surfaceIntegrator, volumeIntegrator, this);
	surfaceIntegrator->Preprocess(this);
	volumeIntegrator->Preprocess(this);
	ProgressReporter progress(sampler->TotalSamples(),
		"Rendering");

	// BLOBS
  if(isCreateBlob == true) {
		cout<<"Blobs being created!"<<endl;
		createBlob(this);
		return;
	}

	// GLOWS
	// Precalculate each light for how much glow they create
	// Precalculate which light actually glow (visible)
	Ray raytmp;
	int count=0;
	for (u_int i = 0; i<lights.size(); ++i) {
		Light *light = lights[i];
		if(light->doGlow) {
			raytmp.O = Point(0,0,0);
			raytmp.D = Vector(0,0,1);
			raytmp.mint = 0.;
			raytmp.maxt = 1.;
			camera->CameraToWorld(raytmp, &raytmp);
			raytmp.D = (light->copyPos-raytmp.O).Hat();
			raytmp.maxt = (light->copyPos-raytmp.O).Length();
			if(IntersectP(raytmp))
				light->doGlow=false;
			else
				count++;
		}
	}
	cout << count << " glowing lights" << endl;

#ifdef ADDSKY
	string tex ("projet/sky3.tif");
	ParamSet geomParams;
	ParamSet surfaceParams;
	SURF_TEX_F(Kd, 1.);
	Ksky = new ScaleTexture<Float, Spectrum>(Kd, new ImageTexture<Spectrum>(new IdentityMapping2D(), tex));
#endif
	
	while (sampler->GetNextSample(sample)) {
		// Compute result for image sample
		// Report rendering progress
		static StatsCounter cameraRaysTraced("Camera",
			"Camera Rays Traced");
		++cameraRaysTraced;
		progress.Update(1);
		// Compute camera ray
		RayDifferential ray;
		camera->GenerateRay(sample, ray);
		// Generate ray differentials for camera ray
		++sample->imageX;
		camera->GenerateRay(sample, ray.rx);
		--sample->imageX;
		++sample->imageY;
		camera->GenerateRay(sample, ray.ry);
		ray.hasDifferentials = true;
		// Evaluate radiance along ray
		Float alpha;

		debugflag=0;
		Float zx=117; Float zy=87;
		//				Float zx=112; Float zy=88;
				
		if(sample->imageX>=zx-0.5 && sample->imageY>=zy-0.5
		  && sample->imageX<=zx+0.5 && sample->imageY<=zy+0.5) {
									debugflag=1;
						cout << "\n" << sample->imageX <<", "<< sample->imageY << "\n"<<endl;
		}
		
		zx=122; zy=77;
		//				Float zx=112; Float zy=88;
		if(sample->imageX>=zx-0.5 && sample->imageY>=zy-0.5
		  && sample->imageX<=zx+0.5 && sample->imageY<=zy+0.5) {
						debugflag=1;
						cout << "\n" << sample->imageX <<", "<< sample->imageY << "\n"<<endl;
		}


		
		zx=180; zy=222;
		//				Float zx=112; Float zy=88;
		if(sample->imageX>=zx-0.5 && sample->imageY>=zy-0.5
		  && sample->imageX<=zx+0.5 && sample->imageY<=zy+0.5) {
									debugflag=1;
						cout << "\n" << sample->imageX <<", "<< sample->imageY << "\n"<<endl;
		}

		//		cout << "\n" << sample->imageX <<", "<< sample->imageY << " -- ";
		Spectrum Ls = L(ray, sample, &alpha);
		if(debugflag==1)
			cout << " ->  " << Ls*256 << "\n";

		if(!(Ls.y()>0 && Ls.y() <1000000000))
			Ls=Spectrum(0.);

		Spectrum Lglow(0.0);
		for (u_int i = 0; i<lights.size(); ++i) {
			Light *light = lights[i];
			if(light->doGlow)
				Lglow += light->Glow(ray);
		}
		alpha=1;
		Ls+=Lglow;

		// Add sample to image
		Float screenz = camera->ScreenDepth(ray(ray.maxt));
		Point Praster(sample->imageX, sample->imageY, screenz);
		sampler->AddSample(camera->film, Praster, Ls, alpha);
	}
	delete sample;
	fprintf(stderr, "\n");
	camera->film->WriteDisplayImage(*displayInfo);
}
Spectrum Scene::L(const RayDifferential &ray,
		const Sample *sample, Float *alpha) const {
	int oldflag=debugflag;
	Spectrum Ls = surfaceIntegrator->L(this, ray,
		sample, alpha);
	debugflag=0;
 	Spectrum T = volumeIntegrator->Transmittance(this, ray,
		sample, alpha);
	Spectrum Lv = volumeIntegrator->L(this, ray,
		sample, alpha);
	debugflag=oldflag;
	return T * Ls + Lv;
}
Spectrum Scene::Transmittance(const Ray &ray) const {
	return volumeIntegrator->Transmittance(this, ray,
		NULL, NULL);
}


#include "shapes/blob.h"
#include <vector>

void createBlob(Scene* scene)
{
  cout<<"In create Blob"<<endl;
  
	//  float Thickness = 1;
	//  float MinHeight = 0;
	//  float MaxHeight = 1.1;
	//  float threshold = 0.6;
	char filename[512];
  
  std::vector<Reference<BLOB_ELEMENT> > elementsVec;
 
	/*
	// Snow bank
  Vector Direction(0,-1,0.9);
  Direction = Direction.Hat();
  int Particles = 10000;
  float size = 15;
	float mult=.33;
	float translatez=-5;
  Point MinPt(-2400,-40,0);
  Point MaxPt(200,40,850);
	sprintf(filename, "projet/snowcylinder2.blob");
	*/
	// Tree 1
  Vector Direction(0,-1,0);
  Direction = Direction.Hat();
  int Particles = 3000;
  float size = 4;
	float mult= .7;
	float translatez=0;
  Point MinPt(-240-460,0-10,-250+460);
  Point MaxPt(240-460,600+10,200+460);
	sprintf(filename, "projet/snowtree1.blob");
	
  Intersection isect;
  Ray ray;
  ray.D = Direction;
	
  int i=0;
  while(i<Particles/3)  {
		ray.O.x = MinPt.x + RandomFloat(0.0f,1.0f)*(MaxPt.x-MinPt.x);
		ray.O.y = MaxPt.y + 100;//(MinHeight+RandomFloat(0.0f,1.0f)*(MaxHeight-MinHeight))*Max.y;
		ray.O.z = MinPt.z + RandomFloat(0.0f,1.0f)*(MaxPt.z-MinPt.z);
		
		ray.mint = 0;
		ray.maxt = 100000;
		
		Normal norm;
		
		if(scene->Intersect(ray, &isect)){
			norm = isect.dgGeom.Nn;
			if(norm.Length()>0) {
				//				float dot   = min(1.0f, Dot(norm.Hat(),-Direction));
				//				float Angle = pow((M_PI-fabs(acos(dot)))/M_PI,2);
				
				//				Point FlakeSize =	Angle*Point(1+RandomFloat(0,1)*1.3,(Angle*0.5+RandomFloat(0,1))*Thickness,1+RandomFloat(0,1)*1.3);
				
				/*	      cout<<FlakeSize.x<<"  ";
									cout<<FlakeSize.y<<"  ";
									cout<<FlakeSize.z<<endl;
				*/
				
				BLOB_ELEMENT* element = new BLOB_ELEMENT(Point(isect.dgGeom.P.x,isect.dgGeom.P.y,isect.dgGeom.P.z+translatez) ,2*size ,1);
				//	      BLOB_ELEMENT* element = new BLOB_ELEMENT(Point(isect.dgGeom.P.x,isect.dgGeom.P.y,isect.dgGeom.P.z) ,2*size ,1);
				//				element->scaleTrans = FlakeSize;
				elementsVec.push_back(element);
				i++;
			}
		}
	}  
	cout <<endl;
	while(i<Particles) {
		ray.O.x = MinPt.x + RandomFloat(0.0f,1.0f)*(MaxPt.x-MinPt.x);
		ray.O.y = MaxPt.y + 100;//(MinHeight+RandomFloat(0.0f,1.0f)*(MaxHeight-MinHeight))*Max.y;
		ray.O.z = MinPt.z + RandomFloat(0.0f,1.0f)*(MaxPt.z-MinPt.z);
		
		ray.mint = 0;
		ray.maxt = 100000;
		
		Normal norm;
		
		if(scene->Intersect(ray, &isect))
			{
				norm = isect.dgGeom.Nn;
				if(norm.Length()>0)
					{
						//						float dot = min(1.0f, Dot(norm.Hat(),-Direction));
						//						float Angle = pow((M_PI-fabs(acos(dot)))/M_PI,2);
						
						//						Point FlakeSize =	Angle*Point(1+RandomFloat(0,1)*1.3,(Angle*0.5+RandomFloat(0,1))*Thickness,1+RandomFloat(0,1)*1.3);
						
						BLOB_ELEMENT* element = new BLOB_ELEMENT(Point(isect.dgGeom.P.x,isect.dgGeom.P.y,isect.dgGeom.P.z+translatez) ,mult*2*size ,1);
						//element->scaleTrans = 0.25*FlakeSize;
						elementsVec.push_back(element);
						i++;
					}
			}
	}  
	FILE* file = fopen(filename,"w");
	fprintf(file,"%d\n",elementsVec.size());
	
	for(unsigned int i=0; i<elementsVec.size(); i++)
		{
			fprintf(file,"%f %f %f %f %f %f %f %f %f %f\n",elementsVec[i]->center.x,elementsVec[i]->center.y,elementsVec[i]->center.z,elementsVec[i]->rad,elementsVec[i]->c[0],elementsVec[i]->c[1],elementsVec[i]->c[2],elementsVec[i]->scaleTrans.x,elementsVec[i]->scaleTrans.y,elementsVec[i]->scaleTrans.z);
		}
	fclose(file);
}

