// scene.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "scene.h"
#include "camera.h"
#include "film.h"
#include "sampling.h"
#include "dynload.h"
#include "transport.h"
#include "volume.h"

bool isCreateBlob = true;

void createBlob(Scene* scene);

// Scene Methods
Scene::~Scene() {
  delete camera;
  delete sampler;
  delete surfaceIntegrator;
  delete volumeIntegrator;
  delete prims;
  delete displayInfo;
  for (u_int i = 0; i < lights.size(); ++i)
    delete lights[i];
  for (u_int i = 0; i < volumeRegions.size(); ++i)
    delete volumeRegions[i];
}
Scene::Scene(Camera *cam, SurfaceIntegrator *si, VolumeIntegrator *vi,
	     Sampler *s, Primitive *accel, const vector<Light *> &lts,
	     const vector<VolumeRegion *> &vrs, DisplayInfo *di) {
  lights = lts;
  prims = accel;
  camera = cam;
  sampler = s;
  surfaceIntegrator = si;
  volumeIntegrator = vi;
  volumeRegions = vrs;
  displayInfo = di;
  if (lts.size() == 0)
    Warning("No light sources defined in scene; "
	    "possibly rendering a black image.");
  bound = prims->WorldBound();

}
void Scene::Render() {
  
  // Prepare to render image
  Sample *sample = new Sample(surfaceIntegrator, volumeIntegrator, this);
  surfaceIntegrator->Preprocess(this);
  volumeIntegrator->Preprocess(this);
  ProgressReporter progress(sampler->TotalSamples(),
			    "Rendering");
  
  if(isCreateBlob == true)
    {
      cout<<"Blob s being created!"<<endl;
      createBlob(this);
      return;
    }

  while (sampler->GetNextSample(sample)) {
    // Compute result for image sample
    // Report rendering progress
    static StatsCounter cameraRaysTraced("Camera",
					 "Camera Rays Traced");
    ++cameraRaysTraced;
    progress.Update(1);
    // Compute camera ray
    RayDifferential ray;
    camera->GenerateRay(sample, ray);
    // Generate ray differentials for camera ray
    ++sample->imageX;
    camera->GenerateRay(sample, ray.rx);
    --sample->imageX;
    ++sample->imageY;
    camera->GenerateRay(sample, ray.ry);
    ray.hasDifferentials = true;
    // Evaluate radiance along ray
    Float alpha;
    Spectrum Ls = L(ray, sample, &alpha);
    // Add sample to image
    Float screenz = camera->ScreenDepth(ray(ray.maxt));
    Point Praster(sample->imageX, sample->imageY, screenz);
    sampler->AddSample(camera->film, Praster, Ls, alpha);
  }
  
  delete sample;
  fprintf(stderr, "\n");
  camera->film->WriteDisplayImage(*displayInfo);
}
Spectrum Scene::L(const RayDifferential &ray,
		  const Sample *sample, Float *alpha) const {
  Spectrum Ls = surfaceIntegrator->L(this, ray,
				     sample, alpha);
  Spectrum T = volumeIntegrator->Transmittance(this, ray,
					       sample, alpha);
  Spectrum Lv = volumeIntegrator->L(this, ray,
				    sample, alpha);
  return T * Ls + Lv;
}
Spectrum Scene::Transmittance(const Ray &ray) const {
  return volumeIntegrator->Transmittance(this, ray,
					 NULL, NULL);
}

#include "shapes/blob.h"
#include <vector>

void createBlob(Scene* scene)
{
  cout<<"In create Blob"<<endl;
  
  int Particles = 2000;
  float size = 0.08;
  float Thickness = 1;
  float MinHeight = 0;
  float MaxHeight = 1.1;
  Vector Direction(0,-1,0);
  Direction = Direction.Hat();
  float threshold = 0.6;
  
  std::vector<Reference<BLOB_ELEMENT> > elementsVec;
 
  Point MinPt(-1,-1,-1);
  Point MaxPt(1,1,1);

  Intersection isect;
  Ray ray;
  ray.D = Direction;

  int i=0;
  while(i<Particles/3)
    {
      ray.O.x = MinPt.x + RandomFloat(0.0f,1.0f)*(MaxPt.x-MinPt.x);
      ray.O.y = MaxPt.y + 10;//(MinHeight+RandomFloat(0.0f,1.0f)*(MaxHeight-MinHeight))*Max.y;
      ray.O.z = 5 + MinPt.z + RandomFloat(0.0f,1.0f)*(MaxPt.z-MinPt.z);
  //translation ****???
      
      ray.mint = 0;
      ray.maxt = 100000;

      Normal norm;
      
      if(scene->Intersect(ray, &isect))
	{
	  norm = isect.dgGeom.Nn;
	  if(norm.Length()>0)
	    {
	      float dot   = min(1.0f, Dot(norm.Hat(),-Direction));
	      float Angle = pow((M_PI-fabs(acos(dot)))/M_PI,2);
	      
	      Point FlakeSize =	Angle*Point(1+RandomFloat(0,1)*1.3,(Angle*0.5+RandomFloat(0,1))*Thickness,1+RandomFloat(0,1)*1.3);

	      /*	      cout<<FlakeSize.x<<"  ";
			      cout<<FlakeSize.y<<"  ";
			      cout<<FlakeSize.z<<endl;
	      */

	      BLOB_ELEMENT* element = new BLOB_ELEMENT(Point(isect.dgGeom.P.x,isect.dgGeom.P.y,isect.dgGeom.P.z-5.0) ,2*size ,1);
	      //element->scaleTrans = FlakeSize;
	      elementsVec.push_back(element);
	      i++;
	    }
	}
    }  

  while(i<Particles)
    {
      ray.O.x = MinPt.x + RandomFloat(0.0f,1.0f)*(MaxPt.x-MinPt.x);
      ray.O.y = 10;//(MinHeight+RandomFloat(0.0f,1.0f)*(MaxHeight-MinHeight))*Max.y;
      ray.O.z = 5 + MinPt.z + RandomFloat(0.0f,1.0f)*(MaxPt.z-MinPt.z);

      ray.mint = 0;
      ray.maxt = 100000;
      
      Normal norm;
      
      if(scene->Intersect(ray, &isect))
	{
	  norm = isect.dgGeom.Nn;
	  if(norm.Length()>0)
	    {
	      float dot = min(1.0f, Dot(norm.Hat(),-Direction));
	      float Angle = pow((M_PI-fabs(acos(dot)))/M_PI,2);

	      Point FlakeSize =	Angle*Point(1+RandomFloat(0,1)*1.3,(Angle*0.5+RandomFloat(0,1))*Thickness,1+RandomFloat(0,1)*1.3);
	      
	      BLOB_ELEMENT* element = new BLOB_ELEMENT(Point(isect.dgGeom.P.x,isect.dgGeom.P.y,isect.dgGeom.P.z-5.0) ,/*.25**/2*size ,1);
	      //element->scaleTrans = 0.25*FlakeSize;
	      elementsVec.push_back(element);
	      i++;
	    }
	}
    }  

  FILE* file = fopen("MakeSnow/data.blob","w");
  fprintf(file,"%d\n",elementsVec.size());
  
  for(unsigned int i=0; i<elementsVec.size(); i++)
    {
      fprintf(file,"%f %f %f %f %f %f %f %f %f %f\n",elementsVec[i]->center.x,elementsVec[i]->center.y,elementsVec[i]->center.z,elementsVec[i]->rad,elementsVec[i]->c[0],elementsVec[i]->c[1],elementsVec[i]->c[2],elementsVec[i]->scaleTrans.x,elementsVec[i]->scaleTrans.y,elementsVec[i]->scaleTrans.z);
    }
  fclose(file);
}

