// shapes.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "shapes.h"
#include "../shapes/trianglemesh.h"
// Shape Method Definitions
Shape::Shape(const Transform &o2w)
	: ObjectToWorld(o2w) {
	WorldToObject = ObjectToWorld.GetInverse();
	// Update shape creation statistics
	static StatsCounter nShapesMade("Geometry",
		"Total shapes created");
	++nShapesMade;
}
void Triangle::GetUVs(Float uv[3][2]) const {
	if (mesh->uvs) {
		uv[0][0] = mesh->uvs[2*v[0]];
		uv[0][1] = mesh->uvs[2*v[0]+1];
		uv[1][0] = mesh->uvs[2*v[1]];
		uv[1][1] = mesh->uvs[2*v[1]+1];
		uv[2][0] = mesh->uvs[2*v[2]];
		uv[2][1] = mesh->uvs[2*v[2]+1];
	}
	else {
	     uv[0][0] = uv[0][1] = uv[1][1] = 0.;
	     uv[1][0] = uv[2][0] = uv[2][1] = 1.;
	}
}
// DifferentialGeometry Method Definitions
void DifferentialGeometry::ComputeDifferentials(
		const RayDifferential &ray) const {
	if (ray.hasDifferentials) {
		// Estimate screen-space change in $P$ and $(u,v)$
		// Compute auxiliary intersection points with plane
		Float D = -Dot(Nn, Vector(P.x, P.y, P.z));
		Vector rxv(ray.rx.O.x, ray.rx.O.y, ray.rx.O.z);
		Float tx = -(Dot(Nn, rxv) + D) / Dot(Nn, ray.rx.D);
		Point Px = ray.rx.O + tx * ray.rx.D;
		Vector ryv(ray.ry.O.x, ray.ry.O.y, ray.ry.O.z);
		Float ty = -(Dot(Nn, ryv) + D) / Dot(Nn, ray.ry.D);
		Point Py = ray.ry.O + ty * ray.ry.D;
		dPdx = Px - P;
		dPdy = Py - P;
		// Compute $(u,v)$ offsets at auxiliary points
		// Initialize [[A]], [[Bx]], and [[By]] matrices for offset computation
		Float A[2][2], Bx[2], By[2], x[2];
		if (fabsf(Nn.x) > fabsf(Nn.y) &&
			fabsf(Nn.x) > fabsf(Nn.z)) {
			// Project onto $yz$ plane to initialize matrices
			#define C1 y
			#define C2 z
			// Initialize matrices for chosen projection plane
			A[0][0] = dPdu.C1;
			A[0][1] = dPdv.C1;
			A[1][0] = dPdu.C2;
			A[1][1] = dPdv.C2;
			Bx[0] = Px.C1 - P.C1;
			Bx[1] = Px.C2 - P.C2;
			By[0] = Py.C1 - P.C1;
			By[1] = Py.C2 - P.C2;
			#undef C1
			#undef C2
		}
		else if (fabsf(Nn.y) > fabsf(Nn.z)) {
			// Project onto $xz$ plane to initialize matrices
			#define C1 x
			#define C2 z
			// Initialize matrices for chosen projection plane
			A[0][0] = dPdu.C1;
			A[0][1] = dPdv.C1;
			A[1][0] = dPdu.C2;
			A[1][1] = dPdv.C2;
			Bx[0] = Px.C1 - P.C1;
			Bx[1] = Px.C2 - P.C2;
			By[0] = Py.C1 - P.C1;
			By[1] = Py.C2 - P.C2;
			#undef C1
			#undef C2
		}
		else {
			// Project onto $xy$ plane to initialize matrices
			#define C1 x
			#define C2 y
			// Initialize matrices for chosen projection plane
			A[0][0] = dPdu.C1;
			A[0][1] = dPdv.C1;
			A[1][0] = dPdu.C2;
			A[1][1] = dPdv.C2;
			Bx[0] = Px.C1 - P.C1;
			Bx[1] = Px.C2 - P.C2;
			By[0] = Py.C1 - P.C1;
			By[1] = Py.C2 - P.C2;
			#undef C1
			#undef C2
		}
		SolveLinearSystem2x2(A, Bx, x);
		dudx = x[0];
		dvdx = x[1];
		SolveLinearSystem2x2(A, By, x);
		dudy = x[0];
		dvdy = x[1];
	}
	else {
		dudx = dvdx = 0.;
		dudy = dvdy = 0.;
		dPdx = dPdy = Vector(0,0,0);
	}
}
