#ifndef SHAPES_H
#define SHAPES_H
#include "lrt.h"
#include "geometry.h"
#include "transform.h"
#include "paramset.h"

// DifferentialGeometry Declarations
struct DifferentialGeometry {
  DifferentialGeometry() { u = v = 0.; shape = NULL; }
  // DifferentialGeometry Method Declarations
  DifferentialGeometry(const Point &p, const Vector &dpdu,
		       const Vector &dpdv, const Vector &dndu,
		       const Vector &dndv, Float uu, Float vv,
		       const Shape *sh)
    : P(p), dPdu(dpdu), dPdv(dpdv), dNdu(dndu), dNdv(dndv) {
    // Initialize [[DifferentialGeometry]] from parameters
    Nb = Normal(Cross(dPdu, dPdv));
    Nn = Nb.Hat();
    S = dPdu.Hat();
    T = Cross(S, Nn);
    u = uu;
    v = vv;
    shape = sh;
    dudx = dvdx = dudy = dvdy = 0;
  }
  Vector WorldToLocal(const Vector &v) const {
    return Vector(Dot(v, S), Dot(v, T), Dot(v, Nn)).Hat();
  }
  Vector LocalToWorld(const Vector &v) const {
    return Vector(S.x * v.x + T.x * v.y + Nn.x * v.z,
		  S.y * v.x + T.y * v.y + Nn.y * v.z,
		  S.z * v.x + T.z * v.y + Nn.z * v.z);
  }
  DifferentialGeometry(const Point &p, const Vector &dpdu,
		       const Vector &dpdv, const Vector &dndu,
		       const Vector &dndv, Float uu, Float vv,
		       const Shape *sh, Float dux, Float dvx,
		       Float duy, Float dvy)
    : P(p), dPdu(dpdu), dPdv(dpdv), dNdu(dndu), dNdv(dndv) {
    // Initialize [[DifferentialGeometry]] from parameters
    Nb = Normal(Cross(dPdu, dPdv));
    Nn = Nb.Hat();
    S = dPdu.Hat();
    T = Cross(S, Nn);
    u = uu;
    v = vv;
    shape = sh;
    dudx = dvdx = dudy = dvdy = 0;
    dudx = dux;
    dvdx = dvx;
    dudy = duy;
    dvdy = dvy;
  }
  void ComputeDifferentials(const RayDifferential &r) const;
  void Shift(Float dx, Float dy, DifferentialGeometry *g) const {
    g->P = P + dx * (dudx * dPdu + dvdx * dPdv) +
      dy * (dudy * dPdu + dvdy * dPdv);
    g->u = u + dx * dudx + dy * dudy;
    g->v = v + dx * dvdx + dy * dvdy;
    g->Nb = Nb + Normal(dx * (dudx * dNdu + dvdx * dNdv) +
			dy * (dudy * dNdu + dvdy * dNdv));
    g->Nn = g->Nb.Hat();
    g->T = Cross(g->S, g->Nb);
    g->S = Cross(g->T, g->Nb);
	
    g->dPdu = dPdu;
    g->dPdv = dPdv;
    g->dPdx = dPdx;
    g->dPdy = dPdy;
    g->dudx = dudx;
    g->dvdx = dvdx;
    g->dudy = dudy;
    g->dvdy = dvdy;
    g->dNdu = dNdu;
    g->dNdv = dNdv;
  }
  void UpdateBasis(const Vector &dpdu, const Vector &dpdv) {
    dPdu = dpdu;
    dPdv = dpdv;
    Nb = Normal(Cross(dPdu, dPdv));
    Nn = Nb.Hat();
    S = dPdu.Hat();
    T = Cross(S, Nn);
  }
  // DifferentialGeometry Data
  Point P;
  Normal Nb, Nn;
  Vector S, T;
  Float u, v;
  const Shape *shape;
  Vector dPdu, dPdv;
  Vector dNdu, dNdv;
  mutable Vector dPdx, dPdy;
  mutable Float dudx, dvdx, dudy, dvdy;
};
// Shape Declarations
class Shape : public ReferenceCounted<Shape> {
 public:
  // Shape Interface
  Shape(const Transform &o2w);
  virtual BBox ObjectBound() const = 0;
  virtual BBox WorldBound() const {
    return ObjectToWorld(ObjectBound());
  }
  virtual bool CanIntersect() const { return true; }
  virtual void Refine(vector<RefinedShape> &refined) const {
    Severe("Unimplemented Shape::Refine() method called");
  }
  virtual bool Intersect(const Ray &ray, Float *thitp,
			 DifferentialGeometry *dg) const {
    Severe("Unimplemented Shape::Intersect() method called");
    return false;
  }
	
  virtual bool IntersectP(const Ray &ray) const {
    Severe("Unimplemented Shape::IntersectP() method called");
    return false;
  }
  virtual Float Area() const {
    Severe("Unimplemented Shape::Area() method called");
    return 0.;
  }
  virtual Point Sample(Float u1, Float u2, Normal *Ns) const {
    Severe("Unimplemented Shape::Sample method called");
    return Point();
  }
  virtual Point Sample(const Point &P,
		       Float u1, Float u2, Normal *Ns) const {
    return Sample(u1, u2, Ns);
  }
  virtual Point Sample(const Point &P, const Normal &N,
		       Float u1, Float u2, Normal *Ns) const {
    return Sample(P, u1, u2, Ns);
  }
  virtual Float Weight(const Point &P, const Normal &N,
		       const Vector &dir) const {
    return Weight(P, dir);
  }
  virtual Float Weight(const Point &P, const Vector &dir) const {
    // Intersect sample ray with area light geometry
    DifferentialGeometry dgLight;
    Ray ray(P, dir);
    Float thit;
    if (!Intersect(ray, &thit, &dgLight)) return 0.;
    ray.maxt = thit;
    // Convert light sample weight to solid angle measure
    Vector dirHat = dir.Hat();
    Float weight = Area() * DistanceSquared(P, ray(ray.maxt)) /
      fabsf(Dot(dgLight.Nn, -dirHat));
    return weight;
  }
  virtual Float Weight(const Point &Pshape) const {
    return 1.f / Area();
  }
  virtual ~Shape() { }
  
  //PRAMOD
  virtual void createTranslucent(Scene* scene)
    {
      cout<<"createTranslucent is not defined!"<<endl;
      return;
    }
  
  virtual Spectrum getTranslucentDE(const Point& pt, const Normal& nl) const
    {
      Spectrum L(0.0);
      cout<<"getTranslucentDE is not defined!"<<endl;
      return L;
    }


 protected:
  // Shape Protected Data
  Transform ObjectToWorld, WorldToObject;
};
struct RefinedShape {
  RefinedShape(const Reference<Shape> &s, Texture<Normal> *ns = NULL,
	       Texture<Vector> *ss = NULL) {
    shape = s;
    Ns = ns;
    Ss = ss;
  }
  Reference<Shape> shape;
  Texture<Normal> *Ns;
  Texture<Vector> *Ss;
};
#endif // SHAPES_H
