// texture.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef TEXTURE_H
#define TEXTURE_H
#include "lrt.h"
#include "color.h"
#include "geometry.h"
#include "transform.h"
#include "mipmap.h"
#include "sampling.h"
#include "shapes/trianglemesh.h"
// Texture Class Declarations
template <class T> class Texture {
public:
	// Texture Interface
	virtual T Evaluate(const DifferentialGeometry &) const = 0;
	virtual ~Texture() { }
};
template <class T>
class ConstantTexture: public Texture<T> {
public:
	ConstantTexture(const T &v) { value = v; }
	T Evaluate(const DifferentialGeometry &) const;
private:
	T value;
};
template <class T1, class T2>
class ScaleTexture : public Texture<T2> {
public:
	ScaleTexture(Texture<T1> *s, Texture<T2> *v) {
		scale = s;
		value = v;
	}
	// ScaleTexture Methods
	T2 Evaluate(const DifferentialGeometry &) const;
	~ScaleTexture() {
		delete scale;
		delete value;
	}
private:
	Texture<T1> *scale;
	Texture<T2> *value;
};
template <class T>
class MixTexture : public Texture<T> {
public:
	MixTexture(Texture<T> *t1, Texture<T> *t2,
			Texture<Float> *amt) {
		tex1 = t1;
		tex2 = t2;
		amount = amt;
	}
	// MixTexture Interface
	T Evaluate(const DifferentialGeometry &) const;
	~MixTexture();
private:
	Texture<T> *tex1, *tex2;
	Texture<Float> *amount;
};
class TextureMapping2D {
public:
	virtual ~TextureMapping2D() { }
	virtual void Map(const DifferentialGeometry &dg,
		Float *s, Float *t, Float *dsdx, Float *dtdx,
		Float *dsdy, Float *dtdy) const = 0;
};
class IdentityMapping2D : public TextureMapping2D {
public:
	IdentityMapping2D(Float su = 1, Float sv = 1,
		Float du = 0, Float dv = 0);
	void Map(const DifferentialGeometry &dg, Float *s, Float *t,
		Float *dsdx, Float *dtdx,
		Float *dsdy, Float *dtdy) const;
private:
	Float su, sv, du, dv;
};
class SphericalMapping2D : public TextureMapping2D {
public:
	SphericalMapping2D(const Transform &toSph)
		: WorldToTexture(toSph) {
	}
	void Map(const DifferentialGeometry &dg, Float *s, Float *t,
		Float *dsdx, Float *dtdx,
		Float *dsdy, Float *dtdy) const;
private:
	void sphere(const Point &P, Float *s, Float *t) const;
	Transform WorldToTexture;
};
class CylindricalMapping2D : public TextureMapping2D {
public:
	CylindricalMapping2D(const Transform &toCyl)
		: WorldToTexture(toCyl) {
		mins=10.0f;
		mint=10.0f;
		maxs=-10.0f;
		maxt=-10.0f;
	}
	void Map(const DifferentialGeometry &dg, Float *s, Float *t,
		Float *dsdx, Float *dtdx,
		Float *dsdy, Float *dtdy) const;
private:
	mutable Float mins, maxs, mint, maxt;
	void cylinder(const Point &P, Float *s, Float *t) const;
	Transform WorldToTexture;
};
class PlanarMapping2D : public TextureMapping2D {
public:
	PlanarMapping2D(const Vector &v1, const Vector &v2,
		Float du = 0, Float dv = 0);
	void Map(const DifferentialGeometry &dg, Float *s, Float *t,
		Float *dsdx, Float *dtdx,
		Float *dsdy, Float *dtdy) const;
private:
	Vector vs, vt;
	Float ds, dt;
};
class TextureMapping3D {
public:
	virtual ~TextureMapping3D() { }
	virtual Point Map(const DifferentialGeometry &dg,
		Vector *dPdx, Vector *dPdy) const = 0;
};
class IdentityMapping3D : public TextureMapping3D {
public:
	IdentityMapping3D(const Transform &x)
		: WorldToTexture(x) { }
	Point Map(const DifferentialGeometry &dg, Vector *dPdx,
		Vector *dPdy) const;
private:
	Transform WorldToTexture;
};
template <class T>
class BilerpTexture : public Texture<T> {
public:
	// BilerpTexture Interface
	BilerpTexture(TextureMapping2D *mapping, const T &t00, const T &t01,
		const T &t10, const T &t11);
	~BilerpTexture();
	T Evaluate(const DifferentialGeometry &) const;
private:
	// BilerpTexture Private Data
	TextureMapping2D *mapping;
	T v00, v01, v10, v11;
};
template <class T>
class VertexTexture : public Texture<T> {
public:
	// VertexTexture Interface
	VertexTexture(const T *vals, int nVertices);
	~VertexTexture();
	T Evaluate(const DifferentialGeometry &) const;
private:
	// VertexTexture Private Data
	int nVertices;
	T *vals;
};
template <class T>
class ImageTexture : public Texture<T> {
public:
	// ImageTexture Interface
	ImageTexture(TextureMapping2D *m, const string &filename);
	T Evaluate(const DifferentialGeometry &) const;
	~ImageTexture();
private:
	// ImageTexture Private Methods
	static MIPMap<T> *GetTexture(const string &filename);
	static void Convert(const Spectrum &s, Spectrum *c) {
		*c = s;
	}
	static void Convert(const Spectrum &s, Float *c) {
		*c = s.y();
	}
	// ImageTexture Private Data
	MIPMap<T> *mipmap;
	TextureMapping2D *mapping;
};
class UVTexture : public Texture<Spectrum> {
public:
	UVTexture(TextureMapping2D *m) {
		mapping = m;
	}
	// UVTexture Interface
	Spectrum Evaluate(const DifferentialGeometry &) const;
	~UVTexture();
private:
	TextureMapping2D *mapping;
};
template <class T> class Checkerboard2D : public Texture<T> {
public:
	// Checkerboard2D Interface
	Checkerboard2D(TextureMapping2D *m, Texture<T> *c1,
			Texture<T> *c2, const string &aa) {
		mapping = m;
		tex1 = c1;
		tex2 = c2;
		// Select anti-aliasing method for [[Checkerboard2D]]
		if (aa == "none") aaMethod = NONE;
		else if (aa == "supersample") aaMethod = SUPERSAMPLE;
		else if (aa == "closedform") aaMethod = CLOSEDFORM;
		else {
			Warning("Anti-aliasing mode \"%s\" not understood "
				"by Checkerboard2D", aa.c_str());
			aaMethod = SUPERSAMPLE;
		}
	}
	~Checkerboard2D();
	T Evaluate(const DifferentialGeometry &) const;
private:
	// Checkerboard2D Private Data
	Texture<T> *tex1, *tex2;
	TextureMapping2D *mapping;
	enum { NONE, SUPERSAMPLE, CLOSEDFORM } aaMethod;
};
template <class T> class Checkerboard3D : public Texture<T> {
public:
	// Checkerboard3D Interface
	Checkerboard3D(TextureMapping3D *m, Texture<T> *c1,
			Texture<T> *c2) {
		mapping = m;
		tex1 = c1;
		tex2 = c2;
	}
	T Evaluate(const DifferentialGeometry &dg) const;
	~Checkerboard3D();
	
private:
	// Checkerboard3D Private Data
	Texture<T> *tex1, *tex2;
	TextureMapping3D *mapping;
};
template <class T> class PolkaDots : public Texture<T> {
public:
	// PolkaDots Interface
	PolkaDots(TextureMapping2D *m, Texture<T> *c1, Texture<T> *c2) {
		mapping = m;
		outsideDot = c1;
		insideDot = c2;
	}
	~PolkaDots();
	T Evaluate(const DifferentialGeometry &) const;
private:
	// PolkaDots Private Data
	Texture<T> *outsideDot, *insideDot;
	TextureMapping2D *mapping;
};
class BumpyTexture : public Texture<Float> {
public:
	// BumpyTexture Interface
	~BumpyTexture();
	BumpyTexture(int oct, Float roughness, TextureMapping3D *map) {
		omega = roughness;
		octaves = oct;
		mapping = map;
	}
	Float Evaluate(const DifferentialGeometry &) const;
private:
	// BumpyTexture Private Data
	int octaves;
	Float omega;
	TextureMapping3D *mapping;
};
class MarbleTexture : public Texture<Spectrum> {
public:
	// MarbleTexture Interface
	~MarbleTexture();
	MarbleTexture(int oct, Float roughness, TextureMapping3D *map) {
		omega = roughness;
		octaves = oct;
		mapping = map;
	}
	Spectrum Evaluate(const DifferentialGeometry &) const;
private:
	// MarbleTexture Private Data
	int octaves;
	Float omega;
	TextureMapping3D *mapping;
};
class Windy : public Texture<Float> {
public:
	// Windy Interface
	~Windy();
	Windy(TextureMapping3D *map) {
		mapping = map;
	}
	Float Evaluate(const DifferentialGeometry &) const;
private:
	// Windy Private Data
	TextureMapping3D *mapping;
};
// Texture Template Method Definitions
template <class T>
T ConstantTexture<T>::Evaluate(
		const DifferentialGeometry &) const {
	return value;
}
template <class T1, class T2>
T2 ScaleTexture<T1, T2>::Evaluate(
		const DifferentialGeometry &dg) const {
	return scale->Evaluate(dg) * value->Evaluate(dg);
}
template <class T>
MixTexture<T>::~MixTexture() {
	delete tex1;
	delete tex2;
	delete amount;
}
template <class T>
T MixTexture<T>::Evaluate(
		const DifferentialGeometry &dg) const {
	T t1 = tex1->Evaluate(dg), t2 = tex2->Evaluate(dg);
	Float amt = amount->Evaluate(dg);
	return (1. - amt) * t1 + amt * t2;
}
template <class T>
BilerpTexture<T>::BilerpTexture(TextureMapping2D *m,
		const T &t00, const T &t01, const T &t10,
		const T &t11) {
	mapping = m;
	v00 = t00;
	v01 = t01;
	v10 = t10;
	v11 = t11;
}
template <class T>
BilerpTexture<T>::~BilerpTexture() {
	delete mapping;
}
template <class T>
T BilerpTexture<T>::Evaluate(const DifferentialGeometry &dg) const {
	Float u, v, dsdx, dtdx, dsdy, dtdy;
	mapping->Map(dg, &u, &v, &dsdx, &dtdx, &dsdy, &dtdy);
	return (1-u)*(1-v) * v00 + (1-u)*v * v01 + u*(1-v) * v10 +
		u*v * v11;
}
template <class T>
VertexTexture<T>::VertexTexture(const T *vs, int nv) {
	nVertices = nv;
	vals = new T[nv];
	for (int i = 0; i < nv; ++i)
		vals[i] = vs[i];
}
template <class T>
VertexTexture<T>::~VertexTexture() {
	delete[] vals;
}
template <class T>
T VertexTexture<T>::Evaluate(
		const DifferentialGeometry &dg) const {
	// Find three vertex texture values, [[v0]], [[v1]], and [[v2]]
	Triangle *tri = (Triangle *)(dg.shape);
	int *v = tri->v;
	const T &v0 = vals[v[0]], &v1 = vals[v[1]], &v2 = vals[v[2]];
	// Compute barycentric coordinates for point
	Float b[3];
	// Initialize A and B matrices for barycentrics
	Float uv[3][2];
	tri->GetUVs(uv);
	Float A[2][2] = { { uv[1][0] - uv[0][0], uv[2][0] - uv[0][0] },
	                  { uv[1][1] - uv[0][1], uv[2][1] - uv[0][1] } };
	Float B[2] = { dg.u - uv[0][0], dg.v - uv[0][1] };
	if (!SolveLinearSystem2x2(A, B, &b[1])) {
		// Handle degenerate parametric mapping
		b[0] = b[1] = b[2] = .3333333333f;
	}
	else
		b[0] = 1.f - b[1] - b[2];
	return b[0] * v0 + b[1] * v1 + b[2] * v2;
}
template <class T>
ImageTexture<T>::ImageTexture(TextureMapping2D *m,
		const string &filename) {
	mapping = m;
	mipmap = GetTexture(filename);

}
template <class T> ImageTexture<T>::~ImageTexture() {
	delete mapping;
}
template <class T> MIPMap<T> *ImageTexture<T>::GetTexture(
		const string &filename) {
	static StringHashTable textures;
	MIPMap<T> *ret = (MIPMap<T> *)textures.Search(filename);
	if (!ret) {
		// Load image map from disk and create MIPMap
		static StatsCounter texLoaded("Texture",
			"Number of image maps loaded");
		++texLoaded;
		int width, height;
		Spectrum *texels = TIFFRead(filename, &width, &height);
		if (texels) {
			// Convert texels to type [[T]] and create [[MIPMap]]
			T *convertedTexels = new T[width*height];
			for (int i = 0; i < width*height; ++i)
				Convert(texels[i], &convertedTexels[i]);
			ret = new MIPMap<T>(width, height, convertedTexels);
			textures.Add(filename, ret);
			delete[] texels;
			delete[] convertedTexels;
		}
	}
	return ret;
}
template <class T>
T ImageTexture<T>::Evaluate(
		const DifferentialGeometry &dg) const {
	Float s, t, dsdx, dtdx, dsdy, dtdy;
	mapping->Map(dg, &s, &t, &dsdx, &dtdx, &dsdy, &dtdy);
	return mipmap->Lookup(s, t, dsdx, dtdx, dsdy, dtdy);
}
template <class T> Checkerboard2D<T>::~Checkerboard2D() {
	delete mapping;
	delete tex1;
	delete tex2;
}
template <class T>
T Checkerboard2D<T>::Evaluate(
		const DifferentialGeometry &dg) const {
	Float s, t, dsdx, dtdx, dsdy, dtdy;
	mapping->Map(dg, &s, &t, &dsdx, &dtdx, &dsdy, &dtdy);
	if (aaMethod == CLOSEDFORM) {
		// Compute closed form box-filtered [[Checkerboard2D]] value
		// Evaluate single check if filter is entirely inside one of them
		Float ds = .5f * max(fabsf(dsdx), fabsf(dsdy));
		Float dt = .5f * max(fabsf(dtdx), fabsf(dtdy));
		Float s0 = s - ds, s1 = s + ds;
		Float t0 = t - dt, t1 = t + dt;
		if (Floor2Int(s0) == Floor2Int(s1) &&
			Floor2Int(t0) == Floor2Int(t1)) {
			// Point sample [[Checkerboard2D]]
			if ((Floor2Int(s) + Floor2Int(t)) % 2 == 0)
				return tex1->Evaluate(dg);
			return tex2->Evaluate(dg);
		}
		// Apply box-filter to checkerboard region
		#define BUMPINT(x) \
			(Floor2Int((x)/2) + 2.f * max((x/2)-Floor2Int(x/2) - .5f, 0.f))
		Float sint = (BUMPINT(s1) - BUMPINT(s0)) / (2. * ds);
		Float tint = (BUMPINT(t1) - BUMPINT(t0)) / (2. * dt);
		Float area2 = sint + tint - 2.f * sint * tint;
		return (1.f - area2) * tex1->Evaluate(dg) +
			area2 * tex2->Evaluate(dg);
	}
	else if (aaMethod == SUPERSAMPLE) {
		// Supersample [[Checkerboard2D]]
		#define N_SAMPLES 4
		Float samples[2*N_SAMPLES*N_SAMPLES];
		StratifiedSample2D(samples, N_SAMPLES, N_SAMPLES);
		T value = 0.;
		Float filterSum = 0.;
		for (int i = 0; i < N_SAMPLES*N_SAMPLES; ++i) {
			// Compute new differential geometry for supersample location
			DifferentialGeometry dgs;
			Float dx = samples[2*i]   - 0.5f;
			Float dy = samples[2*i+1] - 0.5f;
			dg.Shift(dx, dy, &dgs);
			dgs.dudx /= N_SAMPLES;
			dgs.dudy /= N_SAMPLES;
			dgs.dvdx /= N_SAMPLES;
			dgs.dvdy /= N_SAMPLES;
			// Compute $(s,t)$ for supersample and evaluate sub-texture
			Float ss, ts, dsdxs, dtdxs, dsdys, dtdys;
			mapping->Map(dgs, &ss, &ts, &dsdxs, &dtdxs, &dsdys, &dtdys);
			Float wt = expf(-2.f * (dx*dx + dy*dy));
			filterSum += wt;
			if ((Floor2Int(ss) + Floor2Int(ts)) % 2 == 0)
				value += wt * tex1->Evaluate(dgs);
			else
				value += wt * tex2->Evaluate(dgs);
		}
		return value / filterSum;
	}
	// Point sample [[Checkerboard2D]]
	if ((Floor2Int(s) + Floor2Int(t)) % 2 == 0)
		return tex1->Evaluate(dg);
	return tex2->Evaluate(dg);
}
template <class T>
Checkerboard3D<T>::~Checkerboard3D() {
	delete tex1;
	delete tex2;
}
template <class T>
T Checkerboard3D<T>::Evaluate(
		const DifferentialGeometry &dg) const {
	// Supersample [[Checkerboard3D]]
	#define N_SAMPLES 4
	Float samples[2*N_SAMPLES*N_SAMPLES];
	StratifiedSample2D(samples, N_SAMPLES, N_SAMPLES);
	T value = 0.;
	Float filterSum = 0.;
	for (int i = 0; i < N_SAMPLES*N_SAMPLES; ++i) {
		// Compute new differential geometry for supersample location
		DifferentialGeometry dgs;
		Float dx = samples[2*i]   - 0.5f;
		Float dy = samples[2*i+1] - 0.5f;
		dg.Shift(dx, dy, &dgs);
		dgs.dudx /= N_SAMPLES;
		dgs.dudy /= N_SAMPLES;
		dgs.dvdx /= N_SAMPLES;
		dgs.dvdy /= N_SAMPLES;
		// Compute 3D supersample position and evaluate sub-texture
		Vector dPPdx, dPPdy;
		Point PP = mapping->Map(dgs, &dPPdx, &dPPdy);
		Float wt = expf(-2.f * (dx*dx + dy*dy));
		filterSum += wt;
		if ((Floor2Int(PP.x) + Floor2Int(PP.y) + Floor2Int(PP.z)) % 2 == 0)
			value += wt * tex1->Evaluate(dgs);
		else
			value += wt * tex2->Evaluate(dgs);
	}
	return value / filterSum;
}
template <class T> PolkaDots<T>::~PolkaDots() {
	delete mapping;
	delete insideDot;
	delete outsideDot;
}
template <class T>
T PolkaDots<T>::Evaluate(const DifferentialGeometry &dg) const {
	// Compute cell incides for dots
	Float s, t, dsdx, dtdx, dsdy, dtdy;
	mapping->Map(dg, &s, &t, &dsdx, &dtdx, &dsdy, &dtdy);
	int sCell = Floor2Int(s + .5f), tCell = Floor2Int(t + .5f);
	// Return [[insideDot]] result if point is inside dot
	if (Noise(sCell+.5, tCell+.5) > 0) {
		Float radius = .35f;
		Float maxShift = 0.5f - radius;
		Float sCenter = sCell + maxShift *
			Noise(sCell + 1.5f, tCell + 2.8f);
		Float tCenter = tCell + maxShift *
			Noise(sCell + 4.5f, tCell + 9.8f);
		Float ds = fabsf(s - sCenter), dt = fabsf(t - tCenter);
		if (ds*ds + dt*dt < radius*radius)
			return insideDot->Evaluate(dg);
	}
	return outsideDot->Evaluate(dg);
}
#endif // TEXTURE_H
