// tiffio.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "color.h"
#include <tiffio.h>
// TIFF Function Definitions
Spectrum *TIFFRead(const string &name, int *xSize, int *ySize) {
	Spectrum *pixels = NULL;
	Float *fbuf = NULL;
	u_char *ubuf = NULL;
	// Try to open TIFF file
	TIFF *tiff = TIFFOpen(name.c_str(), "r");
	if (!tiff) {
		Error("Unable to open TIFF %s", name.c_str());
		return NULL;
	}
	// Get basic information from TIFF header
	short int nSamples;
	TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, xSize);
	TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, ySize);
	TIFFGetField(tiff, TIFFTAG_SAMPLESPERPIXEL, &nSamples);
	if (nSamples-1 > COLOR_SAMPLES)
		// Allow one extra, e.g. for alpha..
		Warning("TIFF %s has %d samples, > Spectrum (%d samples)",
			name.c_str(), nSamples, COLOR_SAMPLES);
	if (nSamples != 1 && nSamples < COLOR_SAMPLES)
		Warning("TIFF %s has %d samples, < Spectrum (%d samples)",
			name.c_str(), nSamples, COLOR_SAMPLES);
	// Make sure this is a TIFF we can read
	short int bitsPerSample, sampleFormat = SAMPLEFORMAT_UINT;
	if (!TIFFGetField(tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerSample)) {
		Error("TIFFRead: bits per sample not set in TIFF");
		// Clean up after TIFF reading error
		delete[] pixels;
		delete[] ubuf;
		delete[] fbuf;
		TIFFClose(tiff);
		return NULL;
	}
	if (!TIFFGetField(tiff, TIFFTAG_SAMPLEFORMAT, &sampleFormat)) {
		if (bitsPerSample == 32)
			sampleFormat = SAMPLEFORMAT_IEEEFP;
		else
			sampleFormat = SAMPLEFORMAT_UINT;
	}
	
	if (bitsPerSample == 32) {
		if (sampleFormat != SAMPLEFORMAT_IEEEFP) {
			Error("TIFFRead: 32 bit TIFF not stored in floating point format");
			// Clean up after TIFF reading error
			delete[] pixels;
			delete[] ubuf;
			delete[] fbuf;
			TIFFClose(tiff);
			return NULL;
		}
	}
	else {
		if (bitsPerSample != 8 && bitsPerSample != 32) {
			Error("TIFFRead: only 8 and 32 bits per sample supported");
			// Clean up after TIFF reading error
			delete[] pixels;
			delete[] ubuf;
			delete[] fbuf;
			TIFFClose(tiff);
			return NULL;
		}
		if (sampleFormat != SAMPLEFORMAT_UINT) {
			Error("TIFFRead: 8 bit TIFFs must be stored as unsigned ints");
			// Clean up after TIFF reading error
			delete[] pixels;
			delete[] ubuf;
			delete[] fbuf;
			TIFFClose(tiff);
			return NULL;
		}
	}
	int bytesPerSample = bitsPerSample / 8;
	if (nSamples * *xSize * bytesPerSample != TIFFScanlineSize(tiff)) {
		Error("TIFFRead: RGB not interleaved in TIFF %s", name.c_str());
		// Clean up after TIFF reading error
		delete[] pixels;
		delete[] ubuf;
		delete[] fbuf;
		TIFFClose(tiff);
		return NULL;
	}
	// Read TIFF colormap if present
	u_short *mapR = 0, *mapG = 0, *mapB = 0;
	if (nSamples == 1) {
		short photoType;
		TIFFGetField(tiff, TIFFTAG_PHOTOMETRIC, &photoType);
		if (photoType != PHOTOMETRIC_PALETTE) {
			Error("TIFFRead: colormap not found in one-sample image");
			// Clean up after TIFF reading error
			delete[] pixels;
			delete[] ubuf;
			delete[] fbuf;
			TIFFClose(tiff);
			return NULL;
		}
		TIFFGetField(tiff, TIFFTAG_COLORMAP, &mapR, &mapG, &mapB);
	}
	// Allocate space for [[pixels]] and buffers
	pixels = new Spectrum[*xSize * *ySize];
	Spectrum *pixelp = pixels;
	if (bitsPerSample == 32) fbuf = new float[nSamples * *xSize];
	else                     ubuf = new u_char[nSamples * *xSize];
	for (int y = 0; y < *ySize; ++y) {
		// Read a TIFF scanline
		if (fbuf) {
			// Read floating point TIFF scanline
			float *fbufp = fbuf;
			if (TIFFReadScanline(tiff, fbuf, y, 1) == -1) {
				// Clean up after TIFF reading error
				delete[] pixels;
				delete[] ubuf;
				delete[] fbuf;
				TIFFClose(tiff);
				return NULL;
			}
			for (int x = 0; x < *xSize; ++x) {
				Float cs[COLOR_SAMPLES];
				for (int i = 0; i < COLOR_SAMPLES; ++i) {
					if (i < nSamples) cs[i] = fbufp[i];
					else              cs[i] = 0.;
				}
				*pixelp++ = Spectrum(cs);
				fbufp += nSamples;
			}
		}
		else {
			// Read 8-bit TIFF scanline
			u_char *ubufp = ubuf;
			if (TIFFReadScanline(tiff, ubuf, y, 1) == -1) {
				// Clean up after TIFF reading error
				delete[] pixels;
				delete[] ubuf;
				delete[] fbuf;
				TIFFClose(tiff);
				return NULL;
			}
			for (int x = 0; x < *xSize; ++x) {
				if (nSamples == 1) {
					// Decode TIFF colormap entry
					int mapOffset = *ubufp;
					Assert(COLOR_SAMPLES == 3);
					Float cs[3] = { mapR[mapOffset] * INV_255 * INV_255,
						mapG[mapOffset] * INV_255 * INV_255,
						mapB[mapOffset] * INV_255 * INV_255 };
					*pixelp = Spectrum(cs);
				}
				else {
					// Convert standard 8-bit TIFF pixel
					Float cs[COLOR_SAMPLES];
					for (int i = 0; i < COLOR_SAMPLES; ++i)
						cs[i] = ubufp[i] * INV_255;
					*pixelp = Spectrum(cs);
				}
				++pixelp;
				ubufp += nSamples;
			}
		}
	}
	// Close TIFF and return
	delete[] ubuf;
	delete[] fbuf;
	TIFFClose(tiff);
	return pixels;
}
void TIFFWrite8Bit(const string &name, Float *pixels,
		Float *alpha, int XRes, int YRes, int nChannels,
		int totXRes, int totYRes) {
	Assert(pixels);
	// Open 8-bit TIFF file for writing
	TIFF *tiff = TIFFOpen(name.c_str(), "w");
	if (!tiff) {
		Error("Unable to open TIFF %s for writing", name.c_str());
		return;
	}
	
	// Compute and set up samples per pixel
	int sampleCount = 0;
	if (pixels) sampleCount += nChannels;
	if (alpha) ++sampleCount;
	TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, sampleCount);
	if (alpha) {
		short int extra[] = { EXTRASAMPLE_ASSOCALPHA };
		TIFFSetField(tiff, TIFFTAG_EXTRASAMPLES, (short)1, extra);
	}
	TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, XRes);
	TIFFSetField(tiff, TIFFTAG_IMAGELENGTH, YRes);
	if (totXRes != 0) {
		TIFFSetField(tiff, TIFFTAG_PIXAR_IMAGEFULLWIDTH, totXRes);
		TIFFSetField(tiff, TIFFTAG_PIXAR_IMAGEFULLLENGTH, totYRes);
	}
	TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, 8);
	TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
	// Set Generic TIFF Fields
	TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, 1L);
	TIFFSetField(tiff, TIFFTAG_XRESOLUTION, 1.0);
	TIFFSetField(tiff, TIFFTAG_YRESOLUTION, 1.0);
	TIFFSetField(tiff, TIFFTAG_RESOLUTIONUNIT, 1);
	TIFFSetField(tiff, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
	TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(tiff, TIFFTAG_ORIENTATION, (int)ORIENTATION_TOPLEFT);
	// Write 8-bit scanlines
	u_char *buf = new u_char[sampleCount * XRes];
	Float *pixelp = pixels;
	Float *alphap = alpha;
	for (int y = 0; y < YRes; ++y) {
		u_char *bufp = buf;
		for (int x = 0; x < XRes; ++x) {
			// Pack 8-bit pixels samples into buf
			for (int s = 0; s < nChannels; ++s)
				*bufp++ = (u_char)(Clamp(*pixelp++, 0.f, 255.f));
			// Pack 8-bit alpha samples into buf
			if (alphap)
				*bufp++ = (u_char)(Clamp(*alphap++, 0.f, 255.f));
		}
		TIFFWriteScanline(tiff, buf, y, 1);
	}
	// Close 8-bit TIFF file
	delete[] buf;
	TIFFClose(tiff);
}
void TIFFWriteFloat(const string &name, Float *pixels,
		Float *alpha, int XRes, int YRes, int nChannels,
		int totXRes, int totYRes) {
	// Open Float TIFF file for writing
	TIFF *tiff = TIFFOpen(name.c_str(), "w");
	if (!tiff) {
		Error("Unable to open TIFF %s for writing", name.c_str());
		return;
	}
	TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, XRes);
	TIFFSetField(tiff, TIFFTAG_IMAGELENGTH, YRes);
	if (totXRes != 0) {
		TIFFSetField(tiff, TIFFTAG_PIXAR_IMAGEFULLWIDTH, totXRes);
		TIFFSetField(tiff, TIFFTAG_PIXAR_IMAGEFULLLENGTH, totYRes);
	}
	// Compute and set up samples per pixel
	int sampleCount = 0;
	if (pixels) sampleCount += nChannels;
	if (alpha) ++sampleCount;
	TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, sampleCount);
	if (alpha) {
		short int extra[] = { EXTRASAMPLE_ASSOCALPHA };
		TIFFSetField(tiff, TIFFTAG_EXTRASAMPLES, (short)1, extra);
	}
	TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, 32);
	TIFFSetField(tiff, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
	TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
	// Set Generic TIFF Fields
	TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, 1L);
	TIFFSetField(tiff, TIFFTAG_XRESOLUTION, 1.0);
	TIFFSetField(tiff, TIFFTAG_YRESOLUTION, 1.0);
	TIFFSetField(tiff, TIFFTAG_RESOLUTIONUNIT, 1);
	TIFFSetField(tiff, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
	TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(tiff, TIFFTAG_ORIENTATION, (int)ORIENTATION_TOPLEFT);
	// Write Float scanlines
	Float *pixelp = (Float *)pixels;
	Float *alphap = alpha;
	Float *scanline = new Float[sampleCount * XRes];
	for (int y = 0; y < YRes; ++y) {
		Float *sp = scanline;
		for (int x = 0; x < XRes; ++x) {
			if (pixelp)
				for (int c = 0; c < nChannels; ++c)
					*sp++ = *pixelp++;
			if (alphap)
				*sp++ = *alphap++;
		}
		TIFFWriteScanline(tiff, scanline, y, 1);
	}
	delete[] scanline;
	// Close Float TIFF file
	TIFFClose(tiff);
}
