// transport.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "primitives.h"
#include "color.h"
#include "light.h"
#include "scene.h"
#include "reflection.h"
#include "sampling.h"
#include "materials.h"
#include "transport.h"
// Integrator Method Definitions
Integrator::~Integrator() {
}
// Integrator Utility Functions
Spectrum UniformSampleAllLights(const Scene *scene, const Point &P,
		const Normal &N, const Vector &wo, BSDF *bsdf,
		const Sample *sample, int lightSampleOffset,
		int bsdfSampleOffset) {
	Spectrum L(0.);
	for (u_int i = 0; i < scene->lights.size(); ++i) {
		Light *light = scene->lights[i];
		L += EstimateDirect(scene, light, P, N, wo, bsdf,
			sample, lightSampleOffset, bsdfSampleOffset,
			i);
	}
	return L;
}
Spectrum UniformSampleOneLight(const Scene *scene, const Point &P,
		const Normal &N, const Vector &wo, BSDF *bsdf,
		const Sample *sample, int lightSampleOffset,
		int bsdfSampleOffset) {
	int nLights = int(scene->lights.size());
	int lightNum = RandomInt() % nLights;
	Light *light = scene->lights[lightNum];
	return (Float) nLights *
		EstimateDirect(scene, light, P, N, wo, bsdf, sample,
			lightSampleOffset, bsdfSampleOffset, 0);
}
Spectrum WeightedSampleOneLight(const Scene *scene, const Point &P,
		const Normal &N, const Vector &wo, BSDF *bsdf,
		const Sample *sample, int lightSampleOffset,
		int bsdfSampleOffset, Float *&avgY,
		Float *&avgYsample, Float *&cdf, Float &overallAvgY) {
	int nLights = int(scene->lights.size());
	// Initialize [[avgY]] array if necessary
	if (!avgY) {
		avgY = new Float[nLights];
		avgYsample = new Float[nLights];
		cdf = new Float[nLights+1];
		for (int i = 0; i < nLights; ++i)
			avgY[i] = avgYsample[i] = 0.;
	}
	Spectrum L(0.);
	if (overallAvgY == 0.) {
		// Sample one light uniformly and initialize luminance arrays
		L = UniformSampleOneLight(scene, P, N, wo, bsdf, sample, lightSampleOffset,
			bsdfSampleOffset);
		Float luminance = L.y();
		overallAvgY = luminance;
		for (int i = 0; i < nLights; ++i)
			avgY[i] = luminance;
	}
	else {
		// Choose light according to average reflected luminance
		Float c, lightSampleWeight;
		for (int i = 0; i < nLights; ++i)
			avgYsample[i] = max(avgY[i], .1f * overallAvgY);
		ComputeStep1dCDF(avgYsample, nLights, &c, cdf);
		Float t = SampleStep1d(avgYsample, cdf, c, nLights, RandomFloat(),
			&lightSampleWeight);
		int lightNum = min(Float2Int(nLights * t), nLights-1);
		Light *light = scene->lights[lightNum];
		L = EstimateDirect(scene, light, P, N, wo, bsdf,
			sample, lightSampleOffset, bsdfSampleOffset, 0);
		// Update [[avgY]] array
		Float luminance = L.y();
		avgY[lightNum] =
			ExponentialAverage(avgY[lightNum], luminance, .99f);
		overallAvgY =
			ExponentialAverage(overallAvgY, luminance, .999f);
		L /= lightSampleWeight;
	}
	return L;
}
Spectrum EstimateDirect(const Scene *scene, const Light *light,
		const Point &P, const Normal &N, const Vector &wo,
		BSDF *bsdf, const Sample *sample, int lightSamp,
		int bsdfSamp, int sampleNum) {
	Spectrum Ld(0.);
	// Find light and BSDF sample values for direct lighting estimate
	Float ls1, ls2, bs1, bs2;
	if (lightSamp != -1 && bsdfSamp != -1 &&
		sampleNum < sample->n2D[lightSamp] &&
		sampleNum < sample->n2D[bsdfSamp]) {
		ls1 = sample->twoD[lightSamp][2*sampleNum];
		ls2 = sample->twoD[lightSamp][2*sampleNum+1];
		bs1 = sample->twoD[bsdfSamp][2*sampleNum];
		bs2 = sample->twoD[bsdfSamp][2*sampleNum+1];
	}
	else {
		ls1 = RandomFloat();
		ls2 = RandomFloat();
		bs1 = RandomFloat();
		bs2 = RandomFloat();
	}
	// Sample light source with multiple importance sampling
	Vector wi;
	Float lightWeight, bsdfWeight, weight;
	bool deltaLight;
	VisibilityTester visibility;
	Spectrum Li = light->Sample_L(P, N,
		ls1, ls2, &wi, &lightWeight, &deltaLight, &visibility);
	if (lightWeight > 0. && !Li.Black() && visibility.Unoccluded(scene)) {
		if (deltaLight)
			weight = lightWeight;
		else {
			bsdfWeight = bsdf->Weight(wo, wi);
			weight = .5f * lightWeight + .5f * bsdfWeight;
		}
		Ld += bsdf->f(wo, wi) * fabsf(Dot(wi, N)) * Li *
			visibility.Transmittance(scene) / weight;
	}
	// Sample BSDF with multiple importance sampling
	bool specularBounce;
	Spectrum f = bsdf->sample_f(wo, &wi,
		bs1, bs2, &bsdfWeight, &specularBounce);
	if (specularBounce)
		weight = bsdfWeight;
	else {
		lightWeight = light->Weight(P, N, wi);
		weight = .5f * lightWeight + .5f * bsdfWeight;
	}
	// Compute [[Ld]] from selected light
	Intersection lightIsect;
	if (scene->Intersect(Ray(P, wi), &lightIsect) &&
		lightIsect.primitive->areaLight == light)
		Ld += f * lightIsect.Le(-wi) / weight;
	return Ld;
}
