// util.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include <malloc.h>
// Error Reporting Includes
#include <stdarg.h>
// Error Reporting Definitions
#define LRT_ERROR_IGNORE 0
#define LRT_ERROR_CONTINUE 1
#define LRT_ERROR_ABORT 2
// Error Reporting Functions
static void processError(const char *format, va_list args,
		const char *message, int disposition) {
	// Format error string
	#define ERR_BUF_SZ 1024
	static char errorBuf[ERR_BUF_SZ];
	vsprintf(errorBuf, format, args);
	// Report error
	switch (disposition) {
	case LRT_ERROR_IGNORE:
		return;
	case LRT_ERROR_CONTINUE:
		fprintf(stderr, "%s: %s\n", message, errorBuf);
		// Print scene file and line number, if appropriate
		extern int line_num;
		if (line_num != 0) {
			extern string current_file;
			fprintf(stderr, "\tLine %d, file %s\n", line_num,
				current_file.c_str());
		}
		break;
	case LRT_ERROR_ABORT:
		fprintf(stderr, "%s: %s\n", message, errorBuf);
		// Print scene file and line number, if appropriate
		extern int line_num;
		if (line_num != 0) {
			extern string current_file;
			fprintf(stderr, "\tLine %d, file %s\n", line_num,
				current_file.c_str());
		}
		abort();
	}
}
void Info(const char *format, ...) {
	va_list args;
	va_start(args, format);
	processError(format, args, "Notice", LRT_ERROR_CONTINUE);
	va_end(args);
}
void Warning(const char *format, ...) {
	va_list args;
	va_start(args, format);
	processError(format, args, "Warning", LRT_ERROR_CONTINUE);
	va_end(args);
}
void Error(const char *format, ...) {
	va_list args;
	va_start(args, format);
	processError(format, args, "Error", LRT_ERROR_CONTINUE);
	va_end(args);
}
void Severe(const char *format, ...) {
	va_list args;
	va_start(args, format);
	processError(format, args, "Fatal Error", LRT_ERROR_ABORT);
	va_end(args);
}
// Matrix Method Definitions
bool SolveLinearSystem2x2(const Float A[2][2], const Float B[2],
		Float x[2]) {
	Float det = A[0][0]*A[1][1] - A[0][1]*A[1][0];
	if (fabsf(det) < 1e-5)
		return false;
	Float invDet = 1.0f/det;
	x[0] = (A[1][1]*B[0] - A[0][1]*B[1]) * invDet;
	x[1] = (A[0][0]*B[1] - A[1][0]*B[0]) * invDet;
	return true;
}
Reference<Matrix4x4> Matrix4x4::Transpose() const {
	return new Matrix4x4(m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}
Reference<Matrix4x4> Matrix4x4::Inverse() const {
	int indxc[4], indxr[4];
	int ipiv[4] = { 0, 0, 0, 0 };
	Float minv[4][4];
	memcpy(minv, m, 4*4*sizeof(Float));
	for (int i = 0; i < 4; i++) {
		int irow = -1, icol = -1;
		Float big = 0.;
		// Choose pivot
		for (int j = 0; j < 4; j++) {
			if (ipiv[j] != 1) {
				for (int k = 0; k < 4; k++) {
					if (ipiv[k] == 0) {
						if (fabsf(minv[j][k]) >= big) {
							big = Float(fabsf(minv[j][k]));
							irow = j;
							icol = k;
						}
					}
					else if (ipiv[k] > 1)
						Error("Singular matrix in MatrixInvert");
				}
			}
		}
		++ipiv[icol];
		// Swap rows [[irow]] and [[icol]] for pivot
		if (irow != icol) {
			for (int k = 0; k < 4; ++k)
				swap(minv[irow][k], minv[icol][k]);
		}
		indxr[i] = irow;
		indxc[i] = icol;
		if (minv[icol][icol] == 0.)
			Error("Singular matrix in MatrixInvert");
		// Set $m[icol][icol]$ to one by scaling row [[icol]] appropriately
		Float pivinv = 1.f / minv[icol][icol];
		minv[icol][icol] = 1.f;
		for (int j = 0; j < 4; j++)
			minv[icol][j] *= pivinv;
		// Subtract this row from others to zero out their columns
		for (int j = 0; j < 4; j++) {
			if (j != icol) {
				Float save = minv[j][icol];
				minv[j][icol] = 0;
				for (int k = 0; k < 4; k++)
					minv[j][k] -= minv[icol][k]*save;
			}
		}
	}
	// Swap columns to reflect permutation
	for (int j = 3; j >= 0; j--) {
		if (indxr[j] != indxc[j]) {
			for (int k = 0; k < 4; k++)
				swap(minv[k][indxr[j]], minv[k][indxc[j]]);
		}
	}
	return new Matrix4x4(minv);
}
// Statistics Definitions
struct StatTracker {
	StatTracker(const string &cat, const string &n,
	            StatsCounterType *pa, StatsCounterType *pb = NULL,
		    bool percentage = true);
	string category, name;
	StatsCounterType *ptra, *ptrb;
	bool percentage;
};
static vector<StatTracker *> trackers;
static void addTracker(StatTracker *newTracker) {
	for (u_int i = 0; i < trackers.size(); ++i) {
		if (newTracker->category == trackers[i]->category &&
		    newTracker->name == trackers[i]->name)
			return;
	}
	trackers.push_back(newTracker);
}
static void StatsPrintVal(FILE *f, StatsCounterType v);
static void StatsPrintVal(FILE *f, StatsCounterType v1, StatsCounterType v2);
// Statistics Functions
StatTracker::StatTracker(const string &cat, const string &n,
                         StatsCounterType *pa, StatsCounterType *pb, bool p) {
	category = cat;
	name = n;
	ptra = pa;
	ptrb = pb;
	percentage = p;
}
StatsCounter::StatsCounter(const string &category, const string &name) {
	num = 0;
	addTracker(new StatTracker(category, name, &num));
}
StatsRatio::StatsRatio(const string &category, const string &name, bool percent) {
	na = nb = 0;
	addTracker(new StatTracker(category, name, &na, &nb, percent));
}
struct CmpTracker {
	bool operator()(const StatTracker *t1,
			const StatTracker *t2) const {
		if (t1->category == t2->category)
			return (t1->name < t2->name);
		return (t1->category < t2->category);
	}
};

void StatsPrint(FILE *dest) {
	fprintf(dest, "Statistics:\n");
	vector<StatTracker *> t = trackers;
	sort(t.begin(), t.end(), CmpTracker());
	string lastCategory;
	for (u_int i = 0; i < t.size(); ++i) {
		// Print statistic
		if (t[i]->category != lastCategory) {
			fprintf(dest, "%s\n", t[i]->category.c_str());
			lastCategory = t[i]->category;
		}
		fprintf(dest, "    %s", t[i]->name.c_str());
		// Pad out to results column
		int resultsColumn = 56;
		int paddingSpaces = resultsColumn - (int) t[i]->name.size();
		while (paddingSpaces-- > 0)
			putc(' ', dest);
		if (t[i]->ptrb == NULL)
			StatsPrintVal(dest, *t[i]->ptra);
		else {
			if (*t[i]->ptrb > 0) {
				Float ratio = (Float)*t[i]->ptra / (Float)*t[i]->ptrb;
				StatsPrintVal(dest, *t[i]->ptra, *t[i]->ptrb);
				if (t[i]->percentage)
					fprintf(dest, " (%3.2f%%)", 100. * ratio);
				else
					fprintf(dest, " (%.2fx)", ratio);
			}
			else
				StatsPrintVal(dest, *t[i]->ptra, *t[i]->ptrb);
		}
		fprintf(dest, "\n");
	}
}
static void StatsPrintVal(FILE *f, StatsCounterType v) {
	if (v > 1e9) fprintf(f, "%.3fB", v / 1e9f);
	else if (v > 1e6) fprintf(f, "%.3fM", v / 1e6f);
	else if (v > 1e4) fprintf(f, "%.1fk", v / 1e3f);
	else fprintf(f, "%.0f", (float)v);
}
static void StatsPrintVal(FILE *f, StatsCounterType v1,
		StatsCounterType v2) {
	StatsCounterType m = min(v1, v2);
	if (m > 1e9) fprintf(f, "%.3fB:%.3fB", v1 / 1e9f, v2 / 1e9f);
	else if (m > 1e6) fprintf(f, "%.3fM:%.3fM", v1 / 1e6f, v2 / 1e6f);
	else if (m > 1e4) fprintf(f, "%.1fk:%.1fk", v1 / 1e3f, v2 / 1e3f);
	else fprintf(f, "%.0f:%.0f", v1, v2);
}
void StatsCleanup() {
	// Reset user-supplied statistics pointers
	for (u_int i = 0; i < trackers.size(); ++i) {
		trackers[i]->ptra = 0;
		if (trackers[i]->ptrb)
			trackers[i]->ptrb = 0;
	}
	for (u_int i = 0; i < trackers.size(); ++i)
		delete trackers[i];
	trackers.erase(trackers.begin(), trackers.end());
}
// Random Number State
/*
   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote
        products derived from this software without specific prior written
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#define N 624
#define M 397
#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

static unsigned long mt[N]; /* the array for the state vector  */
static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */
// Random Number Functions
static void init_genrand(u_long seed) {
	mt[0]= seed & 0xffffffffUL;
	for (mti=1; mti<N; mti++) {
		mt[mti] =
		(1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti);
		/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
		/* In the previous versions, MSBs of the seed affect   */
		/* only MSBs of the array mt[].                        */
		/* 2002/01/09 modified by Makoto Matsumoto             */
		mt[mti] &= 0xffffffffUL;
		/* for >32 bit machines */
	}
}
unsigned long genrand_int32(void)
{
	unsigned long y;
	static unsigned long mag01[2]={0x0UL, MATRIX_A};
	/* mag01[x] = x * MATRIX_A  for x=0,1 */

	if (mti >= N) { /* generate N words at one time */
		int kk;

		if (mti == N+1)   /* if init_genrand() has not been called, */
			init_genrand(5489UL); /* default initial seed */

		for (kk=0;kk<N-M;kk++) {
			y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
			mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
		}
		for (;kk<N-1;kk++) {
			y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
			mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
		}
		y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
		mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

		mti = 0;
	}

	y = mt[mti++];

	/* Tempering */
	y ^= (y >> 11);
	y ^= (y << 7) & 0x9d2c5680UL;
	y ^= (y << 15) & 0xefc60000UL;
	y ^= (y >> 18);

	return y;
}
/* generates a random number on [0,1]-real-interval */
Float genrand_real1(void)
{
	return genrand_int32()*((Float)1.0/(Float)4294967295.0);
	/* divided by 2^32-1 */
}
/* generates a random number on [0,1)-real-interval */
Float genrand_real2(void)
{
	return genrand_int32()*((Float)1.0/(Float)4294967296.0);
	/* divided by 2^32 */
}
// Memory Allocation Functions
void *AllocL1CacheAligned(size_t size) {
#ifndef L1_CACHE_LINE_SIZE
#define L1_CACHE_LINE_SIZE 32
#endif
	return memalign(L1_CACHE_LINE_SIZE, size);
}
void *AllocL2CacheAligned(size_t size) {
#ifndef L2_CACHE_LINE_SIZE
#define L2_CACHE_LINE_SIZE 256
#endif
	return memalign(L2_CACHE_LINE_SIZE, size);
}
void FreeCacheAligned(void *ptr) {
	free(ptr);
}
// StringHashTable Method Definitions
u_int StringHashTable::Hash(const string &str) const {
	u_int hashValue = 0;
	for (u_int i = 0; i < str.size(); ++i) {
		hashValue <<= 1;
		hashValue ^= str[i];
	}
	return hashValue % NUM_BUCKETS;
}
void *StringHashTable::Search(const string &key) const {
	u_int index = Hash(key);
	for (u_int i = 0; i < buckets[index].size(); ++i)
		if (key == buckets[index][i].first)
			return buckets[index][i].second;
	return NULL;
}
void StringHashTable::Add(const string &key, void *data) {
	u_int index = Hash(key);
	for (u_int i = 0; i < buckets[index].size(); ++i) {
		if (key == buckets[index][i].first) {
			buckets[index][i].second = data;
			return;
		}
	}
	buckets[index].push_back(make_pair(key, data));
}
