// grid.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "primitives.h"
#include "geometry.h"
// GridAccelerator Declarations
class GridAccelerator : public Primitive {
	// GridAccelerator Forward Declarations
	struct MailboxPrim;
public:
	// GridAccelerator Method Declarations
	GridAccelerator(const vector<Reference<Primitive> > &p);
	BBox WorldBound() const { return bounds; }
	bool CanIntersect() const { return true; }
	int x2v(Float x) const { return Float2Int((x - bounds.pMin.x) *
						InvXWidth); }
	int y2v(Float y) const { return Float2Int((y - bounds.pMin.y) *
						InvYWidth); }
	int z2v(Float z) const { return Float2Int((z - bounds.pMin.z) *
						InvZWidth); }
	Float v2x(int x) const { return x * XWidth + bounds.pMin.x; }
	Float v2y(int y) const { return y * YWidth + bounds.pMin.y; }
	Float v2z(int z) const { return z * ZWidth + bounds.pMin.z; }
	~GridAccelerator();
	bool Intersect(const Ray &ray, Intersection *isect) const;
	bool IntersectP(const Ray &ray) const;
private:
	// GridAccelerator Private Data
	BBox bounds;
	int XVoxels, YVoxels, ZVoxels;
	struct Voxel {
		Voxel() { allCanIntersect = false; }
		vector<MailboxPrim *> primitives;
		bool allCanIntersect;
	};
	Float XWidth, YWidth, ZWidth;
	Float InvXWidth, InvYWidth, InvZWidth;
	Voxel **voxels;
	struct MailboxPrim {
		MailboxPrim() {
			primitive = NULL;
			lastMailboxId = -1;
		}
		Reference<Primitive> primitive;
		int lastMailboxId;
	};
	u_int nMailboxes;
	MailboxPrim *mailboxes;
	MemoryArena<Voxel> voxelArena;
	mutable int curMailboxId;
};
// GridAccelerator Method Definitions
GridAccelerator::GridAccelerator(
		const vector<Reference<Primitive> > &prims) {
	// Compute grid bounds
	for (u_int i = 0; i < prims.size(); ++i)
	    bounds = Union(bounds, prims[i]->WorldBound());
	// Choose grid resolution
	// Expand grid bounds by small factor
	Float delta = max(fabsf(bounds.pMin.x),
		max(fabsf(bounds.pMin.y), fabsf(bounds.pMin.z)));
	delta = max(delta, max(fabsf(bounds.pMax.x),
			max(fabsf(bounds.pMax.y), fabsf(bounds.pMax.z))));
	bounds.pMin -= 1e-4f * Vector(delta, delta, delta);
	bounds.pMax += 1e-4f * Vector(delta, delta, delta);
	Vector diag = bounds.pMax - bounds.pMin;
	Float invmaxWidth = 1.0f/max(diag.x, max(diag.y, diag.z));
	Assert(invmaxWidth > 0.f);
	if (prims.size() < 5)
		XVoxels = YVoxels = ZVoxels = 1;
	else {
		Float cubeRoot = powf(Float(prims.size()), 1.f/3.f);
		Float voxScale = 8.f * cubeRoot * invmaxWidth;
		XVoxels = Clamp(Round2Int(diag.x * voxScale), 1, 100);
		YVoxels = Clamp(Round2Int(diag.y * voxScale), 1, 100);
		ZVoxels = Clamp(Round2Int(diag.z * voxScale), 1, 100);
	}
	// Compute voxel widths and allocate voxels
	XWidth = diag.x / XVoxels;
	YWidth = diag.y / YVoxels;
	ZWidth = diag.z / ZVoxels;
	InvXWidth = (XWidth == 0.f) ? 0.f : 1.f / XWidth;
	InvYWidth = (YWidth == 0.f) ? 0.f : 1.f / YWidth;
	InvZWidth = (ZWidth == 0.f) ? 0.f : 1.f / ZWidth;
	int nVoxels = XVoxels * YVoxels * ZVoxels;
	voxels = (Voxel **)AllocL1CacheAligned(nVoxels * sizeof(Voxel *));
	memset(voxels, 0, nVoxels * sizeof(Voxel *));
	// Add primitives to grid voxels
	nMailboxes = prims.size();
	mailboxes = (MailboxPrim *)AllocL1CacheAligned(prims.size() * sizeof(MailboxPrim));
	for (u_int i = 0; i < prims.size(); ++i) {
		new (&mailboxes[i]) MailboxPrim;
		// Find cell extent of primitive
		BBox primBounds = prims[i]->WorldBound();
		int x0 = max(x2v(primBounds.pMin.x), 0);
		int x1 = min(x2v(primBounds.pMax.x), XVoxels-1);
		int y0 = max(y2v(primBounds.pMin.y), 0);
		int y1 = min(y2v(primBounds.pMax.y), YVoxels-1);
		int z0 = max(z2v(primBounds.pMin.z), 0);
		int z1 = min(z2v(primBounds.pMax.z), ZVoxels-1);
		// Add primitive to overlapping cells
		MailboxPrim *mp = mailboxes + i;
		mp->primitive = prims[i];
		for (int z = z0; z <= z1; ++z)
			for (int y = y0; y <= y1; ++y)
				for (int x = x0; x <= x1; ++x) {
					int offset = z*XVoxels*YVoxels + y*XVoxels + x;
					if (!voxels[offset])
						voxels[offset] = new (voxelArena) Voxel;
					voxels[offset]->primitives.push_back(mp);
					// Update grid voxel statistics
					static StatsCounter maxPrimsInVoxel("Acceleration",
						"Max # of primitives in a grid voxel");
					maxPrimsInVoxel.stat_max(voxels[offset]->primitives.size());
				}
		// Update grid/primitive statistics
		static StatsRatio nPrimitiveVoxels("Acceleration",
			"Voxels covered vs # / primitives");
		nPrimitiveVoxels.add((1 + x1-x0) * (1 + y1-y0) * (1 + z1-z0), 1);
	}
	// Update fraction of empty voxels
	static StatsRatio nEmptyVoxels("Acceleration", "Empty voxels");
	nEmptyVoxels.add(0, XVoxels * YVoxels * ZVoxels);
	for (int i = 0; i < XVoxels * YVoxels * ZVoxels; ++i)
		if (!voxels[i]) nEmptyVoxels.add(1, 0);
	// Initialize mailbox
	curMailboxId = 0;
}
GridAccelerator::~GridAccelerator() {
	for (u_int i = 0; i < nMailboxes; ++i)
		mailboxes[i].~MailboxPrim();
	FreeCacheAligned(voxels);
	FreeCacheAligned(mailboxes);
}
bool GridAccelerator::Intersect(const Ray &ray,
		Intersection *isect) const {
	// Check ray against overall grid bounds
	Float rayT = ray.maxt;
	if (bounds.Inside(ray(ray.mint)))
		rayT = ray.mint;
	else if (!bounds.IntersectP(ray, &rayT))
		return false;
	Point gridIntersect = ray(rayT);
	// Set up 3D DDA for this ray
	// Update statistics for ray inside grid
	static StatsRatio rayTests("Acceleration", "Intersection tests per ray", false);
	static StatsRatio rayHits("Acceleration", "Intersections found per ray", false);
	rayTests.add(0, 1);
	rayHits.add(0, 1);
	// Set up X stepping
	// Compute current x voxel
	int x = x2v(gridIntersect.x);
	if (x == XVoxels) x--;
	else if (x < 0) ++x;
	Assert(x >= 0 && x < XVoxels);
	Float NextXCrossing, DeltaX;
	int StepX, OutX;
	if (fabsf(ray.D.x) < 1e-6) {
		// Handle ray perpendicular to x
		NextXCrossing = INFINITY;
		DeltaX = 0;
		OutX = -1;
	}
	else if (ray.D.x > 0) {
		// Handle ray with positive x direction
		NextXCrossing = rayT + (v2x(x+1) - gridIntersect.x)/ray.D.x;
		DeltaX = XWidth / ray.D.x;
		StepX = 1;
		OutX = XVoxels;
	}
	else {
		// Handle ray with negative x direction
		NextXCrossing = rayT + (v2x(x) - gridIntersect.x)/ray.D.x;
		DeltaX = - XWidth / ray.D.x;
		StepX = -1;
		OutX = -1;
	}
	// Set up Y stepping
	int y = y2v( gridIntersect.y );
	if (y == YVoxels) y--;
	if (y < 0) ++y;
	Float NextYCrossing, DeltaY;
	int StepY, OutY;
	
	Assert( y >= 0 && y < YVoxels );
	if (fabsf(ray.D.y) < 1e-6) {
		NextYCrossing = INFINITY;
		DeltaY = 0;
		OutY = -1;
	}
	else if (ray.D.y < 0) {
		NextYCrossing = rayT + ( v2y( y ) - gridIntersect.y )/ray.D.y;
		DeltaY = - YWidth / ray.D.y;
		StepY = OutY = -1;
	}
	else {
		NextYCrossing = rayT + ( v2y( y+1 ) - gridIntersect.y )/ray.D.y;
		DeltaY = YWidth / ray.D.y;
		StepY = 1;
		OutY = YVoxels;
	}
	// Set up Z stepping
	int z = z2v( gridIntersect.z );
	if (z == ZVoxels) z--;
	if (z < 0) ++z;
	Float NextZCrossing, DeltaZ;
	int StepZ, OutZ;
	
	Assert( z >= 0 && z < ZVoxels );
	if (fabsf(ray.D.z) < 1e-6) {
		NextZCrossing = INFINITY;
		DeltaZ = 0;
		OutZ = -1;
	}
	else if (ray.D.z < 0) {
		NextZCrossing = rayT + ( v2z( z ) - gridIntersect.z )/ray.D.z;
		DeltaZ = - ZWidth / ray.D.z;
		StepZ = OutZ = -1;
	}
	else {
		NextZCrossing = rayT + ( v2z( z+1 ) - gridIntersect.z )/ray.D.z;
		DeltaZ = ZWidth / ray.D.z;
		StepZ = 1;
		OutZ = ZVoxels;
	}
	// Walk grid
	bool hitSomething = false;
	// Get mailbox id for the ray
	int rayId = curMailboxId++;
	for (;;) {
		int offset = z*XVoxels*YVoxels + y*XVoxels + x;
		Voxel *voxel = voxels[offset];
		if (voxel != NULL) {
			// Check single voxel
			vector<MailboxPrim *> &primitiveList = voxel->primitives;
			// Refine primitives in voxel if needed
			if (!voxel->allCanIntersect) {
				for (u_int i = 0; i < primitiveList.size(); ++i) {
					MailboxPrim *mp = primitiveList[i];
					// Refine primitive if needed
					while (!mp->primitive->CanIntersect()) {
						vector<Reference<Primitive> > p;
						mp->primitive->Refine(p);
						Assert(p.size() > 0);
						if (p.size() == 1)
							mp->primitive = p[0];
						else
							mp->primitive = new GridAccelerator(p);
					}
				}
				voxel->allCanIntersect = true;
			}
			for (u_int i = 0; i < primitiveList.size(); ++i) {
				MailboxPrim *mp = primitiveList[i];
				// Do mailbox check between ray and primitive
				if (mp->lastMailboxId == rayId)
					continue;
				// Check for ray-primitive intersection
				mp->lastMailboxId = rayId;
				rayTests.add(1, 0);
				if (mp->primitive->Intersect(ray, isect)) {
					rayHits.add(1, 0);
					hitSomething = true;
				}
			}
		}
		// Advance to next voxel
		if (NextXCrossing < NextYCrossing &&
			NextXCrossing < NextZCrossing) {
			// Step in X
			if (ray.maxt < NextXCrossing)
				break;
			x += StepX;
			if (x == OutX)
				break;
			NextXCrossing += DeltaX;
		}
		else if (NextZCrossing < NextYCrossing) {
			// Step in Z
			if (ray.maxt < NextZCrossing)
				break;
			z += StepZ;
			if (z == OutZ)
				break;
			NextZCrossing += DeltaZ;
		}
		else {
			// Step in Y
			if (ray.maxt < NextYCrossing)
				break;
			y += StepY;
			if (y == OutY)
				break;
			NextYCrossing += DeltaY;
		}
	}
	return hitSomething;
}
bool GridAccelerator::IntersectP(const Ray &ray) const {
	// Check ray against overall grid bounds
	Float rayT = ray.maxt;
	if (bounds.Inside(ray(ray.mint)))
		rayT = ray.mint;
	else if (!bounds.IntersectP(ray, &rayT))
		return false;
	Point gridIntersect = ray(rayT);
	// Set up 3D DDA for this ray
	// Update statistics for ray inside grid
	static StatsRatio rayTests("Acceleration", "Intersection tests per ray", false);
	static StatsRatio rayHits("Acceleration", "Intersections found per ray", false);
	rayTests.add(0, 1);
	rayHits.add(0, 1);
	// Set up X stepping
	// Compute current x voxel
	int x = x2v(gridIntersect.x);
	if (x == XVoxels) x--;
	else if (x < 0) ++x;
	Assert(x >= 0 && x < XVoxels);
	Float NextXCrossing, DeltaX;
	int StepX, OutX;
	if (fabsf(ray.D.x) < 1e-6) {
		// Handle ray perpendicular to x
		NextXCrossing = INFINITY;
		DeltaX = 0;
		OutX = -1;
	}
	else if (ray.D.x > 0) {
		// Handle ray with positive x direction
		NextXCrossing = rayT + (v2x(x+1) - gridIntersect.x)/ray.D.x;
		DeltaX = XWidth / ray.D.x;
		StepX = 1;
		OutX = XVoxels;
	}
	else {
		// Handle ray with negative x direction
		NextXCrossing = rayT + (v2x(x) - gridIntersect.x)/ray.D.x;
		DeltaX = - XWidth / ray.D.x;
		StepX = -1;
		OutX = -1;
	}
	// Set up Y stepping
	int y = y2v( gridIntersect.y );
	if (y == YVoxels) y--;
	if (y < 0) ++y;
	Float NextYCrossing, DeltaY;
	int StepY, OutY;
	
	Assert( y >= 0 && y < YVoxels );
	if (fabsf(ray.D.y) < 1e-6) {
		NextYCrossing = INFINITY;
		DeltaY = 0;
		OutY = -1;
	}
	else if (ray.D.y < 0) {
		NextYCrossing = rayT + ( v2y( y ) - gridIntersect.y )/ray.D.y;
		DeltaY = - YWidth / ray.D.y;
		StepY = OutY = -1;
	}
	else {
		NextYCrossing = rayT + ( v2y( y+1 ) - gridIntersect.y )/ray.D.y;
		DeltaY = YWidth / ray.D.y;
		StepY = 1;
		OutY = YVoxels;
	}
	// Set up Z stepping
	int z = z2v( gridIntersect.z );
	if (z == ZVoxels) z--;
	if (z < 0) ++z;
	Float NextZCrossing, DeltaZ;
	int StepZ, OutZ;
	
	Assert( z >= 0 && z < ZVoxels );
	if (fabsf(ray.D.z) < 1e-6) {
		NextZCrossing = INFINITY;
		DeltaZ = 0;
		OutZ = -1;
	}
	else if (ray.D.z < 0) {
		NextZCrossing = rayT + ( v2z( z ) - gridIntersect.z )/ray.D.z;
		DeltaZ = - ZWidth / ray.D.z;
		StepZ = OutZ = -1;
	}
	else {
		NextZCrossing = rayT + ( v2z( z+1 ) - gridIntersect.z )/ray.D.z;
		DeltaZ = ZWidth / ray.D.z;
		StepZ = 1;
		OutZ = ZVoxels;
	}
	// Walk grid for shadow ray
	// Get mailbox id for the ray
	int rayId = curMailboxId++;
	for (;;) {
		int offset = z*XVoxels*YVoxels + y*XVoxels + x;
		Voxel *voxel = voxels[offset];
		if (voxel != NULL) {
			// Check single voxel for shadow ray
			vector<MailboxPrim *> &primitiveList = voxel->primitives;
			// Refine primitives in voxel if needed
			if (!voxel->allCanIntersect) {
				for (u_int i = 0; i < primitiveList.size(); ++i) {
					MailboxPrim *mp = primitiveList[i];
					// Refine primitive if needed
					while (!mp->primitive->CanIntersect()) {
						vector<Reference<Primitive> > p;
						mp->primitive->Refine(p);
						Assert(p.size() > 0);
						if (p.size() == 1)
							mp->primitive = p[0];
						else
							mp->primitive = new GridAccelerator(p);
					}
				}
				voxel->allCanIntersect = true;
			}
			for (u_int i = 0; i < primitiveList.size(); ++i) {
				MailboxPrim *mp = primitiveList[i];
				// Do mailbox check between ray and primitive
				if (mp->lastMailboxId == rayId)
					continue;
				// Check for ray-primitive intersection for shadow ray
				mp->lastMailboxId = rayId;
				rayTests.add(1, 0);
				if (mp->primitive->IntersectP(ray)) {
					rayHits.add(1, 0);
					return true;
				}
			}
		}
		// Advance to next voxel
		if (NextXCrossing < NextYCrossing &&
			NextXCrossing < NextZCrossing) {
			// Step in X
			if (ray.maxt < NextXCrossing)
				break;
			x += StepX;
			if (x == OutX)
				break;
			NextXCrossing += DeltaX;
		}
		else if (NextZCrossing < NextYCrossing) {
			// Step in Z
			if (ray.maxt < NextZCrossing)
				break;
			z += StepZ;
			if (z == OutZ)
				break;
			NextZCrossing += DeltaZ;
		}
		else {
			// Step in Y
			if (ray.maxt < NextYCrossing)
				break;
			y += StepY;
			if (y == OutY)
				break;
			NextYCrossing += DeltaY;
		}
	}
	return false;
}
extern "C" Primitive *CreateAccelerator(const vector<Reference<Primitive> > &prims,
		const ParamSet &ps) {
	return new GridAccelerator(prims);
}
