// kdtree.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "primitives.h"
#include "geometry.h"
// KdTreeAccelerator Declarations
struct KdAccelNode;
struct MailboxPrim;
class KdTreeAccelerator : public Primitive {
public:
	// KdTreeAccelerator Method Declarations
	KdTreeAccelerator(const vector<Reference<Primitive> > &prims);
	~KdTreeAccelerator();
	void buildTree(KdAccelNode **node, const BBox &bounds,
		MailboxPrim *mailboxPrims, const vector<BBox> &primBounds,
		const vector<int> &primNums, int depth);
	bool Intersect(const Ray &ray, Intersection *isect) const;
	BBox WorldBound() const { return bounds; }
	bool CanIntersect() const { return true; }
	bool IntersectP(const Ray &ray) const;
private:
	// KdTreeAccelerator Private Data
	KdAccelNode *root;
	MailboxPrim *mailboxPrims;
	mutable int curMailboxId;
	BBox bounds;
	#define LOWER_NODE_ALLOC_SIZE 2048
	KdAccelNode *topNodes;
	#define MAX_TOP_DEPTH 10
	int topNodeOffset[MAX_TOP_DEPTH];
	KdAccelNode *lowerNodes;
	int lowerNodePos;
	vector<KdAccelNode *> allocedNodeBlocks;
	// KdTreeAccelerator Private Methods
	KdAccelNode *allocNode(int depth, const vector<int> &primNums,
		const BBox &nodeBounds);
	KdAccelNode *allocNode(int depth, int axis, Float split);
	int maxdepth;
	int triggerdepth;
};
struct MailboxPrim {
	MailboxPrim() {
		primitive = NULL;
		lastMailboxId = -1;
	}
	Reference<Primitive> primitive;
	int lastMailboxId;
};
#define SPLIT_X 0
#define SPLIT_Y 1
#define SPLIT_Z 2
struct KdAccelNode {
	// KdAccelNode Constructors
	KdAccelNode(int a, Float s) {
		axis = a;
		isLeaf = 0;
		split = s;
		u.children[0] = u.children[1] = NULL;
	}
	KdAccelNode(MailboxPrim *allPrimitives, const vector<int> &primNums,
			const BBox &nodeBound) {
		nPrimitives = u_int(primNums.size());
		u.primitives = new MailboxPrim *[nPrimitives];
		for (u_int i = 0; i < nPrimitives; ++i)
			u.primitives[i] = &allPrimitives[primNums[i]];
		isLeaf = 1;
	}
	// KdAccelNode Destructor
	~KdAccelNode() {
		if (isLeaf)
			delete[] u.primitives;
	}
	u_int axis:2;
	u_int isLeaf:1;
	u_int nPrimitives:24;
	Float split;
	union {
		MailboxPrim **primitives;
		KdAccelNode *children[2];
	} u;
};
struct KdToDo {
	const KdAccelNode *node;
	Float tmin, tmax;
};
// KdTreeAccelerator Method Definitions
KdTreeAccelerator::KdTreeAccelerator(const vector<Reference<Primitive> > &prims) {
	// Refine all [[prims]] until they are intersectable
	vector<Reference<Primitive> > prefined, todo = prims;
	while (todo.size()) {
		Reference<Primitive> prim = todo.back();
		todo.pop_back();
		if (prim->CanIntersect())
			prefined.push_back(prim);
		else
			prim->Refine(todo);
	}
	// Initialize mailboxes for [[KdTreeAccelerator]]
	curMailboxId = 0;
	mailboxPrims = new MailboxPrim[prefined.size()];
	for (u_int i = 0; i < prefined.size(); ++i)
		mailboxPrims[i].primitive = prefined[i];
	// Set up memory pools for kd tree nodes
	int numTop = (1 << MAX_TOP_DEPTH) - 1;
	topNodes = (KdAccelNode *)AllocL2CacheAligned(numTop * sizeof(KdAccelNode));
	allocedNodeBlocks.push_back(topNodes);
	for (int i = 0; i < MAX_TOP_DEPTH; ++i)
		topNodeOffset[i] = 0;
	lowerNodes = NULL;
	lowerNodePos = LOWER_NODE_ALLOC_SIZE;
	// Build kd tree for accelerator
	vector<BBox> primBounds;
	vector<int> primNums;
	primBounds.reserve(prefined.size());
	primNums.reserve(prefined.size());
	for (u_int i = 0; i < prefined.size(); ++i) {
		BBox b = prefined[i]->WorldBound();
		bounds = Union(bounds, b);
		primBounds.push_back(b);
		primNums.push_back(i);
	}
	maxdepth=0;
	triggerdepth=10;
	if(primNums.size()<20)
		triggerdepth=3;
	buildTree(&root, bounds, mailboxPrims, primBounds, primNums, 0);
	cout << "Created a KdTree, with " << primNums.size() << " primitives and depth=" << maxdepth << ", Bounding Box: " << this->bounds.pMin << " // " <<  this->bounds.pMax  << endl;
}
KdTreeAccelerator::~KdTreeAccelerator() {
	delete[] mailboxPrims;
	// Free memory pools for [[KdTreeNodes]]
	for (u_int i = 0; i < allocedNodeBlocks.size(); ++i)
		FreeCacheAligned(allocedNodeBlocks[i]);
}
void KdTreeAccelerator::buildTree(KdAccelNode **node, const BBox &nodeBounds,
		MailboxPrim *mailboxPrims, const vector<BBox> &allPrimBounds,
		const vector<int> &primNums, int depth) {
	if(maxdepth<depth)
		maxdepth=depth;
	if (!primNums.size()) {
		*node = NULL;
		return;
	}
	// Initialize leaf node if termination criteria met
	if (primNums.size() == 1 ||
		(primNums.size() < 5 && depth > triggerdepth) ||
		depth > 2*triggerdepth) {
		*node = allocNode(depth, primNums, nodeBounds);
		return;
	}
	// Initialize interior node and continue recursion
	// Choose split axis for interior node
	int nextAxis;
	Vector diag = nodeBounds.pMax - nodeBounds.pMin;
	if (diag.x > diag.y && diag.x > diag.z) nextAxis = SPLIT_X;
	else if (diag.y > diag.z) nextAxis = SPLIT_Y;
	else nextAxis = SPLIT_Z;
	// Compute node split position and allocate interior [[KdAccelNode]]
	BBox bounds0 = nodeBounds, bounds1 = nodeBounds;
	Float tsplit;
	if (nextAxis == SPLIT_X) {
		tsplit = Lerp(.5f, nodeBounds.pMin.x, nodeBounds.pMax.x);
		bounds0.pMax.x = bounds1.pMin.x = tsplit;
	}
	else if (nextAxis == SPLIT_Y) {
		tsplit = Lerp(.5f, nodeBounds.pMin.y, nodeBounds.pMax.y);
		bounds0.pMax.y = bounds1.pMin.y = tsplit;
	}
	else {
		tsplit = Lerp(.5f, nodeBounds.pMin.z, nodeBounds.pMax.z);
		bounds0.pMax.z = bounds1.pMin.z = tsplit;
	}
	*node = allocNode(depth, nextAxis, tsplit);
	// Classify primitives with respect to split
	vector<int> prims0, prims1;
	for (u_int i = 0; i < primNums.size(); ++i) {
		int primNum = primNums[i];
		if (bounds0.Overlaps(allPrimBounds[primNum]))
			prims0.push_back(primNum);
		if (bounds1.Overlaps(allPrimBounds[primNum]))
			prims1.push_back(primNum);
	}
	// Recursively initialize children nodes
	buildTree(&((*node)->u.children[0]), bounds0, mailboxPrims, allPrimBounds,
		prims0, depth+1);
	buildTree(&((*node)->u.children[1]), bounds1, mailboxPrims, allPrimBounds,
		prims1, depth+1);
}
KdAccelNode *KdTreeAccelerator::allocNode(int depth, int axis,
		Float split) {
	// Update kd interior node allocation statistics
	static StatsCounter nodesMade("Acceleration",
		"Number of kd-tree interior nodes made");
	++nodesMade;
	// Get pointer to new [[KdAccelNode]]
	KdAccelNode *node;
	if (depth < MAX_TOP_DEPTH) {
		// Allocate kd tree node for top part of tree
		int offset = (1 << depth) - 1 + topNodeOffset[depth];
		Assert(offset < (1 << MAX_TOP_DEPTH) - 1);
		++topNodeOffset[depth];
		node = topNodes + offset;
	}
	else {
		// Allocate kd tree node for bottom part of tree
		if (lowerNodePos == LOWER_NODE_ALLOC_SIZE) {
			lowerNodePos = 0;
			lowerNodes = (KdAccelNode *)AllocL2CacheAligned(
				LOWER_NODE_ALLOC_SIZE*sizeof(KdAccelNode));
			allocedNodeBlocks.push_back(lowerNodes);
		}
		node = lowerNodes + lowerNodePos;
		++lowerNodePos;
	}
	new (node) KdAccelNode(axis, split);
	return node;
}
KdAccelNode *KdTreeAccelerator::allocNode(int depth,
		const vector<int> &primNums, const BBox &nodeBound) {
	// Update kd leaf node allocation statistics
	static StatsCounter numLeafMade("Acceleration", "Number of kd-tree leaf nodes made");
	static StatsCounter maxDepth("Acceleration", "Maximum kd-tree depth");
	static StatsCounter maxLeafPrims("Acceleration",
		"Maximum number of primitives in leaf node");
	++numLeafMade;
	maxDepth.stat_max(depth);
	maxLeafPrims.stat_max((int)primNums.size());
	static StatsRatio leafPrims("Acceleration",
		"Avg. number of primitives in leaf nodes", false);
	leafPrims.add((int)primNums.size(), 1);
	// Get pointer to new [[KdAccelNode]]
	KdAccelNode *node;
	if (depth < MAX_TOP_DEPTH) {
		// Allocate kd tree node for top part of tree
		int offset = (1 << depth) - 1 + topNodeOffset[depth];
		Assert(offset < (1 << MAX_TOP_DEPTH) - 1);
		++topNodeOffset[depth];
		node = topNodes + offset;
	}
	else {
		// Allocate kd tree node for bottom part of tree
		if (lowerNodePos == LOWER_NODE_ALLOC_SIZE) {
			lowerNodePos = 0;
			lowerNodes = (KdAccelNode *)AllocL2CacheAligned(
				LOWER_NODE_ALLOC_SIZE*sizeof(KdAccelNode));
			allocedNodeBlocks.push_back(lowerNodes);
		}
		node = lowerNodes + lowerNodePos;
		++lowerNodePos;
	}
	new (node) KdAccelNode(mailboxPrims, primNums, nodeBound);
	return node;
}
bool KdTreeAccelerator::Intersect(const Ray &ray,
		Intersection *isect) const {
	// Compute initial parametric range of ray inside kd tree extent
	Float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
		return false;
	tmax *= 1.001f;
	// Prepare to traverse kd-tree for ray
	int rayId = curMailboxId++;
	const KdAccelNode *node = root;
	Vector invDir(1.f/ray.D.x, 1.f/ray.D.y, 1.f/ray.D.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;
	bool hit = false;
	while (node != NULL) {
		// Process kd tree node [[node]] for ray traversal
		// Update kd-tree traversal statistics
		static StatsCounter nodesTraversed("Acceleration",
			"Number of kd-tree nodes traversed by normal rays");
		++nodesTraversed;
		// Bail out if we found a hit closer than the current node
		if (ray.maxt < tmin) break;
		if (node->isLeaf) {
			// Check for intersections inside leaf node
			u_int nPrimitives = node->nPrimitives;
			MailboxPrim **prims = node->u.primitives;
			for (u_int i = 0; i < nPrimitives; ++i) {
				MailboxPrim *mp = prims[i];
				if (mp->lastMailboxId != rayId) {
					mp->lastMailboxId = rayId;
					if (mp->primitive->Intersect(ray, isect))
						hit = true;
				}
			}
			// Grab next node to process from todo list
			if (todoPos > 0) {
				--todoPos;
				node = todo[todoPos].node;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
			}
			else
				break;
		}
		else {
			// Process kd tree interior node
			// Compute distance along ray to split plane
			Float tplane;
			if (node->axis == SPLIT_X)
				tplane = (node->split - ray.O.x) * invDir.x;
			else if (node->axis == SPLIT_Y)
				tplane = (node->split - ray.O.y) * invDir.y;
			else
				tplane = (node->split - ray.O.z) * invDir.z;
			// Get near and far child pointers for ray
			KdAccelNode *nearChild, *farChild;
			bool zeroIsNear;
			if (node->axis == SPLIT_X)
				zeroIsNear = (ray.O.x <= node->split);
			else if (node->axis == SPLIT_Y)
				zeroIsNear = (ray.O.y <= node->split);
			else
				zeroIsNear = (ray.O.z <= node->split);
			if (zeroIsNear) {
				nearChild = node->u.children[0];
				farChild  = node->u.children[1];
			}
			else {
				nearChild = node->u.children[1];
				farChild  = node->u.children[0];
			}
			// Advance to next child node, possibly enqueue far child
			if (tplane > tmax || tplane < 0)
				node = nearChild;
			else if (tplane < tmin)
				node = farChild;
			else {
				if (!nearChild) {
					node = farChild;
					tmin = tplane;
				}
				else {
					if (farChild) {
						// Enqueue [[farChild]] in todo list
						todo[todoPos].node = farChild;
						todo[todoPos].tmin = tplane;
						todo[todoPos].tmax = tmax;
						++todoPos;
						Assert(todoPos < MAX_TODO);
					}
					node = nearChild;
					tmax = tplane;
				}
			}
			if (!node) {
				// Grab next node to process from todo list
				if (todoPos > 0) {
					--todoPos;
					node = todo[todoPos].node;
					tmin = todo[todoPos].tmin;
					tmax = todo[todoPos].tmax;
				}
				else
					break;
			}
		}
	}
	return hit;
}
bool KdTreeAccelerator::IntersectP(const Ray &ray) const {
	// Compute initial parametric range of ray inside kd tree extent
	Float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
		return false;
	tmax *= 1.001f;
	// Prepare to traverse kd-tree for ray
	int rayId = curMailboxId++;
	const KdAccelNode *node = root;
	Vector invDir(1.f/ray.D.x, 1.f/ray.D.y, 1.f/ray.D.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;
	while (node != NULL) {
		// Update kd-tree shadow ray traversal statistics
		static StatsCounter nodesTraversed("Acceleration",
			"Number of kd-tree nodes traversed by shadow rays");
		++nodesTraversed;
		if (node->isLeaf) {
			// Check for shadow ray intersections inside leaf node
			u_int nPrimitives = node->nPrimitives;
			MailboxPrim **prims = node->u.primitives;
			for (u_int i = 0; i < nPrimitives; ++i) {
				MailboxPrim *mp = prims[i];
				if (mp->lastMailboxId != rayId) {
					mp->lastMailboxId = rayId;
					if (mp->primitive->IntersectP(ray))
						return true;
				}
			}
			// Grab next node to process from todo list
			if (todoPos > 0) {
				--todoPos;
				node = todo[todoPos].node;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
			}
			else
				break;
		}
		else {
			// Process kd tree interior node
			// Compute distance along ray to split plane
			Float tplane;
			if (node->axis == SPLIT_X)
				tplane = (node->split - ray.O.x) * invDir.x;
			else if (node->axis == SPLIT_Y)
				tplane = (node->split - ray.O.y) * invDir.y;
			else
				tplane = (node->split - ray.O.z) * invDir.z;
			// Get near and far child pointers for ray
			KdAccelNode *nearChild, *farChild;
			bool zeroIsNear;
			if (node->axis == SPLIT_X)
				zeroIsNear = (ray.O.x <= node->split);
			else if (node->axis == SPLIT_Y)
				zeroIsNear = (ray.O.y <= node->split);
			else
				zeroIsNear = (ray.O.z <= node->split);
			if (zeroIsNear) {
				nearChild = node->u.children[0];
				farChild  = node->u.children[1];
			}
			else {
				nearChild = node->u.children[1];
				farChild  = node->u.children[0];
			}
			// Advance to next child node, possibly enqueue far child
			if (tplane > tmax || tplane < 0)
				node = nearChild;
			else if (tplane < tmin)
				node = farChild;
			else {
				if (!nearChild) {
					node = farChild;
					tmin = tplane;
				}
				else {
					if (farChild) {
						// Enqueue [[farChild]] in todo list
						todo[todoPos].node = farChild;
						todo[todoPos].tmin = tplane;
						todo[todoPos].tmax = tmax;
						++todoPos;
						Assert(todoPos < MAX_TODO);
					}
					node = nearChild;
					tmax = tplane;
				}
			}
			if (!node) {
				// Grab next node to process from todo list
				if (todoPos > 0) {
					--todoPos;
					node = todo[todoPos].node;
					tmin = todo[todoPos].tmin;
					tmax = todo[todoPos].tmax;
				}
				else
					break;
			}
		}
	}
	return false;
}
extern "C" Primitive *CreateAccelerator(const vector<Reference<Primitive> > &prims,
		const ParamSet &ps) {
	return new KdTreeAccelerator(prims);
}
