// environment.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "camera.h"
#include "film.h"
#include "paramset.h"
// EnvironmentCamera Declarations
class EnvironmentCamera : public Camera {
public:
	// EnvironmentCamera Method Declarations
	void GenerateRay(const Sample *sample, Ray &) const;
	EnvironmentCamera(const Transform &world2cam, Float hither,
		Float yon, Float sopen, Float sclose, Film *film);
	virtual Float ScreenDepth(const Point &Pworld) const;
private:
	// EnvironmentCamera Private Data
	Point rayOrigin;
};
// EnvironmentCamera Definitions
EnvironmentCamera::EnvironmentCamera(const Transform &world2cam,
		Float hither, Float yon, Float sopen, Float sclose,
		Film *film)
	: Camera(world2cam, hither, yon, sopen, sclose, film) {
	rayOrigin = CameraToWorld(Point(0,0,0));
}
void EnvironmentCamera::GenerateRay(const Sample *sample,
		Ray &ray) const {
	ray.O = rayOrigin;
	// Generate environment camera ray direction
	Float theta = M_PI * sample->imageY / (film->yResolution - 1);
	Float phi = 2 * M_PI * sample->imageX / (film->xResolution - 1);
	Vector dir(sinf(theta) * cosf(phi), cosf(theta),
		sinf(theta) * sinf(phi));
	CameraToWorld(dir, &ray.D);
	// Set ray time value
	ray.time = Lerp(sample->time, ShutterOpen, ShutterClose);
	ray.mint = 0.;
	ray.maxt = INFINITY;
}
Float EnvironmentCamera::ScreenDepth(const Point &Pworld) const {
	return (Distance(Pworld, rayOrigin) - ClipHither) /
		(ClipYon - ClipHither);
}
extern "C" Camera *CreateCamera(const ParamSet &params,
		const Transform &world2cam, Film *film) {
	// Extract common camera parameters from \use{ParamSet}
	Float hither = params.FindOneFloat("_ClipHither", -1);
	Float yon = params.FindOneFloat("_ClipYon", -1);
	Float shutteropen = params.FindOneFloat("_ShutterOpen", -1);
	Float shutterclose = params.FindOneFloat("_ShutterClose", -1);
	Float lensradius = params.FindOneFloat("_LensRadius", -1);
	Float focaldistance = params.FindOneFloat("_FocalDistance", -1);
	Assert(hither != -1 && yon != -1 && shutteropen != -1 &&
		shutterclose != -1 && lensradius != -1 && focaldistance != -1);
	int type, narray;
	const Float *extent = params.FindFloat("_ScreenExtent", &type, &narray);
	Assert(extent && narray == 4);
	Extent2D screen(extent[0], extent[1], extent[2], extent[3]);
	return new EnvironmentCamera(world2cam, hither, yon,
		shutteropen, shutterclose, film);
}
