// orthographic.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "camera.h"
#include "paramset.h"
// OrthographicCamera Declarations
class OrthoCamera : public ProjectiveCamera {
public:
	// OrthoCamera Method Declarations
	OrthoCamera(const Transform &world2cam, const Extent2D &Screen,
		Float hither, Float yon, Float sopen, Float sclose,
		Float lensr, Float focald, Film *film);
	void GenerateRay(const Sample *sample, Ray &) const;
};
// OrthographicCamera Definitions
OrthoCamera::OrthoCamera(const Transform &world2cam,
		const Extent2D &Screen, Float hither, Float yon,
		Float sopen, Float sclose, Float lensr,
		Float focald, Film *f)
	: ProjectiveCamera(world2cam, Orthographic(hither, yon),
		 Screen, hither, yon, sopen, sclose,
		 lensr, focald, f) {
}
void OrthoCamera::GenerateRay(const Sample *sample, Ray &ray) const {
	// Generate raster and camera samples
	Point Pras(sample->imageX, sample->imageY, 0);
	Point Pcamera;
	RasterToCamera(Pras, &Pcamera);
	ray.O = Pcamera;
	ray.D = Vector(0,0,1);
	ray.mint = 0.;
	ray.maxt = INFINITY;
	// Set ray time value
	ray.time = Lerp(sample->time, ShutterOpen, ShutterClose);
	// Modify ray for depth of field
	if (LensRadius > 0.) {
		// Sample point on lens
		Float lu, lv;
		ConcentricSampleDisk(sample->lensX, sample->lensY, &lu, &lv);
		lu *= LensRadius;
		lv *= LensRadius;
		// Compute point on plane of focus
		Float ft = (FocalDistance - ClipHither) / ray.D.z;
		Point Pfocus = ray(ft);
		// Update ray for effect of lens
		ray.O.x += lu;
		ray.O.y += lv;
		ray.D = Pfocus - ray.O;
	}
	CameraToWorld(ray, &ray);
}
extern "C" Camera *CreateCamera(const ParamSet &params,
		const Transform &world2cam, Film *film) {
	// Extract common camera parameters from \use{ParamSet}
	Float hither = params.FindOneFloat("_ClipHither", -1);
	Float yon = params.FindOneFloat("_ClipYon", -1);
	Float shutteropen = params.FindOneFloat("_ShutterOpen", -1);
	Float shutterclose = params.FindOneFloat("_ShutterClose", -1);
	Float lensradius = params.FindOneFloat("_LensRadius", -1);
	Float focaldistance = params.FindOneFloat("_FocalDistance", -1);
	Assert(hither != -1 && yon != -1 && shutteropen != -1 &&
		shutterclose != -1 && lensradius != -1 && focaldistance != -1);
	int type, narray;
	const Float *extent = params.FindFloat("_ScreenExtent", &type, &narray);
	Assert(extent && narray == 4);
	Extent2D screen(extent[0], extent[1], extent[2], extent[3]);
	return new OrthoCamera(world2cam, screen, hither, yon,
		shutteropen, shutterclose, lensradius, focaldistance,
		film);
}
