// realistic.cc

#include "camera.h"
#include "paramset.h"
#include <iostream>
#include <iomanip>
#include <vector>
#include "film.h"

// A lens

struct LensRay {
	Point O;
	Vector D;
};

ostream &operator<<(ostream &os, const LensRay &l) {
	os << l.O << "; " << l.D ;
	return os;
}

class Lens {
public:
	Float radius, radius2;
	Float axepos;
	Float n12;      // from right to left
	Float n12rev;   // from left to right
	Float minz, maxz;
	Float apert, apert2;   // half the aperture actually (radius)
	Float flagrev;
	bool intersect(LensRay &in, LensRay &out) const;
	bool intersectrev(LensRay &ray, LensRay &out) const;
	bool interapert(LensRay &ray) const;
};

// true if the ray passes through a hole of radius z located at z=apert
bool Lens::interapert(LensRay &ray) const {
	Point inter=ray.O+(axepos-ray.O.z)/ray.D.z*ray.D;
	if(inter.x*inter.x+inter.y*inter.y>apert2)
		return false;
	return true;
}

// from right (z-) to left (z+)
bool Lens::intersect(LensRay &ray, LensRay &out) const {
 	Float rayztranslate=ray.O.z-axepos;
	Float A = ray.D.x*ray.D.x + ray.D.y*ray.D.y + ray.D.z*ray.D.z;
	Float B = 2 * (ray.D.x*ray.O.x + ray.D.y*ray.O.y + ray.D.z*rayztranslate);
	Float C = ray.O.x*ray.O.x + ray.O.y*ray.O.y + rayztranslate*rayztranslate - radius2;

	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;

	// We only need to keep the first one that is positive
	if (t0<0) {    // Use t1 for intersection
		out.O = ray.O+ray.D*t1;
		if(out.O.z>maxz || out.O.z<minz)
			return false;
	}
	else {    // Use t0, unless it is actually the wrong side of the sphere
		out.O = ray.O+ray.D*t0;
		if(out.O.z<minz) {
			out.O = ray.O+ray.D*t1;  // in that case, use t1
			if(out.O.z>maxz || out.O.z<minz)
				return false;
		}
		else if (out.O.z>maxz)
			return false;
	}
		
	// The normal is easy in a sphere
	Vector N = Vector(out.O.x/radius,out.O.y/radius,(out.O.z-axepos)/radius);

	// refracted direction
	Float delta,a;   // n1/n2
	Float NdotD=Dot(N,-ray.D);

	delta=1.0-n12*n12*(1.0-NdotD*NdotD);
  // total reflection...
	if(delta<0.0)
		return false; 

	// our linear combination
	a=n12*NdotD-flagrev*sqrtf(delta);
	out.D=a*N+n12*ray.D;
	return true;
}


// from left to right
bool Lens::intersectrev(LensRay &ray, LensRay &out) const {
 	Float rayztranslate=ray.O.z-axepos;
	Float A = ray.D.x*ray.D.x + ray.D.y*ray.D.y + ray.D.z*ray.D.z;
	Float B = 2 * (ray.D.x*ray.O.x + ray.D.y*ray.O.y + ray.D.z*rayztranslate);
	Float C = ray.O.x*ray.O.x + ray.O.y*ray.O.y + rayztranslate*rayztranslate - radius2;

	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;

	// We only need to keep the first one that is positive
	if (t0<0) {    // Use t1 for intersection
		out.O = ray.O+ray.D*t1;
		if(out.O.z>maxz || out.O.z<minz)
			return false;
	}
	else {    // Use t0, unless it is actually the wrong side of the sphere
		out.O = ray.O+ray.D*t0;
		if(out.O.z>maxz) {
			out.O = ray.O+ray.D*t1;  // in that case, use t1
			if(out.O.z>maxz || out.O.z<minz)
				return false;
		}
		else if (out.O.z<minz)
			return false;
	}
			
	Vector N = Vector(out.O.x/radius,out.O.y/radius,(out.O.z-axepos)/radius);

	// refracted direction
	Float delta,a;   // n1/n2
	Float NdotD=Dot(N,-ray.D);

	delta=1.0-n12rev*n12rev*(1.0-NdotD*NdotD);
  // total reflection...
	if(delta<0.0)
		return false;         // Differentiate the kinds of failures ?

	a=n12rev*NdotD+flagrev*sqrtf(delta);
	out.D=a*N+n12rev*ray.D;
	return true;
}

typedef std::vector<Lens> lensList;

ostream &operator<<(ostream &os, const Lens &l) {
	os << "R=" << l.radius << ",\tAxePos=" << l.axepos << ",\tn1/n2=" << l.n12 << ",\tapert/2=" << l.apert;
	os << ",\t " << l.minz << " < z < " << l.maxz;
	return os;
}


// Camera stuff
////////////////////////////////////////////


// Realistic camera

class RealisticCamera : public Camera {
public:
	RealisticCamera(const Transform &world2cam,
									Float hither, Float yon, Float sopen,
									Float sclose, Float focald, Float fstop, string specfile,
									Float filmdiag, Film *film);
	void GenerateRay(const Sample *sample, Ray &) const;
	Float ScreenDepth(const Point &Pworld) const;
	int getsignifline(char *buf, int max, FILE *fp);
	bool getraythrough(LensRay &ray, LensRay &out) const;
	bool getrayhalfthrough(LensRay &ray, LensRay &out) const;
	bool getraythroughspecial(LensRay &ray, LensRay &out) const;
	bool getreverserayhalfthrough(LensRay &ray, LensRay &out) const;
	bool getreverseraythrough(LensRay &ray, LensRay &out) const;
	
private:
	Float zP2, zP1, f2, zfilm;
	Float zExitPupil, radExitPupil; 
	Float filmw, filmh;
	Point rayOrigin;
	int rankapert;
	lensList lenses;
};

// From the film to the world
bool RealisticCamera::getraythrough(LensRay &ray, LensRay &out) const {
	LensRay tmpout;
	// has to be normalized for the formulas to work
	ray.D/=ray.D.Length();
	for(int i=lenses.size()-1;i>=0;i--) {
		if(i==rankapert) {
			if(!lenses[i].interapert(ray))
				return false;
		}
		else if(!lenses[i].intersect(ray,tmpout))
			return false;
		ray=tmpout;
	}
	out=ray;
	return true;
}

// From the film to the world
bool RealisticCamera::getraythroughspecial(LensRay &ray, LensRay &out) const {
	LensRay tmpout;
	// has to be normalized for the formulas to work
	ray.D/=ray.D.Length();
	for(int i=lenses.size()-1;i>=1;i--) {
		if(i==rankapert) {
			if(!lenses[i].interapert(ray))
				return false;
		}
		else if(!lenses[i].intersect(ray,tmpout))
			return false;
		ray=tmpout;
	}
	//	if(!lenses[0].intersectspecial(ray,tmpout))
	//		return false;
	ray=tmpout;

	out=ray;
	return true;
}


// From the film to the aperture
bool RealisticCamera::getrayhalfthrough(LensRay &ray, LensRay &out) const {
	LensRay tmpout;
	// has to be normalized for the formulas to work
	ray.D/=ray.D.Length();
	for(int i=lenses.size()-1;i>rankapert;i--) {
		if(!lenses[i].intersect(ray,tmpout))
			return false;
		ray=tmpout;
	}
	out=ray;
	return true;
}

// From the aperture to the film
bool RealisticCamera::getreverserayhalfthrough(LensRay &ray, LensRay &out) const {
	LensRay tmpout;
	// has to be normalized for the formulas to work
	ray.D/=ray.D.Length();
	for(int i=rankapert+1; i<(int)lenses.size();i++) {
		if(!lenses[i].intersectrev(ray,tmpout))
			return false;
		ray=tmpout;
	}
	out=ray;
	return true;
}

// From the world to the film
bool RealisticCamera::getreverseraythrough(LensRay &ray, LensRay &out) const {
	LensRay tmpout;
	// has to be normalized for the formulas to work
	ray.D/=ray.D.Length();
	for(int i=0;i<(int)lenses.size();i++) {
		//		cout << i << ": " << ray << endl;
		if(i==rankapert) {
			if(!lenses[i].interapert(ray))
				return false;
		}
		else if(!lenses[i].intersectrev(ray,tmpout))
			return false;
		ray=tmpout;
	}
	out=ray;
	return true;
}


// Read one line into buf, of size max, returns the number of bytes read
int RealisticCamera::getsignifline(char *buf, int max, FILE *fp){
  char tmp[2];
  int count=0;
    int num_scanned;

  // scan in a line from script and process it
  while(fgets(buf,max,fp)!=NULL){
    count++;
    // skip blank lines
    num_scanned=sscanf(buf,"%1s",tmp);
    if(num_scanned != 1)
      continue;
    // skip comments
    if(buf[0]=='#')
      continue;
    return count;
  }
  return -1;
}


#define MAXLINESIZE 1024

RealisticCamera::RealisticCamera(const Transform &world2cam,
																 Float hither, Float yon, Float sopen, Float sclose, 
																 Float focald, Float fstop, string specfile, 
																 Float filmdiag, Film *f)
	: Camera(world2cam, hither, yon, sopen, sclose, f) 
{
	char buf[MAXLINESIZE];
	Float currentpos=0.0f;
	Float thickness;
	lenses.clear();

	rayOrigin = CameraToWorld(Point(0,0,0));
	
	cout << "New camera: focald=" << focald << ", fstop=" << fstop << 
		",filmdiag=" << filmdiag << endl;


	// Lens file
	cout << "Opening lens spec file: " << specfile << endl;	
	FILE *fd = fopen(specfile.c_str(),"r");
  if(NULL==fd)
    Severe("Could not open lens file");

	Lens lens;
	Float n,n1,n2,savedn=1.0f;
	while(getsignifline(buf,MAXLINESIZE,fd)!=-1) {
		if(sscanf(buf,"%f %f %f %f",&lens.radius, &thickness, &n, &lens.apert)<4)
			Severe("Could not read your lens file...");

		currentpos-=thickness;
		lens.axepos=currentpos-lens.radius;
		//		lens.minz=currentpos-lens.radius/2.0f;
		//		lens.maxz=currentpos+lens.radius/2.0f;
		lens.apert/=2;    // we store the radius and not the diameter
		lens.apert2=lens.apert*lens.apert;    // square
		lens.radius2=lens.radius*lens.radius;

		if(lens.radius>=0) {   // Put the right ior in the right place
			n1=savedn;   // outside
			n2=n;        // inside
			lens.n12=n2/n1;    // right to left
			lens.n12rev=n1/n2; // left to right
			lens.flagrev=-1;   // reversed
			lens.maxz=currentpos;
			lens.minz=(lens.axepos+sqrtf(lens.radius2-lens.apert2));
		}
		else {
			n1=n;
			n2=savedn;
			lens.n12=n1/n2;
			lens.n12rev=n2/n1;
			lens.flagrev=1;
			lens.radius=-lens.radius;
			lens.minz=currentpos;
			lens.maxz=lens.axepos-sqrtf(lens.radius2-lens.apert2);
		}
		if(lens.radius != 0)
			savedn=n;

		// detecting the aperture
		if(lens.radius==0.0f) {
			lens.minz=currentpos;
			lens.maxz=currentpos;
			cout << "Physical aperture:  " << lens << endl;
			rankapert=lenses.size();
		}
		else
			cout << "Lens:  " << lens << endl;
		lenses.push_back(lens);
	}
  fclose(fd);
	

	
	Float distance;    // distance to the origin axis
	LensRay light,focal;

	Float tmp;
	Float accumzP2=0;
	Float accumzP1=0;
	Float accumf2=0;
	Float compteur1=0.0f;
	Float compteur2=0.0f;
	for(int h=1;h<150;h++) {    // Accumulate statistics over 150 rays
		distance=lenses[0].apert*h/400;    // distance to the origin axis
		light.D=Vector(0.0,0.0,-1.0);
		light.O=Point(0.0,distance,10.0);
		if(getreverseraythrough(light,focal)) {
			tmp=focal.O.z+(distance-focal.O.y)/focal.D.y*focal.D.z;
			accumzP2 +=tmp;
			accumf2 +=(focal.O.z-focal.O.y/focal.D.y*focal.D.z)-tmp;
			compteur2 += 1.0;
		}
		light.D=Vector(0.0,0.0,1.0);
		light.O=Point(0.0,distance,currentpos-10.0);
		if(getraythrough(light,focal)) {
			accumzP1 +=focal.O.z+(distance-focal.O.y)/focal.D.y*focal.D.z;
			compteur1 += 1.0;
		}			
	}
	f2=accumf2/compteur2;
  zP2=accumzP2/compteur2;
  zP1=accumzP1/compteur1;

	// Film distance
	zfilm=1/(1/(focald-zP1)+1/f2)+zP2;
	/*
	cout << "Principal plane 1: " << zP1 << ", 2: " << zP2 << "; fprime: " <<
		f2 << ", focald: "<<focald << "; zfilm: " << zfilm <<endl;
	*/
	light.D=Vector(0.0,0.05,1.0);
	light.O=Point(0.0,0.0,zfilm);
	getraythrough(light,focal);
	//	cout << "Checking the focal distance: " << -focal.O.y/focal.D.y*focal.D.z+focal.O.z << endl;
	
	// Film settings
	Float delta=(Float)film->yResolution/(Float)film->xResolution;
	filmw=filmdiag/sqrtf(1+delta*delta);
	filmh=delta*filmw;

	// Physical aperture
	Float stheta=1/(2*fstop);
	light.D=Vector(0.0,stheta,sqrtf(1-stheta*stheta));
	light.O=Point(0.0,0.0,f2+zP2);
	getrayhalfthrough(light,focal);
	Point inter=focal.O+(lenses[rankapert].axepos-focal.O.z)/focal.D.z*focal.D;
	lenses[rankapert].apert=sqrtf(inter.x*inter.x+inter.y*inter.y);	
	/*
	cout << "Film: " << filmw << " x " << filmh << "; aperture/2: " <<
		lenses[rankapert].apert << ", z=" << inter.z << endl;
	*/

	// arbitrary, does not matter too much
	distance=1.0f;

	// Exit Pupil
	light.D=Vector(0.0,0.0,-1.0);
	light.O=Point(0.0,distance,10.0);
	Float zEPP2, zEPP1, EPf2;
	getreverserayhalfthrough(light,focal);
	zEPP2=focal.O.z+(distance-focal.O.y)/focal.D.y*focal.D.z;
	EPf2=(focal.O.z-focal.O.y/focal.D.y*focal.D.z)-zEPP2;
	light.D=Vector(0.0,0.0,1.0);
	light.O=Point(0.0,distance,currentpos-10.0);
	getrayhalfthrough(light,focal);
	zEPP1=focal.O.z+(distance-focal.O.y)/focal.D.y*focal.D.z;
	
 	Float t=zEPP2-zEPP1;
	Float wprime=(inter.z-zEPP1)/EPf2+1;
	radExitPupil=lenses[rankapert].apert/wprime;
	zExitPupil=zEPP1+((1+t/EPf2)*(inter.z-zEPP1)+t)/wprime;

		
	light.D=Vector(0.0,0.05,1.0);
	light.O=Point(0.0,0.0,zEPP2+EPf2);
	getrayhalfthrough(light,focal);
	cout << "Checking the focal distance: " << focal.D << endl;
	cout << t << " " << wprime <<endl;

	cout << "EPP1: " << zEPP1 << ", 2: " << zEPP2 << "; EPfprime: " <<
		EPf2 << endl;
	cout << "Exit Pupil - radius: " << radExitPupil << ", z: " << zExitPupil << endl;	
	
	//	exit(0);
}

void RealisticCamera::GenerateRay(const Sample *sample, Ray &ray) const {
	// First: get the ray direction
	Float lu, lv;
	ConcentricSampleDisk(sample->lensX, sample->lensY, &lu, &lv);

	LensRay initray,outray;
	initray.O.x=(sample->imageX/film->xResolution-0.5)*filmw;
	initray.O.y=(sample->imageY/film->yResolution-0.5)*filmh;
	initray.O.z=zfilm;
	/*
	initray.D.x=lu*lenses[rankapert].apert-initray.O.x;
	initray.D.y=lv*lenses[rankapert].apert-initray.O.y;
	initray.D.z=lenses[rankapert].axepos-zfilm;
	*/
	initray.D.x=lu*radExitPupil-initray.O.x;
	initray.D.y=lv*radExitPupil-initray.O.y;
	initray.D.z=zExitPupil-zfilm;

		
	if(getraythrough(initray,outray)) {
		ray.mint = 0.;
		ray.maxt = INFINITY;
		ray.O=outray.O;
		ray.D=outray.D;
	}
	else {
		ray.mint = 2.0f;
		ray.maxt = 1.0f;
		ray.D = Vector(0,0,1);
	}
	
	// Set ray time value
	ray.time = Lerp(sample->time, ShutterOpen, ShutterClose);
	CameraToWorld(ray, &ray);
}

Float RealisticCamera::ScreenDepth(const Point &Pworld) const {
	// This is fine for this assignment.
	return (Distance(Pworld, rayOrigin) - ClipHither) /
		(ClipYon - ClipHither);
}

extern "C" Camera *CreateCamera(const ParamSet &params,
																const Transform &world2cam, Film *film) {
	// Extract common camera parameters from \use{ParamSet}
	Float hither = params.FindOneFloat("_ClipHither", -1);
	Float yon = params.FindOneFloat("_ClipYon", -1);
	Float shutteropen = params.FindOneFloat("_ShutterOpen", -1);
	Float shutterclose = params.FindOneFloat("_ShutterClose", -1);
	
	// Realistic camera-specific parameters
	string specfile = params.FindOneString("specfile", "");
	Float focaldistance = params.FindOneFloat("focaldistance", 100.0);
 	Float fstop = params.FindOneFloat("fstop", 1.0);	
	Float filmdiag = params.FindOneFloat("filmdiag", 35.0);
	
	Assert(hither != -1 && yon != -1 && shutteropen != -1 &&
				 shutterclose != -1 && focaldistance != -1);
	if (specfile == "") {
		Severe( "No spec file specified! Did you forget\nto "
						"declare \"specfile\" as \"string\" at the top of your RIB file?" );
	}
	return new RealisticCamera(world2cam, hither, yon,
														 shutteropen, shutterclose, focaldistance, fstop, 
														 specfile, filmdiag, film);
}
