// scene.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "scene.h"
#include "camera.h"
#include "film.h"
#include "sampling.h"
#include "dynload.h"
#include "transport.h"
#include "volume.h"
#include "materials.h"

#include "translucentmap.h"

int debugflag=0;
bool isCreateBlob;
void createBlob(Scene* scene);

// Scene Methods
Scene::~Scene() {
	delete camera;
	delete sampler;
	delete surfaceIntegrator;
	delete volumeIntegrator;
	delete prims;
	delete displayInfo;
	for (u_int i = 0; i < lights.size(); ++i)
		delete lights[i];
	for (u_int i = 0; i < volumeRegions.size(); ++i)
		delete volumeRegions[i];
}
Scene::Scene(Camera *cam, SurfaceIntegrator *si, VolumeIntegrator *vi,
		Sampler *s, Primitive *accel, const vector<Light *> &lts,
		const vector<VolumeRegion *> &vrs, DisplayInfo *di) {
	lights = lts;
	prims = accel;
	camera = cam;
	sampler = s;
	surfaceIntegrator = si;
	volumeIntegrator = vi;
	volumeRegions = vrs;
	displayInfo = di;
	if (lts.size() == 0)
		Warning("No light sources defined in scene; "
			"possibly rendering a black image.");
	bound = prims->WorldBound();

}
void Scene::Render() {
  //create translucent map
  createTranslucentmap(this);
	// Prepare to render image
	Sample *sample = new Sample(surfaceIntegrator, volumeIntegrator, this);
	surfaceIntegrator->Preprocess(this);
	volumeIntegrator->Preprocess(this);
	ProgressReporter progress(sampler->TotalSamples(),
		"Rendering");

	// BLOBS
  if(isCreateBlob == true) {
		cout<<"Blobs being created!"<<endl;
		createBlob(this);
		return;
	}

	// GLOWS
	// Precalculate each light for how much glow they create
	// Precalculate which light actually glow (visible)
	Ray raytmp;
	int count=0;
	for (u_int i = 0; i<lights.size(); ++i) {
		Light *light = lights[i];
		if(light->doGlow) {
			raytmp.O = Point(0,0,0);
			raytmp.D = Vector(0,0,1);
			raytmp.mint = 0.;
			raytmp.maxt = 1.;
			camera->CameraToWorld(raytmp, &raytmp);
			raytmp.D = (light->copyPos-raytmp.O).Hat();
			raytmp.maxt = (light->copyPos-raytmp.O).Length();
			if(IntersectP(raytmp))
				light->doGlow=false;
			else
				count++;
		}
	}
	cout << count << " glowing lights" << endl;

#ifdef ADDSKY
	string tex ("projet/sky3.tif");
	ParamSet geomParams;
	ParamSet surfaceParams;
	SURF_TEX_F(Kd, 1.);
	Ksky = new ScaleTexture<Float, Spectrum>(Kd, new ImageTexture<Spectrum>(new IdentityMapping2D(), tex));
#endif
	
	while (sampler->GetNextSample(sample)) {
		// Compute result for image sample
		// Report rendering progress
		static StatsCounter cameraRaysTraced("Camera",
			"Camera Rays Traced");
		++cameraRaysTraced;
		progress.Update(1);
		// Compute camera ray
		RayDifferential ray;
		camera->GenerateRay(sample, ray);
		// Generate ray differentials for camera ray
		++sample->imageX;
		camera->GenerateRay(sample, ray.rx);
		--sample->imageX;
		++sample->imageY;
		camera->GenerateRay(sample, ray.ry);
		ray.hasDifferentials = true;
		// Evaluate radiance along ray
		Float alpha;

		debugflag=0;
		Float zx=117; Float zy=87;
		//				Float zx=112; Float zy=88;
				
		if(sample->imageX>=zx-0.5 && sample->imageY>=zy-0.5
		  && sample->imageX<=zx+0.5 && sample->imageY<=zy+0.5) {
									debugflag=1;
						cout << "\n" << sample->imageX <<", "<< sample->imageY << "\n"<<endl;
		}
		
		zx=122; zy=77;
		//				Float zx=112; Float zy=88;
		if(sample->imageX>=zx-0.5 && sample->imageY>=zy-0.5
		  && sample->imageX<=zx+0.5 && sample->imageY<=zy+0.5) {
						debugflag=1;
						cout << "\n" << sample->imageX <<", "<< sample->imageY << "\n"<<endl;
		}


		
		zx=180; zy=222;
		//				Float zx=112; Float zy=88;
		if(sample->imageX>=zx-0.5 && sample->imageY>=zy-0.5
		  && sample->imageX<=zx+0.5 && sample->imageY<=zy+0.5) {
									debugflag=1;
						cout << "\n" << sample->imageX <<", "<< sample->imageY << "\n"<<endl;
		}

		//		cout << "\n" << sample->imageX <<", "<< sample->imageY << " -- ";
		Spectrum Ls = L(ray, sample, &alpha);
		if(debugflag==1)
			cout << " ->  " << Ls*256 << "\n";

		if(!(Ls.y()>0 && Ls.y() <1000000000))
			Ls=Spectrum(0.);

		Spectrum Lglow(0.0);
		for (u_int i = 0; i<lights.size(); ++i) {
			Light *light = lights[i];
			if(light->doGlow)
				Lglow += light->Glow(ray);
		}
		alpha=1;
		Ls+=Lglow;

		// Add sample to image
		Float screenz = camera->ScreenDepth(ray(ray.maxt));
		Point Praster(sample->imageX, sample->imageY, screenz);
		sampler->AddSample(camera->film, Praster, Ls, alpha);
	}
	delete sample;
	fprintf(stderr, "\n");
	camera->film->WriteDisplayImage(*displayInfo);
}
Spectrum Scene::L(const RayDifferential &ray,
		const Sample *sample, Float *alpha) const {
	int oldflag=debugflag;
	Spectrum Ls = surfaceIntegrator->L(this, ray,
		sample, alpha);
	debugflag=0;
 	Spectrum T = volumeIntegrator->Transmittance(this, ray,
		sample, alpha);
	Spectrum Lv = volumeIntegrator->L(this, ray,
		sample, alpha);
	debugflag=oldflag;
	return T * Ls + Lv;
}
Spectrum Scene::Transmittance(const Ray &ray) const {
	return volumeIntegrator->Transmittance(this, ray,
		NULL, NULL);
}


#include "shapes/blob.h"
#include <vector>

// Aligning u1 on u2
Transform alignaxis(Vector u1, Vector u2) {
	u1=u1.Hat();
	u2=u2.Hat();
	Vector rotaxis=Cross(u1,u2);
	if(rotaxis.Length()>0) {
		return Rotate(180.0f/M_PI*asin(rotaxis.Length()),rotaxis);
	}
	return Scale(1,1,1);
}

void createBlob(Scene* scene)
{
  cout<<"In create Blob"<<endl;
  
	float Thickness = 1;
	//  float MinHeight = 0;
	//  float MaxHeight = 1.1;
	//  float threshold = 0.6;
	char filename[512];
  
  std::vector<Reference<BLOB_ELEMENT> > elementsVec;
 
	
	// Snow bank
	int which=(int)snowparam/10;
	int layer=snowparam%10;
	int Particles;
	float size;
	float mult;
	float translatez;
	Point MinPt,MaxPt;
	Vector Direction;
	bool issphere=false;

	cout << "Blob " << which << ", layer " << layer <<endl;
	if(which==0) {   // Snowbank
		Direction=Vector(0,-1,0.4);
		Direction = Direction.Hat();
		MinPt=Point(-2400,-40,0);
		MaxPt=Point(200,40,850);
		translatez=0;
		if(layer==0) {
			Particles = 2000;
			size = 15;
			mult= 0.7;
			sprintf(filename, "projet/snowcylinder3.blob");
		}
		else {
			Particles = 2000;
			size = 10;
			mult= 1.0;
			//			issphere=true;
			sprintf(filename, "projet/snowcylinder32.blob");
		}
	}
	else if (which==1){    // Tree 1
		Direction=Vector(0.4,-1,.2);
		Direction = Direction.Hat();
		Particles = 3000;
		size = 6;
		mult= 0.6;
		translatez=0;
		MinPt=Point(-1040-460,0-10,-650+460);
		MaxPt=Point(240-460,700+10,200+460);
		sprintf(filename, "projet/snowtree1.blob");
	}
	else {
		Direction=Vector(0.4,-1,.2);
		Direction = Direction.Hat();
		translatez=0;
		Particles = 4000;
		size = 3;//8;
		mult= 1;//0.8;
		if(which==2) {  // Tree 2
			MinPt=Point(-840-300,0,-550+1000);
			MaxPt=Point(240-300,600,200+1000);
			sprintf(filename, "projet/snowtree2.blob");
		} else if (which==3) { // Tree 3
			MinPt=Point(-840-650,-10+20,-550+950);
			MaxPt=Point(240-650,650+20,200+950);
			sprintf(filename, "projet/snowtree32.blob");
		} else if(which==4) {  // Tree 4
			MinPt=Point(-850-1000,-10+10,-540+800);
			MaxPt=Point(240-1000,650+10,200+800);
			sprintf(filename, "projet/snowtree4.blob");
		} else if(which==5) {	 //Tree 5
			MinPt=Point(-840-1250,-10+20,-550+650);
			MaxPt=Point(240-1250,600+20,200+650);
			sprintf(filename, "projet/snowtree5.blob");
		} else if (which==6) { // Tree6
			MinPt=Point(-860-1450,-10+30,-560+500);
			MaxPt=Point(260-1450,750+30,260+500);
			sprintf(filename, "projet/snowtree6.blob");
		}
	}
  Intersection isect;
  Ray ray;
  ray.D = Direction;

	Transform transfo;
	
  int i=0;
  while(i<Particles/2)  {
		ray.O.x = MinPt.x + RandomFloat(0.0f,1.0f)*(MaxPt.x-MinPt.x);
		ray.O.y = MaxPt.y + 100;//(MinHeight+RandomFloat(0.0f,1.0f)*(MaxHeight-MinHeight))*Max.y;
		ray.O.z = MinPt.z + RandomFloat(0.0f,1.0f)*(MaxPt.z-MinPt.z);
		
		ray.mint = 0;
		ray.maxt = 100000;
		
		Normal norm;
		
		if(scene->Intersect(ray, &isect)){
			norm = isect.dgGeom.Nn;
			if(norm.Length()>0) {
				float dot   = min(1.0f, Dot(norm.Hat(),-Direction));
				float Angle = pow((M_PI-fabs(acos(dot)))/M_PI,2);
				
				Point FlakeSize =	Angle*Point(1+RandomFloat(0,1)*1.3,(Angle*0.5+RandomFloat(0,1))*Thickness,1+RandomFloat(0,1)*1.3);
				
				/*	      cout<<FlakeSize.x<<"  ";
									cout<<FlakeSize.y<<"  ";
									cout<<FlakeSize.z<<endl;
				*/
				//				FlakeSize.y=(FlakeSize.y+1)/2;
				if(issphere)
					FlakeSize = Point(1,1,1);
				BLOB_ELEMENT* element = new BLOB_ELEMENT(Point(isect.dgGeom.P.x,isect.dgGeom.P.y,isect.dgGeom.P.z+translatez) ,2*size ,1);
				//				element->scaleTrans = FlakeSize;
				
				Transform t=alignaxis(Vector(0,1,0),Vector(norm));
				element->transfo=t*Scale(FlakeSize.x,FlakeSize.y,FlakeSize.z);
				elementsVec.push_back(element);
				i++;
			}
		}
	}  
	cout <<endl;
	
	while(i<Particles) {
		ray.O.x = MinPt.x + RandomFloat(0.0f,1.0f)*(MaxPt.x-MinPt.x);
		ray.O.y = MaxPt.y + 100;//(MinHeight+RandomFloat(0.0f,1.0f)*(MaxHeight-MinHeight))*Max.y;
		ray.O.z = MinPt.z + RandomFloat(0.0f,1.0f)*(MaxPt.z-MinPt.z);
		
		ray.mint = 0;
		ray.maxt = 100000;
		
		Normal norm;
		
		if(scene->Intersect(ray, &isect))
			{
				norm = isect.dgGeom.Nn;
				if(norm.Length()>0)
					{
						float dot = min(1.0f, Dot(norm.Hat(),-Direction));
						float Angle = pow((M_PI-fabs(acos(dot)))/M_PI,2);
						
						Point FlakeSize =	Angle*Point(1+RandomFloat(0,1)*1.3,(Angle*0.5+RandomFloat(0,1))*Thickness,1+RandomFloat(0,1)*1.3);

						FlakeSize.y=(FlakeSize.y+1)/2;
						if(issphere)
							FlakeSize = Point(1,1,1);
						
						BLOB_ELEMENT* element = new BLOB_ELEMENT(Point(isect.dgGeom.P.x,isect.dgGeom.P.y,isect.dgGeom.P.z+translatez) ,mult*2*size ,1);

						Transform t=alignaxis(Vector(0,1,0),Vector(norm));
						element->transfo=t*Scale(FlakeSize.x,FlakeSize.y,FlakeSize.z);
						elementsVec.push_back(element);
						i++;
					}
			}
	}  

	FILE* file = fopen(filename,"w");
	fprintf(file,"%d\n",elementsVec.size());
	for(unsigned int i=0; i<elementsVec.size(); i++)
		{
			Matrix4x4 *m=elementsVec[i]->transfo.m.ptr;
			fprintf(file,"%f %f %f %f %f %f %f %f %f %f %f %f %f\n",
							elementsVec[i]->center.x,elementsVec[i]->center.y,elementsVec[i]->center.z,elementsVec[i]->rad,
							m->m[0][0],m->m[0][1],m->m[0][2],
							m->m[1][0],m->m[1][1],m->m[1][2],
							m->m[2][0],m->m[2][1],m->m[2][2]);
		}
	fclose(file);
}

