#ifndef _BLOBHELPER
#define _BLOBHELPER

#include <iostream.h>
#include "lrt.h"
#include "shapes.h"

#include<vector>
#include<list>

#include "blob_element.h"

const int MAX_NUM_COMPONENTS = 10;
const float DEPTH_TOLERANCE = 1.0e-5;
const float Small_Tolerance = 0.001;
const float Max_Distance = 1.0e7;

struct MyKdAccelNode;
struct MailboxPrim;

class MyKdTreeAccelerator : public BLOB_ELEMENT 
{
public:
 // MyKdTreeAccelerator Method Declarations
  MyKdTreeAccelerator(const vector<Reference<BLOB_ELEMENT> > &prims);
  ~MyKdTreeAccelerator();

  void buildTree(MyKdAccelNode **node, const BBox &bounds,
		 MailboxPrim *mailboxPrims, const vector<BBox> &primBounds,
		 const vector<int> &primNums, int depth);
  
  bool Intersect(const Ray &ray,Intersection *isect) const;

  bool IntersectNew(const Ray &ray, std::vector<InterInfo>& intersections) const;

  BBox WorldBound() const { return bounds; }
  bool CanIntersect() const { return true; }
  bool IntersectP(const Ray &ray) const;
private:
  // MyKdTreeAccelerator Private Data
  MyKdAccelNode *root;
  MailboxPrim *mailboxPrims;
  mutable int curMailboxId;
  BBox bounds;
#define LOWER_NODE_ALLOC_SIZE 2048
  MyKdAccelNode *topNodes;
#define MAX_TOP_DEPTH 10
  int topNodeOffset[MAX_TOP_DEPTH];
  MyKdAccelNode *lowerNodes;
  int lowerNodePos;
  vector<MyKdAccelNode *> allocedNodeBlocks;
  // MyKdTreeAccelerator Private Methods
  MyKdAccelNode *allocNode(int depth, const vector<int> &primNums,
			 const BBox &nodeBounds);
  MyKdAccelNode *allocNode(int depth, int axis, Float split);
};

class Blob: public Shape 
{
 public:
  Blob(const Transform &o2w, const BBox &ob);
  ~Blob();
  
  BBox ObjectBound() const;
  bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
  bool IntersectP(const Ray &ray) const;
  Float Area() const;
  
  int _noc;

  mutable std::vector<Reference<BLOB_ELEMENT> > elements;
  mutable BLOB_INTERVAL* intervals;
 
  float _threshold;

 private:
  
  void writeBlob(char* filename);
  bool readBlob(char* filename);

  BBox objBound;
  void MakeBlob();
  
  bool test(Ray& ray, Point& Phit, Normal& Nhit, float* thit) const;

 private:
  void element_normal(Normal& Result, Point P, int ElementIndex) const;
  void Blob_Normal(Normal& Result, Point IPoint) const;
  bool All_Blob_Intersections(Ray& ray, Point& Ipt, float& dist) const;
  int findIntervals(Ray& ray) const;
  void insert_hit(int ElementIndex, float t0, float  t1, int* cnt)const;
  bool intersect_element(int ElementIndex, Ray& ray, float mindist, float  *tmin, float  *tmax)const;
  
  MyKdTreeAccelerator* kdtree;
};


Vector HAT(Vector& v)
{
  float t = 1/sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
  return Vector(v.x *t, v.y *t, v.z *t);
}

void memmove (void *dest, void  *src, size_t length);

#include "lrt.h"
#include "geometry.h"
#include "blob_element.h"

#define SPLIT_X 0
#define SPLIT_Y 1
#define SPLIT_Z 2

struct MailboxPrim 
{
  MailboxPrim() {
    primitive = NULL;
    lastMailboxId = -1;
  }
  Reference<BLOB_ELEMENT> primitive;
  int lastMailboxId;
};

struct MyKdAccelNode
{
  // MyKdAccelNode Constructors
  MyKdAccelNode(int a, Float s)
  {
    axis = a;
    isLeaf = 0;
    split = s;
    u.children[0] = u.children[1] = NULL;
  }
  MyKdAccelNode(MailboxPrim *allPrimitives, const vector<int> &primNums,
	      const BBox &nodeBound) 
  {
    nPrimitives = u_int(primNums.size());
    u.primitives = new MailboxPrim *[nPrimitives];
    for (u_int i = 0; i < nPrimitives; ++i)
      u.primitives[i] = &allPrimitives[primNums[i]];
    isLeaf = 1;
  }
  // MyKdAccelNode Destructor
  ~MyKdAccelNode()
  {
    if (isLeaf)
      delete[] u.primitives;
  }
  u_int axis:2;
  u_int isLeaf:1;
  u_int nPrimitives:24;
  Float split;
  union
  {
    MailboxPrim **primitives;
    MyKdAccelNode *children[2];
  } u;
};
struct MyKdToDo
{
  const MyKdAccelNode *node;
  Float tmin, tmax;
};



MyKdAccelNode *MyKdTreeAccelerator::allocNode(int depth, int axis, Float split) 
{
  // Update kd interior node allocation statistics
  static StatsCounter nodesMade("Acceleration",
				"Number of kd-tree interior nodes made");
  ++nodesMade;
  // Get pointer to new [[MyKdAccelNode]]
  MyKdAccelNode *node;
  if (depth < MAX_TOP_DEPTH)
    {
      // Allocate kd tree node for top part of tree
      int offset = (1 << depth) - 1 + topNodeOffset[depth];
      Assert(offset < (1 << MAX_TOP_DEPTH) - 1);
      ++topNodeOffset[depth];
      node = topNodes + offset;
    }
  else
    {
      // Allocate kd tree node for bottom part of tree
      if (lowerNodePos == LOWER_NODE_ALLOC_SIZE)
	{
	  lowerNodePos = 0;
	  lowerNodes = (MyKdAccelNode *)AllocL2CacheAligned(LOWER_NODE_ALLOC_SIZE*sizeof(MyKdAccelNode));
	  allocedNodeBlocks.push_back(lowerNodes);
	}
      node = lowerNodes + lowerNodePos;
      ++lowerNodePos;
    }
  new (node) MyKdAccelNode(axis, split);
  return node;
}

MyKdAccelNode *MyKdTreeAccelerator::allocNode(int depth,
		const vector<int> &primNums, const BBox &nodeBound) 
{
  // Update kd leaf node allocation statistics
  static StatsCounter numLeafMade("Acceleration", "Number of kd-tree leaf nodes made");
  static StatsCounter maxDepth("Acceleration", "Maximum kd-tree depth");
  static StatsCounter maxLeafPrims("Acceleration",
				   "Maximum number of primitives in leaf node");
  ++numLeafMade;
  maxDepth.stat_max(depth);
  maxLeafPrims.stat_max((int)primNums.size());
  static StatsRatio leafPrims("Acceleration",
			      "Avg. number of primitives in leaf nodes", false);
  leafPrims.add((int)primNums.size(), 1);
  // Get pointer to new [[MyKdAccelNode]]
  MyKdAccelNode *node;
  if (depth < MAX_TOP_DEPTH)
    {
      // Allocate kd tree node for top part of tree
      int offset = (1 << depth) - 1 + topNodeOffset[depth];
      Assert(offset < (1 << MAX_TOP_DEPTH) - 1);
      ++topNodeOffset[depth];
      node = topNodes + offset;
    }
  else 
    {
      // Allocate kd tree node for bottom part of tree
      if (lowerNodePos == LOWER_NODE_ALLOC_SIZE) 
	{
	  lowerNodePos = 0;
	  lowerNodes = (MyKdAccelNode *)AllocL2CacheAligned(LOWER_NODE_ALLOC_SIZE*sizeof(MyKdAccelNode));
	  allocedNodeBlocks.push_back(lowerNodes);
	}
      node = lowerNodes + lowerNodePos;
      ++lowerNodePos;
    }
  new (node) MyKdAccelNode(mailboxPrims, primNums, nodeBound);
  return node;
}

bool MyKdTreeAccelerator::Intersect(const Ray &ray,Intersection *isect) const
{
  // Compute initial parametric range of ray inside kd tree extent
  Float tmin, tmax;
  if (!bounds.IntersectP(ray, &tmin, &tmax))
    return false;
  tmax *= 1.001f;
  // Prepare to traverse kd-tree for ray
  int rayId = curMailboxId++;
  const MyKdAccelNode *node = root;
  Vector invDir(1.f/ray.D.x, 1.f/ray.D.y, 1.f/ray.D.z);
#define MAX_TODO 64
  MyKdToDo todo[MAX_TODO];
  int todoPos = 0;
  bool hit = false;
  while (node != NULL)
    {
      // Process kd tree node [[node]] for ray traversal
      // Update kd-tree traversal statistics
      static StatsCounter nodesTraversed("Acceleration",
					 "Number of kd-tree nodes traversed by normal rays");
      ++nodesTraversed;
      // Bail out if we found a hit closer than the current node
      if (ray.maxt < tmin) break;
 
     if (node->isLeaf) 
	{
	  // Check for intersections inside leaf node
	  u_int nPrimitives = node->nPrimitives;
	  MailboxPrim **prims = node->u.primitives;
	  for (u_int i = 0; i < nPrimitives; ++i)
	    {
	      MailboxPrim *mp = prims[i];
	      if (mp->lastMailboxId != rayId) 
		{
		  mp->lastMailboxId = rayId;
		  if (mp->primitive->Intersect(ray, isect))
		    hit = true;
		}
	    }
	  // Grab next node to process from todo list
	  if (todoPos > 0) 
	    {
	      --todoPos;
	      node = todo[todoPos].node;
	      tmin = todo[todoPos].tmin;
	      tmax = todo[todoPos].tmax;
	    }
	  else
	    break;
	}
      else 
	{
	  // Process kd tree interior node
	  // Compute distance along ray to split plane
	  Float tplane;
	  if (node->axis == SPLIT_X)
	    tplane = (node->split - ray.O.x) * invDir.x;
	  else if (node->axis == SPLIT_Y)
	    tplane = (node->split - ray.O.y) * invDir.y;
	  else
	    tplane = (node->split - ray.O.z) * invDir.z;
	  // Get near and far child pointers for ray
	  MyKdAccelNode *nearChild, *farChild;
	  bool zeroIsNear;
	  if (node->axis == SPLIT_X)
	    zeroIsNear = (ray.O.x <= node->split);
	  else if (node->axis == SPLIT_Y)
	    zeroIsNear = (ray.O.y <= node->split);
	  else
	    zeroIsNear = (ray.O.z <= node->split);
	  if (zeroIsNear)
	    {
	      nearChild = node->u.children[0];
	      farChild  = node->u.children[1];
	    }
	  else 
	    {
	      nearChild = node->u.children[1];
	      farChild  = node->u.children[0];
	    }

	  // Advance to next child node, possibly enqueue far child
	  if (tplane > tmax || tplane < 0)
	    node = nearChild;
	  else if (tplane < tmin)
	    node = farChild;
	  else
	    {
	      if(!nearChild) 
		{
		  node = farChild;
		  tmin = tplane;
		}
	      else 
		{
		  if (farChild)
		    {
		      // Enqueue [[farChild]] in todo list
		      todo[todoPos].node = farChild;
		      todo[todoPos].tmin = tplane;
		      todo[todoPos].tmax = tmax;
		      ++todoPos;
		      Assert(todoPos < MAX_TODO);
		    }
		  node = nearChild;
		  tmax = tplane;
		}
	    }
	  if (!node)
	    {
	      // Grab next node to process from todo list
	      if (todoPos > 0)
		{
		  --todoPos;
		  node = todo[todoPos].node;
		  tmin = todo[todoPos].tmin;
		  tmax = todo[todoPos].tmax;
		}
	      else
		break;
	    }
	}
    }
  return hit;
}

bool MyKdTreeAccelerator::IntersectP(const Ray &ray) const 
{
  // Compute initial parametric range of ray inside kd tree extent
  Float tmin, tmax;
  if (!bounds.IntersectP(ray, &tmin, &tmax))
    return false;
  tmax *= 1.001f;
  // Prepare to traverse kd-tree for ray
  int rayId = curMailboxId++;
  const MyKdAccelNode *node = root;
  Vector invDir(1.f/ray.D.x, 1.f/ray.D.y, 1.f/ray.D.z);
	
#define MAX_TODO 64
  MyKdToDo todo[MAX_TODO];
  int todoPos = 0;
  while (node != NULL)
    {
      // Update kd-tree shadow ray traversal statistics
      static StatsCounter nodesTraversed("Acceleration",
					 "Number of kd-tree nodes traversed by shadow rays");
      ++nodesTraversed;
      if (node->isLeaf) 
	{
	  // Check for shadow ray intersections inside leaf node
	  u_int nPrimitives = node->nPrimitives;
	  MailboxPrim **prims = node->u.primitives;
	  for (u_int i = 0; i < nPrimitives; ++i)
	    {
	      MailboxPrim *mp = prims[i];
	      if (mp->lastMailboxId != rayId)
		{
		  mp->lastMailboxId = rayId;
		  if (mp->primitive->IntersectP(ray))
		    return true;
		}
	    }
	  // Grab next node to process from todo list
	  if (todoPos > 0)
	    {
	      --todoPos;
	      node = todo[todoPos].node;
	      tmin = todo[todoPos].tmin;
	      tmax = todo[todoPos].tmax;
	    }
	  else
	    break;
	}
      else 
	{
	  // Process kd tree interior node
	  // Compute distance along ray to split plane
	  Float tplane;
	  if (node->axis == SPLIT_X)
	    tplane = (node->split - ray.O.x) * invDir.x;
	  else if (node->axis == SPLIT_Y)
	    tplane = (node->split - ray.O.y) * invDir.y;
	  else
	    tplane = (node->split - ray.O.z) * invDir.z;
	  // Get near and far child pointers for ray
	  MyKdAccelNode *nearChild, *farChild;
	  bool zeroIsNear;
	  if (node->axis == SPLIT_X)
	    zeroIsNear = (ray.O.x <= node->split);
	  else if (node->axis == SPLIT_Y)
	    zeroIsNear = (ray.O.y <= node->split);
	  else
	    zeroIsNear = (ray.O.z <= node->split);
	  if (zeroIsNear)
	    {
	      nearChild = node->u.children[0];
	      farChild  = node->u.children[1];
	    }
	  else
	    {
	      nearChild = node->u.children[1];
	      farChild  = node->u.children[0];
	    }
	  // Advance to next child node, possibly enqueue far child
	  if (tplane > tmax || tplane < 0)
	    node = nearChild;
	  else if (tplane < tmin)
	    node = farChild;
	  else 
	    {
	      if (!nearChild) 
		{
		  node = farChild;
		  tmin = tplane;
		}
	      else
		{
		  if (farChild)
		    {
		      // Enqueue [[farChild]] in todo list
		      todo[todoPos].node = farChild;
		      todo[todoPos].tmin = tplane;
		      todo[todoPos].tmax = tmax;
		      ++todoPos;
		      Assert(todoPos < MAX_TODO);
		    }
		  node = nearChild;
		  tmax = tplane;
		}
	    }
	  if (!node)
	    {
	      // Grab next node to process from todo list
	      if (todoPos > 0)
		{
		  --todoPos;
		  node = todo[todoPos].node;
		  tmin = todo[todoPos].tmin;
		  tmax = todo[todoPos].tmax;
		}
	      else
		break;
	    }
	}
    }
  return false;
}

//extern "C" BLOB_ELEMENT *CreateAccelerator(const vector<Reference<BLOB_ELEMENT> > &prims) 
//{
//return new MyKdTreeAccelerator(prims);
//}

#endif
