// bidirectional.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
// Bidirectional Local Declarations
struct BidirVertex;
class BidirIntegrator : public SurfaceIntegrator {
public:
	Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample, Float *alpha) const;
	Sample *AllocateSample(const Scene *scene) const;
private:
	// BidirIntegrator Private Methods
	int generatePath(const Ray &r, BidirVertex *vertices, int maxVerts) const;
	Float weightPath(BidirVertex *eye, int nEye, BidirVertex *light, int nLight) const;
	Spectrum evalPath(const Scene *scene, BidirVertex *eye, int nEye,
	BidirVertex *light, int nLight) const;
	static Float G(const BidirVertex &v0, const BidirVertex &v1);
	static bool visible(const Scene *scene, const Point &P0, const Point &P1);
};
struct BidirVertex {
	BSDF *bsdf;
	Point P;
	Normal N;
	Vector wi, wo;
	Float bsdfWeight, dAWeight;
	bool isSpecular;
};
// Bidirectional Method Definitions
Sample *BidirIntegrator::AllocateSample(const Scene *scene) const {
	Assert(1 == 0);
	return NULL;
}
Spectrum BidirIntegrator::L(const Scene *scene, const RayDifferential &ray,
		const Sample *sample, Float *alpha) const {
	Spectrum L(0.);
	// Generate eye and light sub-paths
	#define MAX_VERTS 8
	BidirVertex eyePath[MAX_VERTS], lightPath[MAX_VERTS];
	int nEye = generatePath(ray, eyePath, MAX_VERTS);
	if (nEye == 0) {
		// XXX handle ray with no intersection and reurn
	}
	// Choose light for bidirectional path
	int lightNum = RandomInt() % scene->lights.size();
	Light *light = scene->lights[lightNum];
	Float lightWeight = Float(scene->lights.size());
	// Sample ray from light source to start light path
	Ray lightRay;
	Float lightSampleWeight;
	bool deltaLight;
	Float u[4];
	for (int i = 0; i < 4; ++i)
		u[i] = RandomFloat();
	Spectrum Le = light->Sample_L(scene, u[0], u[1], u[2], u[3],
		&lightRay, &lightSampleWeight, &deltaLight);
	int nLight = generatePath(lightRay, lightPath, MAX_VERTS);
	// Connect bidirectional path prefixes and evaluate throughput
	for (int i = 1; i <= nEye; ++i) {
		for (int j = 1; j <= nLight; ++j) {
			if (j == 1) {
				// Handle direct lighting for bidirectional integrator
				continue;
			}
			L += evalPath(scene, eyePath, i, lightPath, j) /
				weightPath(eyePath, i, lightPath, j);
	
		}
	}
	return L;
}
int BidirIntegrator::generatePath(const Ray &r, BidirVertex *vertices,
		int maxVerts) const {
// XXX careful if we reuse the same Ray, then make sure mint/maxt are reset!
	return 0; // keep the compiler happy.
}

Float BidirIntegrator::weightPath(BidirVertex *eye, int nEye,
		BidirVertex *light, int nLight) const {
	return Float(nEye + nLight - 1);
}
Spectrum BidirIntegrator::evalPath(const Scene *scene, BidirVertex *eye, int nEye,
		BidirVertex *light, int nLight) const {
	if (!visible(scene, eye[nEye].P, light[nLight].P))
		return 0.;
	Spectrum L(1.);
	for (int i = 0; i < nEye; ++i) {
		BidirVertex *e = &eye[i];
		L *= e->bsdf->f(e->wi, e->wo) * G(eye[i], eye[i+1]) /
			e->dAWeight; // XXX bsdf weight?
	}
	Vector w = light[nLight].P - eye[nEye].P;
	L *= eye[nEye].bsdf->f(eye[nEye].wi, w) *
		G(eye[nEye], light[nLight]) *
		light[nLight].bsdf->f(-w, light[nLight].wi);
	for (int i = nLight-1; i >= 0; --i) {
		BidirVertex *l = &light[i];
		L *= l->bsdf->f(l->wi, l->wo) * G(light[i], light[i-1]) /
			l->dAWeight; // XXXX
	}
	return L;
}
Float BidirIntegrator::G(const BidirVertex &v0, const BidirVertex &v1) {
	Vector w = (v1.P - v0.P).Hat();
	return fabsf(Dot(v0.N, w) * Dot(v1.N, -w)) /
		DistanceSquared(v0.P, v1.P);
}
bool BidirIntegrator::visible(const Scene *scene, const Point &P0,
		const Point &P1) {
	Ray ray(P0, P1-P0, RAY_EPSILON, 1.f - RAY_EPSILON);
	return !scene->IntersectP(ray);
}
extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
	return new BidirIntegrator;
}
