// directlighting.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "transport.h"
#include "scene.h"
// DirectLighting Declarations
class DirectLighting : public SurfaceIntegrator {
public:
	// DirectLighting Methods
	DirectLighting(const string &);
	~DirectLighting();
	Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample,
		Float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene) {
		if (strategy == SAMPLE_ALL_UNIFORM) {
			lightSampleOffset = sample->Add2D(scene->lights.size());
			bsdfSampleOffset = sample->Add2D(scene->lights.size());
		}
		else {
			lightSampleOffset = sample->Add2D(1);
			bsdfSampleOffset = sample->Add2D(1);
		}
	}
private:
	// DirectLighting Private Data
	int lightSampleOffset, bsdfSampleOffset;
	enum LightStrategy { SAMPLE_ALL_UNIFORM, SAMPLE_ONE_UNIFORM,
		SAMPLE_ONE_WEIGHTED } strategy;
	mutable Float *avgY, *avgYsample, *cdf;
	mutable Float overallAvgY;
};
// DirectLighting Method Definitions
DirectLighting::DirectLighting(const string &st) {
	if (st == "one") strategy = SAMPLE_ONE_UNIFORM;
	else if (st == "all") strategy = SAMPLE_ALL_UNIFORM;
	else if (st == "weighted") strategy = SAMPLE_ONE_WEIGHTED;
	else {
		Warning("Strategy \"%s\" for direct lighting unknown. "
			"Using \"all\".", st.c_str());
		strategy = SAMPLE_ALL_UNIFORM;
	}
	// DirectLighting constructor implementation
	avgY = avgYsample = cdf = NULL;
	overallAvgY = 0.;
}
DirectLighting::~DirectLighting() {
	// DirectLighting destructor implementation
	delete[] avgY;
	delete[] avgYsample;
	delete[] cdf;
}
Spectrum DirectLighting::L(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		Float *alpha) const {
	Intersection isect;
	Spectrum L(0.);
	if (scene->Intersect(ray, &isect)) {
		if (alpha) *alpha = 1.;
		// Evaluate BSDF at hit point
		BSDF *bsdf = isect.GetBSDF(ray);
		// Compute emitted light if an area light source
		L += isect.Le(-ray.D);
		// Compute direct lighting for \use{DirectLighting} integrator
		const Point &P = isect.dgShading.P;
		const Normal &N = isect.dgShading.Nn;
		if (scene->lights.size() > 0) {
			Vector wo = -ray.D.Hat();
			// Apply direct lighting strategy
			switch (strategy) {
				case SAMPLE_ALL_UNIFORM: {
					L += UniformSampleAllLights(scene, P, N, wo, bsdf,
						sample, lightSampleOffset, bsdfSampleOffset);
					break;
				}
				case SAMPLE_ONE_UNIFORM: {
					L += UniformSampleOneLight(scene, P, N, wo, bsdf,
						sample, lightSampleOffset, bsdfSampleOffset);
					break;
				}
				case SAMPLE_ONE_WEIGHTED: {
					L += WeightedSampleOneLight(scene, P, N, wo, bsdf,
						sample, lightSampleOffset, bsdfSampleOffset,
						avgY, avgYsample, cdf, overallAvgY);
					break;
				}
			}
		}
		// Clean up from integration
		delete bsdf;
	}
	else {
		// Handle ray with no intersection
		if (alpha) *alpha = 0.;
		for (u_int i = 0; i < scene->lights.size(); ++i)
			L += scene->lights[i]->Le(ray);
		if (alpha && !L.Black()) *alpha = 1.;
		return L;
	}
	return L;

}
extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
	string strategy = params.FindOneString("strategy", "all");
	return new DirectLighting(strategy);
}
