// irradiancecache.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
#include "octree.h"
// IrradianceCache Forward Declarations
struct IrradSample;
struct IrradProcess;
// IrradianceCache Local Declarations
struct IrradSample {
	IrradSample() { }
	IrradSample(const Spectrum &e, const Point &p, const Normal &n,
		Float md, int id) : E(e), N(n), P(p) {
			maxDist = md;
			indirectDepth = id;
	}
	Spectrum E;
	Normal N;
	Point P;
	Float maxDist;
	int indirectDepth;
};
struct IrradProcess {
	IrradProcess(const Normal &n, Float me) {
		N = n;
		maxError = me;
		nFound = samplesChecked = 0;
		sumWt = 0.;
		E = 0.;
	}
	Normal N;
	Float maxError;
	// XXX int indirectLevel;
	mutable int nFound;
	mutable int samplesChecked;
	mutable Float sumWt;
	mutable Spectrum E;
	void operator()(const Point &P, const IrradSample &sample) const;
	bool successful() { return (nFound > 0 && sumWt > 0.); }
};
// IrradianceCache Declarations
class IrradianceCache : public SurfaceIntegrator {
public:
	IrradianceCache(int md, Float maxerr, int nsamples,
		bool ss);
	Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample, Float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene);
private:
	// IrradianceCache Private Data
	Float maxError;
	bool showSamples;
	int nSamples;
	mutable int rayDepth, indirectDepth;
	int maxDepth;
	int lightSampleOffset, bsdfSampleOffset;
	mutable Octree<IrradSample, IrradProcess> *octree;
	// IrradianceCache Private Methods
	bool InterpolateIrradiance(const Scene *scene,
			const Point &P, const Normal &N, Spectrum *E) const;
};
// IrradianceCache Method Definitions
IrradianceCache::IrradianceCache(int md, Float maxerr, int ns,
		bool ss) {
	maxDepth = md;
	maxError = maxerr;
	nSamples = ns;
	showSamples = ss;
	rayDepth = 0;
	indirectDepth = 0;
	// IrradianceCache Constructor Implementation
	octree = NULL;
}
Spectrum IrradianceCache::L(const Scene *scene, const RayDifferential &ray,
		const Sample *sample, Float *alpha) const {
	Intersection isect;
	Spectrum L(0.);
	if (scene->Intersect(ray, &isect)) {
		if (alpha) *alpha = 1.;
		// Compute reflected radiance with irradiance cache
		if (indirectDepth == 0) L += isect.Le(-ray.D);
		// Evaluate BSDF at hit point
		BSDF *bsdf = isect.GetBSDF(ray);
		Vector wo = -ray.D.Hat();
		// Compute direct lighting for irradiance cache
		const Point &P = isect.dgShading.P;
		const Normal &N = isect.dgShading.Nn;
		L += UniformSampleAllLights(scene, P, N, wo, bsdf, sample,
			lightSampleOffset, bsdfSampleOffset);
		if (rayDepth++ < maxDepth) {
			Vector wi;
			// Trace rays for specular reflection and refraction
			for (int i = 0; i < bsdf->NumSpecular(); ++i) {
				Spectrum fr = bsdf->f_delta(i, -ray.D, &wi);
				if (!fr.Black())
					L += scene->L(Ray(isect.dgShading.P, wi), sample) * fr *
						fabsf(Dot(wi, isect.dgShading.Nn));
			}
		}
		if (indirectDepth++ < maxDepth) {
			// Estimate indirect lighting with irradiance cache
			Spectrum E;
			const Point &P = isect.dgShading.P;
			const Normal &N = isect.dgShading.Nn;
			if (!InterpolateIrradiance(scene, P, N, &E)) {
				// Compute irradiance at current point
				// Determine how many samples to take for irradiance estimate
				int ns = nSamples;
				for (int r = indirectDepth; r > 1; --r)
					ns >>= 1;
				if (ns == 0) ns = 1;
				Float *samples = (Float *)alloca(2 * ns * sizeof(Float));
				LatinHypercube(samples, ns, 2);
				Float invSumDists = 0.;
				for (int i = 0; i < ns; ++i) {
					// Trace ray to sample radiance for irradiance estimate
					// Update irradiance statistics for rays traced
					static StatsCounter nIrradianceRays("Integrator",
						"Indirect rays traced for irradiance");
					++nIrradianceRays;
					Vector w = CosineSampleHemisphere(samples[2*i], samples[2*i+1]);
					Ray r(P, bsdf->LocalToWorld(w));
					E += M_PI * scene->L(r, sample);
					invSumDists += 1.f / (r.maxt * r.D.Length());
				}
				E /= Float(ns);
				Float maxDist = ns / invSumDists;
				// Add computed irradiance value to cache
				if (maxError > 0.) {
					BBox sampleExtent(P);
					sampleExtent.Expand(maxDist);
					// Allocate octree if needed
					if (!octree) {
						BBox wb = scene->WorldBound();
						octree = new Octree<IrradSample, IrradProcess>(wb);
					}
					IrradSample sample(E, P, N, maxDist, indirectDepth);
					octree->Add(sample, sampleExtent);
					// Update statistics for new irradiance sample
					static StatsCounter nSamplesComputed("Integrator",
						"Irradiance estimates comupted");
					++nSamplesComputed;
				}
				if (showSamples) E *= 10000;
			}
			else if (showSamples) E = 0.;
			L += E * M_PI * bsdf->rho_r(wo);
		}
		--rayDepth;
		--indirectDepth;
		// Clean up from integration
		delete bsdf;
	}
	else {
		// Handle ray with no intersection
		if (alpha) *alpha = 0.;
		for (u_int i = 0; i < scene->lights.size(); ++i)
			L += scene->lights[i]->Le(ray);
		if (alpha && !L.Black()) *alpha = 1.;
		return L;
	}
	return L;
}
bool IrradianceCache::InterpolateIrradiance(const Scene *scene,
		const Point &P, const Normal &N, Spectrum *E) const {
	// Allocate octree if needed
	if (!octree) {
		BBox wb = scene->WorldBound();
		octree = new Octree<IrradSample, IrradProcess>(wb);
	}
	IrradProcess proc(N, maxError);
	octree->Lookup(P, proc);
	// Update irradiance cache lookup stats
	static StatsRatio nSuccessfulLookups("Integrator",
		"Successful irradiance cache lookups");
	static StatsRatio nSamplesFound("Integrator",
		"Irradiance samples found per successful lookup", false);
	nSuccessfulLookups.add(proc.successful() ? 1 : 0, 1);
	nSamplesFound.add(proc.nFound, 1);
	if (proc.successful()) {
		*E = proc.E / proc.sumWt;
		return true;
	}
	return false;
}
void IrradProcess::operator()(const Point &P, const IrradSample &sample) const{
	++samplesChecked;
	// Skip sample if it is behind point being shaded
	if (Dot(sample.P - P, (sample.N + N).Hat()) > .01)
		return;
	// Skip sample if surface normals are too different
	if (Dot(N, sample.N) < .707f)
		return;
	// Skip sample if it's too far from the sample point
	Float d2 = DistanceSquared(P, sample.P);
	if (d2 > sample.maxDist * sample.maxDist)
		return;
	// Computer estimate error term, [[err]]
	Float err = sqrtf(d2) / (sample.maxDist * Dot(N, sample.N));
	if (err < maxError) {
		++nFound;
		Float wt = 1.f / max(.05f, err);
		E += wt * sample.E;
		sumWt += wt;
	}
}
extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
	Float maxError = params.FindOneFloat("maxerror", .2f);
	int maxDepth = params.FindOneInt("maxdepth", 1);
	int nSamples = params.FindOneInt("nsamples", 4096);
	bool showSamples = params.FindOneInt("showsamples", 0) != 0;
	return new IrradianceCache(maxDepth, maxError,
		nSamples, showSamples);
}
