// photonmap.cc*


#include "lrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
#include "kdtree3.h"
#include "sampling.h"


// Photonmap Local Declarations
//////////////////////////////////////////////////////////////////

struct PhotonData;

struct PhotonProcess;

class PhotonIntegrator : public SurfaceIntegrator {
public:
	// PhotonIntegrator Methods
	PhotonIntegrator(int nStored, int nStoredInd, int nLookup, int mdepth,
									 Float maxdist, int flagfg, int flagpg, int fgs);
	~PhotonIntegrator();
	Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample, Float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene);
	void Preprocess(const Scene *);
	void recursiveGather(KdNode<PhotonData>	*node, KdTree<PhotonData, PhotonProcess> *tree) const;
	PhotonData* PhotonIntegrator::recursiveLookupOne(KdNode<PhotonData>	*node, const Point &P, const Normal &N) const;
private:
	// PhotonIntegrator Private Data
	u_int nStored, nLookup, nStoredInd;
	mutable int pathLength;
	int maxDepth;
	int flagfinalgather, finalgathersamples, flagpregather;
	Float maxDist;
	int lightSampleOffset, bsdfSampleOffset;
	mutable int isfinal;    // Is this ray a final ray
	mutable KdTree<PhotonData, PhotonProcess> *tree;
	mutable KdTree<PhotonData, PhotonProcess> *treeind;
};


struct PhotonData {
	PhotonData() { }
	PhotonData(const DifferentialGeometry &dg, const Spectrum &L,
		const Vector &w, BSDF *bsdf)
		: P(dg.P), N(dg.Nn), dE(L), wi(w) { this->bsdf=bsdf; }
	Point P;
	Normal N;
	Spectrum dE;
	Spectrum Gath;
	BSDF *bsdf;
	Vector wi;
};

struct ClosePhoton {
	ClosePhoton(const PhotonData *d = NULL, Float md2 = HUGE_VAL) {
		data = d;
		maxDist2 = md2;
	}
	bool operator<(const ClosePhoton &p2) const {
		return maxDist2 < p2.maxDist2;
	}
	const PhotonData *data;
	Float maxDist2;
};

struct PhotonProcess {
	PhotonProcess(u_int mp, const Point &p, const Normal &n);
	void operator()(const PhotonData &photon, Float &maxDist) const;

	const Point &P;
	const Normal &N;

	mutable vector<ClosePhoton> photons;
	u_int maxPhotons;
};



// PhotonIntegrator Method Definitions
//////////////////////////////////////////////////////////////////


PhotonIntegrator::PhotonIntegrator(int ns, int nsi, int nl, int mdepth,
		Float mdist, int flagfg, int flagpg, int fgs) {
	nStored = ns;
	nStoredInd = nsi;
	nLookup = nl;
	maxDist = mdist;
	maxDepth = mdepth;
	tree = NULL;
	treeind = NULL;
	pathLength = 0;
	flagfinalgather=flagfg;
	flagpregather=flagpg;
	finalgathersamples=fgs;
}


PhotonIntegrator::~PhotonIntegrator() {
	delete tree;
}


void PhotonIntegrator::RequestSamples(Sample *sample, const Scene *scene) {
	lightSampleOffset = sample->Add2D(scene->lights.size());
	bsdfSampleOffset = sample->Add2D(scene->lights.size());
}


void PhotonIntegrator::Preprocess(const Scene *scene) {
	vector<PhotonData> photons;
	photons.reserve(nStored);
	vector<PhotonData> photonsind;
	photonsind.reserve(nStored);

	printf("Photon Shooting: %d for caustics, %d for global illum., %d used, final gather: %d\n",
				 nStored, nStoredInd, nLookup, (flagfinalgather==1) ?
				 finalgathersamples : 0);

	// Initialize photon shooting statistics
	static StatsCounter nshot("Integrator",
		"Number of photons shot from lights for caustics");
	static StatsCounter nshot2("Integrator",
		"Number of photons shot from lights for GI");

	int nLights = int(scene->lights.size());
	ProgressReporter progress(nStored, "Shooting photons for caustics");
	while (photons.size() < nStored) {
		++nshot;
		// Give up if we're not finding any specular surfaces
		if (nshot > 500000 && (!photons.size() || nshot / photons.size() > 10000)) {
			cerr << "No luck shooting photons !!!" << endl;
			goto gillum;
		}
		// Trace a photon path and store contribution
		Float u[4];
		u[0] = RadicalInverse(nshot+1, 2);
		u[1] = RadicalInverse(nshot+1, 3);
		u[2] = RadicalInverse(nshot+1, 5);
		u[3] = RadicalInverse(nshot+1, 7);
		// Choose light to shoot photon from
		Light *light = scene->lights[RandomInt() % nLights];
		Float lightWeight = 1.f / nLights;
		// Generate [[photonRay]] from light source
		RayDifferential photonRay;
		Float weight;
		bool isDeltaLight;
		Spectrum L = light->Sample_L(scene, u[0], u[1], u[2], u[3],
			&photonRay, &weight, &isDeltaLight);
		L /= weight * lightWeight;
		if (!L.Black()) {
			// Follow photon to non-specular surface and store in [[photons]] array
			bool hitSpecular = false;
			int nBounces = 0;
			Intersection photonIsect;
			BSDF *photonBSDF = 0;
			static StatsRatio specularHits("Integrator",
				"Photons that hit specular surface", true);
			specularHits.add(0, 1);
			while (scene->Intersect(photonRay, &photonIsect)) {
				L *= scene->Transmittance(photonRay);
				delete photonBSDF;
				photonBSDF = photonIsect.GetBSDF(photonRay);
				// Handle non-specular photon hit
				if (hitSpecular &&
					photonBSDF->NumComponents() > photonBSDF->NumSpecular()) {
					photons.push_back(PhotonData(photonIsect.dgGeom, L,
																			 -photonRay.D.Hat(), NULL));
					progress.Update();
				}
				// Handle specular photon hit
				if (photonBSDF->NumSpecular() > 0) {
					// Record photon specular hit statistics
					if (!hitSpecular)
						specularHits.add(1, 0);
					hitSpecular = true;
					int specComponent = RandomInt() % photonBSDF->NumSpecular();
					Vector wi;
					Spectrum fr = photonBSDF->f_delta(specComponent,
						-photonRay.D, &wi);
					if (fr.Black())
						break;
					L *= fr * Float(photonBSDF->NumSpecular()) * fabsf(Dot(wi, photonIsect.dgShading.Nn));
					//					L *= fr * Float(photonBSDF->NumSpecular());
					photonRay = Ray(photonIsect.dgGeom.P, wi);
				}
				// Possibly terminate photon path
				if (photonBSDF->NumSpecular() == 0)
					break;
				if (nBounces++ > 3) {
					Float continueProbability = .2f;
					if (RandomFloat() > continueProbability)
						break;
					L /= continueProbability;
				}
			}
			delete photonBSDF;
		}
	}
	printf("\n");
  // Normalize photon [[dE]] values and build tree
	for (u_int i = 0; i < photons.size(); ++i)
	  photons[i].dE /= Float(nshot);

 gillum:
	// GLOBAL ILLUMINATION
	ProgressReporter progress2(nStoredInd, "Shooting photons for global illumination");
	while (photonsind.size() < nStoredInd) {		
		++nshot2;
		// Trace a photon path and store contribution
		Float u[4];
		u[0] = RadicalInverse(nshot2+1, 2);
		u[1] = RadicalInverse(nshot2+1, 3);
		u[2] = RadicalInverse(nshot2+1, 5);
		u[3] = RadicalInverse(nshot2+1, 7);
		// Choose light to shoot photon from
		int nLights = int(scene->lights.size());
		Light *light = scene->lights[RandomInt() % nLights];
		Float lightWeight = 1.f / nLights;

		// Generate [[photonRay]] from light source
		RayDifferential photonRay;
		Float weight;
		bool isDeltaLight;
		Spectrum L = light->Sample_L(scene, u[0], u[1], u[2], u[3],
			&photonRay, &weight, &isDeltaLight);
		L /= weight * lightWeight;

		if (!L.Black()) {
			int nBounces = 0;
			int hitSpecular = 0;
			Intersection photonIsect;
			BSDF *photonBSDF = 0;

			// Let's go until there is no more intersection and we hit the intersideral void
			while (scene->Intersect(photonRay, &photonIsect)) {
				L *= scene->Transmittance(photonRay);
				//				delete photonBSDF;
				photonBSDF = photonIsect.GetBSDF(photonRay);

				// Handle non-specular photon hit: store the photon in the map
				// except if it is a LS+D

				if ((photonBSDF->NumComponents() > photonBSDF->NumSpecular()) && (hitSpecular!=1)) {
					photonsind.push_back(PhotonData(photonIsect.dgGeom, L,
																					-photonRay.D.Hat(), photonBSDF));
					progress2.Update();
				}

				// Randomly sample BSDF to get new path direction
				//					Vector wo = -photonRay.D.Hat();
				Vector wo = -photonRay.D.Hat();  // Right ??????????????????
				Vector wi;
				Float bs1, bs2;
				bs1 = RandomFloat();
				bs2 = RandomFloat();
				
				Float weight;
				bool specularBounce = false;
				Spectrum f = photonBSDF->sample_f(wo, &wi, bs1, bs2, &weight, &specularBounce);
				if (f == Spectrum(0.) || weight == 0.) {
					break;
				}
				photonRay = Ray(photonIsect.dgGeom.P, wi);
				
				
				if(specularBounce) {
					//						L *= f * Float(photonBSDF->NumSpecular());    // C'est quoi ca ?
						L *= f * Float(photonBSDF->NumSpecular())* fabsf(Dot(wi, photonIsect.dgShading.Nn));     // * cos blah blah blah ?
					if(hitSpecular==0)
						hitSpecular=1;    // we are starting on a series of specular surfaces
				}
				else {
					//					L *= f * fabsf(Dot(wi, photonIsect.dgGeom.Nn)) / weight;   // Shading or geometry ?
					L *= f * fabsf(Dot(wi, photonIsect.dgShading.Nn)) / weight;   // Shading or geometry ?
					hitSpecular=2;     // Exiting the series of spec. surfaces
				}
				
				// Why not max Depth ?
				if (nBounces++ > 3) {
					Float continueProbability = .2f;
					if (RandomFloat() > continueProbability)
						break;
					L /= continueProbability;
				}
			}
			//			delete photonBSDF;
		}
	}
		
  // Normalize photon [[dE]] values and build tree
	for (u_int i = 0; i < photonsind.size(); ++i)
	  photonsind[i].dE /= Float(nshot2);

	if(photons.size()>0)
		tree = new KdTree<PhotonData, PhotonProcess>(photons);
	treeind = new KdTree<PhotonData, PhotonProcess>(photonsind);

	// Pre-gather the photons
	if(flagpregather) {
		printf("\nPrecomputing irradiance\n");
		recursiveGather(treeind->root,treeind);
	}

	if(!flagfinalgather) {
		isfinal=1;    // All the rays are final
	}

	fprintf(stderr, "\n");
}


void PhotonIntegrator::recursiveGather(KdNode<PhotonData>	*node, KdTree<PhotonData, PhotonProcess> *tree) const {
	if (!node) return;

	const Point &P = node->data.P;
	const Normal &N = node->data.N;

	// Gather irradiance at this photon...
	Spectrum Ltmp(0.);
	PhotonProcess proc(nLookup, P, N);
	Float md = maxDist;
	tree->Lookup(P, proc, md);
	// Accumulate light from nearby photons
	vector<ClosePhoton> &photons = proc.photons;
	// Compute photon scale factor with density estimation
	Float scale = photons.size() / (nLookup * M_PI * md * md);
	for (u_int i = 0; i < photons.size(); ++i) {
		Ltmp +=  node->data.bsdf->f(Vector(-node->data.N), photons[i].data->wi) *
			photons[i].data->dE;
		//		Ltmp +=  node->data.bsdf->f(Vector(node->data.N), photons[i].data->wi) *
		//			photons[i].data->dE;
	}
	node->data.Gath=Ltmp*scale;
	//		cout << node->data.Gath << endl;
	//	delete node->data.bsdf;
	recursiveGather(node->children[0],tree);
	recursiveGather(node->children[1],tree);
}

PhotonData* PhotonIntegrator::recursiveLookupOne
(KdNode<PhotonData>	*node, const Point &P, const Normal &N) const
{
	PhotonData *ret;
	if (!node) return NULL;

	// Process Kd node's children based on $x$ split
	if (node->split == SPLIT_X) {
		if (P.x <= node->data.P.x)   // Point on the left
			ret=recursiveLookupOne(node->children[0], P, N);
		else
			ret=recursiveLookupOne(node->children[1], P, N);
	}
	else if (node->split == SPLIT_Y) {
		if (P.y <= node->data.P.y)   // Point on the left
			ret=recursiveLookupOne(node->children[0], P, N);
		else
			ret=recursiveLookupOne(node->children[1], P, N);
	}
	else {
		if (P.z <= node->data.P.z)   // Point on the left
			ret=recursiveLookupOne(node->children[0], P, N);
		else
			ret=recursiveLookupOne(node->children[1], P, N);
	}

	if(Dot(node->data.N, N) >= .707f) {
		if(ret &&	(DistanceSquared(P, ret->P)<DistanceSquared(P,node->data.P)))
			return ret;
		else
			return &node->data;
	}
	else
		return ret;
}

Spectrum PhotonIntegrator::L(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		Float *alpha) const {
	// Compute reflected radiance with photon map
	Spectrum L(0.);
	Intersection isect;
	bool specularBounce;

	if (scene->Intersect(ray, &isect)) {
		if(pathLength==0 && flagfinalgather)
			isfinal=0;    // A ray is always non final when it is cast...

		if (alpha) *alpha = 1.;
		L += isect.Le(-ray.D);
		const Point &P = isect.dgShading.P;
		const Normal &N = isect.dgShading.Nn;
		Vector wo = -ray.D.Hat();
		BSDF *bsdf = isect.GetBSDF(ray);

		// Caustics...
		if (tree) {
			PhotonProcess proc(nLookup, P, N);
			Float md = maxDist;
			tree->Lookup(P, proc, md);
			// Accumulate light from nearby photons
			vector<ClosePhoton> &photons = proc.photons;
			// Compute photon scale factor with density estimation
			Float scale = photons.size() / (nLookup * M_PI * md * md);
			for (u_int i = 0; i < photons.size(); ++i)
				L += scale * bsdf->f(wo, photons[i].data->wi) *
					//					fabsf(Dot(photons[i].data->wi, N)) *
					photons[i].data->dE;
		}
		
		// First bounce, evaluate 
		if(!isfinal) {
			// The ray is not final
			L += UniformSampleAllLights(scene, P, N, wo, bsdf, sample,
																	lightSampleOffset, bsdfSampleOffset);
			if (pathLength++ < maxDepth) {
				Vector wi;
				// Randomly sample BSDF to get new path direction for final gathering ray
				Float mysamples[2*finalgathersamples];
				LatinHypercube(mysamples, finalgathersamples, 2);
				Float weight;
				Spectrum Ltmp(0.);
				for(int kk=0;kk<finalgathersamples;kk++) {
					Spectrum f = bsdf->sample_f(wo, &wi, mysamples[kk*2], mysamples[kk*2+1],
																			&weight, &specularBounce);
					if(specularBounce)
						isfinal=0;
					else
						isfinal=1;
					
					if (!f.Black() && weight!=0)
						Ltmp+= scene->L(Ray(P, wi), sample) * f *	fabsf(Dot(wi, N)) / weight;
				}
				L+=Ltmp/finalgathersamples;
			}
		}
		else {
			// The ray is final...
			// global illumination
			if(treeind){
				if(flagpregather) {
					PhotonData *pd=recursiveLookupOne(treeind->root,P,N);
					if(pd)
						L+=pd->Gath;
				}
				else {
					PhotonProcess proc(nLookup, P, N);
					Float md = maxDist;
					treeind->Lookup(P, proc, md);
					// Accumulate light from nearby photons
					vector<ClosePhoton> &photonsind = proc.photons;
					// Compute photon scale factor with density estimation
					Float scale = photonsind.size() / (nLookup * M_PI * md * md);
					for (u_int i = 0; i < photonsind.size(); ++i)
						L += scale *  bsdf->f(wo, photonsind[i].data->wi) *
							photonsind[i].data->dE;
				}
			}

			// Still tracing for specular surfaces...
			if (pathLength++ < maxDepth) {
				Vector wi;
				// Trace rays for specular reflection and refraction
				for (int i = 0; i < bsdf->NumSpecular(); ++i) {
					Spectrum fr = bsdf->f_delta(i, -ray.D, &wi);
					if (!fr.Black())
						L += scene->L(Ray(P, wi), sample) * fr *
							fabsf(Dot(wi, N));
				}
			}
		}
		--pathLength;
		
		// Clean up from integration
		delete bsdf;
	}
	else {
		// Handle ray with no intersection
		if (alpha) *alpha = 0.;
		for (u_int i = 0; i < scene->lights.size(); ++i)
			L += scene->lights[i]->Le(ray);
		if (alpha && !L.Black()) *alpha = 1.;
		return L;
	}
	return L;
}

// Photonprocess Method Definitions
//////////////////////////////////////////////////////////////////

PhotonProcess::PhotonProcess(u_int mp, const Point &p,
		const Normal &n)
	: P(p), N(n) {
	maxPhotons = mp;
}

void PhotonProcess::operator()(const PhotonData &photon,
		Float &maxDist) const {
	if (Dot(photon.N, N) < .707f) return;
	Float d2 = DistanceSquared(photon.P, P);
	if (photons.size() == 0)
		photons.reserve(maxPhotons);
	if (photons.size() < maxPhotons) {
		// Add photon to unordered array of photons
		photons.push_back(ClosePhoton(&photon, d2));
		if (photons.size() == maxPhotons) {
			std::make_heap(photons.begin(), photons.end());
			maxDist = sqrtf(photons[0].maxDist2);
		}
	}
	else {
		// Remove most distant photon from heap and add new photon
		std::pop_heap(photons.begin(), photons.end());
		photons[maxPhotons-1] = ClosePhoton(&photon, d2);
		std::push_heap(photons.begin(), photons.end());
		maxDist = sqrtf(photons[0].maxDist2);
	}
}

// BBBBBBBBBBLLLLLLLLLLLAAAAAAAAAAAAAHHHHHHHHHHHHHHHHHHHH
//////////////////////////////////////////////////////////////////

extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
  int nStore = params.FindOneInt("nphotons", 20000);
  int nStoreInd = params.FindOneInt("indirectphotons", 100000);
  int nUsed = params.FindOneInt("nused", 50);
  int maxDepth = params.FindOneInt("maxdepth", 5);
  int flagfinalgather = params.FindOneInt("finalgather", 1);
  int flagpregather = params.FindOneInt("pregather", flagfinalgather);
  int finalgathersamples = params.FindOneInt("finalgathersamples", 16);
  Float maxDist = params.FindOneFloat("maxdist", .1f);
  return new PhotonIntegrator(nStore, nStoreInd, nUsed, maxDepth, maxDist,
			      flagfinalgather, flagpregather, finalgathersamples);
}
