// photonmap.cc*


#include "lrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
#include "kdtree.h"
#include "sampling.h"


// Photonmap Local Declarations
//////////////////////////////////////////////////////////////////

struct PhotonData;

struct PhotonProcess;

class PhotonIntegrator : public SurfaceIntegrator {
public:
	// PhotonIntegrator Methods
	PhotonIntegrator(int nStored, int nStoredInd, int nLookup, int mdepth, Float maxdist);
	~PhotonIntegrator();
	Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample, Float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene);
	void Preprocess(const Scene *);
private:
	// PhotonIntegrator Private Data
	u_int nStored, nLookup, nStoredInd;
	mutable int pathLength;
	int maxDepth;
	Float maxDist;
	int lightSampleOffset, bsdfSampleOffset;
	mutable KdTree<PhotonData, PhotonProcess> *tree;
};


struct PhotonData {
	PhotonData() { }
	PhotonData(const DifferentialGeometry &dg, const Spectrum &L,
		const Vector &w)
		: P(dg.P), N(dg.Nn), dE(L), wi(w) { }
	Point P;
	Normal N;
	Spectrum dE;
	Vector wi;
};

struct ClosePhoton {
	ClosePhoton(const PhotonData *d = NULL, Float md2 = HUGE_VAL) {
		data = d;
		maxDist2 = md2;
	}
	bool operator<(const ClosePhoton &p2) const {
		return maxDist2 < p2.maxDist2;
	}
	const PhotonData *data;
	Float maxDist2;
};

struct PhotonProcess {
	PhotonProcess(u_int mp, const Point &p, const Normal &n);
	void operator()(const PhotonData &photon, Float &maxDist) const;

	const Point &P;
	const Normal &N;

	mutable vector<ClosePhoton> photons;
	u_int maxPhotons;
};



// PhotonIntegrator Method Definitions
//////////////////////////////////////////////////////////////////


PhotonIntegrator::PhotonIntegrator(int ns, int nsi, int nl, int mdepth,
		Float mdist) {
	nStored = ns;
	nStoredInd = nsi;
	nLookup = nl;
	maxDist = mdist;
	maxDepth = mdepth;
	tree = NULL;
	pathLength = 0;
}


PhotonIntegrator::~PhotonIntegrator() {
	delete tree;
}


void PhotonIntegrator::RequestSamples(Sample *sample, const Scene *scene) {
	lightSampleOffset = sample->Add2D(scene->lights.size());
	bsdfSampleOffset = sample->Add2D(scene->lights.size());
}


enum {
	FLAG_CONT=0,
	FLAG_FINSPEC=1,
	FLAG_FINGLOB=2,
	FLAG_FIN=3
};

enum {
	FLAG_DEBUT=0,
	FLAG_GOOD=1,
	FLAG_BAD=2
};

void PhotonIntegrator::Preprocess(const Scene *scene) {
	ProgressReporter progress(nStored+nStoredInd, "Shooting photons");
	vector<PhotonData> photons;
	photons.reserve(nStored);
	vector<PhotonData> photonsind;
	photonsind.reserve(nStored);
	int flagfin=FLAG_CONT;
	int flag_forcaustic=FLAG_DEBUT;
	bool specularBounce = false;


	printf(" Yeah ! Photon Shooting : %d for caustics, %d for global illumination\n",nStored, nStoredInd);

	// Initialize photon shooting statistics
	static StatsCounter nshot("Integrator",
		"Number of photons shot from lights");

	while (flagfin!=FLAG_FIN) {
		++nshot;
		// Give up if we're not finding any specular surfaces (after we have finished the global map)
		if ((flagfin&FLAG_FINGLOB) && nshot > 500000 && (!photons.size() || nshot / photons.size() > 10000)) {
			cerr << "No luck shooting photons for specular reflections !!!" << endl;
			return;
		}
		// Trace a photon path and store contribution
		Float u[4];
		u[0] = RadicalInverse(nshot+1, 2);
		u[1] = RadicalInverse(nshot+1, 3);
		u[2] = RadicalInverse(nshot+1, 5);
		u[3] = RadicalInverse(nshot+1, 7);
		// Choose light to shoot photon from
		int nLights = int(scene->lights.size());
		Light *light = scene->lights[RandomInt() % nLights];
		Float lightWeight = 1.f / nLights;

		// Generate [[photonRay]] from light source
		RayDifferential photonRay;
		Float weight;
		bool isDeltaLight;
		Spectrum L = light->Sample_L(scene, u[0], u[1], u[2], u[3],
			&photonRay, &weight, &isDeltaLight);
		L /= weight * lightWeight;
		flag_forcaustic=FLAG_DEBUT;


		if (!L.Black()) {
			// [ add the stat counter ? ] ????????????????????????
			int nBounces = 0;
			Intersection photonIsect;
			BSDF *photonBSDF = 0;

			static StatsRatio specularHits("Integrator",
				"Photons that hit specular surface", true);
			specularHits.add(0, 1);

			// Let's go until there is no more intersection and we hit the intersideral void
			while (scene->Intersect(photonRay, &photonIsect)) {
				L *= scene->Transmittance(photonRay);
				delete photonBSDF;
				photonBSDF = photonIsect.GetBSDF(photonRay);

				// Handle non-specular photon hit: store the photon in the caustic
				// map if there was only specular surfaces til now, and in the 
				// global map in any case
				if (photonBSDF->NumComponents() > photonBSDF->NumSpecular()) {
					if(flag_forcaustic==FLAG_GOOD && !(flagfin & FLAG_FINSPEC)) { 
						photons.push_back(PhotonData(photonIsect.dgGeom, L,	-photonRay.D.Hat()));
						progress.Update();
						flag_forcaustic=FLAG_BAD;
					}
					// In any case: store into photonsind
					if(!(flagfin & FLAG_FINGLOB)) {
						photonsind.push_back(PhotonData(photonIsect.dgGeom, L,	-photonRay.D.Hat()));
						progress.Update();
					}
				}

				if(!(flagfin & FLAG_FINGLOB)) {   // Global photon map not yet completed
					// Randomly sample BSDF to get new path direction
					//					Vector wo = -photonRay.D.Hat();
					Vector wo = -photonRay.D;  // Right ??????????????????
					Vector wi;

					Float bs1, bs2;
					bs1 = RandomFloat();    // Mieux ????????????????????????????
					bs2 = RandomFloat();
					
					Float weight;
					Spectrum f = photonBSDF->sample_f(wo, &wi, bs1, bs2, &weight, &specularBounce);
					if (f == Spectrum(0.) || weight == 0.) {
						break;
					}
					photonRay = Ray(photonIsect.dgGeom.P, wi);

					if(specularBounce) {  
						if(flag_forcaustic==FLAG_DEBUT) {
							specularHits.add(1, 0);
							flag_forcaustic=FLAG_GOOD;
						}
						//						L *= f * Float(photonBSDF->NumSpecular());    // C'est quoi ca ?
						L *= f;    // C'est quoi ca ?
					} else {
						flag_forcaustic=FLAG_BAD;  // not a specular bounce, no caustic here
						L *= f * fabsf(Dot(wi, photonIsect.dgGeom.Nn)) / weight;   // Shading or geometry ?
					}
				}
				
				// Handle specular photon hit
				else if (photonBSDF->NumSpecular() > 0) {    // we just need to complete the caustic map at that point
					if(flag_forcaustic==FLAG_DEBUT) {
						specularHits.add(1, 0);
						flag_forcaustic=FLAG_GOOD;
					}
					int specComponent = RandomInt() % photonBSDF->NumSpecular();
					Vector wi;
					Spectrum fr = photonBSDF->f_delta(specComponent,-photonRay.D, &wi);
					if (fr.Black()) {
						break;
					}
					L *= fr * Float(photonBSDF->NumSpecular());
					photonRay = Ray(photonIsect.dgGeom.P, wi);
				}
				else
					break;    // We have finished global photon map and there is no specularity on this surface
				
				if (nBounces++ > 3) {
					Float continueProbability = .2f;
					if (RandomFloat() > continueProbability)
						break;
					L /= continueProbability;
				}
			}
			delete photonBSDF;
		}
		if(photons.size()>=nStored)
			flagfin|=FLAG_FINSPEC;
		if(photonsind.size()>=nStoredInd)
			flagfin|=FLAG_FINGLOB;
	}
		

			/*	
			if (!L.Black()) {
			// Follow photon to non-specular surface and store in [[photons]] array
			bool hitSpecular = false;
			int nBounces = 0;
			Intersection photonIsect;
			BSDF *photonBSDF = 0;
			static StatsRatio specularHits("Integrator",
				"Photons that hit specular surface", true);
			specularHits.add(0, 1);

			while (scene->Intersect(photonRay, &photonIsect)) {
				L *= scene->Transmittance(photonRay);
				delete photonBSDF;
				photonBSDF = photonIsect.GetBSDF(photonRay);
				// Handle non-specular photon hit
				if (hitSpecular &&
					photonBSDF->NumComponents() > photonBSDF->NumSpecular()) {
					photons.push_back(PhotonData(photonIsect.dgGeom, L,
						-photonRay.D.Hat()));
					progress.Update();
				}

				// Handle specular photon hit
				if (photonBSDF->NumSpecular() > 0) {
					// Record photon specular hit statistics
					if (!hitSpecular)
						specularHits.add(1, 0);
					hitSpecular = true;
					int specComponent = RandomInt() % photonBSDF->NumSpecular();
					Vector wi;
					Spectrum fr = photonBSDF->f_delta(specComponent,
						-photonRay.D, &wi);
					if (fr.Black())
						break;
					L *= fr * Float(photonBSDF->NumSpecular());
					photonRay = Ray(photonIsect.dgGeom.P, wi);
				}

				// Possibly terminate photon path
				if (photonBSDF->NumSpecular() == 0)
					break;
				if (nBounces++ > 3) {
					Float continueProbability = .2f;
					if (RandomFloat() > continueProbability)
						break;
					L /= continueProbability;
				}
			}
			delete photonBSDF;
		}
		//	while (photons.size() < nStored && photonsind.size()<nStoredInd)
		*/

  // Normalize photon [[dE]] values and build tree
	for (u_int i = 0; i < photons.size(); ++i)
	  photons[i].dE /= Float(nshot);
	tree = new KdTree<PhotonData, PhotonProcess>(photons);
	fprintf(stderr, "\n");
}


Spectrum PhotonIntegrator::L(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		Float *alpha) const {
	// Compute reflected radiance with photon map
	Spectrum L(0.);
	Intersection isect;
	if (scene->Intersect(ray, &isect)) {
		if (alpha) *alpha = 1.;
		// Compute emitted light if an area light source
		L += isect.Le(-ray.D);
		// Evaluate BSDF at hit point
		BSDF *bsdf = isect.GetBSDF(ray);
		const Point &P = isect.dgShading.P;
		const Normal &N = isect.dgShading.Nn;
		Vector wo = -ray.D.Hat();
		L += UniformSampleAllLights(scene, P, N, wo, bsdf, sample,
			lightSampleOffset, bsdfSampleOffset);

		// Perform density estimation with photon map
		if (tree) {
			PhotonProcess proc(nLookup, isect.dgShading.P, isect.dgShading.Nn);
			Float md = maxDist;
			tree->Lookup(isect.dgShading.P, proc, md);
			// Accumulate light from nearby photons
			vector<ClosePhoton> &photons = proc.photons;
			if (photons.size() > 0) {
				// Compute photon scale factor with density estimation
				Float scale = photons.size() / (nLookup * M_PI * md * md);
				for (u_int i = 0; i < photons.size(); ++i)
					L += scale * bsdf->f(wo, photons[i].data->wi) *
						fabsf(Dot(photons[i].data->wi, N)) *
						photons[i].data->dE;
			}
		}
		if (pathLength++ < maxDepth) {
			Vector wi;
			// Trace rays for specular reflection and refraction
			for (int i = 0; i < bsdf->NumSpecular(); ++i) {
				Spectrum fr = bsdf->f_delta(i, -ray.D, &wi);
				if (!fr.Black())
					L += scene->L(Ray(isect.dgShading.P, wi), sample) * fr *
						fabsf(Dot(wi, isect.dgShading.Nn));
			}
		}
		--pathLength;
		// Clean up from integration
		delete bsdf;
	}
	else {
		// Handle ray with no intersection
		if (alpha) *alpha = 0.;
		for (u_int i = 0; i < scene->lights.size(); ++i)
			L += scene->lights[i]->Le(ray);
		if (alpha && !L.Black()) *alpha = 1.;
		return L;
	}
	return L;
}


// Photonprocess Method Definitions
//////////////////////////////////////////////////////////////////

PhotonProcess::PhotonProcess(u_int mp, const Point &p,
		const Normal &n)
	: P(p), N(n) {
	maxPhotons = mp;
}

void PhotonProcess::operator()(const PhotonData &photon,
		Float &maxDist) const {
	if (Dot(photon.N, N) < .707f) return;
	Float d2 = DistanceSquared(photon.P, P);
	if (photons.size() == 0)
		photons.reserve(maxPhotons);
	if (photons.size() < maxPhotons) {
		// Add photon to unordered array of photons
		photons.push_back(ClosePhoton(&photon, d2));
		if (photons.size() == maxPhotons) {
			std::make_heap(photons.begin(), photons.end());
			maxDist = sqrtf(photons[0].maxDist2);
		}
	}
	else {
		// Remove most distant photon from heap and add new photon
		std::pop_heap(photons.begin(), photons.end());
		photons[maxPhotons-1] = ClosePhoton(&photon, d2);
		std::push_heap(photons.begin(), photons.end());
		maxDist = sqrtf(photons[0].maxDist2);
	}
}

// BBBBBBBBBBLLLLLLLLLLLAAAAAAAAAAAAAHHHHHHHHHHHHHHHHHHHH
//////////////////////////////////////////////////////////////////

extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
	int nStore = params.FindOneInt("nphotons", 20000);
	int nStoreInd = params.FindOneInt("indirectphotons", 1000000);
	int nUsed = params.FindOneInt("nused", 50);
	int maxDepth = params.FindOneInt("maxdepth", 5);
	Float maxDist = params.FindOneFloat("maxdist", .1f);
	return new PhotonIntegrator(nStore, nStoreInd, nUsed, maxDepth, maxDist);
}
