// path.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "transport.h"
#include "scene.h"
// PathIntegrator Declarations
class PathIntegrator : public SurfaceIntegrator {
public:
	// PathIntegrator Method Declarations
	Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample, Float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene);
private:
	// PathIntegrator Private Data
	#define SAMPLE_DEPTH 3
	int lightSampleOffset[SAMPLE_DEPTH];
	int bsdfSampleOffset[SAMPLE_DEPTH];
};
// PathIntegrator Method Definitions
void PathIntegrator::RequestSamples(Sample *sample,
		const Scene *scene) {
	for (int i = 0; i < SAMPLE_DEPTH; ++i) {
		lightSampleOffset[i] = sample->Add2D(1);
		bsdfSampleOffset[i] = sample->Add2D(1);
	}
}
Spectrum PathIntegrator::L(const Scene *scene,
		const RayDifferential &r, const Sample *sample,
		Float *alpha) const {
	// Declare common path integration variables
	Spectrum pathThroughput = 1.;
	Spectrum L = 0.;
	Ray ray = r;
	bool specularBounce = false;
	int pathLength = 0;
	while (1) {
		// Find next vertex of path
		Intersection isect;
		if (!scene->Intersect(ray, &isect)) {
			for (u_int i = 0; i < scene->lights.size(); ++i)
				L += pathThroughput * scene->lights[i]->Le(ray);
			if (pathLength == 0 && alpha) {
				if (L != 0.) *alpha = 1.;
				else *alpha = 0.;
			}
			break;
		}
		if (pathLength == 0) {
			r.maxt = ray.maxt;
			if (alpha) *alpha = 1.;
		}
		// Add emitted light for first segment only
		if (pathLength == 0 || specularBounce)
			L += pathThroughput * isect.Le(-ray.D);
		// Evaluate BSDF at hit point
		BSDF *bsdf = isect.GetBSDF(ray);
		// Randomly sample illumination from one light source
		const Point &P = isect.dgShading.P;
		const Normal &N = isect.dgShading.Nn;
		Vector wi;
		Vector wo = -ray.D.Hat();
		L += pathThroughput * UniformSampleOneLight(scene, P, N, wo, bsdf, sample,
			pathLength < SAMPLE_DEPTH ? lightSampleOffset[pathLength] : -1,
			pathLength < SAMPLE_DEPTH ? bsdfSampleOffset[pathLength] : -1);
		// Randomly sample BSDF to get new path direction
		Float bs1, bs2;
		if (pathLength < SAMPLE_DEPTH) {
			bs1 = sample->twoD[bsdfSampleOffset[pathLength]][0];
			bs2 = sample->twoD[bsdfSampleOffset[pathLength]][1];
		}
		else {
			bs1 = RandomFloat();
			bs2 = RandomFloat();
		}
		Float weight;
		Spectrum f = bsdf->sample_f(wo, &wi, bs1, bs2,
			&weight, &specularBounce);
		if (f == Spectrum(0.) || weight == 0.) {
			delete bsdf;   // Added by G. Poncin !!!!!!!
			break;
		}
		pathThroughput *= f * fabsf(Dot(wi, N)) / weight;
		ray = Ray(P, wi);
		// Clean up from integration
		delete bsdf;
		// Possibly terminate the path
		if (pathLength > 3) {
			Float continueProbability = .2f;
			if (RandomFloat() > continueProbability)
				break;
			pathThroughput /= continueProbability;
		}
		++pathLength;
	}
	return L;
}
extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
	return new PathIntegrator;
}
