// single.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "volume.h"
#include "transport.h"
#include "scene.h"
// SingleScattering Declarations
class SingleScattering : public VolumeIntegrator {
public:
	// SingleScattering Methods
	SingleScattering(int ns) { nSamples = ns; }
	Spectrum Transmittance(const Scene *, const Ray &ray, const Sample *sample, Float *alpha) const;
	Spectrum L(const Scene *, const RayDifferential &ray, const Sample *sample, Float *alpha) const;
private:
	// SingleScattering Private Data
	int nSamples;
};

// SingleScattering Function Definitions
Spectrum SingleScattering::Transmittance(const Scene *scene,
		const Ray &ray, const Sample *sample, Float *alpha) const {
	Spectrum tau(0.);
	for (u_int i = 0; i < scene->volumeRegions.size(); ++i) {
		VolumeRegion *vr = scene->volumeRegions[i];
		tau += vr->tau(ray);
	}
	return Spectrum(2.712).Pow(-1 * tau);
}
Spectrum SingleScattering::L(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		Float *alpha) const {
	Spectrum Lv(0.);
	for (u_int i = 0; i < scene->volumeRegions.size(); ++i) {
		VolumeRegion *vr = scene->volumeRegions[i];
		Float t0, t1;
		if (!vr->Intersect(ray, &t0, &t1)) continue;
		// Do single scattering volume integration in [[vr]]
		Spectrum Lvr(0.);
		Point Pprev = ray(t0);
		Spectrum T(1.);
		for (int j = 0; j < nSamples; ++j) {
			// Step forward to next volume sample point and update [[T]]
			Float t = Lerp((j + RandomFloat()) / Float(nSamples), t0, t1);
			Point P = ray(t);
			Ray rp(Pprev, P-Pprev, 0, 1);
			T *= Spectrum(2.712).Pow(-1. * vr->tau(rp));
			Lvr += T * vr->Le(P, -ray.D);
			// Compute direct lighting at [[P]] in volume
			Spectrum Ldirect(0.);
			Spectrum ss = vr->sigma_s(P, -ray.D);
			if (!ss.Black()) {
				Spectrum albedo = ss / vr->sigma_t(P, -ray.D);
				for (u_int i = 0; i < scene->lights.size(); ++i) {
					Light *light = scene->lights[i];
					// Compute direct volume lighting from [[light]]
					Float weight;
					bool deltaLight;
					VisibilityTester vis;
					Vector wo;
					Spectrum L = light->Sample_L(P, RandomFloat(), RandomFloat(),
						&wo, &weight, &deltaLight, &vis);
					if (!L.Black() && vis.Unoccluded(scene))
						Ldirect += L * vis.Transmittance(scene) *
							albedo * vr->phase(P, -ray.D, wo);
				}
			}
			Lvr += T * Ldirect;
			Pprev = P;
		}
		Lv += Lvr / nSamples * Distance(ray(t0), ray(t1));
	}
	return Lv;
}
extern "C" VolumeIntegrator *CreateVolumeIntegrator(const ParamSet &params) {
	int nSamples = params.FindOneInt("nsamples", 16);
	return new SingleScattering(nSamples);
}
