// specialtree.cc* integrator
// Based on photonmapping technique...


#include "lrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
#include "kdtree3.h"
#include "sampling.h"
#include "camera.h"
#include "film.h"

// Photonmap Local Declarations
//////////////////////////////////////////////////////////////////

struct PhotonData;

struct PhotonProcess;

class TreeIntegrator : public SurfaceIntegrator {
public:
  // PhotonIntegrator Methods
  TreeIntegrator(int nStoredInd, int nLookup, int mdepth,
		 Float maxdist, int flagfg, int flagpg, int fgs);
  ~TreeIntegrator();
  Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample, Float *alpha) const;
  void RequestSamples(Sample *sample, const Scene *scene);
  void Preprocess(const Scene *);
  void recursiveGather(KdNode<PhotonData>	*node, KdTree<PhotonData, PhotonProcess> *tree) const;
  PhotonData* TreeIntegrator::recursiveLookupOne(KdNode<PhotonData>	*node, const Point &P, const Normal &N) const;
private:
  //PhotonIntegrator Private Data
  u_int nLookup, nStoredInd;
  mutable int pathLength;
  int maxDepth;
  mutable int flagfinalgather, finalgathersamples, flagpregather;
  Float maxDist;
  int lightSampleOffset, bsdfSampleOffset;
  mutable int isfinal;    // Is this ray a final ray
  mutable KdTree<PhotonData, PhotonProcess> *treeind;
  Spectrum skycol;
};


struct PhotonData {
  PhotonData() { }
  PhotonData(const DifferentialGeometry &dg, const Spectrum &L,
	     const Vector &w, BSDF *bsdf)
    : P(dg.P), N(dg.Nn), dE(L), wi(w) { this->bsdf=bsdf; }
  Point P;
  Normal N;
  Spectrum dE;
  Spectrum Gath;
  BSDF *bsdf;
  Vector wi;
};

struct ClosePhoton {
  ClosePhoton(const PhotonData *d = NULL, Float md2 = HUGE_VAL) {
    data = d;
    maxDist2 = md2;
  }
  bool operator<(const ClosePhoton &p2) const {
    return maxDist2 < p2.maxDist2;
  }
  const PhotonData *data;
  Float maxDist2;
};

struct PhotonProcess {
  PhotonProcess(u_int mp, const Point &p, const Normal &n);
  void operator()(const PhotonData &photon, Float &maxDist) const;

  const Point &P;
  const Normal &N;

  mutable vector<ClosePhoton> photons;
  u_int maxPhotons;
};



// PhotonIntegrator Method Definitions
//////////////////////////////////////////////////////////////////


TreeIntegrator::TreeIntegrator(int nsi, int nl, int mdepth,
			       Float mdist, int flagfg, int flagpg, int fgs) {
  nStoredInd = nsi;
  nLookup = nl;
  maxDist = mdist;
  maxDepth = mdepth;
  treeind = NULL;
  pathLength = 0;
  flagfinalgather=flagfg;
  flagpregather=flagpg;
  finalgathersamples=fgs;
  //Sky color
  Float cs[3];
  cs[0]=0.06;
  cs[1]=0.06;
  cs[2]=0.2;
  /*
    cs[0]=0.005;
    cs[1]=0.005;
    cs[2]=0.04;
  */
  skycol=Spectrum(cs);
}


TreeIntegrator::~TreeIntegrator() {
  delete treeind;
}


void TreeIntegrator::RequestSamples(Sample *sample, const Scene *scene) {
  lightSampleOffset = sample->Add2D(scene->lights.size());
  bsdfSampleOffset = sample->Add2D(scene->lights.size());
}


void TreeIntegrator::Preprocess(const Scene *scene) {
  vector<PhotonData> photonsind;
  photonsind.reserve(nStoredInd);

  printf("Photon Shooting: %d for global illum., %d used, final gather: %d\n",
	 nStoredInd, nLookup, (flagfinalgather==1) ?
	 finalgathersamples : 0);

  // Initialize photon shooting statistics
  static StatsCounter nshot2("Integrator",
			     "Number of photons shot from lights for GI");

  int nLights = int(scene->lights.size());
  // GLOBAL ILLUMINATION
  ProgressReporter progress2(nStoredInd, "Shooting photons for global illumination");

  int count=0;

  while (photonsind.size() < nStoredInd) {		
    ++nshot2;
    // Trace a photon path and store contribution
    Float u[4];
    u[0] = RadicalInverse(nshot2+1, 2);
    u[1] = RadicalInverse(nshot2+1, 3);
    u[2] = RadicalInverse(nshot2+1, 5);
    u[3] = RadicalInverse(nshot2+1, 7);
    // Choose light to shoot photon from
    int which=RandomInt() % nLights;
    Light *light = scene->lights[which];
    Float lightWeight = 1.f / nLights;

    // Generate [[photonRay]] from light source
    RayDifferential photonRay;
    Float weight;
    bool isDeltaLight;
    Spectrum L = light->Sample_L(scene, u[0], u[1], u[2], u[3],
				 &photonRay, &weight, &isDeltaLight);
    L /= weight * lightWeight;

    count++;
    //		cout << "Light:" << which << "   D=" << photonRay.D<<"  " <<  count << endl;

    if (!L.Black()) {
      int nBounces = 0;
      int hitSpecular = 0;
      Intersection photonIsect;
      BSDF *photonBSDF = 0;

      // Let's go until there is no more intersection and we hit the intersideral void
      while (scene->Intersect(photonRay, &photonIsect)) {
				
	L *= scene->Transmittance(photonRay);
	//				delete photonBSDF;
	photonBSDF = photonIsect.GetBSDF(photonRay);

	// Handle non-specular photon hit: store the photon in the map
	// except if it is a LS+D

	if ((photonBSDF->NumComponents() > photonBSDF->NumSpecular()) && (hitSpecular!=1) && photonIsect.primitive->material->isphoton) {
	  photonsind.push_back(PhotonData(photonIsect.dgGeom, L,
					  -photonRay.D.Hat(), photonBSDF));
	  //					cout << "P:" << photonIsect.dgGeom.P<< "   L " << L <<endl;
	  progress2.Update();
	}

	// Randomly sample BSDF to get new path direction
	//					Vector wo = -photonRay.D.Hat();
	Vector wo = -photonRay.D.Hat();  // Right ??????????????????
	Vector wi;
	Float bs1, bs2;
	bs1 = RandomFloat();
	bs2 = RandomFloat();

			
	Float weight;
	bool specularBounce = false;
	Spectrum f = photonBSDF->sample_f(wo, &wi, bs1, bs2, &weight, &specularBounce);
	if (f == Spectrum(0.) || weight == 0.) {
	  break;
	}
	photonRay = Ray(photonIsect.dgGeom.P, wi);
	//				cout << "Foireux ? " << photonRay.D;
	if(!(fabs(photonRay.D.x)<100000))  // Bad photon !
	  break;
				
	if(specularBounce) {
	  //						L *= f * Float(photonBSDF->NumSpecular());    // C'est quoi ca ?
	  L *= f * Float(photonBSDF->NumSpecular())* fabsf(Dot(wi, photonIsect.dgShading.Nn));     // * cos blah blah blah ?
	  if(hitSpecular==0)
	    hitSpecular=1;    // we are starting on a series of specular surfaces
	}
	else {
	  //					L *= f * fabsf(Dot(wi, photonIsect.dgGeom.Nn)) / weight;   // Shading or geometry ?
	  L *= f * fabsf(Dot(wi, photonIsect.dgShading.Nn)) / weight;   // Shading or geometry ?
	  hitSpecular=2;     // Exiting the series of spec. surfaces
	}
				

	nBounces++;
	// Why not max Depth ?
	if (nBounces > 1) {
	  Float continueProbability = .2f;
	  if (RandomFloat() > continueProbability)
	    break;
	  L /= continueProbability;
	}
      }
      //			delete photonBSDF;
    }
  }
		
  // Normalize photon [[dE]] values and build tree
  for (u_int i = 0; i < photonsind.size(); ++i)
    photonsind[i].dE /= Float(nshot2)*150;   ///////////////////////////////////////////////HHHHHHHHAAAAAAAAAACKKKKK

  treeind = new KdTree<PhotonData, PhotonProcess>(photonsind);

  // Pre-gather the photons
  if(flagpregather) {
    printf("\nPrecomputing irradiance\n");
    recursiveGather(treeind->root,treeind);
  }

  if(!flagfinalgather) {
    isfinal=1;    // All the rays are final
  }

  fprintf(stderr, "\n");
}


void TreeIntegrator::recursiveGather(KdNode<PhotonData>	*node, KdTree<PhotonData, PhotonProcess> *tree) const {
  if (!node) return;

  const Point &P = node->data.P;
  const Normal &N = node->data.N;

  // Gather irradiance at this photon...
  Spectrum Ltmp(0.);
  PhotonProcess proc(nLookup, P, N);
  Float md = maxDist;
  tree->Lookup(P, proc, md);
  // Accumulate light from nearby photons
  vector<ClosePhoton> &photons = proc.photons;
  // Compute photon scale factor with density estimation
  Float scale = photons.size() / (nLookup * M_PI * md * md);
  for (u_int i = 0; i < photons.size(); ++i) {
    Ltmp +=  node->data.bsdf->f(Vector(-node->data.N), photons[i].data->wi) *
      photons[i].data->dE;
    //		Ltmp +=  node->data.bsdf->f(Vector(node->data.N), photons[i].data->wi) *
    //			photons[i].data->dE;
  }
  node->data.Gath=Ltmp*scale;
  //		cout << node->data.Gath << endl;
  //	delete node->data.bsdf;
  recursiveGather(node->children[0],tree);
  recursiveGather(node->children[1],tree);
}

PhotonData* TreeIntegrator::recursiveLookupOne
(KdNode<PhotonData>	*node, const Point &P, const Normal &N) const
{
  PhotonData *ret;
  if (!node) return NULL;

  // Process Kd node's children based on $x$ split
  if (node->split == SPLIT_X) {
    if (P.x <= node->data.P.x)   // Point on the left
      ret=recursiveLookupOne(node->children[0], P, N);
    else
      ret=recursiveLookupOne(node->children[1], P, N);
  }
  else if (node->split == SPLIT_Y) {
    if (P.y <= node->data.P.y)   // Point on the left
      ret=recursiveLookupOne(node->children[0], P, N);
    else
      ret=recursiveLookupOne(node->children[1], P, N);
  }
  else {
    if (P.z <= node->data.P.z)   // Point on the left
      ret=recursiveLookupOne(node->children[0], P, N);
    else
      ret=recursiveLookupOne(node->children[1], P, N);
  }

  if(Dot(node->data.N, N) >= .707f) {
    if(ret &&	(DistanceSquared(P, ret->P)<DistanceSquared(P,node->data.P)))
      return ret;
    else
      return &node->data;
  }
  else
    return ret;
}

Spectrum TreeIntegrator::L(const Scene *scene,
			   const RayDifferential &ray, const Sample *sample,
			   Float *alpha) const {
  // Compute reflected radiance with photon map
  Spectrum L(0.);
  Intersection isect;
  bool specularBounce;


  if (scene->Intersect(ray, &isect)) {
    if(pathLength==0 && flagfinalgather)
      isfinal=0;    // A ray is always non final when it is cast...

    if (alpha) *alpha = 1.;
    L += isect.Le(-ray.D);
    Vector wo = -ray.D.Hat();
    BSDF *bsdf = isect.GetBSDF(ray);
    const Point &P = isect.dgShading.P;
    const Normal &N = isect.dgShading.Nn;

    if(debugflag)
      cout << "P " << P <<"      N " << N << endl;
    if(debugflag)
      cout << "  Direct light: " << L << endl;
		
    // First bounce, evaluate 
    if(!isfinal) {
      // The ray is not final, we are going to 
      // - get the contributions of the light, 
      // - sample over the BSDF and cast final rays !
      // - if specularities: trace the ray
		  
      // Compute illumination by integrating over the lights
      Vector wi;
      //compute translucency if a translucent object
      if(isect.primitive->material->istranslucent == true)
	{
	  L += isect.primitive->shape->getTranslucentDE(P,N);
	}
      if(debugflag)
	cout << "Translucent L: " << L << endl;
		  
      for (u_int i = 0; i < scene->lights.size(); ++i) {
	VisibilityTester visibility;
	Spectrum dE = scene->lights[i]->dE(P,N, &wi, &visibility);
	if (dE.Black()) continue;
	Spectrum f = bsdf->f(-ray.D, wi);
	if (!f.Black() && visibility.Unoccluded(scene))
	  L += f * dE * visibility.Transmittance(scene);
      }
      if(debugflag)
	cout << "Lighting L: " << L << endl;

      if(debugflag)
	cout << "  First color: " << L << endl;

      // Faster than:
      //L += UniformSampleAllLights(scene, P, N, wo, bsdf, sample,
      //														lightSampleOffset, bsdfSampleOffset);

      if (pathLength++ < maxDepth) {
				
	if(bsdf->NumComponents() > bsdf->NumSpecular()) {  // If the object is a little bit diffuse
	  Vector wi;
	  // Randomly sample BSDF to get new path direction for final gathering ray
	  Float mysamples[2*finalgathersamples];
	  LatinHypercube(mysamples, finalgathersamples, 2);
	  Float weight;
	  Spectrum Ltmp(0.);
	  for(int kk=0;kk<finalgathersamples;kk++) {
	    Spectrum f = bsdf->sample_f(wo, &wi, mysamples[kk*2], mysamples[kk*2+1],
					&weight, &specularBounce);
	    if(specularBounce)
	      isfinal=0;
	    else
	      isfinal=1;
						
	    if (!f.Black() && weight!=0)
	      Ltmp+= scene->L(Ray(P, wi), sample) * f *	fabsf(Dot(wi, N)) / weight;
	  }
	  L+=Ltmp/finalgathersamples;
	  if(debugflag)
	    cout << "  After gather: " << L << endl;
	}
	else{			// If only specular (water...)
	  Vector wi;
	  // Trace rays for specular reflection and refraction
	  for (int i = 0; i < bsdf->NumSpecular(); ++i) {
	    Spectrum fr = bsdf->f_delta(i, -ray.D, &wi);
	    if (!fr.Black())
	      L += scene->L(Ray(P, wi), sample) * fr *
		fabsf(Dot(wi, N));
	  }
	  if(debugflag)
	    cout << "  After spec: " << L << endl;
	}
      }
    }
    else {
      //			flagpregather=false;
      // The ray is final...
      // global illumination
      if(treeind && isect.primitive->material->isphoton){
	if(flagpregather) {
	  PhotonData *pd=recursiveLookupOne(treeind->root,P,N);
	  if(pd && Distance(pd->P,P)<maxDist /*&& pd->Gath.y() < 0.2*/)
	    L+=pd->Gath;
	  if(debugflag)
	    cout << "  After one gathering: " << L << endl;
	}
	else {
	  PhotonProcess proc(nLookup, P, N);
	  Float md = maxDist;
	  treeind->Lookup(P, proc, md);
	  // Accumulate light from nearby photons
	  vector<ClosePhoton> &photonsind = proc.photons;
	  // Compute photon scale factor with density estimation
	  Float scale = photonsind.size() / (nLookup * M_PI * md * md);
	  for (u_int i = 0; i < photonsind.size(); ++i)
	    L += scale *  bsdf->f(wo, photonsind[i].data->wi) *
	      photonsind[i].data->dE;
	}
      }

      pathLength++;
      // Still tracing for specular surfaces...
      if (pathLength < maxDepth) {
	Vector wi;
	// Trace rays for specular reflection and refraction
	for (int i = 0; i < bsdf->NumSpecular(); ++i) {
	  Spectrum fr = bsdf->f_delta(i, -ray.D, &wi);
	  if (!fr.Black()) {
	    if(fr.y()<1000000000)
	      L += scene->L(Ray(P, wi), sample) * fr *
		fabsf(Dot(wi, N));
	    else
	      L+=skycol;
	  }
	  // To compensate for a bug in windy...
	}
	if(debugflag)
	  cout << "  After specularities: " << L << endl;
      }
			
    }
    --pathLength;
		
    // Clean up from integration
    delete bsdf;
  }
  else {
    // Handle ray with no intersection
    if (alpha) *alpha = 0.;
    for (u_int i = 0; i < scene->lights.size(); ++i)
      L += scene->lights[i]->Le(ray);
    if (alpha && !L.Black()) *alpha = 1.;
    // HACKKK !
    else {
#ifdef ADDSKY
      if(pathLength==0) {
	DifferentialGeometry dg;
	dg.u=sample->imageX/scene->camera->film->xResolution;
	dg.v=sample->imageY/scene->camera->film->yResolution/4*5;
	if(dg.v>1)
	  dg.v-=0.5;
	dg.dudx=1/640;
	dg.dvdx=0;
	dg.dudy=0;
	dg.dvdy=1/400;
				
	//			if(debugflag)
	//				cout << dg.u << " " << dg.v << endl;
				
	Spectrum colorsky=scene->Ksky->Evaluate(dg);
				
	//			if(debugflag)
	//				cout << colorsky<< endl;
				
	L+=colorsky/5;
      }
      else
	L+=skycol;
#else
      L+=skycol;
#endif
    }
    if(debugflag)
      cout << "   Return L sky " << L << endl<<endl;
    return L;
  }
  if(debugflag)
    cout << "   Return L " << L << endl<<endl;
  return L;
}

// Photonprocess Method Definitions
//////////////////////////////////////////////////////////////////

PhotonProcess::PhotonProcess(u_int mp, const Point &p,
			     const Normal &n)
  : P(p), N(n) {
  maxPhotons = mp;
}

void PhotonProcess::operator()(const PhotonData &photon,
			       Float &maxDist) const {
  if (Dot(photon.N, N) < .707f) return;
  Float d2 = DistanceSquared(photon.P, P);
  if (photons.size() == 0)
    photons.reserve(maxPhotons);
  if (photons.size() < maxPhotons) {
    // Add photon to unordered array of photons
    photons.push_back(ClosePhoton(&photon, d2));
    if (photons.size() == maxPhotons) {
      std::make_heap(photons.begin(), photons.end());
      maxDist = sqrtf(photons[0].maxDist2);
    }
  }
  else {
    // Remove most distant photon from heap and add new photon
    std::pop_heap(photons.begin(), photons.end());
    photons[maxPhotons-1] = ClosePhoton(&photon, d2);
    std::push_heap(photons.begin(), photons.end());
    maxDist = sqrtf(photons[0].maxDist2);
  }
}

// BBBBBBBBBBLLLLLLLLLLLAAAAAAAAAAAAAHHHHHHHHHHHHHHHHHHHH
//////////////////////////////////////////////////////////////////

extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
  int nStoreInd = params.FindOneInt("indirectphotons", 100000);
  int nUsed = params.FindOneInt("nused", 50);
  int maxDepth = params.FindOneInt("maxdepth", 5);
  int flagfinalgather = params.FindOneInt("finalgather", 1);
  int flagpregather = params.FindOneInt("pregather", flagfinalgather);
  int finalgathersamples = params.FindOneInt("finalgathersamples", 16);
  Float maxDist = params.FindOneFloat("maxdist", .1f);
  return new TreeIntegrator(nStoreInd, nUsed, maxDepth, maxDist,
			    flagfinalgather, flagpregather, finalgathersamples);
}
