#include "lrt.h"
#include "transport.h"
#include "scene.h"
// WhittedIntegrator Declarations
class WhittedIntegrator : public SurfaceIntegrator {
public:
  // WhittedIntegrator Methods
  Spectrum L(const Scene *scene, const RayDifferential &ray, const Sample *sample,
	     Float *alpha) const;
  WhittedIntegrator(int md) {
    maxDepth = md;
    rayDepth = 0;
  }
private:
  // WhittedIntegrator Private Data
  int maxDepth;
  mutable int rayDepth;
};

// WhittedIntegrator Method Definitions
Spectrum WhittedIntegrator::L(const Scene *scene,
			      const RayDifferential &ray, const Sample *sample,
			      Float *alpha) const {
  Intersection isect;
  Spectrum L(0.);
  if (scene->Intersect(ray, &isect)){   
    if (alpha) *alpha = 1.;
    // Compute emitted and reflected light
    // Compute emitted light if an area light source
    L += isect.Le(-ray.D);
    // Evaluate BSDF at hit point
    BSDF *bsdf = isect.GetBSDF(ray);

    //compute translucency if a translucent object
    if(isect.primitive->material->istranslucent == true)
      {
				L += isect.primitive->shape->getTranslucentDE(isect.dgShading.P,
																											isect.dgShading.Nn);
      }
		if(debugflag)
			cout << "Translucent L: " << L << endl;


    // Compute reflection by integrating over the lights
    Vector wi;
    for (u_int i = 0; i < scene->lights.size(); ++i) {
      VisibilityTester visibility;
      Spectrum dE = scene->lights[i]->dE(isect.dgShading.P,
					 isect.dgShading.Nn, &wi, &visibility);
      if (dE.Black()) continue;
      Spectrum f = bsdf->f(-ray.D, wi);
      if (!f.Black() && visibility.Unoccluded(scene))
				L += f * dE * visibility.Transmittance(scene);
    }
		if(debugflag)
			cout << "Lighting L: " << L << endl;

    if (rayDepth++ < maxDepth) {
      // Trace rays for specular reflection and refraction
      for (int i = 0; i < bsdf->NumSpecular(); ++i) {
				Spectrum fr = bsdf->f_delta(i, -ray.D, &wi);
				if (!fr.Black())
					L += scene->L(Ray(isect.dgShading.P, wi), sample) * fr *
						fabsf(Dot(wi, isect.dgShading.Nn));
      }
    }
    --rayDepth;
    // Clean up from integration
    delete bsdf;
  }
  else {
    // Handle ray with no intersection
    if (alpha) *alpha = 0.;
    for (u_int i = 0; i < scene->lights.size(); ++i)
      L += scene->lights[i]->Le(ray);
    if (alpha && !L.Black()) *alpha = 1.;
    return L;
  }
  return L;
}
extern "C" SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params)
{
  int maxDepth = params.FindOneInt("maxdepth", 5);
  return new WhittedIntegrator(maxDepth);
}
