// infinitelight.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "light.h"
#include "texture.h"
#include "shapes.h"
#include "scene.h"
// InfiniteAreaLight Classes
class InfiniteAreaLight : public Light {
public:
	// InfiniteAreaLight Interface
	InfiniteAreaLight(bool shadows, const Transform &light2world,
		const Spectrum &power, Texture<Spectrum> *tex);
	Spectrum Le(const RayDifferential &r) const;
	Spectrum Sample_L(const Point &P, const Normal &N,
		Float u1, Float u2, Vector *wo, Float *weight,
		bool *deltaLight, VisibilityTester *visibility) const;
	Spectrum Sample_L(const Point &P, Float u1, Float u2, Vector *wo, Float *weight,
		bool *deltaLight, VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, Float u1, Float u2,
			Float u3, Float u4, Ray *ray, Float *weight,
			bool *deltaLight) const;
	Float Weight(const Point &, const Normal &, const Vector &) const;
	Float Weight(const Point &, const Vector &) const;
	Spectrum dE(const Point &P, const Normal &N, Vector *w, VisibilityTester *visibility) const;
	~InfiniteAreaLight();
private:
	// InfiniteAreaLight Private Data
	Texture<Spectrum> *radianceMap;
};
// InfiniteAreaLight Function Definitions
InfiniteAreaLight::~InfiniteAreaLight() {
	delete radianceMap;
}
InfiniteAreaLight::InfiniteAreaLight(bool shadows,
		const Transform &light2world,
		const Spectrum &power, Texture<Spectrum> *tex)
	: Light(shadows, light2world, power) {
	radianceMap = tex;
}
Spectrum InfiniteAreaLight::Le(const RayDifferential &r) const {
	Vector w = r.D;
	// Compute infinite light radiance for this direction
	Spectrum L = Power;
	if (radianceMap != NULL) {
		Vector wh = WorldToLight(w).Hat();
		Vector S, T;
		CoordinateSystem(wh, &S, &T);
		Float s = SphericalPhi(wh) / (2.f*M_PI);
		Float t = SphericalTheta(wh) / M_PI;
		DifferentialGeometry dgLight(Point(wh.x, wh.y, wh.z),
			S, T, Vector(0,0,0), Vector(0,0,0), s, t, NULL);
	//	dgLight.ComputeDifferentials(r);
		L *= radianceMap->Evaluate(dgLight);
	}
	return L;
}
Spectrum InfiniteAreaLight::Sample_L(const Point &P,
		const Normal &N, Float u1, Float u2,
		Vector *wo, Float *weight, bool *deltaLight,
		VisibilityTester *visibility) const {
	// Sample cosine-weighted direction on unit sphere
	Float x, y, z;
	ConcentricSampleDisk(u1, u2, &x, &y);
	z = sqrtf(max(0.f, 1.f - x*x - y*y));
	if (RandomFloat() < .5) z *= -1;
	*wo = Vector(x, y, z);
	// Compute [[weight]] for cosine-weighted infinite light direction
	*weight = fabsf(wo->z) / (2.f * M_PI);
	// Transform direction to world space
	Vector v1, v2;
	CoordinateSystem(Vector(N).Hat(), &v1, &v2);
	*wo = Vector(v1.x * wo->x + v2.x * wo->y + N.x * wo->z,
	             v1.y * wo->x + v2.y * wo->y + N.y * wo->z,
	             v1.z * wo->x + v2.z * wo->y + N.z * wo->z);
	*deltaLight = false;
	// XXX yuck
	visibility->SetRay(P, *wo, CastsShadows);
	return Le(Ray(P, *wo));
}
Float InfiniteAreaLight::Weight(const Point &, const Normal &N,
		const Vector &w) const {
	return fabsf(Dot(N, w)) / (2.f * M_PI);
}
Spectrum InfiniteAreaLight::Sample_L(const Point &P,
		Float u1, Float u2, Vector *wo, Float *weight,
		bool *deltaLight, VisibilityTester *visibility) const {
	*wo = UniformSampleSphere(u1, u2);
	*weight = 1.f / (4.f * M_PI);
	*deltaLight = false;
	visibility->SetRay(P, *wo, CastsShadows);
	return Le(Ray(P, *wo));
}
Float InfiniteAreaLight::Weight(const Point &, const Vector &) const {
	return 1.f / (4.f * M_PI);
}
Spectrum InfiniteAreaLight::Sample_L(const Scene *scene,
		Float u1, Float u2, Float u3, Float u4,
		Ray *ray, Float *weight, bool *deltaLight) const {
	// Choose two points [[p1]] and [[p2]] on scene bounding sphere
	Point worldCenter;
	Float worldRadius;
	scene->WorldBound().BoundingSphere(&worldCenter, &worldRadius);
	Point p1 = worldCenter + worldRadius * UniformSampleSphere(u1, u2);
	Point p2 = worldCenter + worldRadius * UniformSampleSphere(u3, u4);
	ray->O = p1;
	ray->D = (p2-p1).Hat();
	*deltaLight = false;
	*weight = 1.f / ((4 * M_PI * worldRadius * worldRadius) *
		(4 * M_PI * worldRadius * worldRadius));
	Spectrum L = Le(*ray);
	ray->D *= -1.;
	return L;
}
Spectrum InfiniteAreaLight::dE(const Point &P, const Normal &N,
		Vector *wo, VisibilityTester *visibility) const {
	*wo = Vector(0,0,0);
	return 0.;
}
extern "C" Light *CreateLight(bool castsShadows,
		const Transform &light2world,
		const ParamSet &paramSet) {
	// Compute infinite light radiance
	Spectrum L = paramSet.FindOneSpectrum("lightcolor", Spectrum(1.)) *
		paramSet.FindOneFloat("intensity", 1.);
	L *= M_PI;
	L = paramSet.FindOneSpectrum("L", L);
	// Create texture map for light
	string texmap = paramSet.FindOneString("mapname", "");
	Texture<Spectrum> *tex = NULL;
	if (texmap != "")
		tex = new ImageTexture<Spectrum>(new IdentityMapping2D,
			texmap.c_str());
	else
		tex = new ConstantTexture<Spectrum>(1.);
	return new InfiniteAreaLight(castsShadows, light2world,
		L, tex);
}
