// pointlight.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "light.h"
#include "shapes.h"
// PointLight Classes
class PointLight : public Light {
public:
  // PointLight Methods
  PointLight(bool shadows, const Transform &light2world,
	     const Spectrum &intensity, bool isglow, Float glow_size, Float glow_cutoff, Float glow_pow,Float glow_mul);
  Spectrum dE(const Point &P, const Normal &N, Vector *w,
	      VisibilityTester *vis) const;
  Spectrum Sample_L(const Point &P, Float u1, Float u2,
		    Vector *wo, Float *weight, bool *deltaLight,
		    VisibilityTester *visibility) const;
  Spectrum Sample_L(const Scene *scene, Float u1, Float u2,
		    Float u3, Float u4, Ray *ray, Float *weight,
		    bool *deltaLight) const;
  Float Weight(const Point &, const Vector &) const;
	Spectrum Glow(const RayDifferential &r) const;
private:
  // PointLight Private Data
  Point lightPos;
  Spectrum Intensity;
	Float glow_size;
	Float glow_cutoff;
	Float glow_pow;
	Float glow_mul;
};
// PointLight Method Definitions
PointLight::PointLight(bool shadows, const Transform &light2world,
		       const Spectrum &intensity, bool isglow, Float glow_size, Float glow_cutoff, Float glow_pow,Float glow_mul)
  : Light(shadows, light2world, intensity * 4. * M_PI) {
	this->doGlow=isglow;
  lightPos = LightToWorld(Point(0,0,0));
	copyPos = lightPos;
  Intensity = intensity;
	this->glow_size=glow_size;
	this->glow_cutoff=glow_cutoff;
	this->glow_pow=glow_pow;
	this->glow_mul=glow_mul;
}
Spectrum PointLight::dE(const Point &P, const Normal &N,
			Vector *w, VisibilityTester *visibility) const {
  *w = (lightPos - P).Hat();
  visibility->SetSegment(P, lightPos, CastsShadows);
  return Intensity * fabs(Dot(*w, N)) /
    DistanceSquared(lightPos, P);
}
Spectrum PointLight::Sample_L(const Point &P, Float u1, Float u2,
			      Vector *wo, Float *weight, bool *deltaLight,
			      VisibilityTester *visibility) const {
  *wo = (lightPos - P).Hat();
  *weight = DistanceSquared(lightPos, P);
  *deltaLight = true;
  visibility->SetSegment(P, lightPos, CastsShadows);
  return Intensity;
}
Spectrum PointLight::Sample_L(const Scene *scene, Float u1, Float u2,
			      Float u3, Float u4, Ray *ray, Float *weight,
			      bool *deltaLight) const {
  ray->O = lightPos;
  ray->D = UniformSampleSphere(u1, u2);
  *weight = 1.f / (4.f * M_PI);
  *deltaLight = true;
  return Intensity;
}
Float PointLight::Weight(const Point &, const Vector &) const {
  return 0.;
}

Spectrum PointLight::Glow(const RayDifferential &r) const {
	// We want the distance between the glow center and the ray  we consider
	Vector camGlow=lightPos-r.O;   // Camera to Glow
	Vector dir=r.D.Hat();
	Float cosangle=Dot(camGlow,dir);  
	Point inter=r.O+cosangle*dir;
	Float dist=(inter-lightPos).Length();
	Float coef=0.0f;
	
	if(glow_cutoff==0 || dist<glow_cutoff) {
		//		if(cosangle>0)
		coef=exp(-dist/glow_size);
		if(glow_pow!=1)
			coef=pow(coef,glow_pow);
	}
	return coef*Intensity*glow_mul;
}

extern "C" Light *CreateLight(bool castsShadows, const Transform &light2world,
			      const ParamSet &paramSet) {
  // Compute point-light intensity
  Spectrum I = paramSet.FindOneSpectrum("lightcolor", Spectrum(1.)) *
    paramSet.FindOneFloat("intensity", 1.);
  I *= M_PI;
  I = paramSet.FindOneSpectrum("I", I);
  Point P = paramSet.FindOnePoint("from", Point(0,0,0));
  int isglow=paramSet.FindOneInt("glow", 0);
	bool doGlow=false;
	Float glow_size=0.0f;
	Float glow_cutoff=0.0f;
	Float glow_pow=0.0f;
	Float glow_mul=0.0f;
	if(isglow) {
		doGlow=true;
		glow_size=paramSet.FindOneFloat("glowsize", 1.);
		glow_cutoff=paramSet.FindOneFloat("glowcutoff", 0.);
		glow_pow=paramSet.FindOneFloat("glowpower", 0.5);
		glow_mul=paramSet.FindOneFloat("glowmul", 0.01);
	}
  Transform l2w = Translate(Vector(P.x, P.y, P.z)) * light2world;
  return new PointLight(castsShadows, l2w, I, doGlow, glow_size, glow_cutoff, glow_pow, glow_mul);
}
