// projectionlight.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "light.h"
#include "shapes.h"
#include "texture.h"
// ProjectionLight Classes
class ProjectionLight : public Light {
public:
	ProjectionLight(bool shadows, const Transform &light2world,
		const Spectrum &intensity, Texture<Spectrum> *tex,
		Float fov);
	~ProjectionLight();
	Spectrum dE(const Point &P, const Normal &N, Vector *w,
		VisibilityTester *vis) const;
	// ProjectionLight Member Functions
	Spectrum Projection(const Vector &w) const;
	Spectrum Sample_L(const Point &P, Float u1, Float u2, Vector *wo,
		Float *weight, bool *deltaLight, VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, Float u1, Float u2,
			Float u3, Float u4, Ray *ray, Float *weight,
			bool *deltaLight) const;
	Float Weight(const Point &, const Vector &) const;
private:
	// ProjectionLight Private Data
	Texture<Spectrum> *projectionMap;
	Transform lightProjection;
	Float hither, yon;
	Point lightPos;
	Spectrum Intensity;
	Float cosTotalWidth;
};
// ProjectionLight Method Definitions
ProjectionLight::~ProjectionLight() { delete projectionMap; }
ProjectionLight::ProjectionLight(bool shadows,
		const Transform &light2world,
		const Spectrum &intensity, Texture<Spectrum> *tex,
		Float fov)
	: Light(shadows, light2world, intensity) {
	projectionMap = tex;
	hither = RAY_EPSILON;
	yon = 1e10;
	lightProjection = Perspective(fov, hither, yon);
	lightPos = LightToWorld(Point(0,0,0));
	Intensity = intensity;
	// Cache [[ProjectionLight]] diagonal fov for sampling
	Float opposite = tanf(Radians(fov) / 2.f);
	Float diag = opposite * sqrtf(2.f);
	Float tanDiag = diag;
	cosTotalWidth = cosf(atanf(tanDiag));
}
Spectrum ProjectionLight::dE(const Point &P, const Normal &N, Vector *w,
	 	VisibilityTester *visibility) const {
	*w = (lightPos - P).Hat();
	visibility->SetSegment(P, lightPos, CastsShadows);
	return Intensity * Projection(-*w) * fabs(Dot(*w, N)) /
		DistanceSquared(lightPos, P);
}
Spectrum ProjectionLight::Projection(const Vector &w) const {
	Vector wl = WorldToLight(w);
	// Discard directions behind projection light
	if (wl.z < hither) return 0.;
	// Project point on to projection plane
	Point Pl = lightProjection(Point(wl.x, wl.y, wl.z));
	if (Pl.x < -1 || Pl.x > 1 || Pl.y < -1 || Pl.y > 1) return 0.;
	Float s = (Pl.x + 1) * 0.5f;
	Float t = (Pl.y + 1) * 0.5f;
	// Compute projected light for direction
	DifferentialGeometry dgMap(Point(Pl.x, Pl.y, 0), Vector(1,0,0),
		Vector(0,1,0), Vector(0,0,0), Vector(0,0,0), s, t, NULL);
	return projectionMap->Evaluate(dgMap);
}
Spectrum ProjectionLight::Sample_L(const Point &P, Float u1, Float u2,
		Vector *wo, Float *weight, bool *deltaLight,
		VisibilityTester *visibility) const {
	*wo = (lightPos - P).Hat();
	*weight = DistanceSquared(lightPos, P);
	*deltaLight = true;
	visibility->SetSegment(P, lightPos, CastsShadows);
	return Intensity * Projection(-*wo);
}
Spectrum ProjectionLight::Sample_L(const Scene *scene, Float u1, Float u2,
		Float u3, Float u4, Ray *ray, Float *weight,
		bool *deltaLight) const {
	ray->O = lightPos;
	Vector v = UniformSampleCone(u1, u2, cosTotalWidth);
	ray->D = LightToWorld(v);
	*weight = UniformConeWeight(cosTotalWidth);
	*deltaLight = true;
	return Intensity * Projection(ray->D);
}
Float ProjectionLight::Weight(const Point &, const Vector &) const {
	return 0.;
}
extern "C" Light *CreateLight(bool castsShadows,
		const Transform &light2world,
		const ParamSet &paramSet) {
	// Compute point-light intensity
	Spectrum I = paramSet.FindOneSpectrum("lightcolor", Spectrum(1.)) *
		paramSet.FindOneFloat("intensity", 1.);
	I *= M_PI;
	I = paramSet.FindOneSpectrum("I", I);
	// Create texture map for light
	string texmap = paramSet.FindOneString("mapname", "");
	Texture<Spectrum> *tex = NULL;
	if (texmap != "")
		tex = new ImageTexture<Spectrum>(new IdentityMapping2D,
			texmap.c_str());
	else
		tex = new ConstantTexture<Spectrum>(1.);
	Float fov = paramSet.FindOneFloat("fov", 45.);
	return new ProjectionLight(castsShadows, light2world, I,
		tex, fov);
}
