// spotlight.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "light.h"
#include "shapes.h"
// SpotLight Classes
class SpotLight : public Light {
public:
	SpotLight(bool shadows, const Transform &light2world,
		const Spectrum &, Float width, Float fall);
	Spectrum dE(const Point &P, const Normal &N, Vector *w,
		VisibilityTester *vis) const;
	// SpotLight Member Functions
	Float Falloff(const Vector &w) const;
	Spectrum Sample_L(const Point &P, Float u1, Float u2,
			Vector *wo, Float *weight, bool *deltaLight,
			VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, Float u1, Float u2,
			Float u3, Float u4, Ray *ray, Float *weight,
			bool *deltaLight) const;
	Float Weight(const Point &, const Vector &) const;
private:
	// SpotLight Private Data
	Float cosTotalWidth, cosFalloffStart;
	Point lightPos;
	Spectrum Intensity;
};
// SpotLight Method Definitions
SpotLight::SpotLight(bool shadows, const Transform &light2world,
		const Spectrum &intensity, Float width, Float fall)
	: Light(shadows, light2world, intensity) {
	lightPos = LightToWorld(Point(0,0,0));
	Intensity = intensity;
	cosTotalWidth = cosf(Radians(width));
	cosFalloffStart = cosf(Radians(fall));
}
Spectrum SpotLight::dE(const Point &P, const Normal &N, Vector *w,
		VisibilityTester *visibility) const {
	*w = (lightPos - P).Hat();
	visibility->SetSegment(P, lightPos, CastsShadows);
	return Intensity * Falloff(-*w) * fabs(Dot(*w, N)) /
		DistanceSquared(lightPos, P);
}
Float SpotLight::Falloff(const Vector &w) const {
	Vector wl = WorldToLight(w).Hat();
	Float costheta = wl.z;
	if (costheta < cosTotalWidth)
		return 0.;
 	if (costheta > cosFalloffStart)
		return 1.;
	// Compute falloff inside spotlight cone
	Float delta = (costheta - cosTotalWidth) /
		(cosFalloffStart - cosTotalWidth);
	return delta*delta*delta*delta;
}
Spectrum SpotLight::Sample_L(const Point &P, Float u1, Float u2,
		Vector *wo, Float *weight, bool *deltaLight,
		VisibilityTester *visibility) const {
	*wo = (lightPos - P).Hat();
	*weight = DistanceSquared(lightPos, P);
	*deltaLight = true;
	visibility->SetSegment(P, lightPos, CastsShadows);
	return Intensity * Falloff(-*wo);
}
Spectrum SpotLight::Sample_L(const Scene *scene, Float u1, Float u2,
		Float u3, Float u4, Ray *ray, Float *weight,
		bool *deltaLight) const {
	ray->O = lightPos;
	Vector v = UniformSampleCone(u1, u2, cosTotalWidth);
	ray->D = LightToWorld(v);
	*weight = UniformConeWeight(cosTotalWidth);
	*deltaLight = true;
	return Intensity * Falloff(ray->D);
}
Float SpotLight::Weight(const Point &, const Vector &) const {
	return 0.;
}
extern "C" Light *CreateLight(bool castsShadows,
		const Transform &l2w, const ParamSet &paramSet) {
	// Compute point-light intensity
	Spectrum I = paramSet.FindOneSpectrum("lightcolor", Spectrum(1.)) *
		paramSet.FindOneFloat("intensity", 1.);
	I *= M_PI;
	I = paramSet.FindOneSpectrum("I", I);
	Float coneangle = paramSet.FindOneFloat("coneangle", 30.);
	Float conedelta = paramSet.FindOneFloat("conedeltaangle", 5.);
	// Compute spotlight world to light transformation
	Point from = paramSet.FindOnePoint("from", Point(0,0,0));
	Point to = paramSet.FindOnePoint("to", Point(0,0,1));
	Vector dir = (to - from).Hat();
	Vector du, dv;
	CoordinateSystem(dir, &du, &dv);
	Transform dirToZ =
		Transform(new Matrix4x4( du.x,  du.y,  du.z, 0.,
	                                 dv.x,  dv.y,  dv.z, 0.,
	                                dir.x, dir.y, dir.z, 0.,
	                                    0,     0,     0, 1.));
	Transform light2world =
	l2w *
	Translate(Vector(from.x, from.y, from.z)) *
	dirToZ.GetInverse();
	return new SpotLight(castsShadows, light2world, I, coneangle,
		coneangle-conedelta);
}
