// translucent.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "materials.h"
// Translucent Class Declarations
class Translucent : public Material {
public:
	// Translucent Interface
	~Translucent();
	Translucent(Texture<Spectrum> *kd, Texture<Spectrum> *ks,
			Texture<Float> *rough, Texture<Float> *refl,
			Texture<Float> *trans, Texture<Float> *disp)
			: Material(disp) {
		Kd = kd;
		Ks = ks;
		roughness = rough;
		reflect = refl;
		transmit = trans;
	}
	BSDF *GetBSDF(const Intersection *isect) const;
private:
	// Translucent Private Data
	Texture<Spectrum> *Kd, *Ks;
	Texture<Float> *roughness;
	Texture<Float> *reflect, *transmit;
};
// Translucent Method Definitions
Translucent::~Translucent() {
	delete Kd;
	delete Ks;
	delete roughness;
	delete reflect;
	delete transmit;
}
BSDF *Translucent::GetBSDF(const Intersection *isect) const {
	BSDF *ret = new BSDF(isect->dgShading, isect->dgGeom);

	Float r = reflect->Evaluate(isect->dgShading);
	Float t = transmit->Evaluate(isect->dgShading);
	if (r == 0. && t == 0.) return ret;

	Spectrum kd = Kd->Evaluate(isect->dgShading);
	if (!kd.Black()) {
		if (r > 0.) ret->Add(new Lambertian(r * kd));
		if (t > 0.) ret->Add(new BRDFToBTDF(new Lambertian(t * kd)));
	}
	Spectrum ks = Ks->Evaluate(isect->dgShading);
	if (!ks.Black()) {
		Float rough = roughness->Evaluate(isect->dgShading);
		if (r > 0.) {
			Fresnel *fresnel = new FresnelDielectric(1.5f, 1.f);
			ret->Add(new Microfacet(r * ks, fresnel,
				new Blinn(1.f / rough)));
		}
		if (t > 0.) {
			Fresnel *fresnel = new FresnelDielectric(1.5f, 1.f);
			ret->Add(new BRDFToBTDF(new Microfacet(t * ks, fresnel,
				new Blinn(1.f / rough))));
		}
	}
	return ret;
}
extern "C" Reference<Material> CreateMaterial(const Transform &xform,
		const ParamSet &geomParams, const ParamSet &surfaceParams,
		const Spectrum &color, Texture<Float> *displace) {
	SURF_TEX_F(Kd, 1.);
	SURF_TEX_F(Ks, 1.);
	SURF_TEX_F(roughness, 0.1f);
	SURF_TEX_S(Cs, color);
	SURF_TEX_S(specularcolor, 1.f);
	SURF_TEX_F(reflect, 0.5f);
	SURF_TEX_F(transmit, 0.5f);
	string tex = geomParams.FindOneString("texturename",
		surfaceParams.FindOneString("texturename", ""));
	Texture<Spectrum> *kd = new ScaleTexture<Float, Spectrum>(Kd, Cs);
	Texture<Spectrum> *ks = new ScaleTexture<Float, Spectrum>(Ks, specularcolor);
	if (tex != "") {
		kd = new ScaleTexture<Float, Spectrum>(Kd,
			new ImageTexture<Spectrum>(new IdentityMapping2D, tex));
		ks = new ScaleTexture<Float, Spectrum>(Ks,
			new ImageTexture<Spectrum>(new IdentityMapping2D, tex));
	}
	return new Translucent(kd, ks, roughness, reflect, transmit, displace);
}
