// geometry.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef GEOMETRY_H
#define GEOMETRY_H
#include "lrt.h"
#include <float.h>
// Geometry Classes
class Vector {
public:
	// Vector Constructors
	Vector()
		: x(0.), y(0.), z(0.) {
	}
	Vector(Float xx, Float yy, Float zz)
		: x(xx), y(yy), z(zz) {
	}
	explicit Vector(const Point &p);
	explicit Vector(const Normal &n);
	// Vector Methods
	Vector operator+(const Vector &v) const {
		return Vector(x + v.x, y + v.y, z + v.z);
	}
	
	Vector& operator+=(const Vector &v) {
		x += v.x; y += v.y; z += v.z;
		return *this;
	}
	Vector operator-(const Vector &v) const {
		return Vector(x - v.x, y - v.y, z - v.z);
	}
	
	Vector& operator-=(const Vector &v) {
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}
	Vector operator*(Float f) const {
		return Vector(f*x, f*y, f*z);
	}
	
	Vector &operator*=(Float f) {
		x *= f; y *= f; z *= f;
		return *this;
	}
	Vector operator/(Float f) const {
		Float inv = 1.f/f;
		return Vector(x * inv, y * inv, z * inv);
	}
	
	Vector &operator/=(Float f) {
		Float inv = 1.f/f;
		x *= inv; y *= inv; z *= inv;
		return *this;
	}
	Vector operator-() const {
		return Vector(-x, -y, -z);
	}
	Float LengthSquared() const { return x*x + y*y + z*z; }
	Float Length() const { return sqrtf( LengthSquared() ); }
	Vector Hat() const { return (*this)/Length(); }
	// Vector Public Data
	Float x, y, z;
};
class Point {
public:
	// Point Constructors
	Point()
		: x(0.), y(0.), z(0.) {
	}
	Point(Float xx, Float yy, Float zz)
		: x(xx), y(yy), z(zz) {
	}
	// Point Methods
	Point operator+(const Vector &v) const {
		return Point(x + v.x, y + v.y, z + v.z);
	}
	
	Point &operator+=(const Vector &v) {
		x += v.x; y += v.y; z += v.z;
		return *this;
	}
	
	Vector operator-(const Point &p) const {
		return Vector(x - p.x, y - p.y, z - p.z);
	}
	
	Point operator-(const Vector &v) const {
		return Point(x - v.x, y - v.y, z - v.z);
	}
	
	Point &operator-=(const Vector &v) {
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}
	Point &operator+=(const Point &p) {
		x += p.x; y += p.y; z += p.z;
		return *this;
	}
	Point operator+(const Point &p) {
		return Point(x + p.x, y + p.y, z + p.z);
	}
	Point operator* (Float f) const {
		return Point(f*x, f*y, f*z);
	}
	Point &operator*=(Float f) {
		x *= f; y *= f; z *= f;
		return *this;
	}
	// Point Public Data
	Float x,y,z;
};
class Normal {
public:
	// Normal Constructors
	Normal()
		: x(0.), y(0.), z(0.) {}
	Normal(Float xx, Float yy, Float zz)
		: x(xx), y(yy), z(zz) {}
	explicit Normal(const Vector &v)
	  : x(v.x), y(v.y), z(v.z) {}
	// Normal Methods
	Normal operator-() const {
		return Normal(-x, -y, -z);
	}
	Normal operator+ (const Normal &v) const {
		return Normal(x + v.x, y + v.y, z + v.z);
	}
	
	Normal& operator+=(const Normal &v) {
		x += v.x; y += v.y; z += v.z;
		return *this;
	}
	Normal operator- (const Normal &v) const {
		return Normal(x - v.x, y - v.y, z - v.z);
	}
	
	Normal& operator-=(const Normal &v) {
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}
	Normal operator* (Float f) const {
		return Normal(f*x, f*y, f*z);
	}
	
	Normal &operator*=(Float f) {
		x *= f; y *= f; z *= f;
		return *this;
	}
	Normal operator/ (Float f) const {
		Float inv = 1.f/f;
		return Normal(x * inv, y * inv, z * inv);
	}
	
	Normal &operator/=(Float f) {
		Float inv = 1.f/f;
		x *= inv; y *= inv; z *= inv;
		return *this;
	}
	Float LengthSquared() const { return x*x + y*y + z*z; }
	Float Length() const        { return sqrtf(LengthSquared()); }
	Normal Hat() const          { return (*this) / Length(); }
	void Normalize()            { (*this) /= Length(); }
	// Normal Public Data
	Float x,y,z;
};
class Ray {
public:
	// Ray Constructor Declarations
	Ray(): mint(RAY_EPSILON), maxt(FLT_MAX), time(0.f) {}
	Ray(const Point &origin, const Vector &direction,
		Float start = RAY_EPSILON, Float end = FLT_MAX, Float t = 0.f)
		: O(origin), D(direction), mint(start), maxt(end), time(t) {
	}
	// Ray Method Declarations
	Point operator()(Float t) const { return O + D * t; }
	// Ray Public Data
	Point O;
	Vector D;
	mutable Float mint, maxt;
	Float time;
};
class RayDifferential : public Ray {
public:
	// RayDifferential Constructors
	RayDifferential() { hasDifferentials = false; }
	RayDifferential(const Ray &ray) : Ray(ray) {
		hasDifferentials = false;
	}
	// RayDifferential Public Data
	bool hasDifferentials;
	Ray rx, ry;
};
struct Extent2D {
	// Extent2D Constructors
	Extent2D() { x0 = x1 = y0 = y1 = 0.; }
	Extent2D(Float xx0, Float xx1, Float yy0, Float yy1) {
		x0 = min(xx0, xx1);
		x1 = max(xx0, xx1);
		y0 = min(yy0, yy1);
		y1 = max(yy0, yy1);
	}
	// Extent2D Data
	Float x0, x1, y0, y1;
};
class BBox {
public:
	// BBox Constructors
	BBox() {
		pMin = Point( INFINITY,  INFINITY,  INFINITY);
		pMax = Point(-INFINITY, -INFINITY, -INFINITY);
	}
	BBox(const Point &p) : pMin(p), pMax(p) { }
	BBox(const Point &p1, const Point &p2) {
		pMin = Point(min(p1.x, p2.x),
					 min(p1.y, p2.y),
					 min(p1.z, p2.z));
		pMax = Point(max(p1.x, p2.x),
					 max(p1.y, p2.y),
					 max(p1.z, p2.z));
	}
	// BBox Method Declarations
	friend inline ostream &operator<<(ostream &os, const BBox &b);
	friend BBox Union(const BBox &b, const Point &p);
	friend BBox Union(const BBox &b, const BBox &b2);
	friend BBox BIntersection(const BBox &b1, const BBox &b2);
	bool Overlaps(const BBox &b) const {
		bool x = (pMax.x >= b.pMin.x) && (pMin.x <= b.pMax.x);
		bool y = (pMax.y >= b.pMin.y) && (pMin.y <= b.pMax.y);
		bool z = (pMax.z >= b.pMin.z) && (pMin.z <= b.pMax.z);
		return (x && y && z);
	}
	bool Inside(const Point &pt) const {
		return (pt.x >= pMin.x && pt.x <= pMax.x &&
	            pt.y >= pMin.y && pt.y <= pMax.y &&
	            pt.z >= pMin.z && pt.z <= pMax.z);
	}
	void Expand(Float delta) {
		pMin -= Vector(delta, delta, delta);
		pMax += Vector(delta, delta, delta);
	}
	Float Volume() const {
		Vector d = pMax - pMin;
		return d.x * d.y * d.z;
	}
	void BoundingSphere(Point *c, Float *rad) const;
	bool IntersectP(const Ray &ray, Float *hitt0 = NULL, Float *hitt1 = NULL) const;
	// BBox Public Data
	Point pMin, pMax;
};
// Geometry Inline Functions
inline Vector::Vector(const Point &p)
	: x(p.x), y(p.y), z(p.z) {
}
inline ostream &operator<<(ostream &os, const Vector &v) {
	os << v.x << ", " << v.y << ", " << v.z;
	return os;
}
inline Vector operator*(Float f, const Vector &v) { return v*f; }
inline Float Dot(const Vector &v1, const Vector &v2) {
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}
inline Vector Cross(const Vector &v1, const Vector &v2) {
	return Vector((v1.y * v2.z) - (v1.z * v2.y),
                  (v1.z * v2.x) - (v1.x * v2.z),
                  (v1.x * v2.y) - (v1.y * v2.x));
}
inline Vector Cross(const Vector &v1, const Normal &v2) {
	return Vector((v1.y * v2.z) - (v1.z * v2.y),
                  (v1.z * v2.x) - (v1.x * v2.z),
                  (v1.x * v2.y) - (v1.y * v2.x));
}
inline Vector Cross(const Normal &v1, const Vector &v2) {
	return Vector((v1.y * v2.z) - (v1.z * v2.y),
                  (v1.z * v2.x) - (v1.x * v2.z),
                  (v1.x * v2.y) - (v1.y * v2.x));
}
inline void CoordinateSystem(const Vector &v1,	Vector *v2,
		Vector *v3) {
	if (fabsf(v1.x) > fabsf(v1.y)) {
	        Float invLen = 1.f / sqrtf(v1.x*v1.x + v1.z*v1.z);
	        *v2 = Vector(-v1.z * invLen, 0.f, v1.x * invLen);
	}
	else {
		Float invLen = 1.f / sqrtf(v1.y*v1.y + v1.z*v1.z);
		*v2 = Vector(0.f, v1.z * invLen, -v1.y * invLen);
	}
	*v3 = Cross(v1, *v2);
}
inline Float Distance(const Point &p1, const Point &p2) {
	return (p1 - p2).Length();
}
inline Float DistanceSquared(const Point &p1, const Point &p2) {
	return (p1 - p2).LengthSquared();
}
inline ostream &operator<<(ostream &os, const Point &v) {
	os << v.x << ", " << v.y << ", " << v.z;
	return os;
}
inline Point operator*(Float f, const Point &p) { return p*f; }
inline Normal operator*(Float f, const Normal &n) {
	return Normal(f*n.x, f*n.y, f*n.z);
}
inline Vector::Vector(const Normal &n)
  : x(n.x), y(n.y), z(n.z) { }
inline Float Dot(const Normal &n1, const Vector &v2) {
	return n1.x * v2.x + n1.y * v2.y + n1.z * v2.z;
}
inline Float Dot(const Vector &v1, const Normal &n2) {
	return v1.x * n2.x + v1.y * n2.y + v1.z * n2.z;
}
inline Float Dot(const Normal &n1, const Normal &n2) {
	return n1.x * n2.x + n1.y * n2.y + n1.z * n2.z;
}
inline ostream &operator<<(ostream &os, const Normal &v) {
	os << v.x << ", " << v.y << ", " << v.z;
	return os;
}
inline ostream &operator<<(ostream &os, Ray &r) {
	os << "org: " << r.O << "dir: " << r.D << " range [" <<
		r.mint << "," << r.maxt << "] time = " <<
		r.time;
	return os;
}
inline ostream &operator<<(ostream &os, const BBox &b) {
	os << b.pMin << " -> " << b.pMax;
	return os;
}
inline Vector SphericalDirection(Float sintheta, Float costheta,
		Float phi) {
	return Vector(sintheta * cosf(phi),
		sintheta * sinf(phi), costheta);
}
inline Vector SphericalDirection(Float sintheta, Float costheta,
		Float phi, const Vector &x, const Vector &y,
		const Vector &z) {
	return sintheta * cosf(phi) * x +
		sintheta * sinf(phi) * y + costheta * z;
}
inline Float SphericalTheta(const Vector &v) {
	return acosf(v.z);
}
inline Float SphericalPhi(const Vector &v) {
	return atan2f(v.y, v.x) + M_PI;
}
#endif // GEOMETRY_H
