// transform.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "transform.h"
#include "shapes.h"
// Transform Methods
ostream &operator<<(ostream &os, const Transform &t) {
	t.m->Print(os);
	return os;
}
Transform Translate(const Vector &delta) {
	Matrix4x4 *m, *minv;
	m = new Matrix4x4(1, 0, 0, delta.x,
                      0, 1, 0, delta.y,
                      0, 0, 1, delta.z,
                      0, 0, 0,       1);
	minv = new Matrix4x4(1, 0, 0, -delta.x,
                         0, 1, 0, -delta.y,
                         0, 0, 1, -delta.z,
                         0, 0, 0,       1);
	return Transform(m, minv);
}
Transform Scale(Float x, Float y, Float z) {
	Matrix4x4 *m, *minv;
	m = new Matrix4x4(x, 0, 0, 0,
                      0, y, 0, 0,
                      0, 0, z, 0,
                      0, 0, 0, 1);
	minv = new Matrix4x4(1.f/x, 0, 0, 0,
                         0, 1.f/y, 0, 0,
                         0, 0, 1.f/z, 0,
                         0, 0, 0, 1);
	return Transform(m, minv);
}
Transform RotateX(Float angle) {
	Float sin_t = sinf(Radians(angle));
	Float cos_t = cosf(Radians(angle));
	Matrix4x4 *m = new Matrix4x4(1,   0,      0,   0,
                                 0, cos_t, -sin_t, 0,
                                 0, sin_t,  cos_t, 0,
                                 0,   0,      0,   1);
	return Transform(m, m->Transpose());
}
Transform RotateY(Float angle) {
	Float sin_t = sinf(Radians(angle));
	Float cos_t = cosf(Radians(angle));
	Matrix4x4 *m = new Matrix4x4(cos_t,   0, -sin_t, 0,
                                 0,   1,      0, 0,
                                 sin_t,   0,  cos_t, 0,
                                 0,   0,   0,    1);
	return Transform(m, m->Transpose());
}

Transform RotateZ(Float angle) {
	Float sin_t = sinf(Radians(angle));
	Float cos_t = cosf(Radians(angle));
	Matrix4x4 *m = new Matrix4x4(cos_t, -sin_t, 0, 0,
                                 sin_t,  cos_t, 0, 0,
                                 0,      0, 1, 0,
                                 0,      0, 0, 1);
	return Transform(m, m->Transpose());
}
Transform Rotate(Float angle, const Vector &axis) {
	Vector a = axis.Hat();
	Float s = sinf(Radians(angle));
	Float c = cosf(Radians(angle));
	Float m[4][4];

	m[0][0] = a.x * a.x + (1.f - a.x * a.x) * c;
	m[0][1] = a.x * a.y * (1.f - c) - a.z * s;
	m[0][2] = a.x * a.z * (1.f - c) + a.y * s;
	m[0][3] = 0;

	m[1][0] = a.x * a.y * (1.f - c) + a.z * s;
	m[1][1] = a.y * a.y + (1.f - a.y * a.y) * c;
	m[1][2] = a.y * a.z * (1.f - c) - a.x * s;
	m[1][3] = 0;

	m[2][0] = a.x * a.z * (1.f - c) - a.y * s;
	m[2][1] = a.y * a.z * (1.f - c) + a.x * s;
	m[2][2] = a.z * a.z + (1.f - a.z * a.z) * c;
	m[2][3] = 0;

	m[3][0] = 0;
	m[3][1] = 0;
	m[3][2] = 0;
	m[3][3] = 1;

	Matrix4x4 *mat = new Matrix4x4(m);
	return Transform(mat, mat->Transpose());
}
Transform LookAt(const Point &pos, const Point &look,
		const Vector &up) {
	Float m[4][4];
	// Initialize fourth column of viewing matrix
	m[0][3] = pos.x;
	m[1][3] = pos.y;
	m[2][3] = pos.z;
	m[3][3] = 1;
	// Initialize first three columns of viewing matrix
	Vector dir = (look - pos).Hat();
	Vector right = Cross(dir, up.Hat());
	Vector newUp = Cross(right, dir);
	m[0][0] = right.x;
	m[1][0] = right.y;
	m[2][0] = right.z;
	m[3][0] = 0.;
	m[0][1] = newUp.x;
	m[1][1] = newUp.y;
	m[2][1] = newUp.z;
	m[3][1] = 0.;
	m[0][2] = dir.x;
	m[1][2] = dir.y;
	m[2][2] = dir.z;
	m[3][2] = 0.;
	Matrix4x4 *camToWorld = new Matrix4x4(m);
	return Transform(camToWorld->Inverse(), camToWorld);
}
BBox Transform::operator()(const BBox &b) const {
	const Transform &M = *this;
	BBox ret(      M(Point(b.pMin.x, b.pMin.y, b.pMin.z)));
	ret = Union(ret, M(Point(b.pMax.x, b.pMin.y, b.pMin.z)));
	ret = Union(ret, M(Point(b.pMin.x, b.pMax.y, b.pMin.z)));
	ret = Union(ret, M(Point(b.pMin.x, b.pMin.y, b.pMax.z)));
	ret = Union(ret, M(Point(b.pMin.x, b.pMax.y, b.pMax.z)));
	ret = Union(ret, M(Point(b.pMax.x, b.pMax.y, b.pMin.z)));
	ret = Union(ret, M(Point(b.pMax.x, b.pMin.y, b.pMax.z)));
	ret = Union(ret, M(Point(b.pMax.x, b.pMax.y, b.pMax.z)));
	return ret;
}
Transform Transform::operator*(const Transform &t2) const {
	Reference<Matrix4x4> m1 = Matrix4x4::Mul(m, t2.m);
	Reference<Matrix4x4> m2 = Matrix4x4::Mul(t2.m_inv, m_inv);
	return Transform(m1, m2);
}
Transform Orthographic(Float znear, Float zfar) {
	return Scale(1.f, 1.f, 1.f / (zfar-znear)) *
		Translate(Vector(0.f, 0.f, -znear));
}
Transform Perspective(Float fov, Float n, Float f) {
	Float invTanAng = 1.f / tanf(Radians(fov) / 2.f);
	Matrix4x4 *persp =
		new Matrix4x4(1, 0,       0,          0,
                      0, 1,       0,          0,
                      0, 0, f/(f-n), -f*n/(f-n),
                      0, 0,       1,          0);
	return Scale(invTanAng, invTanAng, 1) *
               Transform(persp);
}
