// transform.h*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#ifndef TRANSFORM_H
#define TRANSFORM_H
#include "lrt.h"
#include "geometry.h"
// Transform Declarations
class Transform {
public:
	// Transform Constructor Declarations
	Transform() {
		m = m_inv = new Matrix4x4;
	}
	Transform(Float mat[4][4]) {
		m = new Matrix4x4(mat[0][0], mat[0][1], mat[0][2], mat[0][3],
			mat[1][0], mat[1][1], mat[1][2], mat[1][3],
			mat[2][0], mat[2][1], mat[2][2], mat[2][3],
			mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
		m_inv = m->Inverse();
	}
	Transform(const Reference<Matrix4x4> &mat) {
		m = mat;
		m_inv = m->Inverse();
	}
	Transform(const Reference<Matrix4x4> &mat,
			const Reference<Matrix4x4> &minv) {
		m = mat;
		m_inv = minv;
	}
	// Transform Method Declarations
	friend ostream &operator<<(ostream &, const Transform &);
	inline Point operator()(const Point &pt) const;
	inline void operator()(const Point &pt, Point *ptrans) const;
	inline Vector operator()(const Vector &v) const;
	inline void operator()(const Vector &v, Vector *vt) const;
	Transform GetInverse() const {
		return Transform(m_inv, m);
	}
	inline Normal operator()(const Normal &) const;
	inline void operator()(const Normal &, Normal *nt) const;
	inline Ray operator()(const Ray &r) const;
	inline void operator()(const Ray &r, Ray *rt) const;
	BBox operator()(const BBox &b) const;
	Transform operator*(const Transform &t2) const;
public:
	// Transform Private Data
	Reference<Matrix4x4> m, m_inv;
};
// Transform Inline Functions
inline Point Transform::operator()(const Point &pt) const {
    Float x = pt.x, y = pt.y, z = pt.z;

    Float xp = m->m[0][0]*x + m->m[0][1]*y + m->m[0][2]*z + m->m[0][3];
    Float yp = m->m[1][0]*x + m->m[1][1]*y + m->m[1][2]*z + m->m[1][3];
    Float zp = m->m[2][0]*x + m->m[2][1]*y + m->m[2][2]*z + m->m[2][3];
    Float wp = m->m[3][0]*x + m->m[3][1]*y + m->m[3][2]*z + m->m[3][3];

    if (wp == 1.) return Point(xp, yp, zp);
    else          return Point(xp / wp, yp / wp, zp / wp);
}
inline void Transform::operator()(const Point &pt,
		Point *ptrans) const {
	Float x = pt.x, y = pt.y, z = pt.z;
	ptrans->x = m->m[0][0]*x + m->m[0][1]*y + m->m[0][2]*z +
		m->m[0][3];
	ptrans->y = m->m[1][0]*x + m->m[1][1]*y + m->m[1][2]*z +
		m->m[1][3];
	ptrans->z = m->m[2][0]*x + m->m[2][1]*y + m->m[2][2]*z +
		m->m[2][3];
	Float w   = m->m[3][0]*x + m->m[3][1]*y + m->m[3][2]*z +
		m->m[3][3];
	if (w != 1.) {
		ptrans->x /= w;
		ptrans->y /= w;
		ptrans->z /= w;
	}
}
inline Vector Transform::operator()(const Vector &v) const {
	Float x = v.x, y = v.y, z = v.z;
	return Vector(m->m[0][0]*x + m->m[0][1]*y + m->m[0][2]*z,
                  m->m[1][0]*x + m->m[1][1]*y + m->m[1][2]*z,
                  m->m[2][0]*x + m->m[2][1]*y + m->m[2][2]*z);
}
inline void Transform::operator()(const Vector &v,
		Vector *vt) const {
	Float x = v.x, y = v.y, z = v.z;
	vt->x = m->m[0][0] * x + m->m[0][1] * y + m->m[0][2] * z;
	vt->y = m->m[1][0] * x + m->m[1][1] * y + m->m[1][2] * z;
	vt->z = m->m[2][0] * x + m->m[2][1] * y + m->m[2][2] * z;
}
inline Normal Transform::operator()(const Normal &n) const {
	Float x = n.x, y = n.y, z = n.z;
	return Normal(m_inv->m[0][0] * x + m_inv->m[1][0] * y +
                      m_inv->m[2][0] * z,
                  m_inv->m[0][1] * x + m_inv->m[1][1] * y +
                      m_inv->m[2][1] * z,
                  m_inv->m[0][2] * x + m_inv->m[1][2] * y +
                      m_inv->m[2][2] * z);
}
inline void Transform::operator()(const Normal &n,
		Normal *nt) const {
	Float x = n.x, y = n.y, z = n.z;
	nt->x = m_inv->m[0][0] * x + m_inv->m[1][0] * y +
		m_inv->m[2][0] * z;
	nt->y = m_inv->m[0][1] * x + m_inv->m[1][1] * y +
		m_inv->m[2][1] * z;
	nt->z = m_inv->m[0][2] * x + m_inv->m[1][2] * y +
		m_inv->m[2][2] * z;
}
inline Ray Transform::operator()(const Ray &r) const {
	Ray ret;
	(*this)(r.O, &ret.O);
	(*this)(r.D, &ret.D);
	ret.mint = r.mint;
	ret.maxt = r.maxt;
	return ret;
}
inline void Transform::operator()(const Ray &r, Ray *rt) const {
	rt->mint = r.mint;
	rt->maxt = r.maxt;
	(*this)(r.O, &rt->O);
	(*this)(r.D, &rt->D);
}
#endif // TRANSFORM_H
