#include "Globals.h"
#include "Heightfield.h"
#include <math.h>


GLfloat cosTablef[angleMax];
GLfloat sinTablef[angleMax];


Heightfield::Heightfield(const Vector &size, const VectorGLf &dimension, int numwaves) : dim(dimension)
{
	heightw=size.x();
	heighth=size.y();
	stepw=dimension.x()/(heightw-1);
	steph=dimension.z()/(heighth-1);
	yscale=dimension.y();
	posy=-yscale;
	timeline=0;

	heights=new GLfloat[heightw*heighth];
	normals=new VectorGLf[heightw*heighth];
	colors=new GLfloat[heightw*heighth];
	refracted=new VectorGLf[heightw*heighth];

	makeSinCosf();

	wavetab.clear();
	wavetab.reserve(numwaves);

	generateWaves(numwaves);
	this->numwaves=numwaves;
}

Heightfield::~Heightfield() {
	for(int i=0;i<numwaves;i++) {
		delete [] wavetab[i].precalcheight;
		delete [] wavetab[i].precalcdydx;
		delete [] wavetab[i].precalcdydz;
	}
	delete [] heights;
	delete [] normals;
	delete [] colors;
	delete [] refracted;
}

// Does not use numwaves right now, always 2 waves
void Heightfield::generateWaves(int num)
{
	int i,index;
	index=0;
	for(i=0;i<heightw;i++) {
		for(int j=0;j<heighth;j++) {
			heights[index]=0.0f;					// Some plate eau
			normals[index]=VectorGLf(0.0f,0.0f,0.0f);					// Some plate eau
			index++;
		}
	}

	Wave w1;
	w1.amplitude=0.33f*yscale*.5f;
	w1.dirX=13.0f;
	w1.dirZ=8.3f;
	w1.period=3865;
	precalcOneWave(w1);
	wavetab.push_back(w1);

	if(num>1) {
		Wave w2;
		w2.amplitude=0.33f*yscale*.5f;
		w2.dirX=-14.5f;
		w2.dirZ=17.3f;
		w2.period=2869;
		precalcOneWave(w2);
		wavetab.push_back(w2);
	}

	if(num>2) {
		Wave w3;
		w3.amplitude=0.33f*yscale*.5f;
		w3.dirX=1.5f;
		w3.dirZ=-12.3f;
		w3.period=3173;
		precalcOneWave(w3);
		wavetab.push_back(w3);
	}
}

#define HFRDPERIOD 4

// precalculate a huge amount of data for noe wave
void Heightfield::precalcOneWave(Wave &w) {
	w.period-=w.period%HFRDPERIOD;
	w.precalcheight=new GLfloat[heighth*heightw*w.period/HFRDPERIOD];
	w.precalcdydx=new GLfloat[heighth*heightw*w.period/HFRDPERIOD];
	w.precalcdydz=new GLfloat[heighth*heightw*w.period/HFRDPERIOD];

	GLfloat omega=2.0f*M_PI/(float)w.period;
	w.phaseX=Rand(w.period);
	w.phaseZ=Rand(w.period);
	GLfloat x,z;
	GLfloat s,c,arg;

	int index=0;

	x=-dim.x()/2.0f;
	for(int i=0;i<heightw;i++) {
		x+=stepw;

		z=-dim.z()/2.0f;
		for(int j=0;j<heighth;j++) {
			z+=steph;

			for(int k=0;k<w.period;k+=HFRDPERIOD) {
				arg=w.dirX*(x+w.phaseX)+w.dirZ*(z+w.phaseZ)-omega*k;
				getSinCosf(s,c,arg);

//				assert(index<heightw*heighth*w.period/HFRDPERIOD);
				w.precalcheight[index]=w.amplitude*c;
				w.precalcdydx[index]=w.amplitude*w.dirX*s;
				w.precalcdydz[index]=w.amplitude*w.dirZ*s;
				index++;
			}
		}
	}
}

// one frame
void Heightfield::update(int elapsed) {
	int i,j,k;
	timeline+=elapsed;
	int wavetime;
	const int numwaves=wavetab.size();
	int index;

	// Reinit
	index=0;
	for(i=0;i<heightw;i++)
		for(j=0;j<heighth;j++) {
			assert(index<heightw*heighth);
			heights[index]=posy;
			normals[index].x()=0.0f;
			normals[index].z()=0.0f;
			normals[index].y()=(float)numwaves;
			index++;
		}

	// Get new values
	int index2=0,incr=0;

	index2=0;
	for(k=0;k<numwaves;k++) {
		wavetime=timeline%wavetab[k].period;
		incr=wavetab[k].period/HFRDPERIOD;

		index=0;
		index2=wavetime/HFRDPERIOD;
		
		for(i=0;i<heightw;i++) {
			for(j=0;j<heighth;j++) {
//				assert(index2<heightw*heighth*wavetab[k].period/HFRDPERIOD);
//				assert(index<heightw*heighth);
				heights[index]+=wavetab[k].precalcheight[index2];   //(i,j,k): k+(j+i*heighth)*period
				normals[index].x()+=wavetab[k].precalcdydx[index2];
				normals[index].z()+=wavetab[k].precalcdydz[index2];
				index++;
				index2+=incr;
			}
		}
	}

	// Get normalized normals for the vertices
	index=0;
	for(i=0;i<heightw;i++)
		for(j=0;j<heighth;j++) {
			normalize(normals[index]);
			index++;
		}
}

// displays the heightfield
void Heightfield::display() {
	int index;
	GLfloat x,z;

	index=0;
	x=-dim.x()/2.0f;
	for (int i=0;i<heightw-1;i++) {
	  glBegin(GL_TRIANGLE_STRIP);

		z=-dim.z()/2.0f;
		for(int j=0;j<heighth;j++) {
			glColor4f(1.0f,1.0f,1.0f,colors[index]);
		  glNormal3fv(normals[index].c);
		  glVertex3f(x, heights[index],z);
			glColor4f(1.0f,1.0f,1.0f,colors[index+heighth]);

			glNormal3fv(normals[index+heighth].c);			// (i+1,j)
		  glVertex3f(x+stepw, heights[index+heighth],z);
			index++;
			z+=steph;
		}
		glEnd();
		x+=stepw;
	}
}


// Fresnel
// Nvidia paper: http://developer.nvidia.com/docs/IO/3035/ATT/FresnelReflection.pdf
// R(theta)=R0+(1-R0)*(1-cos(theta))^5
// R0=(1-refracindexratio)^2/(1+refracindexratio)^2
// instead of the huge original formula !!!
//
// Refraction
// R=-(index-1)*((N.E)N)-E
//


#define R0 0.0204f  // Water
#define REFRACIND -.33f  // 1-1.33 for water
//#define REFRACIND 0.0f  // 1-1.33 for water
//#define REFRACIND 2.0f  // Reflection formula, just to test...

// per vertex calculation
void Heightfield::calcRefracFresnel(VectorGLf playpos) {
	GLfloat x,z,cost;
	int index;
	int i,j;
	VectorGLf dir;
	x=-dim.x()/2.0f;
	index=0;
	for(i=0;i<heightw;i++) {
		z=-dim.z()/2.0f;
		for(j=0;j<heighth;j++) {
			//Eye vector
			dir.x()=playpos.x()-x;
			dir.y()=playpos.y()-heights[index];
			dir.z()=playpos.z()-z;
			normalize(dir);
			
			//Fresnel alpha
			cost=dir.dotProd(normals[index]);
//			colors[index]=R0+(1-R0)*pow((1-cost),5); Ideal
			colors[index]=R0+(1-R0)*pow((1-cost),1.7);

			//			fprintf(stderr,"%f ",colors[index]);
//			colors[index]=R0+(1-R0)*pow((1-cost),1.3); // best visual result

			//Refracted ray
			refracted[index]=normals[index]*REFRACIND*cost-dir;
			refracted[index].x()+=x;   // trick to get the right point on the cube map...
			refracted[index].y()+=playpos.y()-0.5f;
			refracted[index].z()+=z;

			index++;
			z+=steph;
		}
		x+=stepw;
	}
}


#define REFRACINDREV .33f  // 1.33-1 for water


//refraction from inside
void Heightfield::calcRefracInside(VectorGLf playpos) {
	GLfloat x,z,cost;
	int index;
	int i,j;
	VectorGLf dir;
	x=-dim.x()/2.0f;
	index=0;
	for(i=0;i<heightw;i++) {
		z=-dim.z()/2.0f;
		for(j=0;j<heighth;j++) {
			//Eye vector
			dir.x()=playpos.x()-x;
			dir.y()=playpos.y()-heights[index];
			dir.z()=playpos.z()-z;
			normalize(dir);
			
			cost=dir.dotProd(-normals[index]);

			//Refracted ray
			refracted[index]=-normals[index]*REFRACINDREV*cost-dir;
			refracted[index].x()+=x;
			refracted[index].z()+=z;

			index++;
			z+=steph;
		}
		x+=stepw;
	}
}


void Heightfield::alphadisplay() {
	int index;
	GLfloat x,z;
	index=0;
	x=-dim.x()/2.0f;
	for (int i=0;i<heightw-1;i++) {
	  glBegin(GL_TRIANGLE_STRIP);

		z=-dim.z()/2.0f;
		for(int j=0;j<heighth;j++) {
			glColor4f(1.0f,1.0f,1.0f,colors[index]);
			glNormal3fv(normals[index].c);
		  glVertex3f(x, heights[index],z);

			glColor4f(1.0f,1.0f,1.0f,colors[index+heighth]);
			glNormal3fv(normals[index+heighth].c);			// (i+1,j)
		  glVertex3f(x+stepw, heights[index+heighth],z);
			index++;
			z+=steph;
		}
		glEnd();
		x+=stepw;
	}
}

void Heightfield::refracdisplay() {
	int index;
	GLfloat x,z;
	index=0;
	x=-dim.x()/2.0f;
	for (int i=0;i<heightw-1;i++) {
	  glBegin(GL_TRIANGLE_STRIP);

		z=-dim.z()/2.0f;
		for(int j=0;j<heighth;j++) {
			glTexCoord3f(refracted[index].x(),refracted[index].y(),refracted[index].z());
			glNormal3fv(normals[index].c);
		  glVertex3f(x, heights[index],z);

			glTexCoord3fv(refracted[index+heighth].c);
			glNormal3fv(normals[index+heighth].c);			// (i+1,j)
		  glVertex3f(x+stepw, heights[index+heighth],z);
			index++;
			z+=steph;
		}
		glEnd();
		x+=stepw;
	}
}

void Heightfield::refracdisplay2() {
	int index;
	GLfloat x,z;
	index=0;
	x=-dim.x()/2.0f;
	for (int i=0;i<heightw-1;i++) {
	  glBegin(GL_TRIANGLE_STRIP);

		z=-dim.z()/2.0f;
		for(int j=0;j<heighth;j++) {
			glTexCoord3f(refracted[index].x(),refracted[index].y(),refracted[index].z());
			glNormal3fv((-normals[index]).c);
		  glVertex3f(x, heights[index],z);

			glTexCoord3fv(refracted[index+heighth].c);
			glNormal3fv((-normals[index+heighth]).c);			// (i+1,j)
		  glVertex3f(x+stepw, heights[index+heighth],z);
			index++;
			z+=steph;
		}
		glEnd();
		x+=stepw;
	}
}


//debugging the normals
void Heightfield::showNormals()
{
  GLfloat x, z;
	int index;

  glBegin(GL_LINES);
  glColor3ub(255, 0, 0);
	index=0;
	for (int i=0;i<heightw-1;i++) {
		x=-dim.x()/2.0f+i*stepw;

		for(int j=0;j<heighth;j++) {
			z=-dim.z()/2.0f+j*steph;
		  glVertex3f(x, heights[index],z);
		  glVertex3f(x + normals[index].x()/10.0f, heights[index] + normals[index].y()/10.0f, z + normals[index].z()/10.0f);
			index++;
		}

  }

  glEnd();
}

