// api.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "paramset.h"
#include "ri.h"
#include "lrt.h"
#include "film.h"
#include "primitives.h"
#include "camera.h"
#include "color.h"
#include "light.h"
#include "sampling.h"
#include "materials.h"
#include "transport.h"
#include "scene.h"
#include "texture.h"
#include "dynload.h"
#include <ctype.h>
#ifdef _WIN32
#include <malloc.h>
#else
#include <alloca.h>
#endif
// API Includes
#define MOTION_LEVELS 2
#define MAXTREES 30
int nbKdtrees;
// API Local Classes
struct Token {
	Token(const char *n, int t = 0, int a = 0) {
		token = n;
		type = t;
		arrayLength = a;
	}
	string token;
	int type;
	int arrayLength;
};
struct GfxOptions {
	GfxOptions();
	// Graphics Options
	int PixelSamples[2];
	Float PixelAspectRatio, FrameAspectRatio;
	int XResolution, YResolution;
	Extent2D ScreenExtent;
	bool ScreenExtentSet;
	Extent2D Crop;
	Float ClipHither, ClipYon;
	Float ShutterStart, ShutterEnd;
	Float LensRadius, FocalDistance;
	string cameraName;
	mutable ParamSet cameraParams;
	Transform WorldToCamera[MOTION_LEVELS];
	Float Gain, Gamma;
	string filterName;
	Float FilterXWidth, FilterYWidth;
	ParamSet FilterParams;
	int ColorQuantOne, ColorQuantMin, ColorQuantMax;
	Float ColorQuantDither;
	string DisplayType;
	string DisplayName;
	bool displayRGB, displayA, displayZ;
	string samplerName;
	ParamSet samplerParams;
	string acceleratorName;
	ParamSet acceleratorParams;
	string SearchPath;
	string ToneMapping;
	ParamSet ToneMapParams;
	string surfIntegratorName, volIntegratorName;
	ParamSet surfIntegratorParams, volIntegratorParams;
	mutable vector<Reference<Primitive> > primitives [MAXTREES];
	mutable vector<Light *> lights;
	mutable vector<VolumeRegion *> volumeRegions;
	// GfxOptions Method Declarations
	void WorldEnd() {
		for(int i=0;i<MAXTREES;i++) 
			primitives[i].erase(primitives[i].begin(), primitives[i].end());
		lights.erase(lights.begin(), lights.end());
		volumeRegions.erase(volumeRegions.begin(), volumeRegions.end());
	}
	Scene *MakeScene() const;
};
GfxOptions::GfxOptions() {
	// GfxOptions Constructor Implementation
	PixelSamples[0] = PixelSamples[1] = 2;
	PixelAspectRatio = 1.f;
	FrameAspectRatio = 4.f / 3.f;
	XResolution = 640;
	YResolution = 480;
	ScreenExtent = Extent2D(-4.f / 3.f, 4.f / 3.f, -1, 1);
	ScreenExtentSet = false;
	Crop = Extent2D(0, 1, 0, 1);
	ClipHither = RAY_EPSILON;
	ClipYon = 1e30f;
	ShutterStart = ShutterEnd = 0;
	LensRadius = 0.f;
	FocalDistance = INFINITY;
	cameraName = "orthographic";
	Gain = Gamma = 1.;
	filterName = "mitchell";
	FilterXWidth = FilterYWidth = 2.;
	ColorQuantOne = 255;
	ColorQuantMin = 0;
	ColorQuantMax = 255;
	ColorQuantDither = 0.5;
	DisplayType = "tiff";
	DisplayName = "lrt.tiff";
	displayRGB = displayA = true;
	displayZ = false;
	samplerName = "bestcandidate";
	acceleratorName = "grid";
	#ifdef LRT_BUILDDIR
	SearchPath = ".:" LRT_BUILDDIR ":./texture";
	#else
	SearchPath = ".";
	#endif
	surfIntegratorName = "whitted";
	volIntegratorName = "null";
}
struct GfxState {
	GfxState();
	// Graphics State
	Spectrum color;
	ParamSet surfaceParams;
	string surface;
	ParamSet displaceParams;
	string displacement;
	ParamSet areaLightParams;
	string areaLight;
	// Graphics State Methods
	void AddShape(const Reference<Shape> &shape, ParamSet &geomParams);
};
GfxState::GfxState() {
	// GfxState Constructor Implementation
	color = Spectrum(1.);
	surface = "matte";
	displacement = "";
}
// API Static Data
StringHashTable TokenTable;
#define STATE_UNINITIALIZED  0
#define STATE_BEGIN          1
#define STATE_WORLD_BEGIN    2
static int currentApiState = STATE_UNINITIALIZED;
static GfxOptions *curGfxOptions = NULL;
static vector<Transform> transformStack[MOTION_LEVELS];
static vector<char> hierarchicalState;
static GfxState curGfxState;
static vector<GfxState> gstates;
static int motionLevel = 0;
static bool inMotionBlock = false;
static Transform curTransform[MOTION_LEVELS];
// API Macros
#define VERIFY_STATE(s, func)
// API Static Methods
static int stringToType(const char *strType, int *nArray);
static string typeToString(int type, int narray);
static string UpdateSearchPath(const string &oldpath, const string &newpath) {
	string ret;
	for (u_int i = 0; i < newpath.size(); ++i) {
		if (newpath[i] != '&')
			ret += newpath[i];
		else
			ret += oldpath;
	}
	return ret;
}

//PRAMOD
std::vector<Reference<GeometricPrimitive> > translucentBlobs;

void GfxState::AddShape(const Reference<Shape> &shape,
		ParamSet &geomParams) {
	if (!shape) return;
	AreaLight *area = NULL;
	// Initialize area light for shape
	if (areaLight != "") {
		bool shadows = areaLightParams.FindOneInt("shadows", 1) != 0;
		area = CreateAreaLight(areaLight, curGfxOptions->SearchPath,
			shadows, curTransform[motionLevel],
			areaLightParams, shape);
	}
	// Create [[N]] and [[S]] textures, if needed
	int type, narray;
	Texture<Normal> *Ntex = NULL;
	const Normal *N = geomParams.FindNormal("N", &type, &narray);
	if (N) {
		Assert(type & PARAM_TYPE_NORMAL);
		int count = geomParams.TypeToNum(type);
		Normal *Nw = new Normal[count];
		for (int i = 0; i < count; ++i)
			Nw[i] = curTransform[motionLevel](N[i]);
	
		if (type & PARAM_TYPE_UNIFORM)
			Ntex = new ConstantTexture<Normal>(*Nw);
		else if (type & PARAM_TYPE_VARYING) {
			Ntex = new BilerpTexture<Normal>(new IdentityMapping2D,
				Nw[0], Nw[1], Nw[2], Nw[3]);
		}
		else if (type & PARAM_TYPE_VERTEX) {
			if (geomParams.nVertex != 0)
				Ntex = new VertexTexture<Normal>(Nw,
					geomParams.nVertex);
			else
				Error("Vertex texture not supported for shape");
		}
		delete[] Nw;
	}
	// Create [[S]] texture
	Texture<Vector> *Stex = NULL;
	// Initialize material for shape
	Texture<Float> *displace = CreateBump(displacement, curGfxOptions->SearchPath,
		curTransform[0], geomParams, displaceParams);
	Reference<Material> mtl = CreateMaterial(surface, curGfxOptions->SearchPath,
		curTransform[0], geomParams, surfaceParams, color, displace);
	if (!mtl)
		mtl = CreateMaterial("plastic", curGfxOptions->SearchPath,
			curTransform[0], geomParams, surfaceParams, color,
			displace);
	Assert(mtl);
	Reference<Primitive> prim = new GeometricPrimitive(shape,
		mtl, area, Ntex, Stex);

  //PRAMOD
  
  if(mtl->istranslucent == true){
    GeometricPrimitive* tmp =(GeometricPrimitive* )prim.ptr; 
    translucentBlobs.push_back(tmp);
  }

	curGfxOptions->primitives[nbKdtrees].push_back(prim);
	if (area != NULL) {
		// Create area lights given number of light samples
		int nSamples = areaLightParams.FindOneInt("nsamples", 1);
		if (nSamples == 1)
			curGfxOptions->lights.push_back(area);
		else
			for (int i = 0; i < nSamples; ++i)
				curGfxOptions->lights.push_back(new MultiAreaLight(area, nSamples));
	}
	geomParams.ReportUnused();
}
static int stringToType(const char *strType, int *narray) {
	if (strType == NULL) return PARAM_TYPE_VOID;
	int type;
	// Decide if the type is [[uniform]] or [[vertex]]
	const char *strp = strType;
	while (*strp && isspace(*strp))
		++strp;
	if (!*strp) return PARAM_TYPE_ERROR;
	if (strncmp("uniform", strp, strlen("uniform")) == 0) {
		type = PARAM_TYPE_UNIFORM;    strp += strlen("uniform");
	}
	else if	(strncmp("constant", strp, strlen("constant")) == 0) {
		type = PARAM_TYPE_UNIFORM;    strp += strlen("constant");
	}
	else if (strncmp("varying", strp, strlen("varying")) == 0) {
		type = PARAM_TYPE_VARYING;    strp += strlen("varying");
	}
	else if (strncmp("vertex", strp, strlen("vertex")) == 0) {
		type = PARAM_TYPE_VERTEX;     strp += strlen("vertex");
	}
	else type = PARAM_TYPE_UNIFORM;
	// Decode the storage class of the type
	while (*strp && isspace(*strp)) ++strp;
	if (!*strp) return PARAM_TYPE_ERROR;
	
	#define TRY_DECODING_TYPE(name, mask) \
		if (strncmp(name, strp, strlen(name)) == 0) { \
			type |= mask; strp += strlen(name); \
		}
	
	// *INDENT-OFF*
	     TRY_DECODING_TYPE("float",    PARAM_TYPE_FLOAT)
	else TRY_DECODING_TYPE("integer",  PARAM_TYPE_INT)
	else TRY_DECODING_TYPE("point",    PARAM_TYPE_POINT)
	else TRY_DECODING_TYPE("hpoint",   PARAM_TYPE_HPOINT)
	else TRY_DECODING_TYPE("vector",   PARAM_TYPE_VECTOR)
	else TRY_DECODING_TYPE("normal",   PARAM_TYPE_NORMAL)
	else TRY_DECODING_TYPE("string",   PARAM_TYPE_STRING)
	else TRY_DECODING_TYPE("color",    PARAM_TYPE_COLOR)
	else {
		Error("RiDeclare: unable to decode type \"%s\"", strType);
		return PARAM_TYPE_ERROR;
	}
	// *INDENT-ON*
	// Decode array length declaration, if any
	while (*strp && isspace(*strp)) ++strp;
	if (*strp == '[') {
		++strp;
		Assert(*strp); // XXX error message
		sscanf(strp, "%d", narray);
		while (*strp && isdigit(*strp) || isspace(*strp))
			++strp;
		Assert(*strp);  // XXX error message
		Assert(*strp == ']');
		++strp;
	}
	else *narray = 1;
	// Warn if there is additional non-whitespace
	while (*strp && isspace(*strp)) ++strp;
	if (*strp)
		Warning("RiDeclare: unknown text at end of declared "
			"type \"%s\".", strType);
	return type;
}
static string typeToString(int type, int nArray) {
	if (type == PARAM_TYPE_VOID)
		return "NULL";

	string ret;
	if (type & PARAM_TYPE_UNIFORM)
		ret = "uniform ";
	else if (type & PARAM_TYPE_VARYING)
		ret = "varying ";
	else if (type & PARAM_TYPE_VERTEX)
		ret = "vertex ";
	else
		Severe("typeToString: unknown type %d", type);

	if (type & PARAM_TYPE_FLOAT)
		ret += "float";
	else if (type & PARAM_TYPE_INT)
		ret += "integer";
	else if (type & PARAM_TYPE_POINT)
		ret += "point";
	else if (type & PARAM_TYPE_HPOINT)
		ret += "hpoint";
	else if (type & PARAM_TYPE_VECTOR)
		ret += "vector";
	else if (type & PARAM_TYPE_NORMAL)
		ret += "normal";
	else if (type & PARAM_TYPE_STRING)
		ret += "string";
	else if (type & PARAM_TYPE_COLOR)
		ret += "color";
	else Severe("typeToString: unknown type %d", type);

	if (nArray > 1) {
		char buf[64];
		sprintf(buf, "[%d]", nArray);
		ret += buf;
	}
	return ret;
}
// API Function Definitions
const char *RiDeclare(char *name, char *type) {
	// Compute integer type code for variable
	int narray;
	int tokenType = stringToType(type, &narray);
	if (tokenType == PARAM_TYPE_ERROR)
		return NULL;
	// Search for token in token table
	Token *token = (Token *)TokenTable.Search(name);
	if (token) {
		// Complain if token was redeclared to a different type
		if (token->type != tokenType || token->arrayLength != narray) {
			string s1 = typeToString(token->type, token->arrayLength);
			string s2 = typeToString(tokenType, narray);
			Warning("RiDeclare: Token '%s' redeclared from %s to %s.",
				name, s1.c_str(), s2.c_str());
		}
		return token->token.c_str();
	}
	// Add token to table if not found
	Token *newToken = new Token(name, tokenType, narray);
	TokenTable.Add(name, newToken);
	return newToken->token.c_str();
}
void InitAPI() {
	// API Initialization
	RiDeclare("Cs", "vertex color");
	RiDeclare("Os", "float");
	RiDeclare("N", "vertex normal");
	RiDeclare("P", "vertex point");
	RiDeclare("Pw", "vertex float[4]");
	RiDeclare("Pz", "vertex float");
	RiDeclare("s", "vertex float");
	RiDeclare("st", "vertex float[2]");
	RiDeclare("t", "vertex float");
	RiDeclare("shadows", "float");
	//if (currentApiState != STATE_UNINITIALIZED)
	//	Severe("RiBegin() has already been called.");
	//currentApiState = STATE_BEGIN;
	curGfxOptions = new GfxOptions;
	for (int i = 0; i < MOTION_LEVELS; ++i)
		curTransform[i] = Transform();
}
void CleanupAPI() {
	// API Cleanup
	delete curGfxOptions;
	curGfxOptions = NULL;
}
void RiPixelSamples(Float x, Float y) {
	VERIFY_STATE(STATE_BEGIN, "RiPixelSamples");
	curGfxOptions->PixelSamples[0] = max(1, Round2Int(x));
	curGfxOptions->PixelSamples[1] = max(1, Round2Int(y));
}
void RiFormat(int x, int y, Float pixelAspect) {
	VERIFY_STATE(STATE_BEGIN, "RiFormat");
	// Set negative format values to defaults
	if (x < 0) x = 640;
	if (y < 0) y = 480;
	if (pixelAspect < 0) pixelAspect = 1;
	// Check values of format values
	if (x < 2) {
		Error("RiFormat: x resolution must be >= 2.  "
			"Value %d invalid", x);
		x = 2;
	}
	if (y < 2) {
		Error("RiFormat: y resolution must be >= 2.  "
			"Value %d invalid", y);
		y = 2;
	}

	curGfxOptions->XResolution = x;
	curGfxOptions->YResolution = y;
	curGfxOptions->PixelAspectRatio = pixelAspect;
	curGfxOptions->FrameAspectRatio = x * pixelAspect / y;
	// Update screen window from format values
	if (curGfxOptions->ScreenExtentSet == false) {
		if (curGfxOptions->FrameAspectRatio >= 1)
			curGfxOptions->ScreenExtent =
				Extent2D(-curGfxOptions->FrameAspectRatio,
					curGfxOptions->FrameAspectRatio, -1, 1);
		else
			curGfxOptions->ScreenExtent =
				Extent2D(-1, 1, -1.f / curGfxOptions->FrameAspectRatio,
					1.f / curGfxOptions->FrameAspectRatio);
	}
}
void RiFrameAspectRatio(Float aspect) {
	VERIFY_STATE(STATE_BEGIN, "RiFrameAspectRatio");
	curGfxOptions->FrameAspectRatio = aspect;
}
void RiScreenWindow(Float left, Float right,
		Float bottom, Float top) {
	VERIFY_STATE(STATE_BEGIN, "RiScreenWindow");
	curGfxOptions->ScreenExtent = Extent2D(left, right,
		bottom, top);
	curGfxOptions->ScreenExtentSet = true;
}
void RiCropWindow(Float left, Float right,
		Float top, Float bottom) {
	VERIFY_STATE(STATE_BEGIN, "RiCropWindow");
	left = Clamp(left, 0., 1.);
	right = Clamp(right, 0., 1.);
	top = Clamp(top, 0., 1.);
	bottom = Clamp(bottom, 0., 1.);
	curGfxOptions->Crop = Extent2D(left, right, top, bottom);
}
void RiClipping(Float hither, Float yon) {
	VERIFY_STATE(STATE_BEGIN, "RiClipping");
	if (hither < RAY_EPSILON) {
		Warning("RiClipping: hither value of %f is too low",
			hither);
		hither = RAY_EPSILON;
	}
	if (yon <= hither) {
		Warning("RiClipping: yon value must be greater "
			"than hither");
		yon = 1e30f;
	}
	curGfxOptions->ClipHither = hither;
	curGfxOptions->ClipYon = yon;
}
void RiShutter(Float time0, Float time1) {
	VERIFY_STATE(STATE_BEGIN, "RiShutter");
	curGfxOptions->ShutterStart = time0;
	curGfxOptions->ShutterEnd = time1;
}
void RiDepthOfField(Float fstop, Float focallen,
		Float focaldist) {
	VERIFY_STATE(STATE_BEGIN, "RiDepthOfField");
	curGfxOptions->LensRadius = .5f * focallen / fstop;
	curGfxOptions->FocalDistance = focaldist;
}
void RiProjectionV(const char * name, int nArgs, const char * tokens[],
		void * params[]) {
	VERIFY_STATE(STATE_BEGIN, "RiProjection");
	curGfxOptions->cameraName = name;
	curGfxOptions->cameraParams.init(nArgs, tokens, params);
	curTransform[0] = Transform();
}
void RiExposure(Float gain, Float gamma) {
	curGfxOptions->Gain = gain;
	curGfxOptions->Gamma = gamma;
}
void RiPixelFilter(const char *filterName, Float xwidth,
		Float ywidth, int nArgs, const char *tokens[],
		void *params[]) {
	curGfxOptions->filterName = filterName;
	curGfxOptions->FilterXWidth = xwidth;
	curGfxOptions->FilterYWidth = ywidth;
	curGfxOptions->FilterParams.init(nArgs, tokens, params);
}
void RiQuantize(const char * type, int one, int minimum,
		int maximum, Float ditheramp) {
	if (strcmp(type, "rgba") == 0) {
		curGfxOptions->ColorQuantOne = one;
		curGfxOptions->ColorQuantMin = minimum;
		curGfxOptions->ColorQuantMax = maximum;
		curGfxOptions->ColorQuantDither = ditheramp;
	}
	else if (strcmp(type, "z") == 0) {
		RI_UNIMP();
	}
	else
		Error("Unknown type %s passed to RiQuantize()", type);
}
void RiDisplayV(char *name, const char *type, const char *displayMode,
		int nArgs, const char *tokens[],
		void *parameters[]) {
	// Complain about unknown display [[type]]
	if (strcmp(type, "tiff") && strcmp(type, "file")) {
		Error("Display type %s not supported.", type);
		return;
	}
	// Initialize display options
	curGfxOptions->DisplayType = type;
	curGfxOptions->DisplayName = name;
	curGfxOptions->displayRGB = !strncmp(displayMode, "rgb", 3);
	curGfxOptions->displayA = strchr(displayMode, 'a') != NULL;
	curGfxOptions->displayZ = strchr(displayMode, 'z') != NULL;
	if (!curGfxOptions->displayRGB && !curGfxOptions->displayA &&
		!curGfxOptions->displayZ) {
		Error("Display mode %s not recognized.", displayMode);
		return;
	}
}
void RiSamplerV(const char * name, int n, const char * tokens[],
		void * params[]) {
	curGfxOptions->samplerName = name;
	curGfxOptions->samplerParams.init(n, tokens, params);
}
void RiAcceleratorV(const char * name, int n, const char * tokens[],
		void * params[]) {
	curGfxOptions->acceleratorName = name;
	curGfxOptions->acceleratorParams.init(n, tokens, params);
}
Scene *GfxOptions::MakeScene() const {
	// Initialize [[filter]] with pixel filter
	Filter *filter = CreateFilter(filterName, SearchPath, FilterParams,
		FilterXWidth, FilterYWidth);
	// Initialize [[sampler]] from API settings
	Sampler *sampler = CreateSampler(samplerName, SearchPath, samplerParams,
			XResolution, YResolution, PixelSamples[0],
			PixelSamples[1], Crop, filter);
	// Initialize [[film]] and [[camera]] from API settings
	Film *film = new Film(XResolution, YResolution, Crop);
	cameraParams.AddFloat("_ShutterOpen", &ShutterStart);
	cameraParams.AddFloat("_ShutterClose", &ShutterEnd);
	cameraParams.AddFloat("_ClipHither", &ClipHither);
	cameraParams.AddFloat("_ClipYon", &ClipYon);
	cameraParams.AddFloat("_LensRadius", &LensRadius);
	cameraParams.AddFloat("_FocalDistance", &FocalDistance);
	cameraParams.AddFloat("_PixelAspectRatio", &PixelAspectRatio);
	cameraParams.AddFloat("_ScreenExtent", (Float *)&ScreenExtent,
		PARAM_TYPE_UNIFORM, 4);
	Camera *camera = CreateCamera(cameraName, SearchPath, cameraParams,
		WorldToCamera[0], film);
	// Initialize [[displayInfo]] from API settings
	DisplayInfo *displayInfo = new DisplayInfo;
	if (ToneMapping != "")
		displayInfo->toneMap = CreateToneMap(ToneMapping, SearchPath,
			ToneMapParams);
	displayInfo->gain = Gain;
	displayInfo->invGamma = 1.f / Gamma;
	displayInfo->integerFormat = ColorQuantOne != 0;
	displayInfo->clampToGamut = displayInfo->integerFormat;
	displayInfo->maxDisplayValue = ColorQuantMax;
	displayInfo->ditherAmount = ColorQuantDither;
	displayInfo->filename = DisplayName;
	// Initialize [[surfaceIntegrator]] from API settings
	SurfaceIntegrator *surfaceIntegrator = CreateSurfaceIntegrator(surfIntegratorName,
		curGfxOptions->SearchPath, surfIntegratorParams);
	// Initialize [[volumeIntegrator]] from API settings
	VolumeIntegrator *volumeIntegrator = CreateVolumeIntegrator(volIntegratorName,
		curGfxOptions->SearchPath, volIntegratorParams);
	// Initialize [[accelerator]] from API settings
	Primitive *toptree;
	vector<Reference<Primitive> > accelerators;
	for(int i=0;i<MAXTREES;i++) {
		if(primitives[i].size()) {
			Reference<Primitive> prim=CreateAccelerator(acceleratorName, SearchPath, primitives[i], acceleratorParams);
			accelerators.push_back(prim);
		}
	}
	toptree=CreateAccelerator(acceleratorName, SearchPath, accelerators, acceleratorParams);

	if (!toptree) {
		ParamSet ps;
		toptree = CreateAccelerator("grid",	SearchPath,	primitives[0], ps);
	}
	Assert(toptree);
	for(int i=0;i<MAXTREES;i++)
		primitives[i].erase(primitives[i].begin(), primitives[i].end());
	if (!camera || !sampler || !film || !camera || !toptree ||
		!displayInfo || !surfaceIntegrator || !volumeIntegrator) {
		Severe("Unable to create scene due to missing DSOs");
		return NULL;
	}
	Scene *ret = new Scene(camera, surfaceIntegrator, volumeIntegrator,
		sampler, toptree, lights, volumeRegions, displayInfo);
	lights.erase(lights.begin(), lights.end());
	volumeRegions.erase(volumeRegions.begin(), volumeRegions.end());
	return ret;
}
void RiSearchPath(const char *path) {
	UpdateSearchPath(curGfxOptions->SearchPath, path);
}
void RiToneMap(const char *name, int n, const char *tokens[],
		void *params[]) {
	curGfxOptions->ToneMapping = name;
	curGfxOptions->ToneMapParams.init(n, tokens, params);
}
void RiSurfaceIntegratorV(const char *name, int n, const char * tokens[],
		void * params[]) {
	curGfxOptions->surfIntegratorName = name;
	curGfxOptions->surfIntegratorParams.init(n, tokens, params);
}
void RiVolumeIntegratorV(const char *name, int n, const char * tokens[],
		void * params[]) {
	curGfxOptions->volIntegratorName = name;
	curGfxOptions->volIntegratorParams.init(n, tokens, params);
}
void RiWorldBegin() {
	VERIFY_STATE(STATE_BEGIN, "RiWorldBegin");
	currentApiState = STATE_WORLD_BEGIN;
	// RiWorldBegin initialization
	for (int i = 0; i < MOTION_LEVELS; ++i) {
		curGfxOptions->WorldToCamera[i] = curTransform[i];
		curTransform[i] = Transform();
	}
	curGfxState = GfxState();
	nbKdtrees=0;
}
void RiWorldEnd() {
	// Check for valid WorldEnd state
	while (hierarchicalState.size()) {
		char c = hierarchicalState.back();
		if (c == 't') Error("Missing end to RiTransformBegin");
		else if (c == 'a') Error("Missing end to RiAttributeBegin");
		else Severe("Internal error in gfx state management");
		hierarchicalState.pop_back();
	}
	// Create scene and render
	Scene *scene = curGfxOptions->MakeScene();
	scene->Render();
	delete scene;
	currentApiState = STATE_BEGIN;
	curGfxOptions->WorldEnd();
	// Print per-frame statistics
	StatsPrint(stderr);
}
void RiTransformBegin() {
	for (int i = 0; i < MOTION_LEVELS; ++i)
		transformStack[i].push_back(curTransform[i]);
	hierarchicalState.push_back('t');
}
void RiTransformEnd() {
	if (!transformStack[0].size() ||
		hierarchicalState.back() != 't') {
		Error("Unmatched RiTransformEnd encountered.  Ignoring it.");
		return;
	}
	for (int i = 0; i < MOTION_LEVELS; ++i) {
		curTransform[i] = transformStack[i].back();
		transformStack[i].pop_back();
	}
	hierarchicalState.pop_back();
}
void RiAttributeBegin() {
	RiTransformBegin();
	VERIFY_STATE(STATE_WORLD_BEGIN, "RiAttributeBegin");
	gstates.push_back(curGfxState);
	hierarchicalState.push_back('a');
}
void RiAttributeEnd() {
	VERIFY_STATE(STATE_WORLD_BEGIN, "RiAttributeEnd");
	if (!gstates.size() || hierarchicalState.back() != 'a') {
		Error("Unmatched RiAttributeEnd encountered.  Ignoring it.");
		return;
	}
	curGfxState = gstates.back();
	gstates.pop_back();
	hierarchicalState.pop_back();
	RiTransformEnd();
	//MEGA BIG HACK !!!!!
	// Every time we add a new top level AttributeBegin, start a new KdTree !!!!
	if(gstates.size()==0) {
		cout << "Added tree "<< nbKdtrees <<endl;
		nbKdtrees++;
	}
}
void RiColor(Float *Cs) {
	Assert(COLOR_SAMPLES == 3);
	curGfxState.color = Spectrum(Cs);
}
void RiOpacity(Float *Os) {
	RI_UNIMP();
}
void RiSurfaceV(const char * name, int n, const char * tokens[],
		void * params[]) {
	curGfxState.surfaceParams.init(n, tokens, params);
	curGfxState.surface = name;
}
void RiDisplacementV(const char * name, int n, const char * tokens[],
	void * params[]) {
	curGfxState.displaceParams.init(n, tokens, params);
	curGfxState.displacement = name;
}
void RiMotionBeginV(int N, Float times[]) {
	Assert(!inMotionBlock);
	inMotionBlock = true;
	motionLevel = 0;
	if (N > 2)
		Warning("Only two levels in motion block will be used.");
}
void RiMotionEnd() {
	if (!inMotionBlock)
		Error("Unmatched MotionEnd statement");
	inMotionBlock = false;
}
void RiIdentity() {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	curTransform[xform] = Transform();
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiTransform(Float tr[16]) {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	curTransform[xform] = Transform(new Matrix4x4(
		tr[0], tr[4], tr[8], tr[12],
		tr[1], tr[5], tr[9], tr[13],
		tr[2], tr[6], tr[10], tr[14],
		tr[3], tr[7], tr[11], tr[15]));
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiConcatTransform(Float tr[16]) {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	curTransform[xform] = curTransform[xform] *
		Transform(new Matrix4x4(tr[0], tr[4], tr[8], tr[12],
				 tr[1], tr[5], tr[9], tr[13],
				 tr[2], tr[6], tr[10], tr[14],
				 tr[3], tr[7], tr[11], tr[15]));
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiPerspective(Float fov) {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	// XXX?
	curTransform[xform] = curTransform[xform] *
		Perspective(fov, curGfxOptions->ClipHither, curGfxOptions->ClipYon);
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiTranslate(Float dx, Float dy, Float dz) {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	curTransform[xform] = curTransform[xform] * Translate(Vector(dx, dy, dz));
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiRotate(Float angle, Float dx, Float dy, Float dz) {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	curTransform[xform] = curTransform[xform] * Rotate(angle, Vector(dx, dy, dz));
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiScale(Float sx, Float sy, Float sz) {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	curTransform[xform] = curTransform[xform] * Scale(sx, sy, sz);
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiLookAt(Float ex, Float ey, Float ez, Float lx, Float ly,
	Float lz, Float ux, Float uy, Float uz) {
	// Prepare for motion transform
	int xform = 0;
	if (inMotionBlock)
		xform = motionLevel;
	if (motionLevel > MOTION_LEVELS) {
		Warning("Only %d motion levels are supported. Ignoring.",
			motionLevel);
		return;
	}
	curTransform[xform] = curTransform[xform] *
		LookAt(Point(ex, ey, ez), Point(lx, ly, lz),
			Vector(ux, uy, uz));
	// Update transform for motion block
	if (inMotionBlock)
		++motionLevel;
}
void RiGeneralPolygonV(int nloops, int nverts[], int n,
	const char * tokens[], void * parms[]) {
	Warning("Treating GeneralPolygon as regular Polygon; ignoring holes");
	RiPolygonV(nverts[0], n, tokens, parms);
}
void RiPolygonV(int nverts, int n, const char * tokens[],
		void * params[]) {
	ParamSet geomParams(n, tokens, params, nverts);
	// Get vertex positions from parameters
	Point *P = NULL;
	int i;
	for (i = 0; i < n; ++i)
		if (strcmp(tokens[i], "P") == 0)
			P = (Point *)params[i];
	// Tessellate single convex polygon
	int nTris = nverts-2;
	int nIndices = 3*nTris;
	int *indices = (int *)alloca(nIndices * sizeof(int));
	int *nv = (int *)alloca(nTris * sizeof(int));
	for (i = 0; i < nTris; ++i)
		nv[i] = 3;
	for (i = 0 ; i < nverts-2 ; i++) {
		indices[3*i] = 0;
		indices[3*i+1] = i+2;
		indices[3*i+2] = i+1;
	}
	geomParams.AddInt("_ntris", &nTris);
	geomParams.AddInt("_nverts", &nverts);
	geomParams.AddInt("_vertexIndices", indices,
		PARAM_TYPE_UNIFORM, nIndices);
	curGfxState.AddShape(
		CreateShape("trianglemesh", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams), geomParams);
}
void RiSphereV(Float radius, Float zmin, Float zmax,
		Float thetaMax, int n, const char * tokens[],
		void * params[]) {
	ParamSet geomParams(n, tokens, params);
	geomParams.AddFloat("_radius", &radius);
	geomParams.AddFloat("_zmin", &zmin);
	geomParams.AddFloat("_zmax", &zmax);
	geomParams.AddFloat("_thetamax", &thetaMax);
	curGfxState.AddShape(
		CreateShape("sphere", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams);
}
void *RiLightSourceV(const char * name, int nArgs,
     		const char *tokens[], void * params[]) {
	ParamSet paramSet(nArgs, tokens, params);
	bool shadows = paramSet.FindOneInt("shadows", 1) != 0;
	Light *lt = CreateLight(name, curGfxOptions->SearchPath,
		shadows, curTransform[motionLevel],
		paramSet);
	// Add new light to graphics state
	if (lt == NULL)
		Error("RiLightSource: light type '%s' unknown.", name);
	else
		curGfxOptions->lights.push_back(lt);
	return lt;
}
void *RiAreaLightSourceV(const char *name, int n,
		const char *tokens[], void *params[] ) {
	curGfxState.areaLightParams.init(n, tokens, params);
	curGfxState.areaLight = name;
	return NULL;
}
void RiVolumeV(const char *name, int nArgs, const char *tokens[],
		void *params[]) {
	ParamSet paramSet(nArgs, tokens, params);
	VolumeRegion *vr = CreateVolumeRegion(name, curGfxOptions->SearchPath,
		curTransform[motionLevel], paramSet);
	if (vr) curGfxOptions->volumeRegions.push_back(vr);
}
void RiPixelVariance(Float) { }
void RiTextureCoordinates(Float s1, Float t1, Float s2, Float t2,
		Float s3, Float t3, Float s4, Float t4) { RI_UNIMP( ); }
void RiShadingRate(Float r) { }
void RiShadingInterpolation(const char * type) { }
void RiRelativeDetail(Float relativedetail) { }
void RiMatte(bool onoff) { RI_UNIMP(); }
void RiBound(Float bound[6]) {RI_UNIMP( ); }
void RiDetail(Float bound[6]) {RI_UNIMP( ); }
extern void RiDetailRange(Float minvis, Float lowtran,
	Float uptran, Float maxvis) {RI_UNIMP( ); }
void RiGeometricApproximation(const char * type, Float value) { }
void RiGeometricRepresentation(const char * type) { RI_UNIMP(); }
void RiSkew(Float, Float, Float, Float, Float, Float, Float) {
	RI_UNIMP();
}
extern void RiDeformationV(const char * name, int n, const char * tokens[],
	void * parms[]) {RI_UNIMP( ); }
void RiMakeTextureV(char *pic, char *tex, const char * swrap, const char * twrap,
                    const char * filterfunc, Float swidth, Float twidth,
                    int n, const char * tokens[], void * params[]) { RI_UNIMP(); }
void RiMakeBumpV(char *pic, char *tex, const char * swrap, const char * twrap,
                 const char * filterfunc, Float swidth, Float twidth,
                 int n, const char * tokens[], void * params[]) { RI_UNIMP(); }
void RiMakeLatLongEnvironmentV(char *pic, char *tex, const char * filterfunc,
                              Float swidth, Float twidth,
                               int n, const char * tokens[], void * params[]) { RI_UNIMP(); }
void RiMakeCubeFaceEnvironmentV(char *px, char *nx, char *py,
                                char *ny, char *pz, char *nz,
                                char *tex, Float fov, const char * filterfunc,
                                Float swidth, Float twidth,
                                int n, const char * tokens[], void * params[]) { RI_UNIMP(); }
void RiMakeShadowV(char *pic, char *tex,
                   int n, const char * tokens[], void * params[]) { RI_UNIMP(); }
extern void RiColorSamples(int N, Float *nRGB, Float *RGBn) {RI_UNIMP( ); }
extern void RiAtmosphereV(const char * name, int n, const char * tokens[],
	void * parms[]) {RI_UNIMP( ); }
extern void RiInteriorV(const char * name, int n, const char * tokens[],
	void * parms[]) {RI_UNIMP( ); }
extern void RiExteriorV(const char * name, int n, const char * tokens[],
       void * parms[]) {RI_UNIMP( ); }
void RiSolidBegin(const char *) { RI_UNIMP(); }
void RiSolidEnd() { RI_UNIMP(); }
void *RiObjectBegin() { RI_UNIMP(); return NULL; }
void RiObjectEnd() { RI_UNIMP(); }
void RiNuPatchV(int nu, int uorder, Float *uknot,
		Float umin, Float umax, int nv, int vorder,
		Float *vknot, Float vmin, Float vmax,
		int nArgs, const char * tokens[], void * params[]) {
	ParamSet geomParams(nArgs, tokens, params, nu*nv);
	geomParams.AddInt("_nu", &nu);
	geomParams.AddInt("_uorder", &uorder);
	geomParams.AddFloat("_uknot", uknot, PARAM_TYPE_UNIFORM, nu+uorder);
	geomParams.AddFloat("_u0", &umin);
	geomParams.AddFloat("_u1", &umax);
	geomParams.AddInt("_nv", &nv);
	geomParams.AddInt("_vorder", &vorder);
	geomParams.AddFloat("_vknot", vknot, PARAM_TYPE_UNIFORM, nv+vorder);
	geomParams.AddFloat("_v0", &vmin);
	geomParams.AddFloat("_v1", &vmax);
	curGfxState.AddShape(
		CreateShape("nurbs", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams );
}

void RiTrimCurve(int nloops, int *ncurves, int *order,
	Float *knot, Float *amin, Float *amax,
	int *n, Float *u, Float *v, Float *w) {
	RI_UNIMP();
}
void RiBasis(Float ubasis[16], int ustep, Float vbasis[16], int vstep) {
}

void RiPatchV(const char * type, int n, const char * tokens[], void * params[]) {
	RI_UNIMP();
}
void RiOrientation(const char * orientation) { RI_UNIMP(); }
void RiReverseOrientation() { RI_UNIMP(); }
void RiSides(int sides) { RI_UNIMP(); }
void RiFrameBegin(int num) {
}
void RiFrameEnd() {
}
void RiImagerV(const char * name, int n, const char * tokens[],
		void * params[]) {
	RI_UNIMP();
}
void RiCoordinateSystem(const char *) { RI_UNIMP( ); }
void RiIlluminate(void *light, bool on) {
	RI_UNIMP();
}
void RiHiderV(const char * type, int nArgs,
		const char * tokens[], void * parms[]) {
}
bool lookupType(const char *token, int *type, int *narray, string &name) {
	Assert(token != NULL);
	*narray = 1;
	// Try to decode inline parameter declaration
	const char *end = &token[strlen(token) - 1];
	while (isspace(*end) && end > token)
		--end;
	while (!isspace(*end) && end > token)
		--end;
	if (end != token) {
		name = string(end+1);
		string typeString(token, end+1);
		end = &token[strlen(token) - 1];
		while (isspace(*end) && end > token)
			--end;
		if (end == token) return false;
		if (*end == ']') {
			while (*end != '[' && end > token)
				--end;
			if (end == token) return false;
			typeString += string(end);
		}
		// Decode the type substring
		*type = stringToType(typeString.c_str(), narray);
		if (*type != PARAM_TYPE_ERROR)
			return true;
	}
	// Lookup token and type in token table
	Token *tok = (Token *)TokenTable.Search(token);
	if (tok) {
		name = tok->token;
		*type = tok->type;
		*narray = tok->arrayLength;
		return true;
	}
	return false;
}
void RiOptionV(const char * name, int n, const char * tokens[], void * parms[]) {
	RI_UNIMP();
}
void RiAttributeV(const char * name, int n, const char * tokens[], void * parms[]) {
	RI_UNIMP();
}
void RiPointsGeneralPolygonsV(int npolys, int nloops[],
		int nverts[],	int verts[], int n, const char * tokens[],
		void * params[]) {
	Warning("Treating PointsGeneralPolygons as regular PointsPolygons;"
		"ignoring holes");
	int *newNumVerts = new int[npolys];
	vector<int> newVerts;
	int *nvp = nverts, *vp = verts;
	for (int i = 0; i < npolys; ++i) {
		int j;
		newNumVerts[i] = *nvp++;
		for (j = 0; j < newNumVerts[i]; ++j)
			newVerts.push_back(*vp++);
		for (j = 1; j < nloops[i]; ++j) {
			vp += *nvp;
			++nvp;
		}
	}
	RiPointsPolygonsV(npolys, newNumVerts, &newVerts[0],
		n, tokens, params);
	delete[] newNumVerts;
}
void RiPointsPolygonsV( int npolys, int nvertices[],
		int vertices[], int n, const char * tokens[],
		void * params[])  {
	// Get vertex positions from parameters
	Point *P = NULL;
	int i;
	for (i = 0; i < n; ++i)
		if (strcmp(tokens[i], "P") == 0)
			P = (Point *)params[i];
	// Tessellate convex polygon mesh
	int nVertices = 0;
	int nIndices = 0;
	int vIndex = 0;
	int nTris = 0;
	for (int i = 0; i < npolys; ++i) {
		nTris += nvertices[i]-2;
		nIndices += 3*(nvertices[i]-2);
		for (int j = 0 ; j < nvertices[i] ; j++) {
			nVertices = max(nVertices, vertices[vIndex] + 1);
			vIndex++;
		}
	}
	int *indices = new int[ nIndices ];
	
	int whichIndex = 0;
	vIndex = 0;
	for (int i = 0; i < npolys; ++i) {
		int anchorIndex = vIndex;
		for (int j = 0 ; j < nvertices[i]-2 ; j++) {
			indices[whichIndex++] = vertices[anchorIndex];
			indices[whichIndex++] = vertices[vIndex+1];
			indices[whichIndex++] = vertices[vIndex+2];
			vIndex++;
		}
		vIndex += 2;
	}
	ParamSet geomParams(n, tokens, params, nVertices);
	geomParams.AddInt("_ntris", &nTris);
	geomParams.AddInt("_nverts", &nVertices);
	geomParams.AddInt("_vertexIndices", indices, PARAM_TYPE_UNIFORM, nIndices);
	curGfxState.AddShape(
		CreateShape("trianglemesh", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams);
}
void RiDiskV(Float height, Float radius, Float thetamax,
	int n, const char * tokens[], void * params[]) {
	ParamSet geomParams(n, tokens, params);
	geomParams.AddFloat( "_height", &height );
	geomParams.AddFloat( "_radius", &radius );
	geomParams.AddFloat( "_thetamax", &thetamax );
	curGfxState.AddShape(
		CreateShape("disk", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams);
}
void RiGeometryV(const char * type, int n, const char * tokens[], void * params[])
{
	ParamSet geomParams(n, tokens, params);
	curGfxState.AddShape(
		CreateShape(type, curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams );
}

void RiConeV(Float height, Float radius, Float tmax,
	int n, const char * tokens[], void * params[]) {
	ParamSet geomParams(n, tokens, params);
	geomParams.AddFloat( "_height", &height );
	geomParams.AddFloat( "_radius", &radius );
	geomParams.AddFloat( "_thetamax", &tmax );
	curGfxState.AddShape(
		CreateShape("cone", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams );
}

void RiCylinderV(Float radius, Float zmin, Float zmax, Float tmax,
	int n, const char * tokens[], void * params[]) {
	ParamSet geomParams(n, tokens, params);
	geomParams.AddFloat( "_radius", &radius );
	geomParams.AddFloat( "_zmin", &zmin );
	geomParams.AddFloat( "_zmax", &zmax );
	geomParams.AddFloat( "_thetamax", &tmax );
	curGfxState.AddShape(
		CreateShape("cylinder", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams );
}

void RiHyperboloidV(Float p1[3], Float p2[3], Float tmax,
	int n, const char * tokens[], void * params[]) {
	ParamSet geomParams(n, tokens, params);
        Point P1(p1[0], p1[1], p1[2]), P2(p2[0], p2[1], p2[2]);
	geomParams.AddPoint( "_p1", &P1 );
	geomParams.AddPoint( "_p2", &P2 );
	geomParams.AddFloat( "_thetamax", &tmax );
	curGfxState.AddShape(
		CreateShape("hyperboloid", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams );
}

void RiParaboloidV(Float rmax, Float zmin, Float zmax, Float tmax,
	int n, const char * tokens[], void * params[]) {
	ParamSet geomParams(n, tokens, params);
	geomParams.AddFloat( "_radius", &rmax );
	geomParams.AddFloat( "_zmin", &zmin );
	geomParams.AddFloat( "_zmax", &zmax );
	geomParams.AddFloat( "_thetamax", &tmax );
	curGfxState.AddShape(
		CreateShape("paraboloid", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams),
		geomParams );
}

void RiTorusV(Float majrad, Float minrad, Float phimin,
	Float phimax, Float tmax, int n, const char * tokens[],
	void * params[]) {
	RI_UNIMP();
}
void RiPatchMeshV(const char * type, int nu, const char * uwrap, int nv,
		const char * vwrap, int n, const char * tokens[],
		void * params[]) {
	if (!strcmp(type, "bilinear") &&
		!strcmp(uwrap, "nonperiodic") &&
		!strcmp(vwrap, "nonperiodic")) {
		ParamSet geomParams(n, tokens, params, nu*nv);
		geomParams.AddInt("_nu", &nu);
		geomParams.AddInt("_nv", &nv);
		curGfxState.AddShape(
			CreateShape("heightfield", curGfxOptions->SearchPath,
				curTransform[motionLevel],
				geomParams),
			geomParams);
	}
	else {
		RI_UNIMP();
	}
}
void RiSubdivisionMeshV (const char * scheme, int nfaces,
			 int nvertices[], int vertices[],
			 int ntags, const char * tags[], int nargs[],
			 int intargs[], Float floatargs[],
			 int n, const char * tokens[], void * params[]) {
	int i;
	int totVertices = 0, maxVertNum = 0;
	for (i = 0; i < nfaces; ++i)
		totVertices += nvertices[i];
	for (i = 0; i < totVertices; ++i)
		maxVertNum = max(maxVertNum, vertices[i]);
	++maxVertNum;

	ParamSet geomParams(n, tokens, params, maxVertNum);
	string s(scheme);
	geomParams.AddString("_scheme", &s);
	geomParams.AddInt("_nfaces", &nfaces);
	geomParams.AddInt("_nvertices", &maxVertNum);
	geomParams.AddInt("_faceNumVertices", nvertices, PARAM_TYPE_UNIFORM,
		nfaces);
	geomParams.AddInt("_vertices", vertices, PARAM_TYPE_UNIFORM,
		totVertices);
	curGfxState.AddShape(CreateShape("subdiv", curGfxOptions->SearchPath,
			curTransform[motionLevel], geomParams), geomParams);
}
