/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
%{
#include "ri.h"
#include "lrt.h"

#include <stdarg.h>

extern int yylex( void );
int line_num = 0;
string current_file;

#define YYMAXDEPTH 100000000

void yyerror( char *str ) {
	Error( "RIB parsing error: %s", str);
}

void ParseError( const char *format, ... ) PRINTF_FORMAT;

void ParseError( const char *format, ... ) {
	char error[4096];
	va_list args;
	va_start( args, format );
	vsprintf( error, format, args );
	yyerror(error);
	va_end( args );
}

#define MAX_LIGHT_HANDLE 65536
static void *lights[MAX_LIGHT_HANDLE];

int cur_paramlist_allocated = 0;
int cur_paramlist_size = 0;
const char **cur_paramlist_tokens = NULL;
void **cur_paramlist_args = NULL;
int *cur_paramlist_sizes = NULL;

#define CPS cur_paramlist_size
#define CPT cur_paramlist_tokens
#define CPA cur_paramlist_args
#define CPSZ cur_paramlist_sizes

typedef struct RIB_array {
	int element_size;
	int allocated;
	int nelems;
	void *array;
} RIB_array;

RIB_array *cur_array = NULL;

#define NA(r) ((Float *) r->array)
#define SA(r) ((const char **) r->array)

void AddArrayElement( void *elem ) {
	if (cur_array->nelems >= cur_array->allocated) {
		cur_array->allocated = 2*cur_array->allocated + 1;
		cur_array->array = realloc( cur_array->array,
			cur_array->allocated*cur_array->element_size );
	}
	char *next = ((char *)cur_array->array) + cur_array->nelems *
		cur_array->element_size;
	memcpy( next, elem, cur_array->element_size );
	cur_array->nelems++;
}

RIB_array *ArrayDup( RIB_array *ra )
{
	RIB_array *ret = new RIB_array;
	ret->element_size = ra->element_size;
	ret->allocated = ra->allocated;
	ret->nelems = ra->nelems;
	ret->array = malloc(ra->nelems * ra->element_size);
	memcpy( ret->array, ra->array, ra->nelems * ra->element_size );
	return ret;
}

void ArrayFree( RIB_array *ra )
{
	free(ra->array);
	delete ra;
}

void FreeArgs()
{
	for (int i = 0; i < cur_paramlist_size; ++i)
		delete[] ((char *)cur_paramlist_args[i]);
}

static Float * FindBasis( const char *str )
{
	/*
	if (!strcmp( str, "bezier" )) return (Float *) RiBezierBasis;
	else if (!strcmp( str, "b-spline" )) return (Float *) RiBSplineBasis;
	else if (!strcmp( str, "catmull-rom" )) return (Float *) RiCatmullRomBasis;
	else if (!strcmp( str, "hermite" )) return (Float *) RiHermiteBasis;
	else if (!strcmp( str, "power" )) return (Float *) RiPowerBasis;
	else ParseError( "Unknown basis: %s", str );
	*/

	Warning("Named bases aren't supported yet...");

	return NULL;
}
static bool VerifyArrayLength( RIB_array *arr, int required,
	const char *command ) {
	if (arr->nelems != required) {
		ParseError( "%s requires a(n) %d element array!", command, required);
		return false;
	}
	return true;
}

static int *MakeIntArray( RIB_array *arr )
{
	int *ret = new int[arr->nelems];
	for (int i = 0 ; i < arr->nelems ; i++)
		ret[i] = int(NA(arr)[i]);
	return ret;
}

%}

%union {
char string[1024];
Float num;
RIB_array *ribarray;
}
%token <string> STRING ID
%token <num> NUM
%token LBRACK RBRACK DASH


%token ACCELERATOR AREALIGHTSOURCE ATMOSPHERE ATTRIBUTE
%token ATTRIBUTEBEGIN ATTRIBUTEEND BASIS
%token BOUND CLIPPING COLOR COLORSAMPLES CONCATTRANSFORM
%token CONE COORDINATESYSTEM CROPWINDOW CYLINDER DECLARE
%token DEFORMATION DEPTHOFFIELD DETAIL DETAILRANGE DISK
%token DISPLACEMENT DISPLAY EXPOSURE EXTERIOR FORMAT
%token FRAMEASPECTRATIO FRAMEBEGIN FRAMEEND GENERALPOLYGON
%token GEOMETRICAPPROXIMATION GEOMETRY HIDER HYPERBOLOID
%token IDENTITY ILLUMINATE IMAGER INTERIOR LIGHTSOURCE LOOKAT
%token MAKEBUMP MAKECUBEFACEENVIRONMENT MAKELATLONGENVIRONMENT
%token MAKESHADOW MAKETEXTURE MATTE MOTIONBEGIN MOTIONEND
%token NUPATCH OBJECTBEGIN OBJECTEND OBJECTINSTANCE OPACITY
%token OPTION ORIENTATION PARABOLOID PATCH PATCHMESH PERSPECTIVE PIXELFILTER
%token PIXELSAMPLES PIXELVARIANCE POINTSGENERALPOLYGONS
%token POINTSPOLYGONS POLYGON PROJECTION QUANTIZE
%token RELATIVEDETAIL REVERSEORIENTATION ROTATE SAMPLER SCALE
%token SCREENWINDOW SEARCHPATH SHADINGINTERPOLATION SHADINGRATE SHUTTER
%token SIDES SKEW SOLIDBEGIN SOLIDEND SPHERE SUBDIVISIONMESH SURFACE
%token SURFACEINTEGRATOR
%token TEXTURECOORDINATES TONEMAP TORUS TRANSFORM TRANSFORMBEGIN
%token TRANSFORMEND TRANSLATE TRIMCURVE
%token VERSION VOLUME VOLUMEINTEGRATOR WORLDBEGIN WORLDEND

%token HIGH_PRECEDENCE

%type<ribarray> array num_array string_array
%type<ribarray> real_num_array real_string_array
%%
start: ri_stmt_list

array_init: %prec HIGH_PRECEDENCE
{
	if (cur_array) ArrayFree( cur_array );
	cur_array = new RIB_array;
	cur_array->allocated = 0;
	cur_array->nelems = 0;
	cur_array->array = NULL;
}

string_array_init: %prec HIGH_PRECEDENCE
{
	cur_array->element_size = sizeof( const char * );
}

num_array_init: %prec HIGH_PRECEDENCE
{
	cur_array->element_size = sizeof( Float );
}

array: string_array
{
	$$ = $1;
}
| num_array
{
	$$ = $1;
}

string_array: real_string_array
{
	$$ = $1;
}
| single_element_string_array
{
	$$ = ArrayDup(cur_array);
}

real_string_array: array_init LBRACK string_list RBRACK
{
	$$ = ArrayDup(cur_array);
}
single_element_string_array: array_init string_list_entry

string_list: string_list string_list_entry
| string_list_entry

string_list_entry: string_array_init STRING
{
	char *to_add = strdup($2);
	AddArrayElement( &to_add );
}

num_array: real_num_array
{
	$$ = $1;
}
| single_element_num_array
{
	$$ = ArrayDup(cur_array);
}

real_num_array: array_init LBRACK num_list RBRACK
{
	$$ = ArrayDup(cur_array);
}
single_element_num_array: array_init num_list_entry

num_list: num_list num_list_entry
	| num_list_entry

num_list_entry: num_array_init NUM
{
	Float to_add = $2;
	AddArrayElement( &to_add );
}

paramlist: paramlist_init paramlist_contents

paramlist_init: %prec HIGH_PRECEDENCE
{
	cur_paramlist_size = 0;
}

paramlist_contents: paramlist_entry paramlist_contents
	|

paramlist_entry: STRING array
{
	void *arg = new char[ $2->nelems * $2->element_size ];
	memcpy(arg, $2->array, $2->nelems * $2->element_size);
	if (cur_paramlist_size >= cur_paramlist_allocated) {
		cur_paramlist_allocated = 2*cur_paramlist_allocated + 1;
		cur_paramlist_tokens = (const char **) realloc(cur_paramlist_tokens, cur_paramlist_allocated*sizeof(const char *) );
		cur_paramlist_args = (void * *) realloc( cur_paramlist_args, cur_paramlist_allocated*sizeof(void *) );
		cur_paramlist_sizes = (int *) realloc( cur_paramlist_sizes, cur_paramlist_allocated*sizeof(int) );
	}
	cur_paramlist_tokens[cur_paramlist_size] = $1;
	cur_paramlist_sizes[cur_paramlist_size] = $2->nelems;
	cur_paramlist_args[cur_paramlist_size++] = arg;
	ArrayFree( $2 );
}

ri_stmt_list: ri_stmt_list ri_stmt
	| ri_stmt

ri_stmt: VERSION NUM { }
| ACCELERATOR STRING paramlist
{
	RiAcceleratorV($2, CPS, CPT, CPA);
	FreeArgs();
}
| AREALIGHTSOURCE STRING NUM paramlist
{
    int num = (int)$3;
	if (num < 0 || num >= MAX_LIGHT_HANDLE)
		Error("Invalid light handle %d passed to AreaLightSource. Ignoring.", num);
	else
		lights[num] = RiAreaLightSourceV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| ATMOSPHERE STRING paramlist
{
	RiAtmosphereV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| ATTRIBUTE STRING paramlist
{
	RiAttributeV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| ATTRIBUTEBEGIN
{
	RiAttributeBegin();
}
| ATTRIBUTEEND
{
	RiAttributeEnd();
}
| BASIS STRING NUM STRING NUM
{
	Float * basis1 = FindBasis( $2 );
	Float * basis2 = FindBasis( $4 );
	RiBasis( basis1, int($3), basis2, int($5) );
}
| BASIS STRING NUM real_num_array NUM
{
	if (VerifyArrayLength( $4, 16, "Basis" )) {
		Float * basis1 = FindBasis( $2 );
		RiBasis( basis1, int($3), (Float *) $4->array, int($5) );
	}
	ArrayFree( $4 );
}
| BASIS real_num_array NUM STRING NUM
{
	if (VerifyArrayLength( $2, 16, "Basis" )) {
		Float * basis2 = FindBasis( $4 );
		RiBasis( (Float *) $2->array, int($3), basis2, int($5) );
	}
	ArrayFree( $2 );
}
| BASIS real_num_array NUM real_num_array NUM
{
	if (VerifyArrayLength($2, 16, "Basis") && VerifyArrayLength($4, 16, "Basis"))
		RiBasis( (Float *) $2->array, int($3), (Float *) $4->array, int($5) );
	ArrayFree( $2 );
	ArrayFree( $4 );
}
| BOUND NUM NUM NUM NUM NUM NUM
{
	Float bound[6];
	bound[0] = $2; bound[1] = $3; bound[2] = $4;
	bound[3] = $5; bound[4] = $6; bound[5] = $7;
	RiBound( bound );
}
| BOUND real_num_array
{
	if (VerifyArrayLength( $2, 6, "Bound" ))
		RiBound( (Float *) $2->array );
	ArrayFree( $2 );
}
| CLIPPING NUM NUM
{
	RiClipping( $2, $3 );
}
| COLOR NUM NUM NUM
{
	Assert(COLOR_SAMPLES == 3);
	Float color[3];
	color[0] = $2; color[1] = $3; color[2] = $4;
	RiColor( color );
}
| COLOR real_num_array
{
	if (VerifyArrayLength( $2, 3, "Color" ))
		RiColor( (Float *) $2->array );
	ArrayFree( $2 );
}
| COLORSAMPLES real_num_array real_num_array
{
	if ($2->nelems != $3->nelems) {
		ParseError( "The two arrays should be the same size in ColorSamples" );
		goto noGoodColor;
	}
	if ($2->nelems % 3 != 0) {
		ParseError( "The two arrays should be the multiples of 3 in ColorSamples" );
		goto noGoodColor;
	}
	RiColorSamples( $2->nelems / 3, NA($2), NA($3) );
noGoodColor:
	ArrayFree( $2 );
	ArrayFree( $3 );
}
| CONE NUM NUM NUM paramlist
{
	RiConeV( $2, $3, $4, CPS, CPT, CPA );
	FreeArgs();
}
| CONE real_num_array paramlist
{
	if (VerifyArrayLength( $2, 3, "Cone" ))
		RiConeV( NA($2)[0], NA($2)[1], NA($2)[2], CPS, CPT, CPA );
	FreeArgs();
}
| CONCATTRANSFORM num_array
{
	if (VerifyArrayLength( $2, 16, "ConcatTransform" ))
		RiConcatTransform( (Float *) $2->array );
	ArrayFree( $2 );
}
| COORDINATESYSTEM STRING
{
	RiCoordinateSystem( $2 );
}
| CROPWINDOW NUM NUM NUM NUM
{
	RiCropWindow( $2, $3, $4, $5 );
}
| CROPWINDOW real_num_array
{
	if (VerifyArrayLength( $2, 4, "CropWindow" ))
		RiCropWindow( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3] );
	ArrayFree( $2 );
}
| CYLINDER NUM NUM NUM NUM paramlist
{
	RiCylinderV( $2, $3, $4, $5, CPS, CPT, CPA );
	FreeArgs();
}
| CYLINDER real_num_array paramlist
{
	if (VerifyArrayLength( $2, 4, "Cylinder" ))
		RiCylinderV( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3], CPS, CPT, CPA );
	ArrayFree( $2 );
	FreeArgs();
}
| DECLARE STRING STRING
{
	RiDeclare( $2, $3 );
}
| DEFORMATION STRING paramlist
{
	RiDeformationV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| DEPTHOFFIELD NUM NUM NUM
{
	RiDepthOfField( $2, $3, $4 );
}
| DETAIL NUM NUM NUM NUM NUM NUM
{
	Float bound[6];
	bound[0] = $2; bound[1] = $3; bound[2] = $4;
	bound[3] = $5; bound[4] = $6; bound[5] = $7;
	RiDetail( bound );
}
| DETAIL real_num_array
{
	if (VerifyArrayLength( $2, 6, "Detail" ))
		RiDetail( (Float *) $2->array );
	ArrayFree( $2 );
}
| DETAILRANGE NUM NUM NUM NUM
{
	RiDetailRange( $2, $3, $4, $5 );
}
| DETAILRANGE real_num_array
{
	if (VerifyArrayLength( $2, 4, "DetailRange" ))
		RiDetailRange( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3] );
	ArrayFree( $2 );
}
| DISK NUM NUM NUM paramlist
{
	RiDiskV( $2, $3, $4, CPS, CPT, CPA );
	FreeArgs();
}
| DISK real_num_array paramlist
{
	if (VerifyArrayLength( $2, 3, "Disk" ))
		RiDiskV( NA($2)[0], NA($2)[1], NA($2)[2], CPS, CPT, CPA );
	ArrayFree( $2 );
	FreeArgs();
}
| DISPLACEMENT STRING paramlist
{
	RiDisplacementV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| DISPLAY STRING STRING STRING paramlist
{
	RiDisplayV( $2, $3, $4, CPS, CPT, CPA );
	FreeArgs();
}
| EXPOSURE NUM NUM
{
	RiExposure( $2, $3 );
}
| EXTERIOR STRING paramlist
{
	RiExteriorV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| FORMAT NUM NUM NUM
{
	RiFormat( int($2), int($3), $4 );
}
| FRAMEASPECTRATIO NUM
{
	RiFrameAspectRatio( $2 );
}
| FRAMEBEGIN NUM
{
	RiFrameBegin( int($2) );
}
| FRAMEEND
{
	RiFrameEnd();
}
| GENERALPOLYGON num_array paramlist
{
	int *verts = MakeIntArray( $2 );
	RiGeneralPolygonV( $2->nelems, verts, CPS, CPT, CPA );
	delete [] verts;
	ArrayFree( $2 );
	FreeArgs();
}
| GEOMETRICAPPROXIMATION STRING NUM
{
	RiGeometricApproximation( $2, $3 );
}
| GEOMETRY STRING paramlist
{
	RiGeometryV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| HIDER STRING paramlist
{
	RiHiderV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| HYPERBOLOID NUM NUM NUM NUM NUM NUM NUM paramlist
{
	Float point1[3], point2[3];
	point1[0] = $2; point1[1] = $3; point1[2] = $4;
	point2[0] = $5; point2[1] = $6; point2[2] = $7;
	RiHyperboloidV( point1, point2, $8, CPS, CPT, CPA );
	FreeArgs();
}
| HYPERBOLOID real_num_array paramlist
{
	Float point1[3], point2[3];
	if (VerifyArrayLength( $2, 7, "Hyperboloid" )) {
		point1[0] = NA($2)[0]; point1[1] = NA($2)[1]; point1[2] = NA($2)[2];
		point2[0] = NA($2)[3]; point2[1] = NA($2)[4]; point2[2] = NA($2)[5];
		RiHyperboloidV( point1, point2, NA($2)[6], CPS, CPT, CPA );
	}
	ArrayFree( $2 );
	FreeArgs();
}
| IDENTITY
{
	RiIdentity();
}
| ILLUMINATE NUM NUM
{
    int num = (int)$2;
	bool light_on = ($3 == 1.0f) ? true : false;
	if (num < 0 || num >= MAX_LIGHT_HANDLE)
		Error("Invalid light handle %d passed to Illuminate. Ignoring.", num);
	else {
		void *light = lights[num];
		if (!light)
			ParseError( "Unknown light number: %d", num);
		else
			RiIlluminate( light, light_on );
	}
}
| IMAGER STRING paramlist
{
	RiImagerV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| INTERIOR STRING paramlist
{
	RiInteriorV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| LIGHTSOURCE STRING NUM paramlist
{
    int num = (int)$3;
	if (num < 0 || num >= MAX_LIGHT_HANDLE)
		Error("Invalid light handle %d passed to LightSource. Ignoring.", num);
	else
		lights[num] = RiLightSourceV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| LOOKAT NUM NUM NUM NUM NUM NUM NUM NUM NUM
{
	RiLookAt($2, $3, $4, $5, $6, $7, $8, $9, $10);
}
| MAKEBUMP STRING STRING STRING STRING STRING NUM NUM paramlist
{
	RiMakeBumpV( $2, $3, $4, $5, $6, $7, $8, CPS, CPT, CPA );
	FreeArgs();
}
| MAKECUBEFACEENVIRONMENT STRING STRING STRING STRING STRING STRING STRING NUM STRING NUM NUM paramlist
{
	RiMakeCubeFaceEnvironmentV( $2, $3, $4, $5, $6, $7, $8, $9, $10,
		$11, $12, CPS, CPT, CPA );
	FreeArgs();
}
| MAKELATLONGENVIRONMENT STRING STRING STRING NUM NUM paramlist
{
	RiMakeLatLongEnvironmentV( $2, $3, $4, $5, $6, CPS, CPT, CPA );
	FreeArgs();
}
| MAKESHADOW STRING STRING paramlist
{
	RiMakeShadowV( $2, $3, CPS, CPT, CPA );
	FreeArgs();
}
| MAKETEXTURE STRING STRING STRING STRING STRING NUM NUM paramlist
{
	RiMakeTextureV( $2, $3, $4, $5, $6, $7, $8, CPS, CPT, CPA );
	FreeArgs();
}
| MATTE NUM
{
	RiMatte( $2 != 0. );
}
| MOTIONBEGIN num_array
{
	RiMotionBeginV( $2->nelems, NA($2) );
	ArrayFree( $2 );
}
| MOTIONEND
{
	RiMotionEnd();
}
| NUPATCH NUM NUM real_num_array NUM NUM NUM NUM real_num_array NUM NUM paramlist
{
	RiNuPatchV( int($2), int($3), NA($4), $5, $6,
				int($7), int($8), NA($9), $10, $11, CPS, CPT, CPA );
	ArrayFree($4);
	ArrayFree($9);
	FreeArgs();
}
| OBJECTBEGIN NUM
{
	Warning( "Don't know how to parse ObjectInstance" );
}
| OBJECTEND
{
	RiObjectEnd();
}
| OBJECTINSTANCE NUM
{
	Warning( "Don't know how to parse ObjectInstance" );
}
| OPACITY NUM NUM NUM
{
	Float color[3];
	color[0] = $2; color[1] = $3; color[2] = $4;
	RiOpacity( color );
}
| OPACITY real_num_array
{
	if (VerifyArrayLength( $2, 3, "Opacity" ))
		RiOpacity( (Float *) $2->array );
	ArrayFree( $2 );
}
| OPTION STRING paramlist
{
	RiOptionV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| ORIENTATION STRING
{
	RiOrientation( $2 );
}
| PARABOLOID NUM NUM NUM NUM paramlist
{
	RiParaboloidV( $2, $3, $4, $5, CPS, CPT, CPA );
	FreeArgs();
}
| PARABOLOID real_num_array paramlist
{
	if (VerifyArrayLength( $2, 4, "Paraboloid" ))
		RiParaboloidV( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3], CPS, CPT, CPA );
	ArrayFree( $2 );
	FreeArgs();
}
| PATCH STRING paramlist
{
	RiPatchV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| PATCHMESH STRING NUM STRING NUM STRING paramlist
{
	RiPatchMeshV( $2, int($3), $4, int($5), $6, CPS, CPT, CPA );
	FreeArgs();
}
| PERSPECTIVE NUM
{
	RiPerspective( $2 );
}
| PIXELFILTER STRING NUM NUM paramlist
{
	RiPixelFilter( $2, $3, $4, CPS, CPT, CPA );
	FreeArgs();
}
| PIXELSAMPLES NUM NUM
{
	RiPixelSamples( $2, $3 );
}
| PIXELVARIANCE NUM
{
	RiPixelVariance( $2 );
}
| POINTSGENERALPOLYGONS num_array num_array num_array paramlist
{
	int *nloops, *nvertices, *vertices;

	nloops = MakeIntArray( $2 );
	nvertices = MakeIntArray( $3 );
	vertices = MakeIntArray( $4 );

	RiPointsGeneralPolygonsV( $2->nelems, nloops, nvertices, vertices, CPS, CPT, CPA );

	delete [] nloops;
	delete [] nvertices;
	delete [] vertices;
	ArrayFree( $2 );
	ArrayFree( $3 );
	ArrayFree( $4 );
	FreeArgs();
}
| POINTSPOLYGONS num_array num_array paramlist
{
	// Gotta convert these bad boys to arrays of integers.  that sucks.
	int *nverts, *verts;

	nverts = MakeIntArray( $2 );
	verts = MakeIntArray( $3 );
	int tot = 0;
	for (int i = 0 ; i < $2->nelems ; i++)
		tot += nverts[i];
	if (tot != $3->nelems)
		ParseError("Wrong number of vertex elements in PointsPolygons. Ignoring.");
	else
		RiPointsPolygonsV( $2->nelems, nverts, verts, CPS, CPT, CPA );

	delete [] nverts;
	delete [] verts;
	ArrayFree( $2 );
	ArrayFree( $3 );
	FreeArgs();
}
| POLYGON paramlist
{
	int num_verts = -1;

	for (int i = 0 ; i < CPS ; i++) {
		if (strcmp(CPT[i], "P") == 0) {
			num_verts = CPSZ[i];
			if (num_verts % 3 != 0) {
				ParseError("Polygon P array must be 3*n in length." );
				goto noGoodPolygon;
			}
			num_verts /= 3;
			break;
		}
	}
	if (num_verts == -1)
		ParseError( "Polygon without a \"P\" parameter???" );
	else
		RiPolygonV( num_verts, CPS, CPT, CPA );
noGoodPolygon:
	FreeArgs();
}
| PROJECTION STRING paramlist
{
	RiProjectionV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| QUANTIZE STRING NUM NUM NUM NUM
{
	RiQuantize( $2, int($3), int($4), int($5), $6 );
}
| RELATIVEDETAIL NUM
{
	RiRelativeDetail( $2 );
}
| REVERSEORIENTATION
{
	RiReverseOrientation();
}
| ROTATE NUM NUM NUM NUM
{
	RiRotate( $2, $3, $4, $5 );
}
| SAMPLER STRING paramlist
{
	RiSamplerV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| SCALE NUM NUM NUM
{
	RiScale( $2, $3, $4 );
}
| SCREENWINDOW NUM NUM NUM NUM
{
	RiScreenWindow( $2, $3, $4, $5 );
}
| SCREENWINDOW real_num_array
{
	if (VerifyArrayLength( $2, 4, "ScreenWindow" ))
		RiScreenWindow( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3] );
	ArrayFree( $2 );
}
| SEARCHPATH STRING
{
	RiSearchPath($2);
}
| SHADINGINTERPOLATION STRING
{
	RiShadingInterpolation( $2 );
}
| SHADINGRATE NUM
{
	RiShadingRate( $2 );
}
| SHUTTER NUM NUM
{
	RiShutter( $2, $3 );
}
| SIDES NUM
{
	RiSides( int($2) );
}
| SKEW NUM NUM NUM NUM NUM NUM NUM
{
	RiSkew( $2, $3, $4, $5, $6, $7, $8 );
}
| SKEW real_num_array
{
	if (VerifyArrayLength( $2, 7, "Skew" ))
		RiSkew( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3],
				NA($2)[4], NA($2)[5], NA($2)[6] );
	ArrayFree( $2 );
}
| SOLIDBEGIN STRING
{
	RiSolidBegin( $2 );
}
| SOLIDEND
{
	RiSolidEnd();
}
| SPHERE NUM NUM NUM NUM paramlist
{
	RiSphereV( $2, $3, $4, $5, CPS, CPT, CPA );
	FreeArgs();
}
| SPHERE real_num_array paramlist
{
	if (VerifyArrayLength( $2, 4, "Sphere" ))
		RiSphereV( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3], CPS, CPT, CPA );
	ArrayFree( $2 );
	FreeArgs();
}
| SUBDIVISIONMESH STRING num_array num_array paramlist
{
	int *nvertices = MakeIntArray($3);
	int nfaces = $3->nelems;
	int *verts = MakeIntArray($4);
	int tot = 0;
	for (int i = 0 ; i < nfaces; i++)
		tot += nvertices[i];
	if (tot != $4->nelems)
		ParseError("Wrong number of vertex elements in SubdivisionMesh. Ignoring.");
	else {
		// XXX ignore tags here
		const char *noTags[] = { NULL };
		int noTagsCount[2] = { 0, 0 };
		int noIntTags[] = { 0 };
		Float noFloatTags[] = { 0 };
		RiSubdivisionMeshV($2, nfaces, nvertices, verts, 0, noTags,
			noTagsCount, noIntTags, noFloatTags, CPS, CPT, CPA);
	}
	ArrayFree($3);
	ArrayFree($4);
	delete[] nvertices;
	delete[] verts;
	FreeArgs();
}
| SURFACE STRING paramlist
{
	RiSurfaceV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| SURFACEINTEGRATOR STRING paramlist
{
	RiSurfaceIntegratorV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| TEXTURECOORDINATES NUM NUM NUM NUM NUM NUM NUM NUM
{
	RiTextureCoordinates( $2, $3, $4, $5, $6, $7, $8, $9 );
}
| TEXTURECOORDINATES real_num_array
{
	if (VerifyArrayLength( $2, 8, "TextureCoordinates" ))
		RiTextureCoordinates( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3],
							  NA($2)[4], NA($2)[5], NA($2)[7], NA($2)[8] );
	ArrayFree( $2 );
}
| TONEMAP STRING paramlist
{
	RiToneMap($2, CPS, CPT, CPA);
	FreeArgs();
}
| TORUS NUM NUM NUM NUM NUM paramlist
{
	RiTorusV( $2, $3, $4, $5, $6, CPS, CPT, CPA );
	FreeArgs();
}
| TORUS real_num_array paramlist
{
	if (VerifyArrayLength( $2, 5, "Torus" ))
		RiTorusV( NA($2)[0], NA($2)[1], NA($2)[2], NA($2)[3], NA($2)[4],
			CPS, CPT, CPA );
    ArrayFree($2);
	FreeArgs();
}
| TRANSFORM real_num_array
{
	if (VerifyArrayLength( $2, 16, "Transform" ))
		RiTransform( (Float *) $2->array );
	ArrayFree( $2 );
}
| TRANSFORMBEGIN
{
	RiTransformBegin();
}
| TRANSFORMEND
{
	RiTransformEnd();
}
| TRANSLATE NUM NUM NUM
{
	RiTranslate( $2, $3, $4 );
}
| TRIMCURVE num_array num_array num_array num_array num_array num_array num_array num_array num_array
{
	int *ncurves, *order, *n;

	ncurves = MakeIntArray( $2 );
	order = MakeIntArray( $3 );
	n = MakeIntArray( $7 );

	RiTrimCurve( $2->nelems, ncurves, order, NA($4), NA($5), NA($6), n, NA($8), NA($9), NA($10) );
	delete [] ncurves;
	delete [] order;
	delete [] n;

	ArrayFree( $2 );
	ArrayFree( $3 );
	ArrayFree( $4 );
	ArrayFree( $5 );
	ArrayFree( $6 );
	ArrayFree( $7 );
	ArrayFree( $8 );
	ArrayFree( $9 );
	ArrayFree( $10 );
}
| VOLUMEINTEGRATOR STRING paramlist
{
	RiVolumeIntegratorV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| VOLUME STRING paramlist
{
	RiVolumeV( $2, CPS, CPT, CPA );
	FreeArgs();
}
| WORLDBEGIN
{
	RiWorldBegin();
}
| WORLDEND
{
	RiWorldEnd();
}
%%
