// bestcandidate.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "sampling.h"
#include "paramset.h"
// BestCandidateSampler Declarations
class BestCandidateSampler : public Sampler {
public:
	// BestCandidateSampler Method Declarations
	BestCandidateSampler(int xres, int yres, int xs, int ys,
		const Extent2D &crop, Filter *f);
	~BestCandidateSampler() {
		delete[] scrambles;
		delete[] pixelSamplesDone;
	}
	bool GetNextSample(Sample *sample);
private:
	// BestCandidateSampler Private Data
	int tableOffset;
	Float xTablePos, yTablePos;
	Float xTableWidth, yTableWidth;
	u_int *scrambles, *pixelSamplesDone;
	static const Float sampleTable[SAMPLE_TABLE_SIZE][5];
	Float sampleOffsets[3];
};
// BestCandidateSampler Method Definitions
BestCandidateSampler::BestCandidateSampler(int xres, int yres, int xSamp,
		int ySamp, const Extent2D &crop, Filter *f)
	: Sampler(xres, yres, xSamp, ySamp, crop, f) {
	xTablePos = xPixelStart;
	yTablePos = yPixelStart;
	xTableWidth = (Float)SQRT_SAMPLE_TABLE_SIZE / xPixelSamples;
	yTableWidth = (Float)SQRT_SAMPLE_TABLE_SIZE / yPixelSamples;
	tableOffset = 0;
	// Update sample shifts
	for (int i = 0; i < 3; ++i)
		sampleOffsets[i] = RandomFloat();
	scrambles = pixelSamplesDone = NULL;
}
#include "sampledata.cc"
bool BestCandidateSampler::GetNextSample(Sample *sample) {
again:
	// Return [[false]] if \use{BestCandidateSampler} is done
	if (yTablePos >= yPixelEnd)
		return false;
	// Compute raster sample from table
	#define WRAP(x) ((x) > 1 ? ((x)-1) : (x))
	sample->imageX = xTablePos + xTableWidth * sampleTable[tableOffset][0];
	sample->imageY = yTablePos + yTableWidth * sampleTable[tableOffset][1];
	sample->time  = WRAP(sampleOffsets[0] + sampleTable[tableOffset][2]);
	sample->lensX = WRAP(sampleOffsets[1] + sampleTable[tableOffset][3]);
	sample->lensY = WRAP(sampleOffsets[2] + sampleTable[tableOffset][4]);
	// Advance to next sample table position
	if (++tableOffset == SAMPLE_TABLE_SIZE) {
		// Update sample shifts
		for (int i = 0; i < 3; ++i)
			sampleOffsets[i] = RandomFloat();
		tableOffset = 0;
		xTablePos += xTableWidth;
		if (xTablePos >= xPixelEnd) {
			xTablePos = xPixelStart;
			yTablePos += yTableWidth;
		}
	}
	// Check sample against crop window, goto [[again]] if outside
	if (sample->imageX <  xPixelStart ||
	    sample->imageX >= xPixelEnd   ||
	    sample->imageY <  yPixelStart ||
	    sample->imageY >= yPixelEnd)
		goto again;
	sample->imageX -= .5f;
	sample->imageY -= .5f;
	// Compute low-discrepancy integrator samples
	if (!scrambles) {
		// Allocate and initialize arrays for per-pixel records
		int nPix = (xPixelEnd - xPixelStart) * (yPixelEnd - yPixelStart);
		pixelSamplesDone = new u_int[nPix];
		for (int i = 0; i < nPix; ++i)
			pixelSamplesDone[i] = 0;
		int nScrambles = nPix * (sample->n1D.size() + sample->n2D.size());
		scrambles = new u_int[nScrambles];
		for (int i = 0; i < nScrambles; ++i)
			scrambles[i] = RandomInt();
	}
	// Get pointer to scramble values for current pixel
	int xPixel = Round2Int(sample->imageX);
	int yPixel = Round2Int(sample->imageY);
	u_int pixelNum = (xPixel - xPixelStart) +
		(yPixel - yPixelStart) * (xPixelEnd - xPixelStart);
	u_int *scramble = &scrambles[pixelNum *
		(sample->n1D.size() + sample->n2D.size())];
	int pixelSamples = pixelSamplesDone[pixelNum]++;
	// Compute low-discrepancy 1D samples
	for (u_int i = 0; i < sample->n1D.size(); ++i) {
		u_int nextPow2 = RoundUpPow2(sample->n1D[i]);
		for (u_int j = 0; j < sample->n1D[i]; ++j) {
			int sampNum = pixelSamples * nextPow2 + j + 1;
			sample->oneD[i][j] = VanDerCorput(sampNum, *scramble);
		}
		++scramble;
	}
	// Compute low-discrepancy 2D samples
	for (u_int i = 0; i < sample->n2D.size(); ++i) {
		u_int nextPow2 = RoundUpPow2(sample->n2D[i]);
		int sampStart = pixelSamples * nextPow2;
		for (u_int j = 0; j < sample->n2D[i]; ++j) {
			int sampNum = sampStart + j + 1;
			sample->twoD[i][2*j]   = VanDerCorput(sampNum, *scramble);
			sample->twoD[i][2*j+1] = Sobol2(sampNum, *scramble);
		}
		++scramble;
	}
	return true;
}
extern "C" Sampler *CreateSampler(const ParamSet &params,
		int xres, int yres, int xsamp, int ysamp,
		const Extent2D &crop, Filter *filter) {
	return new BestCandidateSampler(xres, yres, xsamp, ysamp,
		crop, filter);
}
