// lowdiscrepancy.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "sampling.h"
#include "paramset.h"
// LDSampler Declarations
class LDSampler : public Sampler {
public:
	// LDSampler Method Declarations
	LDSampler(int xResolution, int yResolution, int xSamples,
		int ySamples, const Extent2D &crop, Filter *f, bool useFolded);
	~LDSampler() {
		delete[] scrambles;
		delete[] pixelSamplesDone;
	}
	bool GetNextSample(Sample *sample);
private:
	// LDSampler Private Data
	int scale, nSamples, curSample;
	Float invNSamples;
	bool useFolded;
	u_int *scrambles, *pixelSamplesDone;
};
// LDSampler Method Definitions
LDSampler::LDSampler(int xres, int yres, int xpix,
		int ypix, const Extent2D &crop, Filter *f, bool uf)
	: Sampler(xres, yres, xpix, ypix, crop, f) {
	scale = max(xPixelEnd - xPixelStart, yPixelEnd - yPixelStart);
	nSamples = xPixelSamples * yPixelSamples * scale * scale;
	invNSamples = 1.f / nSamples;
	curSample = 1;
	useFolded = uf;
	scrambles = pixelSamplesDone = NULL;
}
bool LDSampler::GetNextSample(Sample *sample) {
tryAgain:
	if (curSample > nSamples) return false;
	// Compute Hammersley $(x,y)$ image sample location
	Float x = curSample * invNSamples;
	Float y = useFolded ? FoldedRadicalInverse(curSample, 2) :
		RadicalInverse(curSample, 2);
	sample->imageX = xPixelStart + x * scale;
	sample->imageY = yPixelStart + y * scale;
	if (sample->imageX >= xPixelEnd || sample->imageY >= yPixelEnd) {
		++curSample;
		goto tryAgain;
	}
	sample->imageX -= .5f;
	sample->imageY -= .5f;
	// Compute remaining dimensions of Hammersley sample
	if (useFolded) {
		sample->time  = FoldedRadicalInverse(curSample, 3);
		sample->lensX = FoldedRadicalInverse(curSample, 5);
		sample->lensY = FoldedRadicalInverse(curSample, 7);
	}
	else {
		sample->time  = RadicalInverse(curSample, 3);
		sample->lensX = RadicalInverse(curSample, 5);
		sample->lensY = RadicalInverse(curSample, 7);
	}
	// Compute low-discrepancy integrator samples
	if (!scrambles) {
		// Allocate and initialize arrays for per-pixel records
		int nPix = (xPixelEnd - xPixelStart) * (yPixelEnd - yPixelStart);
		pixelSamplesDone = new u_int[nPix];
		for (int i = 0; i < nPix; ++i)
			pixelSamplesDone[i] = 0;
		int nScrambles = nPix * (sample->n1D.size() + sample->n2D.size());
		scrambles = new u_int[nScrambles];
		for (int i = 0; i < nScrambles; ++i)
			scrambles[i] = RandomInt();
	}
	// Get pointer to scramble values for current pixel
	int xPixel = Round2Int(sample->imageX);
	int yPixel = Round2Int(sample->imageY);
	u_int pixelNum = (xPixel - xPixelStart) +
		(yPixel - yPixelStart) * (xPixelEnd - xPixelStart);
	u_int *scramble = &scrambles[pixelNum *
		(sample->n1D.size() + sample->n2D.size())];
	int pixelSamples = pixelSamplesDone[pixelNum]++;
	// Compute low-discrepancy 1D samples
	for (u_int i = 0; i < sample->n1D.size(); ++i) {
		u_int nextPow2 = RoundUpPow2(sample->n1D[i]);
		for (u_int j = 0; j < sample->n1D[i]; ++j) {
			int sampNum = pixelSamples * nextPow2 + j + 1;
			sample->oneD[i][j] = VanDerCorput(sampNum, *scramble);
		}
		++scramble;
	}
	// Compute low-discrepancy 2D samples
	for (u_int i = 0; i < sample->n2D.size(); ++i) {
		u_int nextPow2 = RoundUpPow2(sample->n2D[i]);
		int sampStart = pixelSamples * nextPow2;
		for (u_int j = 0; j < sample->n2D[i]; ++j) {
			int sampNum = sampStart + j + 1;
			sample->twoD[i][2*j]   = VanDerCorput(sampNum, *scramble);
			sample->twoD[i][2*j+1] = Sobol2(sampNum, *scramble);
		}
		++scramble;
	}
	++curSample;
	return true;
}
extern "C" Sampler *CreateSampler(const ParamSet &params,
		int xres, int yres, int xsamp, int ysamp,
		const Extent2D &crop, Filter *filter) {
	bool useFolded = params.FindOneInt("zaremba", 0) != 0;
	return new LDSampler(xres, yres, xsamp, ysamp,
		crop, filter, useFolded);
}
