// stratified.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "sampling.h"
#include "paramset.h"
// StratifiedSampler Declarations
class StratifiedSampler : public Sampler {
public:
	// StratifiedSampler Method Declarations
	StratifiedSampler(int xResolution, int yResolution, int xSamples,
		int ySamples, bool jitter, const Extent2D &crop,
		Filter *f);
	~StratifiedSampler() {
		delete[] imageSamples;
		delete[] lensSamples;
		delete[] timeSamples;
	}
	bool GetNextSample(Sample *sample);
private:
	// StratifiedSampler Private Data
	bool jitterSamples;
	int xPos, yPos;
	int samplePos;
	Float *imageSamples, *lensSamples, *timeSamples;
};
// StratifiedSampler Method Definitions
StratifiedSampler::StratifiedSampler(int xres, int yres,
		int xpix, int ypix, bool jitter,
		const Extent2D &crop, Filter *f)
	: Sampler(xres, yres, xpix, ypix, crop, f) {
	jitterSamples = jitter;
	xPos = xPixelStart;
	yPos = yPixelStart;
	// Allocate storage for a pixel's worth of stratified samples
	imageSamples = new Float[xPixelSamples * yPixelSamples * 2];
	lensSamples = new Float[xPixelSamples * yPixelSamples * 2];
	timeSamples = new Float[xPixelSamples * yPixelSamples];
	// Generate stratified camera samples for ([[xPos]],[[yPos]])
	StratifiedSample2D(imageSamples, xPixelSamples, yPixelSamples);
	StratifiedSample2D(lensSamples, xPixelSamples, yPixelSamples);
	StratifiedSample1D(timeSamples, xPixelSamples*yPixelSamples);
	// Shift stratified image samples to pixel coordinates
	for (int y = 0; y < yPixelSamples; ++y) {
		for (int x = 0; x < xPixelSamples; ++x) {
			int o = (x + y * xPixelSamples) * 2;
			imageSamples[o]   += xPos - .5f;
			imageSamples[o+1] += yPos - .5f;
		}
	}
	// Decorrelate sample dimensions
	Shuffle(lensSamples, xPixelSamples*yPixelSamples, 2);
	Shuffle(timeSamples, xPixelSamples*yPixelSamples, 1);
	samplePos = 0;
}
bool StratifiedSampler::GetNextSample(Sample *sample) {
	// Compute new set of samples if needed for next pixel
	if (samplePos == xPixelSamples * yPixelSamples) {
		// Advance to next pixel for stratified sampling
		if (++xPos == xPixelEnd) {
			xPos = xPixelStart;
			++yPos;
		}
		if (yPos == yPixelEnd)
			return false;
		// Generate stratified camera samples for ([[xPos]],[[yPos]])
		StratifiedSample2D(imageSamples, xPixelSamples, yPixelSamples);
		StratifiedSample2D(lensSamples, xPixelSamples, yPixelSamples);
		StratifiedSample1D(timeSamples, xPixelSamples*yPixelSamples);
		// Shift stratified image samples to pixel coordinates
		for (int y = 0; y < yPixelSamples; ++y) {
			for (int x = 0; x < xPixelSamples; ++x) {
				int o = (x + y * xPixelSamples) * 2;
				imageSamples[o]   += xPos - .5f;
				imageSamples[o+1] += yPos - .5f;
			}
		}
		// Decorrelate sample dimensions
		Shuffle(lensSamples, xPixelSamples*yPixelSamples, 2);
		Shuffle(timeSamples, xPixelSamples*yPixelSamples, 1);
		samplePos = 0;
	}
	// Return next \use{StratifiedSampler} sample point
	sample->imageX = imageSamples[2*samplePos];
	sample->imageY = imageSamples[2*samplePos+1];
	sample->lensX = lensSamples[2*samplePos];
	sample->lensY = lensSamples[2*samplePos+1];
	sample->time = timeSamples[samplePos];
	// Generate stratified lens and BSDF samples
	for (u_int i = 0; i < sample->n1D.size(); ++i)
		LatinHypercube(sample->oneD[i], sample->n1D[i], 1);
	for (u_int i = 0; i < sample->n2D.size(); ++i)
		LatinHypercube(sample->twoD[i], sample->n2D[i], 2);
	++samplePos;
	return true;
}
extern "C" Sampler *CreateSampler(const ParamSet &params,
		int xres, int yres, int xsamp, int ysamp,
		const Extent2D &crop, Filter *filter) {
	bool jitter = params.FindOneInt("jitter", 0) != 0;
	return new StratifiedSampler(xres, yres, xsamp, ysamp,
		jitter, crop, filter);
}
