// algebraic.cc
// Ray tracing algebraic surfaces

#include "lrt.h"
#include "shapes.h"

// This file includes the implementation of the Polynomial class that you
// should use for root-finding
#include "polynomial_mod.h"

// Term describes a term of an algebraic equation.  For example, the unit
// sphere x^2+y^2+z^2-1 = 0 is described by four terms, 
// Term(1,2,0,0), Term(1,0,2,0), Term(1,0,0,2), and Term(-1,0,0,0)

struct Term {
	Term(Float s = 0, int x=0, int y=0, int z=0) {
		scale = s;
		xPower = x;
		yPower = y;
		zPower = z;
	}
	int GetTotal() const {
		return xPower+yPower+zPower;
	}
	Float scale;
	int xPower, yPower, zPower;
};

struct PreCalc {
	bool Equal(const PreCalc &pref) const {
		return (nx0==pref.nx0 && nx1==pref.nx1 && ny0==pref.ny0 &&
						ny1==pref.ny1 && nz0==pref.nz0 && nz1==pref.nz1);
	}
	Float scale;
	int nx0, nx1;
	int ny0, ny1;
	int nz0, nz1;
};

ostream &operator<<(ostream &os, const PreCalc &p) {
	if(p.nx0)
		os << "x0^" << p.nx0 << " * " ;
	if(p.nx1)
		os << "x1^" << p.nx1 << " * " ;
	if(p.ny0)
		os << "y0^" << p.ny0 << " * " ;
	if(p.ny1)
		os << "y1^" << p.ny1 << " * " ;
	if(p.nz0)
		os << "z0^" << p.nz0 << " * " ;
	if(p.nz1)
		os << "z1^" << p.nz1 << " * " ;
	os << p.scale;
	return os;
}

class AlgebraicSurf: public Shape {
public:
	AlgebraicSurf(const Transform &o2w, const BBox &ob, Term *terms, int nTerms);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
	
private:
	Float evalderiv(const int &i, const Point &p) const;
	bool CommonIntersect(const Ray &r, Ray *ray, Float *thitp) const;
	BBox objBound;
	Polynomial poly;
	Term *t;
	Term *dt[3];
	vector<PreCalc> *coef;
	int nt, ndt[3];
	int maxdeg;
	int maxxdeg, maxydeg, maxzdeg;
	Float *powx0, *powx1,*powy0,*powy1,*powz0,*powz1;
	int *nCr;
};

AlgebraicSurf::AlgebraicSurf(const Transform &o2w, const BBox &ob, 
			     Term *terms, int nTerms)
	: Shape(o2w) {
	// An object-space bounding box is passed into the constructor, so you
	// don't have to worry about computing bounding boxes of
	// (potentially-infinitely-large!) algebraic surfaces.  Therefore, you
	// don't need to worry about intersections outside the given bounding
	// box.
	objBound = ob;
	
	maxdeg=-1; maxxdeg=-1; maxydeg=-1; maxzdeg=-1;
	// Find the polynomial degree
	for(int i=0;i<nTerms;i++) {
		int degcur=terms[i].xPower+terms[i].yPower+terms[i].zPower;
		if(degcur>maxdeg)
			maxdeg=degcur;
		if(terms[i].xPower>maxxdeg)
			maxxdeg=terms[i].xPower;
		if(terms[i].yPower>maxydeg)
			maxydeg=terms[i].yPower;
		if(terms[i].zPower>maxzdeg)
			maxzdeg=terms[i].zPower;
	}
	poly.SetDegree(maxdeg);

	// Float or polyFloat here ?
	// Will contain all the powers of x,y,z
	powx0=new Float[maxxdeg+1];
	powx1=new Float[maxxdeg+1];
	powy0=new Float[maxydeg+1];
	powy1=new Float[maxydeg+1];
	powz0=new Float[maxzdeg+1];
	powz1=new Float[maxzdeg+1];
	powx0[0]=1.0f; powx1[0]=1.0f;
	powy0[0]=1.0f; powy1[0]=1.0f;
	powz0[0]=1.0f; powz1[0]=1.0f;

	// Allocate data
	t=new Term[nTerms];
	dt[0]=new Term[nTerms];  // for the derivatives
	dt[1]=new Term[nTerms];
	dt[2]=new Term[nTerms];

	for(int i=0;i<3;i++)
		ndt[i]=0;
	// Getting the derivatives in terms of Terms
	for(int j=0;j<nTerms;j++) {
		memcpy(&t[j],&terms[j],sizeof(Term));
		for(int k=0;k<3;k++) {
			const int n=ndt[k];
			memcpy(&dt[k][n],&terms[j],sizeof(Term));
			if(k==0 && terms[j].xPower) {
				dt[k][n].scale*=dt[k][n].xPower;
				dt[k][n].xPower--;
				ndt[k]++;
			}
			if(k==1 && terms[j].yPower) {
				dt[k][n].scale*=dt[k][n].yPower;
				dt[k][n].yPower--;
				ndt[k]++;
			}
			if(k==2 && terms[j].zPower) {
				dt[k][n].scale*=dt[k][n].zPower;
				dt[k][n].zPower--;
				ndt[k]++;
			}
		}
	}
	nt=nTerms;

	// Precalculate binomial coefficients (Pascal Triangle)
#define nCrInd(n,p) ((n)*sznCr+(p))
	int sznCr=maxdeg+1;
	nCr=new int[sznCr*sznCr];

	for(int n=0;n<=maxdeg;n++)
		nCr[nCrInd(n,0)]=1;
	for(int p=1;p<=maxdeg;p++)
		nCr[nCrInd(0,p)]=0;
	for(int n=1;n<=maxdeg;n++) {
		for(int p=1;p<=maxdeg;p++) {
			nCr[nCrInd(n,p)]=nCr[nCrInd(n-1,p)]+nCr[nCrInd(n-1,p-1)];
			//			cout << nCr[nCrInd(n,p)] << " ";
		}
		//		cout << "\n";
	}

	// An array of nTerms x maxdeg that contains which component of the ray
	// to multiply together to get each coefficient in the 
	coef=new (vector<PreCalc>)[maxdeg+1];
	for(int i=0;i<maxdeg+1;i++)
		coef[i].reserve(10);
	

	for(int i=0;i<nTerms;i++) {
		const int totalt=terms[i].GetTotal();
		//		printf("\nYo man %d %d, %d %d %d\n", i, totalt, terms[i].xPower,terms[i].yPower, terms[i].zPower);
		for(int t=0;t<=totalt;t++)
			for(int nx=0;nx<=terms[i].xPower;nx++)
				for(int ny=0;ny<=terms[i].yPower;ny++) {
					int nz=t-nx-ny;
					//					printf("i=%d, t=%d, %d %d %d \n",i,t,nx,ny,nz);
					if(nz>=0 && nz<=terms[i].zPower) {
						PreCalc pc;
						pc.scale=terms[i].scale*nCr[nCrInd(terms[i].xPower,nx)]*nCr[nCrInd(terms[i].yPower,ny)]*nCr[nCrInd(terms[i].zPower,nz)];
						pc.nx0=terms[i].xPower-nx;
						pc.ny0=terms[i].yPower-ny;
						pc.nz0=terms[i].zPower-nz;
						pc.nx1=nx;
						pc.ny1=ny;
						pc.nz1=nz;
						coef[t].push_back(pc);
						cout << " (t^" << t << ") " << pc << "\n";
					}
				}
	}
}

BBox AlgebraicSurf::ObjectBound() const {
	return objBound;
}


// Get the derivative of the surface equation along each axis at the
// intersection point
inline Float AlgebraicSurf::evalderiv(const int &i,const Point &p) const {
	Float ret=0;
	Float tmp;

	for(int j=0;j<ndt[i];j++) {
		tmp=dt[i][j].scale;
		for(int k=0;k<dt[i][j].xPower;k++)
			tmp*=p.x;
		for(int k=0;k<dt[i][j].yPower;k++)
			tmp*=p.y;
		for(int k=0;k<dt[i][j].zPower;k++)
			tmp*=p.z;
		ret+=tmp;
	}
	return ret;
}

// We do all the intesection work here
inline bool AlgebraicSurf::CommonIntersect(const Ray &r, Ray *ray, Float *thit) const {
	// Transform [[Ray]] to object space
	WorldToObject(r, ray);

	if(debugflag)
		cout << ray->D << " \\ " <<ray->O<< "\n";
	// Precalc the powers of x0, x1, etc.
	for(int i=1;i<=maxxdeg;i++) {
		powx0[i]=powx0[i-1]*ray->O.x;
		powx1[i]=powx1[i-1]*ray->D.x;
		//		if(debugflag)
		//			cout << "i="<<i<< " x0:" << powx0[i] << " x1:" <<powx1[i]<<"\n";
	}
	if(debugflag)	cout<< "\n";
	for(int i=1;i<=maxydeg;i++) {
		powy0[i]=powy0[i-1]*ray->O.y;
		powy1[i]=powy1[i-1]*ray->D.y;
		//		if(debugflag)
		//			cout << "i="<<i<< " y0:" << powy0[i] << " y1:" <<powy1[i]<<"\n";
	}
	if(debugflag)	cout<< "\n";
	for(int i=1;i<=maxzdeg;i++) {
		powz0[i]=powz0[i-1]*ray->O.z;
		powz1[i]=powz1[i-1]*ray->D.z;
		//		if(debugflag)
		//			cout << "i="<<i<< " z0:" << powz0[i] << " z1:" <<powz1[i]<<"\n";
	}
	//	if(debugflag)
	//		cout<<"powy0[0]="<<powy0[0]<<"\n";

	/*		
	// One way to do it: with polynomials...
	Polynomial total(0);
	total[0]=0;
	Polynomial accum;
	Polynomial tmp(1);
	Polynomial tmp2;
	Polynomial tmpterm;
	for(int i=0;i<nt;i++) {
		accum.SetDegree(0);
		accum[0]=t[i].scale;

		if(t[i].xPower) {
			tmp[0]=ray->O.x;
			tmp[1]=ray->D.x;
			tmp2=tmp;
			for(int j=1;j<t[i].xPower;j++)
				tmp2=tmp*tmp2;
			accum=accum*tmp2;
		}
		if(t[i].yPower) {
			tmp[0]=ray->O.y;
			tmp[1]=ray->D.y;
			tmp2=tmp;
			for(int j=1;j<t[i].yPower;j++)
				tmp2=tmp*tmp2;
			accum=accum*tmp2;
		}
		if(t[i].zPower) {
			tmp[0]=ray->O.z;
			tmp[1]=ray->D.z;
			tmp2=tmp;
			for(int j=1;j<t[i].zPower;j++)
				tmp2=tmp*tmp2;
			accum=accum*tmp2;
		}
		
		total=total+accum;
	}  // We have the polynomial here
	*/

	// Another way to do : by expanding the polynomial first, and then
	// perform all the computation for each ray
	Polynomial total(maxdeg);

	// Compute the polynomial in t
	vector<PreCalc>::iterator iter;
	for(int i=0;i<=maxdeg;i++) {
		total[i]=0;
		if(debugflag)
			cout << " \n Calcul de t=" << i << "\n";
		for(iter=coef[i].begin();iter!=coef[i].end();++iter) {
			total[i]+=(*iter).scale
				*powx0[(*iter).nx0]*powx1[(*iter).nx1]
				*powy0[(*iter).ny0]*powy1[(*iter).ny1]
				*powz0[(*iter).nz0]*powz1[(*iter).nz1];
			/*
			if(debugflag) {
				cout << "scale " << (*iter).scale << ", res=" << (*iter).scale
					*powx0[(*iter).nx0]*powx1[(*iter).nx1]
					*powy0[(*iter).ny0]*powy1[(*iter).ny1]
					*powz0[(*iter).nz0]*powz1[(*iter).nz1] << " // x0:"
						 << (*iter).nx0 << " x1:" << (*iter).nx1<< " -- y0:"
						 << (*iter).ny0 << " y1:" << (*iter).ny1<< " -- z0:"
						 << (*iter).nz0 << " z1:" << (*iter).nz1 << "\n";
				cout << "scale " << (*iter).scale << ", res=" << (*iter).scale
					*powx0[(*iter).nx0]*powx1[(*iter).nx1]
					*powy0[(*iter).ny0]*powy1[(*iter).ny1]
					*powz0[(*iter).nz0]*powz1[(*iter).nz1] << " // x0:"
						 << powx0[(*iter).nx0] << " x1:" << powx1[(*iter).nx1]<< " -- y0:"
						 << powy0[(*iter).ny0] << " y1:" << powy1[(*iter).ny1]<< " -- z0:"
						 << powz0[(*iter).nz0] << " z1:" << powz1[(*iter).nz1] <<	"\n\n";
			}
			*/
		}
	}

	//	if(debugflag)
	//			cout << total[0] << ", " << total[1] << ", " << total[2] << ", "<< total[3] << ", " << total[4] << " => \n";

	bool hit = false;
	int nroots;
	
	// Get the roots
	double roots[maxdeg];
	switch(maxdeg) {
		// Not precise enough or so it seems...
		//	case 4:
		//    total.RootsDegree4(nroots,roots);
		//    break;
	case 2:
		total.RootsDegree2(nroots,roots);
		break;
	case 3:
		total.RootsDegree3(nroots,roots);
		break;
	case 1:
		nroots=1;
		roots[0]=-total[0]/total[1];
		break;
	default:
    total.GetRootsOnInterval (ray->mint, ray->maxt, nroots, roots);
		break;
	}

	// Select the mini root in maxt, mint and in the Bounding box
	Float tmax=ray->maxt;
	//	ray->mint*=10;      // HACK to get rid of artefacts...
	Point Phit;
	for(int i=0;i<nroots;i++) {
		//		if(debugflag)
		//			cout << roots[i] << ",";
		if(roots[i]>=ray->mint && roots[i]<=tmax) {
			Phit=(*ray)(roots[i]);
			if(objBound.Inside(Phit)) {
				tmax=roots[i];
				hit=true;
			}
		}
	}
	//	if(debugflag)
	//		cout << "\n";

	*thit=tmax;
	return hit;
}

bool AlgebraicSurf::Intersect(const Ray &r, Float *thit, DifferentialGeometry *dg) const {
	Ray ray;
	// Intersection ?
	const bool hit=CommonIntersect(r,&ray,thit);

	// Initialize Differential geometry if there was an intersection
	if (hit) {
		Point Phit=ray(*thit);
		
		// Normal direction
		Normal Nhit;
		Nhit.x=evalderiv(0,Phit);
		Nhit.y=evalderiv(1,Phit);
		Nhit.z=evalderiv(2,Phit);

		//		if(debugflag)
		//			cout << " Normal " << Nhit.x << ", " << Nhit.y << ", " << Nhit.z << "\n";
		
		// Because the algebraic surface may not have a parameterization
		// (and it's not in general possible to automatically compute one,
		// just use the CoordinateSystem() utility method to compute dPdu
		// and dPdv vectors as two vectors orthogonal to the normal you
		// computed above.
		Vector dPdu, dPdv;
		CoordinateSystem((Vector)Nhit, &dPdu, &dPdv);
		
		// Don't worry about computing these values for the algebraic
		// surface, either.
		Vector dNdu(0,0,0), dNdv(0,0,0);
		Float u = 0, v = 0;
		
		*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
															 ObjectToWorld(dPdv), ObjectToWorld(dNdu),
															 ObjectToWorld(dNdv), u, v, this);
	}
	
	// Return true if there was an intersection and false otherwise
	return hit;
}

bool AlgebraicSurf::IntersectP(const Ray &r) const {
	Ray ray;
	Float thit;
	return CommonIntersect(r, &ray,&thit);
}

Float AlgebraicSurf::Area() const {
	// It's not necessary to implement this method for this assignment.
	return 0.;
}

extern "C" Reference<Shape>CreateShape(const Transform &obj2world, 
																			 const ParamSet &params) {
	// Extract information about the terms of the equation specified in the
	// input file from the ParamSet class.
	int nTermItems, type;
	const Float *termData = params.FindFloat("terms", &type, &nTermItems);
	assert(termData != NULL);
	assert((nTermItems % 4) == 0);
	
	int nTerms = nTermItems / 4;
	Term *terms = new Term[nTerms];
	for (int i = 0; i < nTerms; ++i) {
		terms[i] = Term(termData[0], (int)termData[1], (int)termData[2], 
										(int)termData[3]);
		termData += 4;
	}
	
	BBox bbox(params.FindOnePoint("p0", Point(0,0,0)),
						params.FindOnePoint("p1", Point(1,1,1)));
	
	return new AlgebraicSurf(obj2world, bbox, terms, nTerms);
}

