// algebraic.cc
// Ray tracing algebraic surfaces

#include "lrt.h"
#include "shapes.h"

// This file includes the implementation of the Polynomial class that you
// should use for root-finding
#include "polynomial.h"

// Term describes a term of an algebraic equation.  For example, the unit
// sphere x^2+y^2+z^2-1 = 0 is described by four terms, 
// Term(1,2,0,0), Term(1,0,2,0), Term(1,0,0,2), and Term(-1,0,0,0)

struct Term {
	Term(Float s = 0, int x=0, int y=0, int z=0) {
		scale = s;
		xPower = x;
		yPower = y;
		zPower = z;
	}
	Float scale;
	int xPower, yPower, zPower;
};

class AlgebraicSurf: public Shape {
public:
	AlgebraicSurf(const Transform &o2w, const BBox &ob, Term *terms, int nTerms);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
	
private:
	Float evalderiv(const int &i, const Point &p) const;
	BBox objBound;
	Polynomial poly;
	int *coef;
	Term t[15];
	Term dt[3][15];
	int nt, ndt[3];
	int maxdeg;
};

AlgebraicSurf::AlgebraicSurf(const Transform &o2w, const BBox &ob, 
			     Term *terms, int nTerms)
	: Shape(o2w) {
	// An object-space bounding box is passed into the constructor, so you
	// don't have to worry about computing bounding boxes of
	// (potentially-infinitely-large!) algebraic surfaces.  Therefore, you
	// don't need to worry about intersections outside the given bounding
	// box.
	objBound = ob;
	
	maxdeg=-1;
	for(int i=0;i<nTerms;i++) {
		int degcur=terms[i].xPower+terms[i].yPower+terms[i].zPower;
		if(degcur>maxdeg)
			maxdeg=degcur;
	}
	poly.SetDegree(maxdeg);

	for(int i=0;i<3;i++)
		ndt[i]=0;
	// Getting the derivatives in terms of Terms
	for(int j=0;j<nTerms;j++) {
		memcpy(&t[j],&terms[j],sizeof(Term));
		for(int k=0;k<3;k++) {
			const int n=ndt[k];
			memcpy(&dt[k][n],&terms[j],sizeof(Term));
			if(k==0 && terms[j].xPower) {
				dt[k][n].scale*=dt[k][n].xPower;
				dt[k][n].xPower--;
				ndt[k]++;
			}
			if(k==1 && terms[j].yPower) {
				dt[k][n].scale*=dt[k][n].yPower;
				dt[k][n].yPower--;
				ndt[k]++;
			}
			if(k==2 && terms[j].zPower) {
				dt[k][n].scale*=dt[k][n].zPower;
				dt[k][n].zPower--;
				ndt[k]++;
			}
		}
	}
	nt=nTerms;

	// An array of nTerms x maxdeg that contains which component of the ray
	// to multiply together to get each coefficient in the 
	//	coef=(int)malloc(sizeof(int)*maxdeg*nTerms);

}

BBox AlgebraicSurf::ObjectBound() const {
	return objBound;
}


// Get the derivative of the surface equation along each axis at the
// intersection point
Float AlgebraicSurf::evalderiv(const int &i,const Point &p) const {
	Float ret=0;
	Float tmp;
	for(int j=0;j<ndt[i];j++) {
		tmp=dt[i][j].scale;
		for(int k=0;k<dt[i][j].xPower;k++)
			tmp*=p.x;
		for(int k=0;k<dt[i][j].yPower;k++)
			tmp*=p.y;
		for(int k=0;k<dt[i][j].zPower;k++)
			tmp*=p.z;
		ret+=tmp;
	}
	return ret;
}

bool AlgebraicSurf::Intersect(const Ray &r, Float *thit,
															DifferentialGeometry *dg) const {
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);

	Polynomial total(0);
	total[0]=0;
	Polynomial accum;
	Polynomial tmp(1);
	Polynomial tmp2;
	Polynomial tmpterm;
	if(debugflag)
		cout << ray.D << " \\ " <<ray.O<< "\n";
	for(int i=0;i<nt;i++) {
		accum.SetDegree(0);
		accum[0]=t[i].scale;

		if(t[i].xPower) {
			tmp[0]=ray.O.x;
			tmp[1]=ray.D.x;
			tmp2=tmp;
			for(int j=1;j<t[i].xPower;j++)
				tmp2=tmp*tmp2;
			accum=accum*tmp2;
		}
		if(t[i].yPower) {
			tmp[0]=ray.O.y;
			tmp[1]=ray.D.y;
			tmp2=tmp;
			for(int j=1;j<t[i].yPower;j++)
				tmp2=tmp*tmp2;
			accum=accum*tmp2;
		}
		if(t[i].zPower) {
			tmp[0]=ray.O.z;
			tmp[1]=ray.D.z;
			tmp2=tmp;
			for(int j=1;j<t[i].zPower;j++)
				tmp2=tmp*tmp2;
			accum=accum*tmp2;
		}
		
		total=total+accum;
	}  // We have the polynomial here

	if(debugflag)
		cout << total[0] << ", " << total[1] << ", " << total[2] << ", " <<
		total[3] << ", " << total[4] << " => ";

	bool hit = false;
	int nroots;
	
	Float roots[maxdeg];
	switch(maxdeg) {
	case 4:
		total.RootsDegree4(nroots,roots);
		break;
	case 2:
		total.RootsDegree2(nroots,roots);
		break;
	case 3:
		total.RootsDegree3(nroots,roots);
		break;
	case 1:
		nroots=1;
		roots[0]=-total[0]/total[1];
		break;
	default:
    total.GetRootsOnInterval (ray.mint, ray.maxt, nroots, roots);
		break;
	}
	Float tmax=ray.maxt;
	Point Phit;
	for(int i=0;i<nroots;i++) {
		if(debugflag)
			cout << roots[i] << ",";
		if(roots[i]>=ray.mint && roots[i]<=tmax) {
			Phit=ray(roots[i]);
			if(objBound.Inside(Phit)) {
				tmax=roots[i];
				hit=true;
			}
		}
	}
	if(debugflag)
		cout << "\n";

	// TODO: Compute t values for ray intersection, find smallest one that
	// is inside ray's mint/maxt range (if any).  Initialize *thit with
	// that value.  Be careful to ignore any hits outside of the given
	// bounding box.
	
	// Initialize Differential geometry if there was an intersection
	if (hit) {
		*thit=tmax;
		Phit=ray(*thit);
		
		Normal Nhit;  // TODO: Compute the normal direction at the intersection
		Nhit.x=evalderiv(0,Phit);
		Nhit.y=evalderiv(1,Phit);
		Nhit.z=evalderiv(2,Phit);

		if(debugflag)
			cout << " Normal " << Nhit.x << ", " << Nhit.y << ", " << Nhit.z << "\n";
		
		// Because the algebraic surface may not have a parameterization
		// (and it's not in general possible to automatically compute one,
		// just use the CoordinateSystem() utility method to compute dPdu
		// and dPdv vectors as two vectors orthogonal to the normal you
		// computed above.
		Vector dPdu, dPdv;
		CoordinateSystem((Vector)Nhit, &dPdu, &dPdv);
		
		// Don't worry about computing these values for the algebraic
		// surface, either.
		Vector dNdu(0,0,0), dNdv(0,0,0);
		Float u = 0, v = 0;
		
		*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
															 ObjectToWorld(dPdv), ObjectToWorld(dNdu),
															 ObjectToWorld(dNdv), u, v, this);
	}
	
	// Return true if there was an intersection and false otherwise
	return hit;
}

bool AlgebraicSurf::IntersectP(const Ray &r) const {
	// Default implementation: just call the regular intersect method and
	// unnecessarily do the extra work to initialize the
	// DifferentialGeometry
	DifferentialGeometry dg;
	Float thit;
	return Intersect(r, &thit, &dg);
}

Float AlgebraicSurf::Area() const {
	// It's not necessary to implement this method for this assignment.
	return 0.;
}

extern "C" Reference<Shape>CreateShape(const Transform &obj2world, 
																			 const ParamSet &params) {
	// Extract information about the terms of the equation specified in the
	// input file from the ParamSet class.
	int nTermItems, type;
	const Float *termData = params.FindFloat("terms", &type, &nTermItems);
	assert(termData != NULL);
	assert((nTermItems % 4) == 0);
	
	int nTerms = nTermItems / 4;
	Term *terms = new Term[nTerms];
	for (int i = 0; i < nTerms; ++i) {
		terms[i] = Term(termData[0], (int)termData[1], (int)termData[2], 
										(int)termData[3]);
		termData += 4;
	}
	
	BBox bbox(params.FindOnePoint("p0", Point(0,0,0)),
						params.FindOnePoint("p1", Point(1,1,1)));
	
	return new AlgebraicSurf(obj2world, bbox, terms, nTerms);
}

