// cone.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
// Cone Declarations
class Cone: public Shape {
public:
	// Cone Interface
	Cone(const Transform &o2w, Float height, Float rad, Float tm );
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
protected:
	// Cone Data
	Float radius, height, phiMax;
};
// Cone Methods
Cone::Cone(const Transform &o2w, Float ht, Float rad, Float tm )
	: Shape(o2w) {
	radius = rad;
	height = ht;
	phiMax = Radians( Clamp( tm, 0.0f, 360.0f ) );
}
BBox Cone::ObjectBound() const {
	Point p1 = Point( -radius, -radius, 0 );
	Point p2 = Point(  radius,  radius, height );
	return BBox( p1, p2 );
}
bool Cone::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic cone coefficients
	Float k = radius / height;
	k = k*k;
	Float A = ray.D.x * ray.D.x + ray.D.y * ray.D.y -
		k * ray.D.z * ray.D.z;
	Float B = 2 * (ray.D.x * ray.O.x + ray.D.y * ray.O.y -
		k * ray.D.z * (ray.O.z-height) );
	Float C = ray.O.x * ray.O.x + ray.O.y * ray.O.y -
		k * (ray.O.z -height) * (ray.O.z-height);
	
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute cone inverse mapping
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test cone intersection against clipping parameters
	if (Phit.z < 0 || Phit.z > height || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute cone inverse mapping
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < 0 || Phit.z > height || phi > phiMax)
			return false;
	}
	// Fill in [[DifferentialGeometry]] from cone hit
	Float u = phi / phiMax;
	Float v = Phit.z / height;
	// Compute cone \dPdu and \dPdv
	Vector dPdu(-phiMax * Phit.y, phiMax * Phit.x, 0);
	Vector dPdv(-Phit.x / (1.f - v), Phit.y / (1.f - v), height);
	// Compute cone \dNdu and \dNdv
	Vector d2Pduu = -phiMax * phiMax * Vector(Phit.x, Phit.y, 0.);
	Vector d2Pduv = phiMax / (1.f - v) * Vector(Phit.y, -Phit.x, 0.);
	Vector d2Pdvv(0, 0, 0);
	// Compute coefficients for fundamental forms
	Float E = Dot(dPdu, dPdu);
	Float F = Dot(dPdu, dPdv);
	Float G = Dot(dPdv, dPdv);
	Vector N = Cross(dPdu, dPdv);
	Float e = Dot(N, d2Pduu);
	Float f = Dot(N, d2Pduv);
	Float g = Dot(N, d2Pdvv);
	// Compute \dNdu and \dNdv from fundamental form coefficients
	Float invEGF2 = 1.f / (E*G - F*F);
	Vector dNdu = (f*F - e*G) * invEGF2 * dPdu +
		(e*F - f*E) * invEGF2 * dPdv;
	Vector dNdv = (g*F - f*G) * invEGF2 * dPdu +
		(f*F - g*E) * invEGF2 * dPdv;
	*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
		ObjectToWorld(dPdv), ObjectToWorld(dNdu), ObjectToWorld(dNdv),
		u, v, this);
	// Update [[thitp]] for quadric intersection
	*thitp = thit;
	return true;
}

bool Cone::IntersectP(const Ray &r) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic cone coefficients
	Float k = radius / height;
	k = k*k;
	Float A = ray.D.x * ray.D.x + ray.D.y * ray.D.y -
		k * ray.D.z * ray.D.z;
	Float B = 2 * (ray.D.x * ray.O.x + ray.D.y * ray.O.y -
		k * ray.D.z * (ray.O.z-height) );
	Float C = ray.O.x * ray.O.x + ray.O.y * ray.O.y -
		k * (ray.O.z -height) * (ray.O.z-height);
	
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute cone inverse mapping
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test cone intersection against clipping parameters
	if (Phit.z < 0 || Phit.z > height || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute cone inverse mapping
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < 0 || Phit.z > height || phi > phiMax)
			return false;
	}
	return true;
}
Float Cone::Area() const {
	return phiMax*height*height*
		sqrtf((height*height)+(radius*radius))/(2.0f*radius);
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	Float radius = params.FindOneFloat( "_radius", 1 );
	Float height = params.FindOneFloat( "_height", 1 );
	Float thetamax = params.FindOneFloat( "_thetamax", 360 );
	return new Cone( o2w, height, radius, thetamax );
}
