// cylinder.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
// Cylinder Declarations
class Cylinder: public Shape {
public:
	// Cylinder Interface
	Cylinder(const Transform &o2w, Float rad, Float zmin, Float zmax, Float phiMax);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
	Point Sample(Float u1, Float u2, Normal *Ns) const;
protected:
	// Cylinder Data
	Float radius;
	Float zmin, zmax;
	Float phiMax;
};
// Cylinder Methods
Cylinder::Cylinder(const Transform &o2w, Float rad, Float z0,
		Float z1, Float pm)
	: Shape(o2w) {
	radius = rad;
	zmin = min(z0, z1);
	zmax = max(z0, z1);
	phiMax = Radians(Clamp(pm, 0.0f, 360.0f));
}
BBox Cylinder::ObjectBound() const {
	Point p1 = Point(-radius, -radius, zmin);
	Point p2 = Point( radius,  radius, zmax);
	return BBox(p1, p2);
}
bool Cylinder::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic cylinder coefficients
	Float A = ray.D.x*ray.D.x + ray.D.y*ray.D.y;
	Float B = 2 * (ray.D.x*ray.O.x + ray.D.y*ray.O.y);
	Float C = ray.O.x*ray.O.x + ray.O.y*ray.O.y - radius*radius;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute cylinder hit point and $\phi$
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test cylinder intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute cylinder hit point and $\phi$
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phi > phiMax)
			return false;
	}
	// Fill in [[DifferentialGeometry]] from cylinder hit
	Float u = phi / phiMax;
	Float v = (Phit.z - zmin) / (zmax - zmin);
	// Compute cylinder \dPdu and \dPdv
	Vector dPdu(-phiMax * Phit.y, phiMax * Phit.x, 0);
	Vector dPdv(0, 0, zmax - zmin);
	// Compute cylinder \dNdu and \dNdv
	Vector d2Pduu = -phiMax * phiMax * Vector(Phit.x, Phit.y, 0);
	Vector d2Pduv(0, 0, 0), d2Pdvv(0, 0, 0);
	// Compute coefficients for fundamental forms
	Float E = Dot(dPdu, dPdu);
	Float F = Dot(dPdu, dPdv);
	Float G = Dot(dPdv, dPdv);
	Vector N = Cross(dPdu, dPdv);
	Float e = Dot(N, d2Pduu);
	Float f = Dot(N, d2Pduv);
	Float g = Dot(N, d2Pdvv);
	// Compute \dNdu and \dNdv from fundamental form coefficients
	Float invEGF2 = 1.f / (E*G - F*F);
	Vector dNdu = (f*F - e*G) * invEGF2 * dPdu +
		(e*F - f*E) * invEGF2 * dPdv;
	Vector dNdv = (g*F - f*G) * invEGF2 * dPdu +
		(f*F - g*E) * invEGF2 * dPdv;
	*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
		ObjectToWorld(dPdv), ObjectToWorld(dNdu), ObjectToWorld(dNdv),
		u, v, this);

	// Update [[thitp]] for quadric intersection
	*thitp = thit;
	//	if(debugflag)
	//		cout << "   cyl:"<< thit << " P " << Phit << " transformed " <<ObjectToWorld(Phit) <<endl;
	return true;
}
bool Cylinder::IntersectP(const Ray &r) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic cylinder coefficients
	Float A = ray.D.x*ray.D.x + ray.D.y*ray.D.y;
	Float B = 2 * (ray.D.x*ray.O.x + ray.D.y*ray.O.y);
	Float C = ray.O.x*ray.O.x + ray.O.y*ray.O.y - radius*radius;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute cylinder hit point and $\phi$
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test cylinder intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute cylinder hit point and $\phi$
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phi > phiMax)
			return false;
	}
	return true;
}
Float Cylinder::Area() const {
	return (zmax-zmin)*phiMax*radius;
}
Point Cylinder::Sample(Float u1, Float u2,
		Normal *Ns) const {
	Float h = Lerp(u1, zmin, zmax);
	Float t = u2 * phiMax;
	Point p = Point(radius * cosf(t), radius * sinf(t), h);
	*Ns = ObjectToWorld(Normal(p.x, p.y, 0.)).Hat();
	return ObjectToWorld(p);
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	Float radius = params.FindOneFloat( "_radius", 1 );
	Float zmin = params.FindOneFloat( "_zmin", -1 );
	Float zmax = params.FindOneFloat( "_zmax", 1 );
	Float thetamax = params.FindOneFloat( "_thetamax", 360 );
	return new Cylinder( o2w, radius, zmin, zmax, thetamax );
}
