// disk.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
// Disk Declarations
class Disk : public Shape {
public:
	// Disk Interface
	Disk(const Transform &o2w, Float height, Float radius, Float phiMax);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
	virtual Point Sample(Float u1, Float u2, Normal *Ns) const;
private:
	// Disk Private Data
	Float height, radius, phiMax;
};
// Disk Methods
Disk::Disk(const Transform &o2w, Float ht, Float r, Float tmax)
	: Shape(o2w) {
	height = ht;
	radius = r;
	phiMax = Radians(Clamp(tmax, 0.0f, 360.0f));
}
BBox Disk::ObjectBound() const {
	return BBox(Point(-radius,-radius,height),
				Point(radius, radius,height));
}
bool Disk::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute plane intersection for disk
	if (fabsf(ray.D.z) < 1e-7) return false;
	Float thit = (height - ray.O.z) / ray.D.z;
	if (thit < ray.mint || thit > ray.maxt)
		return false;
	// See if hit point is inside disk radius and \phimax
	Point Phit = ray(thit);
	Float Dist2 = Phit.x * Phit.x + Phit.y * Phit.y;
	if (Dist2 > radius * radius)
		return false;
	// Test disk $\phi$ value against \phimax
	Float phi = atan2f(Phit.y, Phit.x);
	if (phi < 0) phi += 2. * M_PI;
	if (phi > phiMax)
		return false;
	// Fill in [[DifferentialGeometry]] from disk hit
	Float u = phi / phiMax;
	Float v = 1.f - (sqrtf(Dist2) / radius);
	Vector dPdu(-phiMax * Phit.y, phiMax * Phit.x, 0.);
	Vector dPdv(-Phit.x / (1-v), -Phit.y / (1-v), 0.);
	*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
		ObjectToWorld(dPdv), Vector(0,0,0), Vector(0,0,0), u, v, this);
	// Update [[thitp]] for quadric intersection
	*thitp = thit;
	return true;
}
bool Disk::IntersectP(const Ray &r) const {
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute plane intersection for disk
	if (fabsf(ray.D.z) < 1e-7) return false;
	Float thit = (height - ray.O.z) / ray.D.z;
	if (thit < ray.mint || thit > ray.maxt)
		return false;
	// See if hit point is inside disk radius and \phimax
	Point Phit = ray(thit);
	Float Dist2 = Phit.x * Phit.x + Phit.y * Phit.y;
	if (Dist2 > radius * radius)
		return false;
	// Test disk $\phi$ value against \phimax
	Float phi = atan2f(Phit.y, Phit.x);
	if (phi < 0) phi += 2. * M_PI;
	if (phi > phiMax)
		return false;
	return true;
}
Float Disk::Area() const {
	return phiMax * 0.5f * radius * radius;
}
Point Disk::Sample(Float u1, Float u2, Normal *Ns) const {
	Float r = radius * sqrtf(u2);
	Float phi = phiMax * u1;
	Point p = Point(r * cosf(phi), r * sinf(phi), height);
	*Ns = ObjectToWorld(Normal(0,0,1)).Hat();
	return ObjectToWorld(p);
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	Float height = params.FindOneFloat( "_height", 1 );
	Float radius = params.FindOneFloat( "_radius", 1 );
	Float thetamax = params.FindOneFloat( "_thetamax", 360 );
	return new Disk( o2w, height, radius, thetamax );
}
