// heightfield.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
#include "paramset.h"
#include "dynload.h"
// Heightfield Declarations
class Heightfield : public Shape {
public:
	// Heightfield Method Declarations
	Heightfield(const Transform &o2w, int nu, int nv, const Float *zs);
	~Heightfield();
	bool CanIntersect() const;
	void Refine(vector<RefinedShape> &refined) const;
	BBox ObjectBound() const;
private:
	// Heightfield Data
	Float *z;
	int nx, ny;
};
// Heightfield Methods
Heightfield::Heightfield(const Transform &o2w, int x, int y,
		const Float *zs)
	: Shape(o2w) {
	nx = x;
	ny = y;
	z = new Float[nx*ny];
	memcpy(z, zs, nx*ny*sizeof(Float));
}
Heightfield::~Heightfield() {
	delete[] z;
}
BBox Heightfield::ObjectBound() const {
	Float minz = z[0], maxz = z[0];
	for (int i = 1; i < nx*ny; ++i) {
		if (z[i] < minz) minz = z[i];
		if (z[i] > maxz) maxz = z[i];
	}
	return BBox(Point(0,0,minz), Point(1,1,maxz));
}
bool Heightfield::CanIntersect() const {
	return false;
}
void Heightfield::Refine(vector<RefinedShape> &refined) const {
	int ntris = 2*(nx-1)*(ny-1);
	refined.reserve(ntris);
	int *verts = new int[3*ntris];
	Point *P = new Point[nx*ny];
	Float *uvs = new Float[2*nx*ny];
	int nverts = nx*ny;
	int x, y;
	// Compute heightfield vertex positions
	int pos = 0;
	for (y = 0; y < ny; ++y) {
		for (x = 0; x < nx; ++x) {
			P[pos].x = uvs[2*pos]   = (Float)x / (Float)(nx-1);
			P[pos].y = uvs[2*pos+1] = (Float)y / (Float)(ny-1);
			P[pos].z = z[pos];
			++pos;
		}
	}
	// Fill in heightfield vertex offset array
	int *vp = verts;
	for (y = 0; y < ny-1; ++y) {
		for (x = 0; x < nx-1; ++x) {
	#define VERT(x,y) ((x)+(y)*nx)
			*vp++ = VERT(x, y);
			*vp++ = VERT(x+1, y);
			*vp++ = VERT(x+1, y+1);
	
			*vp++ = VERT(x, y);
			*vp++ = VERT(x+1, y+1);
			*vp++ = VERT(x, y+1);
		}
	#undef VERT
	}
	ParamSet paramSet(nx*ny);
	paramSet.AddInt("_ntris", &ntris);
	paramSet.AddInt("_nverts", &nverts);
	paramSet.AddInt("_vertexIndices", verts, PARAM_TYPE_UNIFORM,
		3*ntris);
	paramSet.AddFloat("uv", uvs, PARAM_TYPE_VERTEX, 2);
	paramSet.AddPoint("P", P, PARAM_TYPE_VERTEX);
	refined.push_back(RefinedShape(
		CreateShape("trianglemesh", LRT_BUILDDIR,
			ObjectToWorld, paramSet)));
	delete[] P;
	delete[] uvs;
	delete[] verts;
}
extern "C" Reference<Shape>CreateShape(const Transform &o2w,
		const ParamSet &params) {
	int nu = params.FindOneInt("_nu", -1);
	int nv = params.FindOneInt("_nv", -1);
	int type, narray;
	const Float *Pz = params.FindFloat("Pz", &type, &narray);
	Assert(nu != -1 && nv != -1 && Pz != NULL);
	return new Heightfield(o2w, nu, nv, Pz);
}
