// hyperboloid.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
// Hyperboloid Declarations
class Hyperboloid: public Shape {
public:
	// Hyperboloid Interface
	Hyperboloid(const Transform &o2w, const Point &point1, const Point &point2,
		Float tm);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
protected:
	// Hyperboloid Data
	Point p1, p2;
	Float zmin, zmax;
	Float phiMax;
	Float rmax;
	Float a, c;
};
// Hyperboloid Methods
Hyperboloid::Hyperboloid(const Transform &o2w, const Point &point1,
		const Point &point2, Float tm)
	: Shape(o2w) {
	p1 = point1;
	p2 = point2;
	phiMax = Radians(Clamp(tm, 0.0f, 360.0f));
	Float rad1 = sqrtf(p1.x*p1.x + p1.y*p1.y);
	Float rad2 = sqrtf(p2.x*p2.x + p2.y*p2.y);
	rmax = max(rad1,rad2);
	zmin = min(p1.z,p2.z);
	zmax = max(p1.z,p2.z);
	// Compute implicit function coefficients for hyperboloid
	if (p2.z == 0.) swap(p1, p2);
	Point pp = p1;
	Float xy1, xy2;
	do {
		xy1 = pp.x*pp.x + pp.y*pp.y;
		xy2 = p2.x*p2.x + p2.y*p2.y;
		a = (1.f/xy1 - (pp.z*pp.z)/(xy1*p2.z*p2.z)) / (1 - (xy2*pp.z*pp.z)/(xy1*p2.z*p2.z));
		c = (a * xy2 - 1) / (p2.z*p2.z);
		pp += 2.f * (p2-p1);
	} while (isinf(a) || isnan(a));
}
BBox Hyperboloid::ObjectBound() const {
	Point p1 = Point( -rmax, -rmax, zmin );
	Point p2 = Point(  rmax,  rmax, zmax );
	return BBox( p1, p2 );
}
bool Hyperboloid::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	Float phi, v;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic hyperboloid coefficients
	Float A = a*ray.D.x*ray.D.x + a*ray.D.y*ray.D.y - c*ray.D.z*ray.D.z;
	Float B = 2.f * (a*ray.D.x*ray.O.x + a*ray.D.y*ray.O.y - c*ray.D.z*ray.O.z);
	Float C = a*ray.O.x*ray.O.x + a*ray.O.y*ray.O.y - c*ray.O.z*ray.O.z - 1;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute hyperboloid inverse mapping
	Phit = ray(thit);
	v = (Phit.z - p1.z)/(p2.z - p1.z);
	Point pr = (1.f-v) * p1 + v * p2;
	phi = atan2f(pr.x*Phit.y - Phit.x*pr.y,
		Phit.x*pr.x + Phit.y*pr.y);
	if (phi < 0)
		phi += 2*M_PI;
	// Test hyperboloid intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute hyperboloid inverse mapping
		Phit = ray(thit);
		v = (Phit.z - p1.z)/(p2.z - p1.z);
		Point pr = (1.f-v) * p1 + v * p2;
		phi = atan2f(pr.x*Phit.y - Phit.x*pr.y,
			Phit.x*pr.x + Phit.y*pr.y);
		if (phi < 0)
			phi += 2*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phi > phiMax)
			return false;
	}
	// Fill in [[DifferentialGeometry]] from hyperboloid hit
	Float u = phi / phiMax;
	// Compute hyperboloid \dPdu and \dPdv
	Float cosphi = cosf(phi), sinphi = sinf(phi);
	Vector dPdu(-phiMax * Phit.y, phiMax * Phit.x, 0.);
	Vector dPdv((p2.x-p1.x) * cosphi - (p2.y-p1.y) * sinphi,
		(p2.x-p1.x) * sinphi + (p2.y-p1.y) * cosphi,
		p2.z-p1.z);
	// Compute hyperboloid \dNdu and \dNdv
	Vector d2Pduu = -phiMax * phiMax * Vector(Phit.x, Phit.y, 0);
	Vector d2Pduv = phiMax * Vector(-dPdv.y, dPdv.x, 0.);
	Vector d2Pdvv(0, 0, 0);
	// Compute coefficients for fundamental forms
	Float E = Dot(dPdu, dPdu);
	Float F = Dot(dPdu, dPdv);
	Float G = Dot(dPdv, dPdv);
	Vector N = Cross(dPdu, dPdv);
	Float e = Dot(N, d2Pduu);
	Float f = Dot(N, d2Pduv);
	Float g = Dot(N, d2Pdvv);
	// Compute \dNdu and \dNdv from fundamental form coefficients
	Float invEGF2 = 1.f / (E*G - F*F);
	Vector dNdu = (f*F - e*G) * invEGF2 * dPdu +
		(e*F - f*E) * invEGF2 * dPdv;
	Vector dNdv = (g*F - f*G) * invEGF2 * dPdu +
		(f*F - g*E) * invEGF2 * dPdv;
	*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
		ObjectToWorld(dPdv), ObjectToWorld(dNdu), ObjectToWorld(dNdv),
		u, v, this);
	// Update [[thitp]] for quadric intersection
	*thitp = thit;
	return true;
}
bool Hyperboloid::IntersectP(const Ray &r) const {
	Float phi, v;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic hyperboloid coefficients
	Float A = a*ray.D.x*ray.D.x + a*ray.D.y*ray.D.y - c*ray.D.z*ray.D.z;
	Float B = 2.f * (a*ray.D.x*ray.O.x + a*ray.D.y*ray.O.y - c*ray.D.z*ray.O.z);
	Float C = a*ray.O.x*ray.O.x + a*ray.O.y*ray.O.y - c*ray.O.z*ray.O.z - 1;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute hyperboloid inverse mapping
	Phit = ray(thit);
	v = (Phit.z - p1.z)/(p2.z - p1.z);
	Point pr = (1.f-v) * p1 + v * p2;
	phi = atan2f(pr.x*Phit.y - Phit.x*pr.y,
		Phit.x*pr.x + Phit.y*pr.y);
	if (phi < 0)
		phi += 2*M_PI;
	// Test hyperboloid intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute hyperboloid inverse mapping
		Phit = ray(thit);
		v = (Phit.z - p1.z)/(p2.z - p1.z);
		Point pr = (1.f-v) * p1 + v * p2;
		phi = atan2f(pr.x*Phit.y - Phit.x*pr.y,
			Phit.x*pr.x + Phit.y*pr.y);
		if (phi < 0)
			phi += 2*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phi > phiMax)
			return false;
	}
	return true;
}
#define SQR(a) ((a)*(a))
#define QUAD(a) ((SQR(a))*(SQR(a)))
Float Hyperboloid::Area() const {
	return phiMax/6.f * (2.f*QUAD(p1.x) - 2.f*p1.x*p1.x*p1.x*p2.x +
		2.f*QUAD(p2.x) + 2.f*(p1.y*p1.y + p1.y*p2.y + p2.y*p2.y)*
			(SQR(p1.y - p2.y) + SQR(p1.z - p2.z)) +
		   p2.x*p2.x*(5.f*p1.y*p1.y + 2.f*p1.y*p2.y - 4.f*p2.y*p2.y +
			  2.f*SQR(p1.z - p2.z)) +
		   p1.x*p1.x*(-4.f*p1.y*p1.y + 2.f*p1.y*p2.y + 5.f*p2.y*p2.y +
			  2.f*SQR(p1.z - p2.z)) -
		   2.f*p1.x*p2.x*(p2.x*p2.x - p1.y*p1.y + 5.f*p1.y*p2.y -
			  p2.y*p2.y - p1.z*p1.z + 2.f*p1.z*p2.z - p2.z*p2.z));
}
#undef SQR
#undef QUAD
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	Point p1 = params.FindOnePoint( "_p1", Point(0,0,0) );
	Point p2 = params.FindOnePoint( "_p2", Point(1,1,1) );
	Float thetamax = params.FindOneFloat( "_thetamax", 360 );
	return new Hyperboloid( o2w, p1, p2, thetamax );
}
