// paraboloid.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
// Paraboloid Declarations
class Paraboloid: public Shape {
public:
	// Paraboloid Interface
	Paraboloid(const Transform &o2w, Float rad, Float z0, Float z1, Float tm );
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
protected:
	// Paraboloid Data
	Float radius;
	Float zmin, zmax;
	Float phiMax;
};
// Paraboloid Methods
Paraboloid::Paraboloid(const Transform &o2w, Float rad, Float z0,
		Float z1, Float tm)
	: Shape(o2w) {
	radius = rad;
	zmin = min(z0,z1);
	zmax = max(z0,z1);
	phiMax = Radians( Clamp( tm, 0.0f, 360.0f ) );
}
BBox Paraboloid::ObjectBound() const {
	Point p1 = Point( -radius, -radius, zmin );
	Point p2 = Point(  radius,  radius, zmax );
	return BBox( p1, p2 );
}
bool Paraboloid::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic paraboloid coefficients
	Float k = zmax/(radius*radius);
	Float A =   k*(ray.D.x * ray.D.x + ray.D.y * ray.D.y );
	Float B = 2*k*(ray.D.x * ray.O.x + ray.D.y * ray.O.y ) - ray.D.z;
	Float C =   k*(ray.O.x * ray.O.x + ray.O.y * ray.O.y ) - ray.O.z;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute paraboloid inverse mapping
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test paraboloid intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute paraboloid inverse mapping
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phi > phiMax)
			return false;
	}
	// Fill in [[DifferentialGeometry]] from paraboloid hit
	Float u = phi / phiMax;
	Float v = (Phit.z-zmin) / (zmax-zmin);
	// Compute parabaloid \dPdu and \dPdv
	Vector dPdu(-phiMax * Phit.y, phiMax * Phit.x, 0.);
	Vector dPdv = (zmax - zmin) *
		Vector(Phit.x / Phit.z, Phit.y / Phit.z, 1.);
	// Compute parabaloid \dNdu and \dNdv
	Vector d2Pduu = -phiMax * phiMax * Vector(Phit.x, Phit.y, 0);
	Vector d2Pduv = (zmax - zmin) * phiMax *
		Vector(-Phit.y / Phit.z, Phit.x / Phit.z, 0);
	Vector d2Pdvv = -2.f * (zmax - zmin) * (zmax - zmin) *
		Vector(Phit.x/(Phit.z*Phit.z), Phit.y/Phit.z*Phit.z, 0.);
	// Compute coefficients for fundamental forms
	Float E = Dot(dPdu, dPdu);
	Float F = Dot(dPdu, dPdv);
	Float G = Dot(dPdv, dPdv);
	Vector N = Cross(dPdu, dPdv);
	Float e = Dot(N, d2Pduu);
	Float f = Dot(N, d2Pduv);
	Float g = Dot(N, d2Pdvv);
	// Compute \dNdu and \dNdv from fundamental form coefficients
	Float invEGF2 = 1.f / (E*G - F*F);
	Vector dNdu = (f*F - e*G) * invEGF2 * dPdu +
		(e*F - f*E) * invEGF2 * dPdv;
	Vector dNdv = (g*F - f*G) * invEGF2 * dPdu +
		(f*F - g*E) * invEGF2 * dPdv;
	*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
		ObjectToWorld(dPdv), ObjectToWorld(dNdu), ObjectToWorld(dNdv),
		u, v, this);
	// Update [[thitp]] for quadric intersection
	*thitp = thit;
	return true;
}

bool Paraboloid::IntersectP(const Ray &r) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic paraboloid coefficients
	Float k = zmax/(radius*radius);
	Float A =   k*(ray.D.x * ray.D.x + ray.D.y * ray.D.y );
	Float B = 2*k*(ray.D.x * ray.O.x + ray.D.y * ray.O.y ) - ray.D.z;
	Float C =   k*(ray.O.x * ray.O.x + ray.O.y * ray.O.y ) - ray.O.z;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute paraboloid inverse mapping
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test paraboloid intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		thit = t1;
		if (t1 > ray.maxt) return false;
		// Compute paraboloid inverse mapping
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phi > phiMax)
			return false;
	}
	return true;
}
Float Paraboloid::Area() const {
	return phiMax/12.0f *
		(powf(1+4*zmin, 1.5f) - powf(1+4*zmax, 1.5f));
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	Float radius = params.FindOneFloat( "_radius", 1 );
	Float zmin = params.FindOneFloat( "_zmin", 0 );
	Float zmax = params.FindOneFloat( "_zmax", 1 );
	Float thetamax = params.FindOneFloat( "_thetamax", 360 );
	return new Paraboloid( o2w, radius, zmin, zmax, thetamax );
}
