// Magic Software, Inc.
// http://www.magic-software.com
// Copyright (c) 2000-2003.  All Rights Reserved
//
// Source code from Magic Software is supplied under the terms of a license
// agreement and may not be copied or disclosed except in accordance with the
// terms of that agreement.  The various license agreements may be found at
// the Magic Software web site.  This file is subject to the license
//
// FREE SOURCE CODE
// http://www.magic-software.com/License/free.pdf

class Polynomial
{
public:
    // construction and destruction
    Polynomial (int iDegree = -1);
    Polynomial (const Polynomial& rkPoly);
    ~Polynomial ();

    // assignment
    Polynomial& operator= (const Polynomial& rkPoly);

    // coefficient access
    void SetDegree (int iDegree);
    int GetDegree () const;
    Float& operator[] (int i) const;

    // evaluation
    Float operator() (Float fT) const;

    // derivation
    Polynomial GetDerivative () const;

    // inversion ( invpoly[i] = poly[degree-i] for 0 <= i <= degree )
    Polynomial GetInversion () const;

    // Reduce degree by eliminating all (nearly) zero leading coefficients
    // and by making the leading coefficient one.  The input parameter is
    // the threshold for specifying that a coefficient is effectively zero.
    void Compress (Float fEpsilon = 1e-08f);

    // arithmetic
    Polynomial operator+ (const Polynomial& rkPoly) const;
    Polynomial operator- (const Polynomial& rkPoly) const;
    Polynomial operator* (const Polynomial& rkPoly) const;
    Polynomial operator* (Float fScalar) const;
    Polynomial operator- () const;
    friend Polynomial operator* (Float fScalar,
        const Polynomial& rkPoly);

    // If 'this' is P(t) and the divisor is D(t) with degree(P) >= degree(D),
    // then P(t) = Q(t)*D(t)+R(t) where Q(t) is the quotient with
    // degree(Q) = degree(P) - degree(D) and R(t) is the remainder with
    // degree(R) < degree(D).  If this routine is called with
    // degree(P) < degree(D), then Q = 0 and R = P are returned.  The value
    // of REMAINDER_EPSILON is used as a threshold on the coefficients of
    // the remainder polynomial.  If smaller, the coefficient is assumed to
    // be zero.
    static Float REMAINDER_EPSILON;
    void Divide (const Polynomial& rkDiv, Polynomial& rkQuot,
        Polynomial& rkRem);

    // root finding
    static int DIGITS_ACCURACY;
    static Float ZERO_TOLERANCE;
    bool Bisection (Float fXMin, Float fXMax, Float& rfRoot) const;
    void GetRootsOnInterval (Float fXMin, Float fXMax,
        int& riCount, Float* afRoot) const;
    void GetAllRoots (int& riCount, Float* afRoot) const;

    Float GetRootBound () const;  // real roots must be in [-bound,bound]
    bool AllFloatPartsNegative () const;
    bool AllFloatPartsPositive () const;

    // low-degree root finding (degree must be correct)
    bool RootsDegree2 (int& riCount, Float afRoot[2]) const;
    bool RootsDegree3 (int& riCount, Float afRoot[3]) const;
    bool RootsDegree4 (int& riCount, Float afRoot[4]) const;

    // solve A*r^3 + B*r = C where A > 0 and B > 0
    Float SpecialCubeRoot (Float fA, Float fB, Float fC);

    // Count the number of roots on [t0,t1].  Uses Sturm sequences to do the
    // counting.  It is allowed to pass in t0 = -MgcMath::MAXREAL or
    // t1 = +MgcMath::MAXREAL.  The value of STURM_EPSILON is used as a
    // threshold on the value of a sturm polynomial at an end point.  If
    // smaller, that value is assumed to be zero.  The return value is the
    // number of roots.  If there are infinitely many, -1 is returned.
    static Float STURM_EPSILON;
    int GetRootCount (Float fT0, Float fT1);

protected:
    int m_iDegree;
    Float* m_afCoeff;

    // root finding
    static const Float ms_fInvLog2;
    static const Float ms_fLog10;
    static const Float ms_fThird;
    static const Float ms_fSqrt3;
    static const Float ms_fTwentySeventh;
    static bool AllFloatPartsNegative (int iDegree, Float* afCoeff);
};

////////////////////////////////////////////////////////////////////
// Polynomial class implementation (don't worry about this code)
////////////////////////////////////////////////////////////////////

#include <vector>
using namespace std;

/*
int Polynomial::DIGITS_ACCURACY = 10;
Float Polynomial::ZERO_TOLERANCE = 1e-10f;
Float Polynomial::REMAINDER_EPSILON = 1e-10f;
Float Polynomial::STURM_EPSILON = 1e-08f;
*/

int Polynomial::DIGITS_ACCURACY = 6;
Float Polynomial::ZERO_TOLERANCE = 1e-10f;
Float Polynomial::REMAINDER_EPSILON = 1e-06f;
Float Polynomial::STURM_EPSILON = 1e-04f;


const Float Polynomial::ms_fInvLog2       = 1.442695041f;
const Float Polynomial::ms_fLog10         = 2.302585093f;
const Float Polynomial::ms_fThird         = 0.333333333f;
const Float Polynomial::ms_fSqrt3         = 1.732050808f;
const Float Polynomial::ms_fTwentySeventh = 0.037037037f;

//----------------------------------------------------------------------------
Polynomial::Polynomial (int iDegree)
{
    if ( iDegree >= 0 )
    {
        m_iDegree = iDegree;
        m_afCoeff = new Float[m_iDegree+1];
    }
    else
    {
        // default creation
        m_iDegree = -1;
        m_afCoeff = NULL;
    }
}
//----------------------------------------------------------------------------
Polynomial::Polynomial (const Polynomial& rkPoly)
{
    m_iDegree = rkPoly.m_iDegree;
    m_afCoeff = new Float[m_iDegree+1];
    for (int i = 0; i <= m_iDegree; i++)
        m_afCoeff[i] = rkPoly.m_afCoeff[i];
}
//----------------------------------------------------------------------------
Polynomial::~Polynomial ()
{
    delete[] m_afCoeff;
}
//----------------------------------------------------------------------------
Float& Polynomial::operator[] (int i) const
{
    assert( 0 <= i && i <= m_iDegree );
    return ((Float*)m_afCoeff)[i];
}
//----------------------------------------------------------------------------
Polynomial& Polynomial::operator= (const Polynomial& rkPoly)
{
    delete[] m_afCoeff;
    m_iDegree = rkPoly.m_iDegree;

    if ( m_iDegree >= 0 )
    {
        m_afCoeff = new Float[m_iDegree+1];
        for (int i = 0; i <= m_iDegree; i++)
            m_afCoeff[i] = rkPoly.m_afCoeff[i];
    }

    return *this;
}
//----------------------------------------------------------------------------
void Polynomial::SetDegree (int iDegree)
{
    m_iDegree = iDegree;
    delete[] m_afCoeff;

    if ( m_iDegree >= 0 )
        m_afCoeff = new Float[m_iDegree+1];
    else
        m_afCoeff = NULL;
}
//----------------------------------------------------------------------------
int Polynomial::GetDegree () const
{
    return m_iDegree;
}
//----------------------------------------------------------------------------
Float Polynomial::operator() (Float fT) const
{
    assert( m_iDegree >= 0 );

    Float fResult = m_afCoeff[m_iDegree];
    for (int i = m_iDegree-1; i >= 0; i--)
    {
        fResult *= fT;
        fResult += m_afCoeff[i];
    }
    return fResult;
}
//----------------------------------------------------------------------------
Polynomial Polynomial::GetDerivative () const
{
    if ( m_iDegree > 0 )
    {
        Polynomial kDeriv(m_iDegree-1);
        for (int i0 = 0, i1 = 1; i0 < m_iDegree; i0++, i1++)
            kDeriv[i0] = i1*m_afCoeff[i1];
        return kDeriv;
    }
    else
    {
        return Polynomial(-1);
    }
}
//----------------------------------------------------------------------------
Polynomial Polynomial::GetInversion () const
{
    Polynomial kInvPoly(m_iDegree);
    for (int i = 0; i <= m_iDegree; i++)
        kInvPoly[i] = m_afCoeff[m_iDegree-i];
    return kInvPoly;
}
//----------------------------------------------------------------------------
void Polynomial::Compress (Float fEpsilon)
{
	int i;
	for (i = m_iDegree; i >= 0; i--)
	{
        if ( fabsf(m_afCoeff[i]) <= fEpsilon )
			m_iDegree--;
		else
			break;
	}

	for (i = 0; i <= m_iDegree; i++)
		m_afCoeff[i] /= m_afCoeff[m_iDegree];
}
//----------------------------------------------------------------------------
Polynomial Polynomial::operator+ (const Polynomial& rkPoly) const
{
    assert( m_iDegree >= 0 && rkPoly.m_iDegree >= 0 );

    Polynomial kSum(-1);
    int i;

    if ( m_iDegree > rkPoly.m_iDegree )
    {
        kSum.SetDegree(m_iDegree);
        for (i = 0; i <= rkPoly.m_iDegree; i++)
            kSum[i] = m_afCoeff[i] + rkPoly.m_afCoeff[i];
        for (i = rkPoly.m_iDegree+1; i <= m_iDegree; i++)
            kSum[i] = m_afCoeff[i];

    }
    else
    {
        kSum.SetDegree(rkPoly.m_iDegree);
        for (i = 0; i <= m_iDegree; i++)
            kSum[i] = m_afCoeff[i] + rkPoly.m_afCoeff[i];
        for (i = m_iDegree+1; i <= rkPoly.m_iDegree; i++)
            kSum[i] = rkPoly.m_afCoeff[i];
    }

    return kSum;
}
//----------------------------------------------------------------------------
Polynomial Polynomial::operator- (const Polynomial& rkPoly) const
{
    assert( m_iDegree >= 0 && rkPoly.m_iDegree >= 0 );

    Polynomial kDiff(-1);
    int i;

    if ( m_iDegree > rkPoly.m_iDegree )
    {
        kDiff.SetDegree(m_iDegree);
        for (i = 0; i <= rkPoly.m_iDegree; i++)
            kDiff[i] = m_afCoeff[i] - rkPoly.m_afCoeff[i];
        for (i = rkPoly.m_iDegree+1; i <= m_iDegree; i++)
            kDiff[i] = m_afCoeff[i];

    }
    else
    {
        kDiff.SetDegree(rkPoly.m_iDegree);
        for (i = 0; i <= m_iDegree; i++)
            kDiff[i] = m_afCoeff[i] - rkPoly.m_afCoeff[i];
        for (i = m_iDegree+1; i <= rkPoly.m_iDegree; i++)
            kDiff[i] = -rkPoly.m_afCoeff[i];
    }

    return kDiff;
}
//----------------------------------------------------------------------------
Polynomial Polynomial::operator* (const Polynomial& rkPoly) const
{
    assert( m_iDegree >= 0 && rkPoly.m_iDegree >= 0 );

    Polynomial kProd(m_iDegree + rkPoly.m_iDegree);

    memset(kProd.m_afCoeff,0,(kProd.m_iDegree+1)*sizeof(Float));
    for (int i0 = 0; i0 <= m_iDegree; i0++)
    {
        for (int i1 = 0; i1 <= rkPoly.m_iDegree; i1++)
        {
            int i2 = i0 + i1;
            kProd.m_afCoeff[i2] += m_afCoeff[i0]*rkPoly.m_afCoeff[i1];
        }
    }

    return kProd;
}
//----------------------------------------------------------------------------
Polynomial Polynomial::operator* (Float fScalar) const
{
    assert( m_iDegree >= 0 );

    Polynomial kProd(m_iDegree);
    for (int i = 0; i <= m_iDegree; i++)
        kProd[i] = fScalar*m_afCoeff[i];

    return kProd;
}
//----------------------------------------------------------------------------
Polynomial Polynomial::operator- () const
{
    assert( m_iDegree >= 0 );

    Polynomial kNeg(m_iDegree);
    for (int i = 0; i <= m_iDegree; i++)
        kNeg[i] = -m_afCoeff[i];

    return kNeg;
}
//----------------------------------------------------------------------------
Polynomial operator* (Float fScalar, const Polynomial& rkPoly)
{
    assert( rkPoly.m_iDegree >= 0 );

    Polynomial kProd(rkPoly.m_iDegree);
    for (int i = 0; i <= rkPoly.m_iDegree; i++)
        kProd[i] = fScalar*rkPoly.m_afCoeff[i];

    return kProd;
}
//----------------------------------------------------------------------------
void Polynomial::Divide (const Polynomial& rkDiv, Polynomial& rkQuot,
    Polynomial& rkRem)
{
    int iQuotDegree = m_iDegree - rkDiv.m_iDegree;
    if ( iQuotDegree >= 0 )
    {
        rkQuot.SetDegree(iQuotDegree);

        // temporary storage for the remainder
        Polynomial kTmp = *this;

        // do the division (Euclidean algorithm)
        Float fInv = 1.0f/rkDiv[rkDiv.m_iDegree];
        for (int iQ = iQuotDegree; iQ >= 0; iQ--)
        {
            int iR = rkDiv.m_iDegree + iQ;
            rkQuot[iQ] = fInv*kTmp[iR];
            for (iR--; iR >= iQ; iR--)
                kTmp[iR] -= rkQuot[iQ]*rkDiv[iR-iQ];
        }

        // calculate the correct degree for the remainder
        int iRemDeg = rkDiv.m_iDegree - 1;
        while ( iRemDeg > 0 && fabsf(kTmp[iRemDeg]) < REMAINDER_EPSILON )
            iRemDeg--;

        if ( iRemDeg == 0 && fabsf(kTmp[0]) < REMAINDER_EPSILON )
            kTmp[0] = 0.0f;

        rkRem.SetDegree(iRemDeg);
        memcpy(rkRem.m_afCoeff,kTmp.m_afCoeff,(iRemDeg+1)*sizeof(Float));
    }
    else
    {
        rkQuot.SetDegree(0);
        rkQuot[0] = 0.0f;
        rkRem = *this;
    }
}
//----------------------------------------------------------------------------
bool Polynomial::Bisection (Float fXMin, Float fXMax,
    Float& rfRoot) const
{
    Float fP0 = (*this)(fXMin);
    Float fP1 = (*this)(fXMax);

    // check for endpoint zeros
    if ( fabsf(fP0) <= ZERO_TOLERANCE )
    {
        rfRoot = fXMin;
        return true;
    }
    if ( fabsf(fP1) <= ZERO_TOLERANCE )
    {
        rfRoot = fXMax;
        return true;
    }

    if ( fP0*fP1 > 0.0f )
        return false;

    // determine number of iterations to get 'digits' accuracy.
    Float fTmp0 = logf(fXMax-fXMin);
    Float fTmp1 = DIGITS_ACCURACY*ms_fLog10;
    Float fArg = (fTmp0 + fTmp1)*ms_fInvLog2;
    int iMaxIter = (int)ceilf(fArg);
    
    for (int i = 0; i < iMaxIter; i++)
    {
        rfRoot = 0.5f*(fXMin + fXMax);
        Float fP = (*this)(rfRoot);
        if ( fabsf(fP) <= ZERO_TOLERANCE )
            return true;
        
        if ( fP*fP0 < 0.0f )
        {
            fXMax = rfRoot;
            fP1 = fP;
        }
        else
        {
            fXMin = rfRoot;
            fP0 = fP;
        }
    }

    return true;
}
//----------------------------------------------------------------------------
void Polynomial::GetRootsOnInterval (Float fXMin, Float fXMax,
    int& riCount, Float* afRoot) const
{
    Float fRoot;

    if ( m_iDegree == 1 )
    {
        if ( Bisection(fXMin,fXMax,fRoot) )
        {
            riCount = 1;
            afRoot[0] = fRoot;
        }
        else
        {
            riCount = 0;
        }
        return;
    }

    // get roots of derivative polynomial
    Polynomial kDeriv = GetDerivative();
    kDeriv.GetRootsOnInterval(fXMin,fXMax,riCount,afRoot);

    int i, iNewCount = 0;
    Float* afNewRoot = new Float[riCount+1];

    if ( riCount > 0 )
    {
        // find root on [xmin,root[0]]
        if ( Bisection(fXMin,afRoot[0],fRoot) )
            afNewRoot[iNewCount++] = fRoot;

        // find root on [root[i],root[i+1]] for 0 <= i <= count-2
        for (i = 0; i <= riCount-2; i++)
        {
            if ( Bisection(afRoot[i],afRoot[i+1],fRoot) )
                afNewRoot[iNewCount++] = fRoot;
        }

        // find root on [root[count-1],xmax]
        if ( Bisection(afRoot[riCount-1],fXMax,fRoot) )
            afNewRoot[iNewCount++] = fRoot;
    }
    else
    {
        // polynomial is monotone on [xmin,xmax], has at most one root
        if ( Bisection(fXMin,fXMax,fRoot) )
            afNewRoot[iNewCount++] = fRoot;
    }

    // copy to old buffer
    if ( iNewCount > 0 )
    {
        riCount = 1;
        afRoot[0] = afNewRoot[0];
        for (i = 1; i < iNewCount; i++)
        {
            if ( fabsf(afNewRoot[i]-afNewRoot[i-1]) > ZERO_TOLERANCE )
                afRoot[riCount++] = afNewRoot[i];
        }
    }
    else
    {
        riCount = 0;
    }

    delete[] afNewRoot;
}
//----------------------------------------------------------------------------
void Polynomial::GetAllRoots (int& riCount, Float* afRoot) const
{
    Float fBound, fTmp;
    int i;

    Float fAbs0 = fabsf(m_afCoeff[0]);
    Float fAbsD = fabsf(m_afCoeff[m_iDegree]);

    if ( fAbsD >= fAbs0 )
    {
        fBound = fAbs0;
        for (i = 0; i < m_iDegree; i++)
        {
            fTmp = fabsf(m_afCoeff[i]) + fAbsD;
            if ( fBound < fTmp )
                fBound = fTmp;
        }
        fBound /= fAbsD;

        GetRootsOnInterval(-fBound,fBound,riCount,afRoot);
    }
    else
    {
        fBound = fAbsD;
        for (i = 0; i < m_iDegree; i++)
        {
            fTmp = fabsf(m_afCoeff[i]) + fAbs0;
            if ( fBound < fTmp )
                fBound = fTmp;
        }
        fBound /= fAbs0;

        // use inverted polynomial to find inverse roots
        Float* afInvRoot = new Float[m_iDegree];
        Polynomial kInvPoly = GetInversion();
        kInvPoly.GetRootsOnInterval(-fBound,fBound,riCount,afInvRoot);

        // Adjust the tolerance to your liking.  Any inverse root whose
        // magnitude is smaller than the tolerance is assumed to be zero.  In
        // this case the original polynomial has a "root at infinity", but
        // only finite roots are included in the final results.
        const Float fEpsilon = 1e-08f;
        int iFiniteRoots = 0;
        for (i = 0; i < riCount; i++)
        {
            Float fTmp = afInvRoot[riCount-1-i];
            if ( fabsf(fTmp) >= fEpsilon )
            {
                afRoot[i] = 1.0f/afInvRoot[riCount-1-i];
                iFiniteRoots++;
            }
        }
        delete[] afInvRoot;
        riCount = iFiniteRoots;
    }
}
//----------------------------------------------------------------------------
Float Polynomial::GetRootBound () const
{
    Float fBound, fTmp;
    int i;

    Float fAbs0 = fabsf(m_afCoeff[0]);
    Float fAbsD = fabsf(m_afCoeff[m_iDegree]);

    if ( fAbsD >= fAbs0 )
    {
        fBound = fAbs0;
        for (i = 0; i < m_iDegree; i++)
        {
            fTmp = fabsf(m_afCoeff[i]) + fAbsD;
            if ( fBound < fTmp )
                fBound = fTmp;
        }
        fBound /= fAbsD;
    }
    else
    {
        fBound = fAbsD;
        for (i = 0; i < m_iDegree; i++)
        {
            fTmp = fabsf(m_afCoeff[i]) + fAbs0;
            if ( fBound < fTmp )
                fBound = fTmp;
        }
        fBound /= fAbs0;
    }

    return fBound;
}
//----------------------------------------------------------------------------
bool Polynomial::AllFloatPartsNegative (int iDegree, Float* afCoeff)
{
    // assert:  afCoeff[iDegree] = 1

    if ( afCoeff[iDegree-1] <= 0.0f )
        return false;
    if ( iDegree == 1 )
        return true;

    Float* afTmpCoeff = new Float[iDegree];
    afTmpCoeff[0] = 2.0f*afCoeff[0]*afCoeff[iDegree-1];
    int i;
    for (i = 1; i <= iDegree-2; i++) 
    {
        afTmpCoeff[i] = afCoeff[iDegree-1]*afCoeff[i];
        if ( ((iDegree-i) % 2) == 0 )
            afTmpCoeff[i] -= afCoeff[i-1];
        afTmpCoeff[i] *= 2.0f;
    }
    afTmpCoeff[iDegree-1] = 2.0f*afCoeff[iDegree-1]*afCoeff[iDegree-1];

    int iNextDegree;
    for (iNextDegree = iDegree-1; iNextDegree >= 0; iNextDegree--)
    {
        if ( afTmpCoeff[iNextDegree] != 0.0f )
            break;
    }
    for (i = 0; i <= iNextDegree-1; i++)
        afCoeff[i] = afTmpCoeff[i]/afTmpCoeff[iNextDegree];
    delete[] afTmpCoeff;

    return AllFloatPartsNegative(iNextDegree,afCoeff);
}
//----------------------------------------------------------------------------
bool Polynomial::AllFloatPartsNegative () const
{
    // make a copy of coefficients, later calls will change the copy
    Float* afCoeff = new Float[m_iDegree+1];
    memcpy(afCoeff,m_afCoeff,(m_iDegree+1)*sizeof(Float));

    // make polynomial monic
    if ( afCoeff[m_iDegree] != 1.0f )
    {
        Float fInv = 1.0f/afCoeff[m_iDegree];
        for (int i = 0; i < m_iDegree; i++)
            afCoeff[i] *= fInv;
        afCoeff[m_iDegree] = 1.0f;
    }

    return AllFloatPartsNegative(m_iDegree,afCoeff);
}
//----------------------------------------------------------------------------
bool Polynomial::AllFloatPartsPositive () const
{
    // make a copy of coefficients, later calls will change the copy
    Float* afCoeff = new Float[m_iDegree+1];
    memcpy(afCoeff,m_afCoeff,(m_iDegree+1)*sizeof(Float));

    // make polynomial monic
    int i;
    if ( afCoeff[m_iDegree] != 1.0f )
    {
        Float fInv = 1.0f/afCoeff[m_iDegree];
        for (i = 0; i < m_iDegree; i++)
            afCoeff[i] *= fInv;
        afCoeff[m_iDegree] = 1.0f;
    }

    // reflect z -> -z
    int iSign = -1;
    for (i = m_iDegree-1; i >= 0; i--, iSign = -iSign)
        afCoeff[i] *= iSign;

    return AllFloatPartsNegative(m_iDegree,afCoeff);
}
//----------------------------------------------------------------------------
bool Polynomial::RootsDegree2 (int& riCount, Float afRoot[2]) const
{
    // compute real roots to x^2+c[1]*x+c[0] = 0
    if ( m_iDegree != 2 )
        return false;

    // make polynomial monic
    Float afCoeff[2] = { m_afCoeff[0], m_afCoeff[1] };
    if ( m_afCoeff[2] != 1.0f )
    {
        Float fInv = 1.0f/m_afCoeff[2];
        afCoeff[0] *= fInv;
        afCoeff[1] *= fInv;
    }

    Float fDiscr = afCoeff[1]*afCoeff[1]-4.0f*afCoeff[0];
    if ( fabsf(fDiscr) <= ZERO_TOLERANCE )
        fDiscr = 0.0f;

    if ( fDiscr >= 0.0f )
    {
        fDiscr = sqrtf(fDiscr);
        afRoot[0] = 0.5f*(-afCoeff[1]-fDiscr);
        afRoot[1] = 0.5f*(-afCoeff[1]+fDiscr);
        riCount = 2;
    }
    else
    {
        riCount = 0;
    }

    return true;
}
//----------------------------------------------------------------------------
bool Polynomial::RootsDegree3 (int& riCount, Float afRoot[3]) const
{
    // compute real roots to x^3+c[2]*x^2+c[1]*x+c[0] = 0
    if ( m_iDegree != 3 )
        return false;

    // make polynomial monic
    Float afCoeff[3] = { m_afCoeff[0], m_afCoeff[1], m_afCoeff[2] };
    if ( m_afCoeff[3] != 1.0f )
    {
        Float fInv = 1.0f/m_afCoeff[3];
        afCoeff[0] *= fInv;
        afCoeff[1] *= fInv;
        afCoeff[2] *= fInv;
    }

    // convert to y^3+a*y+b = 0 by x = y-c[2]/3 and
    Float fA = ms_fThird*(3.0f*afCoeff[1]-afCoeff[2]*afCoeff[2]);
    Float fB = ms_fTwentySeventh*(2.0f*afCoeff[2]*afCoeff[2]*afCoeff[2] -
        9.0f*afCoeff[1]*afCoeff[2]+27.0f*afCoeff[0]);
    Float fOffset = ms_fThird*afCoeff[2];

    Float fDiscr = 0.25f*fB*fB + ms_fTwentySeventh*fA*fA*fA;
    if ( fabsf(fDiscr) <= ZERO_TOLERANCE )
        fDiscr = 0.0f;

    Float fHalfB = 0.5f*fB;
    if ( fDiscr > 0.0f )  // 1 real, 2 complex roots
    {
        fDiscr = sqrtf(fDiscr);
        Float fTemp = -fHalfB + fDiscr;
        if ( fTemp >= 0.0f )
            afRoot[0] = powf(fTemp,ms_fThird);
        else
            afRoot[0] = -powf(-fTemp,ms_fThird);
        fTemp = -fHalfB - fDiscr;
        if ( fTemp >= 0.0f )
            afRoot[0] += powf(fTemp,ms_fThird);
        else
            afRoot[0] -= powf(-fTemp,ms_fThird);
        afRoot[0] -= fOffset;
        riCount = 1;
    }
    else if ( fDiscr < 0.0f ) 
    {
        Float fDist = sqrtf(-ms_fThird*fA);
        Float fAngle = ms_fThird*atan2(sqrtf(-fDiscr),-fHalfB);
        Float fCos = cosf(fAngle);
        Float fSin = sinf(fAngle);
        afRoot[0] = 2.0f*fDist*fCos-fOffset;
        afRoot[1] = -fDist*(fCos+ms_fSqrt3*fSin)-fOffset;
        afRoot[2] = -fDist*(fCos-ms_fSqrt3*fSin)-fOffset;
        riCount = 3;
    }
    else 
    {
        Float fTemp;
        if ( fHalfB >= 0.0f )
            fTemp = -powf(fHalfB,ms_fThird);
        else
            fTemp = powf(-fHalfB,ms_fThird);
        afRoot[0] = 2.0f*fTemp-fOffset;
        afRoot[1] = -fTemp-fOffset;
        afRoot[2] = afRoot[1];
        riCount = 3;
    }

    return true;
}
//----------------------------------------------------------------------------
bool Polynomial::RootsDegree4 (int& riCount, Float afRoot[4]) const
{
    // compute real roots to x^4+c[3]*x^3+c[2]*x^2+c[1]*x+c[0] = 0
    if ( m_iDegree != 4 )
        return false;

    // make polynomial monic
    Float afCoeff[4] = { m_afCoeff[0], m_afCoeff[1], m_afCoeff[2],
        m_afCoeff[3] };
    if ( m_afCoeff[4] != 1.0f )
    {
        Float fInv = 1.0f/m_afCoeff[4];
        afCoeff[0] *= fInv;
        afCoeff[1] *= fInv;
        afCoeff[2] *= fInv;
        afCoeff[3] *= fInv;
    }

    // reduction to resolvent cubic polynomial
    Polynomial kResolve(3);
    kResolve[3] = 1.0f;
    kResolve[2] = -afCoeff[2];
    kResolve[1] = afCoeff[3]*afCoeff[1]-4.0f*afCoeff[0];
    kResolve[0] = -afCoeff[3]*afCoeff[3]*afCoeff[0] +
        4.0f*afCoeff[2]*afCoeff[0]-afCoeff[1]*afCoeff[1];
    int iResolveCount;
    Float afResolveRoot[3];
    kResolve.RootsDegree3(iResolveCount,afResolveRoot);
    Float fY = afResolveRoot[0];

    riCount = 0;
    Float fDiscr = 0.25f*afCoeff[3]*afCoeff[3]-afCoeff[2]+fY;
    if ( fabsf(fDiscr) <= ZERO_TOLERANCE )
        fDiscr = 0.0f;

    if ( fDiscr > 0.0f ) 
    {
        Float fR = sqrtf(fDiscr);
        Float fT1 = 0.75f*afCoeff[3]*afCoeff[3]-fR*fR-2.0f*afCoeff[2];
        Float fT2 = (4.0f*afCoeff[3]*afCoeff[2]-8.0f*afCoeff[1]-
            afCoeff[3]*afCoeff[3]*afCoeff[3])/(4.0f*fR);

        Float fTplus = fT1+fT2;
        Float fTminus = fT1-fT2;
        if ( fabsf(fTplus) <= ZERO_TOLERANCE ) 
            fTplus = 0.0f;
        if ( fabsf(fTminus) <= ZERO_TOLERANCE ) 
            fTminus = 0.0f;

        if ( fTplus >= 0.0f )
        {
            Float fD = sqrtf(fTplus);
            afRoot[0] = -0.25f*afCoeff[3]+0.5f*(fR+fD);
            afRoot[1] = -0.25f*afCoeff[3]+0.5f*(fR-fD);
            riCount += 2;
        }
        if ( fTminus >= 0.0f )
        {
            Float fE = sqrtf(fTminus);
            afRoot[riCount++] = -0.25f*afCoeff[3]+0.5f*(fE-fR);
            afRoot[riCount++] = -0.25f*afCoeff[3]-0.5f*(fE+fR);
        }
    }
    else if ( fDiscr < 0.0f )
    {
        riCount = 0;
    }
    else
    {
        Float fT2 = fY*fY-4.0f*afCoeff[0];
        if ( fT2 >= -ZERO_TOLERANCE ) 
        {
            if ( fT2 < 0.0f ) // round to zero
                fT2 = 0.0f;
            fT2 = 2.0f*sqrtf(fT2);
            Float fT1 = 0.75f*afCoeff[3]*afCoeff[3]-2.0f*afCoeff[2];
            if ( fT1+fT2 >= ZERO_TOLERANCE ) 
            {
                Float fD = sqrtf(fT1+fT2);
                afRoot[0] = -0.25f*afCoeff[3]+0.5f*fD;
                afRoot[1] = -0.25f*afCoeff[3]-0.5f*fD;
                riCount += 2;
            }
            if ( fT1-fT2 >= ZERO_TOLERANCE ) 
            {
                Float fE = sqrtf(fT1-fT2);
                afRoot[riCount++] = -0.25f*afCoeff[3]+0.5f*fE;
                afRoot[riCount++] = -0.25f*afCoeff[3]-0.5f*fE;
            }
        }
    }

    return true;
}
//----------------------------------------------------------------------------
Float Polynomial::SpecialCubeRoot (Float fA, Float fB, Float fC)
{
    // Solve A*r^3 + B*r = C where A > 0 and B > 0.
    //
    // Let r = D*sinh(u) where D = sqrt(4*B/(3*A)).  Then
    // sinh(3*u) = 4*[sinh(u)]^3+3*sinh(u) = E where E = 4*C/(A*D^3).
    // sinh(3*u) = E has solution u = (1/3)*log(E+sqrt(E^2+1)).  This
    // leads to sinh(u) = ((E+sqrt(E^2+1))^{1/3}-(E+sqrt(E^2+1))^{-1/3})/2.
    // Therefore,
    //
    //     r = D*((E+sqrt(E^2+1))^{1/3}-(E+sqrt(E^2+1))^{-1/3})/2.

    Float fD = sqrtf(4.0f*ms_fThird*fB/fA);
    Float fE = 4.0f*fC/(fA*fD*fD*fD);
    Float fF = powf(fE+sqrtf(fE*fE+1.0f),ms_fThird);
    Float fRoot = 0.5f*fD*(fF-1.0f/fF);

    return fRoot;
}
//----------------------------------------------------------------------------
int Polynomial::GetRootCount (Float fT0, Float fT1)
{
    if ( m_iDegree == 0 )
    {
        // polynomial is constant on the interval
        if ( m_afCoeff[0] != 0.0f )
            return 0;
        else
            return -1;  // to indicate "infinitely many"
    }

    // generate the Sturm sequence
    vector<Polynomial*> kSturm;
    Polynomial* pkF0 = new Polynomial(*this);
    Polynomial* pkF1 = new Polynomial(pkF0->GetDerivative());
    kSturm.push_back(pkF0);
    kSturm.push_back(pkF1);

    while ( pkF1->GetDegree() > 0 )
    {
        Polynomial* pkF2 = new Polynomial;
        Polynomial kQuot;
        pkF0->Divide(*pkF1,kQuot,*pkF2);
        *pkF2 = -(*pkF2);
        kSturm.push_back(pkF2);
        pkF0 = pkF1;
        pkF1 = pkF2;
    }

    int i;
    Float fValue0, fValue1;

    // count the sign changes at t0
    int iSignChanges0 = 0;
    if ( fT0 == -1e30 )
    {
        pkF0 = kSturm[0];
        if ( pkF0->GetDegree() & 1 )
            fValue0 = -(*pkF0)[pkF0->GetDegree()];
        else
            fValue0 = (*pkF0)[pkF0->GetDegree()];

        if ( fabsf(fValue0) < STURM_EPSILON )
            fValue0 = 0.0f;

        for (i = 1; i < (int)kSturm.size(); i++)
        {
            pkF1 = kSturm[i];

            if ( pkF1->GetDegree() & 1 )
                fValue1 = -(*pkF1)[pkF1->GetDegree()];
            else
                fValue1 = (*pkF1)[pkF1->GetDegree()];

            if ( fabsf(fValue1) < STURM_EPSILON )
                fValue1 = 0.0f;

            if ( fValue0*fValue1 < 0.0f || fValue0 == 0.0f )
                iSignChanges0++;

            fValue0 = fValue1;
            pkF0 = pkF1;
        }
    }
    else
    {
        pkF0 = kSturm[0];
        fValue0 = (*pkF0)(fT0);
        if ( fabsf(fValue0) < STURM_EPSILON )
            fValue0 = 0.0f;

        for (i = 1; i < (int)kSturm.size(); i++)
        {
            pkF1 = kSturm[i];
            fValue1 = (*pkF1)(fT0);
            if ( fabsf(fValue1) < STURM_EPSILON )
                fValue1 = 0.0f;

            if ( fValue0*fValue1 < 0.0f || fValue0 == 0.0f )
                iSignChanges0++;

            fValue0 = fValue1;
            pkF0 = pkF1;
        }
    }

    // count the sign changes at t1
    int iSignChanges1 = 0;
    if ( fT1 == 1e30 )
    {
        pkF0 = kSturm[0];
        fValue0 = (*pkF0)[pkF0->GetDegree()];
        if ( fabsf(fValue0) < STURM_EPSILON )
            fValue0 = 0.0f;

        for (i = 1; i < (int)kSturm.size(); i++)
        {
            pkF1 = kSturm[i];
            fValue1 = (*pkF1)[pkF1->GetDegree()];
            if ( fabsf(fValue1) < STURM_EPSILON )
                fValue1 = 0.0f;

            if ( fValue0*fValue1 < 0.0f || fValue0 == 0.0f )
                iSignChanges1++;

            fValue0 = fValue1;
            pkF0 = pkF1;
        }
    }
    else
    {
        pkF0 = kSturm[0];
        fValue0 = (*pkF0)(fT1);
        if ( fabsf(fValue0) < STURM_EPSILON )
            fValue0 = 0.0f;

        for (i = 1; i < (int)kSturm.size(); i++)
        {
            pkF1 = kSturm[i];
            fValue1 = (*pkF1)(fT1);
            if ( fabsf(fValue1) < STURM_EPSILON )
                fValue1 = 0.0f;

            if ( fValue0*fValue1 < 0.0f || fValue0 == 0.0f )
                iSignChanges1++;

            fValue0 = fValue1;
            pkF0 = pkF1;
        }
    }

    // clean up
    for (i = 0; i < (int)kSturm.size(); i++)
        delete kSturm[i];

    if ( iSignChanges0 >= iSignChanges1 )
        return iSignChanges0 - iSignChanges1;

    // theoretically we should not get here
    assert( false );
    return 0;
}
//----------------------------------------------------------------------------
