// sphere.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
#include "mc.h"
// Sphere Declarations
class Sphere: public Shape {
public:
	// Sphere Interface
	Sphere(const Transform &o2w, Float rad, Float zmin, Float zmax, Float phiMax);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const;
	virtual Point Sample(Float u1, Float u2, Normal *Ns) const;
	virtual Point Sample(const Point &P, Float u1, Float u2,
		Normal *Ns) const;
	virtual Float Weight(const Point &P, const Vector &dir) const;
private:
	// Sphere Data
	Float radius;
	Float phiMax;
	Float zmin, zmax;
	Float thetaMin, thetaMax;
};
// Sphere Methods
Sphere::Sphere(const Transform &o2w, Float rad, Float z0,
		Float z1, Float pm)
	: Shape(o2w) {
	radius = rad;
	zmin = Clamp(min(z0, z1), -radius, radius);
	zmax = Clamp(max(z0, z1), -radius, radius);
	thetaMin = asinf(zmin/radius);
	thetaMax = asinf(zmax/radius);
	phiMax = Radians(Clamp(pm, 0.0f, 360.0f));
}
BBox Sphere::ObjectBound() const {
	return BBox(Point(-radius, -radius, zmin),
                    Point( radius,  radius, zmax));
}
bool Sphere::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic sphere coefficients
	Float A = ray.D.x*ray.D.x + ray.D.y*ray.D.y + ray.D.z*ray.D.z;
	Float B = 2 * (ray.D.x*ray.O.x + ray.D.y*ray.O.y +
				   ray.D.z*ray.O.z);
	Float C = ray.O.x*ray.O.x + ray.O.y*ray.O.y +
			  ray.O.z*ray.O.z - radius*radius;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute sphere hit position and $\phi$
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test sphere intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		if (t1 > ray.maxt) return false;
		thit = t1;
		// Compute sphere hit position and $\phi$
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phiMax)
			return false;
	}
	// Fill in [[DifferentialGeometry]] from sphere hit
	Float u = phi / phiMax;
	Float theta = asinf(Phit.z / radius);
	Float v = (theta - thetaMin) / (thetaMax - thetaMin);
	// Compute sphere \dPdu and \dPdv
	Float cosphi = cosf(phi), sinphi = sinf(phi);
	Vector dPdu(-phiMax * Phit.y, phiMax * Phit.x, 0);
	Vector dPdv = (thetaMax-thetaMin) *
		Vector(-Phit.z * cosphi, -Phit.z * sinphi,
			radius * cosf(thetaMin + v * (thetaMax - thetaMin)));
	// Compute sphere \dNdu and \dNdv
	Vector d2Pduu = -phiMax * phiMax * Vector(Phit.x, Phit.y, 0);
	Vector d2Pduv = (zmax - zmin) * Phit.z * phiMax *
		Vector(sinphi, -cosphi, 0.);
	Vector d2Pdvv = -(thetaMax - thetaMin) * (thetaMax - thetaMin) *
		Vector(Phit.x, Phit.y, Phit.z);
	// Compute coefficients for fundamental forms
	Float E = Dot(dPdu, dPdu);
	Float F = Dot(dPdu, dPdv);
	Float G = Dot(dPdv, dPdv);
	Vector N = Cross(dPdu, dPdv);
	Float e = Dot(N, d2Pduu);
	Float f = Dot(N, d2Pduv);
	Float g = Dot(N, d2Pdvv);
	// Compute \dNdu and \dNdv from fundamental form coefficients
	Float invEGF2 = 1.f / (E*G - F*F);
	Vector dNdu = (f*F - e*G) * invEGF2 * dPdu +
		(e*F - f*E) * invEGF2 * dPdv;
	Vector dNdv = (g*F - f*G) * invEGF2 * dPdu +
		(f*F - g*E) * invEGF2 * dPdv;
	*dg = DifferentialGeometry(ObjectToWorld(Phit), ObjectToWorld(dPdu),
		ObjectToWorld(dPdv), ObjectToWorld(dNdu), ObjectToWorld(dNdv),
		u, v, this);
	// Update [[thitp]] for quadric intersection
	*thitp = thit;
	return true;
}
bool Sphere::IntersectP(const Ray &r) const {
	Float phi;
	Point Phit;
	// Transform [[Ray]] to object space
	Ray ray;
	WorldToObject(r, &ray);
	// Compute quadratic sphere coefficients
	Float A = ray.D.x*ray.D.x + ray.D.y*ray.D.y + ray.D.z*ray.D.z;
	Float B = 2 * (ray.D.x*ray.O.x + ray.D.y*ray.O.y +
				   ray.D.z*ray.O.z);
	Float C = ray.O.x*ray.O.x + ray.O.y*ray.O.y +
			  ray.O.z*ray.O.z - radius*radius;
	// Solve quadratic equation for [[t]] values
	Float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
		return false;
	// Compute intersection distance along ray
	if (t0 > ray.maxt || t1 < ray.mint)
		return false;
	Float thit = t0;
	if (t0 < ray.mint) {
		thit = t1;
		if (thit > ray.maxt) return false;
	}
	// Compute sphere hit position and $\phi$
	Phit = ray(thit);
	phi = atan2f(Phit.y, Phit.x);
	if (phi < 0.) phi += 2.f*M_PI;
	// Test sphere intersection against clipping parameters
	if (Phit.z < zmin || Phit.z > zmax || phi > phiMax) {
		if (thit == t1) return false;
		if (t1 > ray.maxt) return false;
		thit = t1;
		// Compute sphere hit position and $\phi$
		Phit = ray(thit);
		phi = atan2f(Phit.y, Phit.x);
		if (phi < 0.) phi += 2.f*M_PI;
		if (Phit.z < zmin || Phit.z > zmax || phiMax)
			return false;
	}
	return true;
}
Float Sphere::Area() const {
	return phiMax * radius * (zmax-zmin);
}
Point Sphere::Sample(Float u1, Float u2, Normal *Ns) const {
	Point P = Point(0,0,0) + radius * UniformSampleSphere(u1, u2);
	*Ns = ObjectToWorld(Normal(P.x, P.y, P.z)).Hat();
	return ObjectToWorld(P);
}
Point Sphere::Sample(const Point &P,
		Float u1, Float u2, Normal *Ns) const {
	// Compute coordinate system for sphere sampling
	Point Pcenter = ObjectToWorld(Point(0,0,0));
	Vector wc = (Pcenter - P).Hat();
	Vector wcX, wcY;
	CoordinateSystem(wc, &wcX, &wcY);
	// Sample uniformly on sphere if [[P]] is inside it
	if (DistanceSquared(P, Pcenter) < radius*radius)
		return Sample(u1, u2, Ns);
	// Sample sphere uniformly inside subtended cone
	Float cosThetaMax = sqrtf(max(0.f, 1.f - radius*radius /
		DistanceSquared(P, Pcenter)));
	DifferentialGeometry dgSphere;
	Float thit;
	Point Ps;
	Ray r(P, UniformSampleCone(u1, u2, cosThetaMax, wcX, wcY, wc));
	if (!Intersect(r, &thit, &dgSphere)) Ps = Pcenter - radius * wc; // !@$!$
	else                                 Ps = r(thit);
	*Ns = Normal(Ps - Pcenter).Hat();
	return Ps;
}
Float Sphere::Weight(const Point &P, const Vector &dir) const {
	Point Pcenter = ObjectToWorld(Point(0,0,0));
	if (DistanceSquared(P, Pcenter) < radius*radius)
		return Shape::Weight(P, dir);
	Float cosThetaMax = sqrtf(max(0.f, 1.f - radius*radius /
		DistanceSquared(P, Pcenter)));
	return UniformConeWeight(cosThetaMax);
}
extern "C" Reference<Shape> CreateShape(const Transform &o2w,
		const ParamSet &params) {
	Float radius = params.FindOneFloat("_radius", 1);
	Float zmin = params.FindOneFloat("_zmin", -radius);
	Float zmax = params.FindOneFloat("_zmax", radius);
	Float thetamax = params.FindOneFloat("_thetamax", 360);
	return new Sphere(o2w, radius, zmin, zmax, thetamax);
}
