// subdiv.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
#include "paramset.h"
#include "dynload.h"
#include "texture.h"
#include <set>
#include <map>
using std::set;
using std::map;
// SubdivisionMesh Macros
#define NEXT(i) (((i)+1)%3)
#define PREV(i) (((i)+2)%3)
// SubdivisionMesh Local Structures
struct SDFace;
struct SDFace;
struct SDVertex {
	// SDVertex Constructor
	SDVertex(Point pt = Point(0,0,0))
		: P(pt) {
		startFace = NULL;
		child = NULL;
		regular = false;
		boundary = false;
	}
	// SDVertex Methods
	int valence();
	void oneRing(Point *P);
	Point P;
	SDFace *startFace;
	SDVertex *child;
	bool regular, boundary;
};
struct SDFace {
	// SDFace Constructor
	SDFace() {
		int i;
		for (i = 0; i < 3; ++i) {
			v[i] = NULL;
			f[i] = NULL;
		}
		for (i = 0; i < 4; ++i)
			children[i] = NULL;
	}
	// SDFace Methods
	SDFace *nextFace(SDVertex *vert) {
		return f[vnum(vert)];
	}
	SDFace *prevFace(SDVertex *vert) {
		return f[PREV(vnum(vert))];
	}
	int vnum(SDVertex *vert) const {
		for (int i = 0; i < 3; ++i)
			if (v[i] == vert) return i;
		Assert(1 == 0);
		return -1;
	}
	SDVertex *nextVert(SDVertex *vert) {
		return v[NEXT(vnum(vert))];
	}
	SDVertex *prevVert(SDVertex *vert) {
		return v[PREV(vnum(vert))];
	}
	SDVertex *otherVert(SDVertex *v0, SDVertex *v1) {
		for (int i = 0; i < 3; ++i)
			if (v[i] != v0 && v[i] != v1)
				return v[i];
		Assert(1 == 0);
		return NULL;
	}
	SDVertex *v[3];
	SDFace *f[3];
	SDFace *children[4];
};
struct SDEdge {
	// SDEdge Constructor
	SDEdge(SDVertex *v0 = NULL, SDVertex *v1 = NULL) {
		v[0] = min(v0, v1);
		v[1] = max(v0, v1);
		f[0] = f[1] = NULL;
		fptr = NULL;
	}
	// SDEdge Comparison Function
	bool operator<(const SDEdge &e2) const {
		if (v[0] == e2.v[0]) return v[1] < e2.v[1];
		return v[0] < e2.v[0];
	}
	SDVertex *v[2];
	SDFace *f[2];
	SDFace **fptr;
};
// SubdivisionMesh Declarations
class LoopSubdiv : public Shape {
public:
	// LoopSubdiv Method Declarations
	LoopSubdiv(const Transform &o2w, int nt, int nv, const int *vi,
		const Point *P, int nlevels);
	~LoopSubdiv();
	bool CanIntersect() const;
	void Refine(vector<RefinedShape> &refined) const;
	BBox ObjectBound() const;
	BBox WorldBound() const;
private:
	// LoopSubdiv Private Methods
	static Float beta(int val) {
		if (val == 3) return 3.f/16.f;
		else return 3.f / (8.f * val);
	}
	static Point weightOneRing(SDVertex *vert, Float beta);
	static Point weightBoundary(SDVertex *vert, Float beta);
	static Float gamma(int val) {
		return 1.f / (val + 3.f / (8.f * beta(val)));
	}
	// LoopSubdiv Private Data
	int nLevels;
	vector<SDVertex *> vertices;
	vector<SDFace *> faces;
};
// SubdivisionMesh Inline Functions
inline int SDVertex::valence() {
	SDFace *f = startFace;
	if (!boundary) {
		// Compute valence of interior vertex
		int nf = 1;
		while ((f = f->nextFace(this)) != startFace)
			++nf;
		return nf;
	}
	else {
		// Compute valence of boundary vertex
		int nf = 1;
		while ((f = f->nextFace(this)) != NULL)
			++nf;
		f = startFace;
		while ((f = f->prevFace(this)) != NULL)
			++nf;
		return nf+1;
	}
}
// SubdivisionMesh Methods
LoopSubdiv::LoopSubdiv(const Transform &o2w, int nfaces,
		int nvertices, const int *vertexIndices,
		const Point *P, int nl)
	: Shape(o2w) {
	nLevels = nl;
	// Allocate [[LoopSubdiv]] vertices and faces
	int i;
	SDVertex *verts = new SDVertex[nvertices];
	for (i = 0; i < nvertices; ++i) {
		verts[i] = SDVertex(P[i]);
		vertices.push_back(&verts[i]);
	}
	SDFace *fs = new SDFace[nfaces];
	for (i = 0; i < nfaces; ++i)
		faces.push_back(&fs[i]);
	// Set face to vertex pointers
	const int *vp = vertexIndices;
	for (i = 0; i < nfaces; ++i) {
		SDFace *f = faces[i];
		for (int j = 0; j < 3; ++j) {
			SDVertex *v = vertices[vp[j]];
			f->v[j] = v;
			v->startFace = f;
		}
		vp += 3;
	}
	// Set neighbor pointers in [[faces]]
	set<SDEdge> edges;
	for (i = 0; i < nfaces; ++i) {
		SDFace *f = faces[i];
		for (int edge = 0; edge < 3; ++edge) {
			// Update neighbor pointer for [[edge]]
			int v0 = edge, v1 = NEXT(edge);
			SDEdge e(f->v[v0], f->v[v1]);
			if (edges.find(e) == edges.end()) {
				// Handle new edge
				e.f[0] = f;
				e.fptr = &(f->f[edge]);
				edges.insert(e);
			}
			else {
				// Handle previously-seen edge
				e = *edges.find(e);
				*e.fptr = f;
				f->f[edge] = e.f[0];
				edges.erase(e);
			}
		}
	}
	// Finish vertex initialization
	for (i = 0; i < nvertices; ++i) {
		SDVertex *v = vertices[i];
		SDFace *f = v->startFace;
		do {
			f = f->nextFace(v);
		} while (f && f != v->startFace);
		v->boundary = (f == NULL);
		v->regular = (!v->boundary && (v->valence() == 6) ||
	                 (v->boundary && (v->valence() == 4)));
	}
}
LoopSubdiv::~LoopSubdiv() {
	delete[] vertices[0];
	delete[] faces[0];
}
BBox LoopSubdiv::ObjectBound() const {
	BBox b;
	for (u_int i = 0; i < vertices.size(); i++)
		b = Union(b, vertices[i]->P);
	return b;
}
BBox LoopSubdiv::WorldBound() const {
	BBox b;
	for (u_int i = 0; i < vertices.size(); i++)
		b = Union(b, ObjectToWorld(vertices[i]->P));
	return b;
}
bool LoopSubdiv::CanIntersect() const {
	return false;
}
void LoopSubdiv::Refine(vector<RefinedShape> &refined) const {
	vector<SDFace *> f = faces;
	vector<SDVertex *> v = vertices;
	MemoryArena<SDVertex> vertexArena;
	MemoryArena<SDFace> faceArena;
	u_int i;
	for (i = 0; i < nLevels; ++i) {
		// Update [[f]] and [[v]] for next level of subdivision
		vector<SDFace *> newFaces;
		vector<SDVertex *> newVertices;
		// Allocate next level of children in mesh tree
		u_int j;
		for (j = 0; j < v.size(); ++j) {
			v[j]->child = new (vertexArena) SDVertex;
			v[j]->child->regular = v[j]->regular;
			v[j]->child->boundary = v[j]->boundary;
			newVertices.push_back(v[j]->child);
		}
		for (j = 0; j < f.size(); ++j)
			for (int k = 0; k < 4; ++k) {
				f[j]->children[k] = new (faceArena) SDFace;
				newFaces.push_back(f[j]->children[k]);
			}
		// Update vertex positions and create new edge vertices
		// Update vertex positions for even vertices
		for (j = 0; j < v.size(); ++j) {
			if (!v[j]->boundary) {
				// Apply one-ring rule for even vertex
				if (v[j]->regular)
					v[j]->child->P = weightOneRing(v[j], 1.f/16.f);
				else
					v[j]->child->P = weightOneRing(v[j], beta(v[j]->valence()));
			}
			else {
				// Apply boundary rule for even vertex
				v[j]->child->P = weightBoundary(v[j], 1.f/8.f);
			}
		}
		// Compute new odd edge vertices
		map<SDEdge, SDVertex *> splitEdges;
		for (j = 0; j < f.size(); ++j) {
			SDFace *face = f[j];
			for (int k = 0; k < 3; ++k) {
				// Compute odd vertex on [[k]]th edge
				SDEdge edge(face->v[k], face->v[NEXT(k)]);
				SDVertex *vert = splitEdges[edge];
				if (!vert) {
					// Create and initialize new odd vertex
					vert = new (vertexArena) SDVertex;
					newVertices.push_back(vert);
					vert->regular = true;
					vert->boundary = (face->f[k] == NULL);
					vert->startFace = face->children[3];
					// Apply edge rules to compute new vertex position
					if (vert->boundary) {
						vert->P =  0.5f * edge.v[0]->P;
						vert->P += 0.5f * edge.v[1]->P;
					}
					else {
						vert->P =  3.f/8.f * edge.v[0]->P;
						vert->P += 3.f/8.f * edge.v[1]->P;
						vert->P += 1.f/8.f *
							face->otherVert(edge.v[0], edge.v[1])->P;
						vert->P += 1.f/8.f *
							face->f[k]->otherVert(edge.v[0], edge.v[1])->P;
					}
					splitEdges[edge] = vert;
				}
			}
		}
		// Update new mesh topology
		// Update even vertex face pointers
		for (j = 0; j < v.size(); ++j) {
			SDVertex *vert = v[j];
			int vertNum = vert->startFace->vnum(vert);
			vert->child->startFace = vert->startFace->children[vertNum];
		}
		// Update face neighbor pointers
		for (j = 0; j < f.size(); ++j) {
			SDFace *face = f[j];
			for (int k = 0; k < 3; ++k) {
				// Update children [[f]] pointers for siblings
				face->children[3]->f[k] = face->children[NEXT(k)];
				face->children[k]->f[NEXT(k)] = face->children[3];
				// Update children [[f]] pointers for neighbor children
				SDFace *f2 = face->f[k];
				face->children[k]->f[k] =
					f2 ? f2->children[f2->vnum(face->v[k])] : NULL;
				f2 = face->f[PREV(k)];
				face->children[k]->f[PREV(k)] =
					f2 ? f2->children[f2->vnum(face->v[k])] : NULL;
			}
		}
		// Update face vertex pointers
		for (j = 0; j < f.size(); ++j) {
			SDFace *face = f[j];
			for (int k = 0; k < 3; ++k) {
				// Update child vertex pointer to new even vertex
				face->children[k]->v[k] = face->v[k]->child;
				// Update child vertex pointer to new odd vertex
				SDVertex *vert = splitEdges[SDEdge(face->v[k], face->v[NEXT(k)])];
				face->children[k]->v[NEXT(k)] = vert;
				face->children[NEXT(k)]->v[k] = vert;
				face->children[3]->v[k] = vert;
			}
		}
		// Prepare for next level of subdivision
		#if 0
		if (i != 0) {
			for (u_int j = 0; j < f.size(); ++j)
				delete f[j];
			for (u_int j = 0; j < v.size(); ++j)
				delete v[j];
		}
		#endif
		f = newFaces;
		v = newVertices;
	}
	// Push vertices to limit surface
	Point *Plimit = new Point[v.size()];
	for (i = 0; i < v.size(); ++i) {
		if (v[i]->boundary)
			Plimit[i] = weightBoundary(v[i], 1.f/5.f);
		else
			Plimit[i] = weightOneRing(v[i], gamma(v[i]->valence()));
	}
	for (i = 0; i < v.size(); ++i)
		v[i]->P = Plimit[i];
	
	// Compute vertex tangents on limit surface
	vector<Normal> Ns;
	Ns.reserve(v.size());
	for (i = 0; i < v.size(); ++i) {
		SDVertex *vert = v[i];
		Vector S(0,0,0), T(0,0,0);
		// Put [[vert]] one-ring in [[Pring]]
		int val = vert->valence();
		Point *Pring = (Point *)alloca(val * sizeof(Point));
		vert->oneRing(Pring);
		if (!vert->boundary) {
			// Compute tangents of interior face
			for (int k = 0; k < val; ++k) {
				S += cosf(2.f*M_PI*k/val) * Vector(Pring[k]);
				T += sinf(2.f*M_PI*k/val) * Vector(Pring[k]);
			}
		}
		else {
			// Compute tangents of boundary face
			S = Pring[val-1] - Pring[0];
			if (val == 2)
				T = Vector(Pring[0] + Pring[1] - 2 * vert->P);
			else if (val == 3)
				T = Pring[1] - vert->P;
			else if (val == 4) // regular
				T = Vector(-1*Pring[0] + 2*Pring[1] + 2*Pring[2] +
					-1*Pring[3] + -2*vert->P);
			else {
				Float theta = M_PI / float(val-1);
				T = Vector(sinf(theta) * (Pring[0] + Pring[val-1]));
				for (int k = 1; k < val-1; ++k) {
					Float wt = (2 * cosf(theta) - 2) * sinf((k) * theta);
					T += Vector(wt * Pring[k]);
				}
				T = -T;
			}
		}
		Ns.push_back(Normal(Cross(S, T)));
	}
	// Create [[TriangleMesh]] from subdivision mesh
	u_int ntris = u_int(f.size());
	int *verts = new int[3*ntris];
	int *vp = verts;
	u_int totVerts = u_int(v.size());
	map<SDVertex *, int> usedVerts;
	for (i = 0; i < totVerts; ++i)
		usedVerts[v[i]] = i;
	for (i = 0; i < ntris; ++i) {
		for (int j = 0; j < 3; ++j) {
			*vp = usedVerts[f[i]->v[j]];
			++vp;
		}
	}
	for (i = 0; i < Ns.size(); ++i)
		Ns[i] = ObjectToWorld(Ns[i]);
	ParamSet paramSet(totVerts);
	paramSet.AddInt("_ntris", (int *) &ntris);
	paramSet.AddInt("_nverts", (int *) &totVerts);
	paramSet.AddInt("_vertexIndices", verts, PARAM_TYPE_UNIFORM,
		3*ntris);
	paramSet.AddPoint("P", Plimit, PARAM_TYPE_VERTEX);
	Texture<Normal> *Ntex = new VertexTexture<Normal>(&Ns[0], int(Ns.size()));
	refined.push_back(RefinedShape(
		CreateShape("trianglemesh", LRT_BUILDDIR, ObjectToWorld,
			paramSet), Ntex));
	Assert(refined[0].shape);
	delete[] verts;
	delete[] Plimit;
}
Point LoopSubdiv::weightOneRing(SDVertex *vert, Float beta) {
	// Put [[vert]] one-ring in [[Pring]]
	int val = vert->valence();
	Point *Pring = (Point *)alloca(val * sizeof(Point));
	vert->oneRing(Pring);
	Point P = (1 - val * beta) * vert->P;
	for (int i = 0; i < val; ++i)
		P += beta * Pring[i];
	return P;
}
void SDVertex::oneRing(Point *P) {
	if (!boundary) {
		// Get one ring vertices for interior vertex
		SDFace *face = startFace;
		do {
			*P++ = face->nextVert(this)->P;
			face = face->nextFace(this);
		} while (face != startFace);
	}
	else {
		// Get one ring vertices for boundary vertex
		SDFace *face = startFace, *f2;
		while ((f2 = face->nextFace(this)) != NULL)
			face = f2;
		*P++ = face->nextVert(this)->P;
		do {
			*P++ = face->prevVert(this)->P;
			face = face->prevFace(this);
		} while (face != NULL);
	}
}
Point LoopSubdiv::weightBoundary(SDVertex *vert, Float beta) {
	// Put [[vert]] one-ring in [[Pring]]
	int val = vert->valence();
	Point *Pring = (Point *)alloca(val * sizeof(Point));
	vert->oneRing(Pring);
	Point P = (1-2*beta) * vert->P;
	P += beta * Pring[0];
	P += beta * Pring[val-1];
	return P;
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	int nfaces = params.FindOneInt("_nfaces", -1);
	int nv = params.FindOneInt("_nvertices", -1);
	int nlevels = params.FindOneInt("nlevels", 3);
	int type, narray;
	const int *vi = params.findInt("_vertices", &type, &narray);
	const int *numvs = params.findInt("_faceNumVertices", &type, &narray);
	Assert(narray == nfaces);
	const Point *P = params.FindPoint("P", &type, &narray);
	if (nfaces == -1 || nv == -1 || !vi || !numvs || !P) return NULL;

	vector<int> vi2;
	for (int i = 0; i < nfaces; ++i) {
		Assert(numvs[i] >= 3);
		for (int j = 0; j < numvs[i]-2; ++j) {
			vi2.push_back(vi[0]);
			vi2.push_back(vi[j+1]);
			vi2.push_back(vi[j+2]);
		}
		vi += numvs[i];
	}
	return new LoopSubdiv(o2w, int(vi2.size())/3, nv, &vi2[0], P, nlevels);
}
