// trianglemesh.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "lrt.h"
#include "shapes.h"
#include "paramset.h"
#include "trianglemesh.h"
// TriangleMesh Methods
TriangleMesh::TriangleMesh(const Transform &o2w, int nt, int nv,
		const int *vi, const Point *P, const Float *uv)
	: Shape(o2w) {
	ntris = nt;
	nverts = nv;
	vertexIndex = new int[3 * ntris];
	memcpy(vertexIndex, vi, 3 * ntris * sizeof(int));
	if (uv) {
		uvs = new Float[2*nverts];
		memcpy(uvs, uv, 2*nverts*sizeof(Float));
	}
	else uvs = NULL;
	p = new Point[nverts];
	// Transform mesh vertices to world space
	for (int i  = 0; i < nverts; ++i)
		p[i] = ObjectToWorld(P[i]);
}
TriangleMesh::~TriangleMesh() {
	delete[] vertexIndex;
	delete[] p;
}
BBox TriangleMesh::ObjectBound() const {
	BBox bobj;
	for (int i = 0; i < nverts; i++)
		bobj = Union(bobj, WorldToObject(p[i]));
	return bobj;
}
BBox TriangleMesh::WorldBound() const {
	BBox worldBounds;
	for (int i = 0; i < nverts; i++)
		worldBounds = Union(worldBounds, p[i]);
	return worldBounds;
}
void TriangleMesh::Refine(vector<RefinedShape> &refined) const {
	for (int i = 0; i < ntris; ++i)
		refined.push_back(RefinedShape(
			new Triangle(ObjectToWorld, (TriangleMesh *)this, i)));
}
BBox Triangle::ObjectBound() const {
	// Get triangle vertices in [[p1]], [[p2]], and [[p3]]
	const Point &p1 = mesh->p[v[0]];
	const Point &p2 = mesh->p[v[1]];
	const Point &p3 = mesh->p[v[2]];
	return Union(BBox(WorldToObject(p1), WorldToObject(p2)),
		WorldToObject(p3));
}

BBox Triangle::WorldBound() const {
	// Get triangle vertices in [[p1]], [[p2]], and [[p3]]
	const Point &p1 = mesh->p[v[0]];
	const Point &p2 = mesh->p[v[1]];
	const Point &p3 = mesh->p[v[2]];
	return Union(BBox(p1, p2), p3);
}
bool Triangle::Intersect(const Ray &ray, Float *thitp,
		DifferentialGeometry *dg) const {
	// Initialize triangle intersection statistics
	static StatsRatio triangleHits("Geometry",
		"Triangle Ray Intersections");
	// Update triangle tests count
	triangleHits.add(0, 1);
	// Compute $\VEC{s}_1$
	// Get triangle vertices in [[p1]], [[p2]], and [[p3]]
	const Point &p1 = mesh->p[v[0]];
	const Point &p2 = mesh->p[v[1]];
	const Point &p3 = mesh->p[v[2]];
	Vector E1 = p2 - p1;
	Vector E2 = p3 - p1;
	Vector S_1 = Cross(ray.D, E2);
	Float divisor = Dot(S_1, E1);
	if (divisor == 0.)
		return false;
	Float invDivisor = 1.f / divisor;
	// Compute first barycentric coordinate
	Vector T = ray.O - p1;
	Float b1 = Dot(T, S_1) * invDivisor;
	if (b1 < 0. || b1 > 1.)
		return false;
	// Compute second barycentric coordinate
	Vector S_2 = Cross(T, E1);
	Float b2 = Dot(ray.D, S_2) * invDivisor;
	if (b2 < 0. || b1 + b2 > 1.)
		return false;
	// Compute [[t]] to intersection point
	Float t = Dot(E2, S_2) * invDivisor;
	if (t < ray.mint || t > ray.maxt)
		return false;
	triangleHits.add(1, 0);
	// Fill in [[DifferentialGeometry]] from triangle hit
	// Compute triangle partial derivatives
	Vector dPdu, dPdv;
	Float uvs[3][2];
	GetUVs(uvs);
	// Compute deltas for triangle partial derivatives
	Float du1 = uvs[1][0] - uvs[0][0];
	Float du2 = uvs[2][0] - uvs[0][0];
	Float dv1 = uvs[1][1] - uvs[0][1];
	Float dv2 = uvs[2][1] - uvs[0][1];
	Float dx1 = p2.x - p1.x;
	Float dx2 = p3.x - p1.x;
	Float dy1 = p2.y - p1.y;
	Float dy2 = p3.y - p1.y;
	Float dz1 = p2.z - p1.z;
	Float dz2 = p3.z - p1.z;
	Float determinant = du1 * dv2 - dv1 * du2;
	if (determinant == 0) {
		// Handle zero determinant for triangle partial derivative matrix
		CoordinateSystem(Cross(E2, E1).Hat(), &dPdu, &dPdv);
	}
	else {
		Float invdet = 1.f / determinant;
		dPdu = Vector((dx1 * dv2 - dv1 * dx2) * invdet,
			(dy1 * dv2 - dv1 * dy2) * invdet,
			(dz1 * dv2 - dv1 * dz2) * invdet);
		dPdv = Vector((du1 * dx2 - dx1 * du2) * invdet,
			(du1 * dy2 - dy1 * du2) * invdet,
			(du1 * dz2 - dz1 * du2) * invdet);
	}
	// Interpolate $(u,v)$ triangle parametric coordinates
	Float b0 = 1 - b1 - b2;
	Float tu = b0*uvs[0][0] + b1*uvs[1][0] + b2*uvs[2][0];
	Float tv = b0*uvs[0][1] + b1*uvs[1][1] + b2*uvs[2][1];
	*dg = DifferentialGeometry(ray(t), dPdu, dPdv, Vector(0,0,0),
		Vector(0,0,0), tu, tv, this);
	*thitp = t;
	return true;
}
bool Triangle::IntersectP(const Ray &ray) const {
	// Initialize triangle intersection statistics
	static StatsRatio triangleHits("Geometry",
		"Triangle Ray Intersections");
	// Update triangle tests count
	triangleHits.add(0, 1);
	// Compute $\VEC{s}_1$
	// Get triangle vertices in [[p1]], [[p2]], and [[p3]]
	const Point &p1 = mesh->p[v[0]];
	const Point &p2 = mesh->p[v[1]];
	const Point &p3 = mesh->p[v[2]];
	Vector E1 = p2 - p1;
	Vector E2 = p3 - p1;
	Vector S_1 = Cross(ray.D, E2);
	Float divisor = Dot(S_1, E1);
	if (divisor == 0.)
		return false;
	Float invDivisor = 1.f / divisor;
	// Compute first barycentric coordinate
	Vector T = ray.O - p1;
	Float b1 = Dot(T, S_1) * invDivisor;
	if (b1 < 0. || b1 > 1.)
		return false;
	// Compute second barycentric coordinate
	Vector S_2 = Cross(T, E1);
	Float b2 = Dot(ray.D, S_2) * invDivisor;
	if (b2 < 0. || b1 + b2 > 1.)
		return false;
	// Compute [[t]] to intersection point
	Float t = Dot(E2, S_2) * invDivisor;
	if (t < ray.mint || t > ray.maxt)
		return false;
	triangleHits.add(1, 0);
	return true;
}
Float Triangle::Area() const {
	// Get triangle vertices in [[p1]], [[p2]], and [[p3]]
	const Point &p1 = mesh->p[v[0]];
	const Point &p2 = mesh->p[v[1]];
	const Point &p3 = mesh->p[v[2]];
	return 0.5f * Cross(p2-p1, p3-p1).Length();
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	int nt = params.FindOneInt("_ntris", -1);
	int nv = params.FindOneInt("_nverts", -1);
	int type, narray;
	const int *vi = params.findInt("_vertexIndices", &type, &narray);
	const Point *P = params.FindPoint("P", &type, &narray);
	const Float *uvs = params.FindFloat("uv", &type, &narray);
	if (nt == -1 || nv == -1 || !vi || !P) return NULL;
	return new TriangleMesh( o2w, nt, nv, vi, P, uvs );
}
