// trianglemesh.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */

#include "lrt.h"
#include "shapes.h"
#include "paramset.h"


typedef struct {
	Float period;
	Float dirX;
	Float dirZ;
	Float amplitude;
	Float phaseX;
	Float phaseZ;
} Wave;


class WaterSurf: public Shape {
public:
	// Sphere Interface
	WaterSurf(const Transform &o2w, ,Float xs, Float ys, Float zs, Float xscale2, Float zscale2);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const {return 0.0f;};
private:
	// Sphere Data
	std::vector<Wave> wavetab;
	Float xs,zs,ys, xs2, zs2;
};

// WaterSurf Methods
WaterSurf::WaterSurf(const Transform &o2w,Float xs, Float ys, Float zs, Float xscale2, Float zscale2) 
	: Shape(o2w) {
	this->xs=xs;
	this->ys=ys;
	this->zs=zs;
	this->xs2=xscale2;
	this->zs2=zscale2;

	Wave w1;
	w1.amplitude=0.33f*ys*.5f;
	w1.dirX=13.0f;
	w1.dirZ=8.3f;
	w1.period=3865;
	wavetab.push_back(w1);

	Wave w2;
	w2.amplitude=0.33f*ys*.5f;
	w2.dirX=-14.5f;
	w2.dirZ=17.3f;
	w2.period=2869;
	wavetab.push_back(w2);
	
	Wave w3;
	w3.amplitude=0.33f*ys*.5f;
	w3.dirX=1.5f;
	w3.dirZ=-12.3f;
	w3.period=3173;
	wavetab.push_back(w3);
}

WaterSurf::~WaterSurf() {
}

BBox WaterSurf::ObjectBound() const {
	return BBox(Point(-xs, -0.1, -zs),
                    Point( xs,  0.1, zs));
}

bool WaterSurf::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	Ray ray;
	WorldToObject(r, &ray);
	Float t=-ray.O.y/ray.D.y;
	if (t > ray.maxt || t < ray.mint)
		return false;
	Phit P=ray(t);
	if(P.x<-xs || P.x>xs || P.z<-zs || P.z>zs)
		return false;
	// We got an intersection
	
	height=0.0f;
	dydu=0.0f;
	dydv=0.0f;
	for(k=0;k<wavetab.size();k++) {
		const Wave &w=wavetab[k];
		Float omega=2.0f*M_PI/w.period;
		Float arg=w.dirX*P.x/xs2+w.dirZ*P.z/zs2-omega*k;
		Float s=sin(arg);
		Float c=cos(arg);

		Float height+=w.amplitude*ys*c;
		dydu+=w.amplitude*ys*w.dirX*s/xs2;
		dydv+=w.amplitude*ys*w.dirZ*s/xz2;
	}
	tu=P.x;
	tv=P.z
	*dg = DifferentialGeometry(ray(t), Vector(1,dydu,0), Vector(0,dydv,1),
														 Vector(0,0,0),Vector(0,0,0), tu, tv, this);
	*thitp = t;
	return true;
}

bool WaterSurf::IntersectP(const Ray &ray) const {
	Ray ray;
	WorldToObject(r, &ray);
	Float t=-ray.O.y/ray.D.y;
	if (t > ray.maxt || t < ray.mint)
		return false;
	Phit P=ray(t);
	if(P.x<-xs || P.x>xs || P.z<-zs || P.z>zs)
		return false;
	// We got an intersection
	
	return tru;
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	Float xscale=params.findOneFloat("x",1);
	Float yscale=params.findOneFloat("y",1);
	Float zscale=params.findOneFloat("z",1);
	Float xscale2=params.findOneFloat("xs",1);
	Float zscale2=params.findOneFloat("zs",1);
	return new WaterSurf( xscale, yscale , zscale, xscale2, zscale2);
}
