// trianglemesh.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */

#include "lrt.h"
#include "shapes.h"
#include "paramset.h"


typedef struct {
	Float period;
	Float dirX;
	Float dirZ;
	Float amplitude;
	Float phaseX;
	Float phaseZ;
} Wave;


class WaterSurf: public Shape {
public:
	// Sphere Interface
	WaterSurf(const Transform &o2w ,Float xs, Float ys, Float zs, Float xscale2, Float zscale2);
	BBox ObjectBound() const;
	bool Intersect(const Ray &ray, Float *thitp, DifferentialGeometry *dg) const;
	bool IntersectP(const Ray &ray) const;
	Float Area() const {return 0.0f;};
private:
	// Sphere Data
	std::vector<Wave> wavetab;
	Float xs,zs,ys, xs2, zs2;
};

#define randf ((Float)rand()/RAND_MAX)

// WaterSurf Methods
WaterSurf::WaterSurf(const Transform &o2w,Float xs, Float ys, Float zs, Float xscale2, Float zscale2) 
	: Shape(o2w) {
	this->xs=xs;
	this->ys=ys;
	this->zs=zs;
	this->xs2=xscale2;
	this->zs2=zscale2;

	Wave w1;
	w1.amplitude=0.33f*ys*.5f;
	w1.dirX=13.0f;
	w1.dirZ=8.3f;
	w1.period=3865;
	w1.phaseX=randf*w1.period;
	w1.phaseZ=randf*w1.period;
	wavetab.push_back(w1);

	Wave w2;
	w2.amplitude=0.33f*ys*.5f;
	w2.dirX=-14.5f;
	w2.dirZ=17.3f;
	w2.period=2869;
	w2.phaseX=randf*w2.period;
	w2.phaseZ=randf*w2.period;
	wavetab.push_back(w2);
	
	Wave w3;
	w3.amplitude=0.33f*ys*.5f;
	w3.dirX=1.5f;
	w3.dirZ=-12.3f;
	w3.period=3173;
	w3.phaseX=randf*w3.period;
	w3.phaseZ=randf*w3.period;
	wavetab.push_back(w3);
}


BBox WaterSurf::ObjectBound() const {
	return BBox(Point(-xs, -0.1, -zs),
                    Point( xs,  0.1, zs));
}

bool WaterSurf::Intersect(const Ray &r, Float *thitp,
		DifferentialGeometry *dg) const {
	Ray ray;
	WorldToObject(r, &ray);
	Float t=-ray.O.y/ray.D.y;
	if (!(t < ray.maxt && t > ray.mint))
		return false;
	//	if(debugflag) 
	//		cout << "t: " << t << "   rayd:" << ray.D << "   rayo:" << ray.O <<endl;
	Point P=ray(t);
	if(P.x<-xs || P.x>xs || P.z<-zs || P.z>zs)
		return false;
	// We got an intersection
	
	Float height=0.0f;
	Float dydu=0.0f;
	Float dydv=0.0f;
	Float noise=3*Noise(P.x/100,P.y/100,P.z/100);
	//	cout << noise << " ";
	for(u_int k=0;k<wavetab.size();k++) {
		const Wave &w=wavetab[k];
		Float omega=2.0f*M_PI/w.period;
		Float arg=w.dirX*P.x/xs2+w.dirZ*P.z/zs2-omega*k+noise;
		Float s=sin(arg);
		Float c=cos(arg);

		height+=w.amplitude*ys*c;
		dydu+=w.amplitude*ys*w.dirX*s/xs2;
		dydv+=w.amplitude*ys*w.dirZ*s/zs2;
	}
	Float tu=P.z;
	Float tv=P.x;
	//	if(debugflag) 
	//		cout << "P:" << P << ",   dydu:" <<dydu<< ",   dydv:"<< dydv << ",   height:" << height<<endl;
	//	*dg = DifferentialGeometry(ObjectToWorld(P), ObjectToWorld(Vector(-1,-dydu,0)), ObjectToWorld(Vector(0,dydv,1)),
	//														 ObjectToWorld(Vector(0,0,0)),ObjectToWorld(Vector(0,0,0)), tu, tv, this);
	*dg = DifferentialGeometry(ObjectToWorld(P), ObjectToWorld(Vector(1,0,0)), ObjectToWorld(Vector(0,0,-1)),
														 ObjectToWorld(Vector(0,0,0)),ObjectToWorld(Vector(0,0,0)), tu, tv, this);
	//	if (debugflag)
	//		cout << "Nn: " << dg->Nn<<endl;
	*thitp = t;
	return true;
}

bool WaterSurf::IntersectP(const Ray &r) const {
	Ray ray;
	WorldToObject(r, &ray);
	Float t=-ray.O.y/ray.D.y;
	if (t > ray.maxt || t < ray.mint)
		return false;
	Point P=ray(t);
	if(P.x<-xs || P.x>xs || P.z<-zs || P.z>zs)
		return false;
	// We got an intersection
	
	return true;
}
extern "C" Reference<Shape>CreateShape( const Transform &o2w, const ParamSet &params )
{
	Float xscale=params.FindOneFloat("xs",1.0f);
	Float yscale=params.FindOneFloat("ys",1.0f);
	Float zscale=params.FindOneFloat("zs",1.0f);
	Float xscale2=params.FindOneFloat("xs2",1.0f);
	Float zscale2=params.FindOneFloat("zs2",1.0f);
	return new WaterSurf(o2w, xscale, yscale , zscale, xscale2, zscale2);
}
