// highcontrast.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "tonemap.h"
#include "mipmap.h"
// HighContrastOp Declarations
class HighContrastOp : public ToneMap {
public:
	void Map(Float *xyz, int xRes, int yRes, const DisplayInfo &di) const;
private:
	// HighContrastOp Utility Methods
	static Float C(Float y) {
		if (y < 0.0034f) return y / 0.0014f;
		else if (y < 1) return 2.4483f + log10f(y/0.0034f)/0.4027f;
		else if (y < 7.2444f) return 16.563f + (y - 1)/0.4027f;
		else return 32.0693f + log10f(y / 7.2444f)/0.0556f;
	}
	static Float T(Float y, Float CYmin, Float CYmax,
			Float maxDisplayY) {
		return maxDisplayY * (C(y) - CYmin) / (CYmax - CYmin);
	}
};
// HighContrastOp Method Definitions
void HighContrastOp::Map(Float *xyz, int xRes, int yRes,
		const DisplayInfo &di) const {
	// Find minimum and maximum image luminances
	Float minY = Y(0), maxY = Y(0);
	for (int i = 0; i < xRes * yRes; ++i) {
		minY = min(minY, Y(i));
		maxY = max(maxY, Y(i));
	}
	Float CYmin = C(minY), CYmax = C(maxY);
	// Build luminance image pyramid
	Float *Yadapt = new Float[xRes * yRes];
	for (int i = 0; i < xRes * yRes; ++i)
		Yadapt[i] = Y(i);
	MIPMap<Float> pyramid(xRes, yRes, Yadapt);
	delete[] Yadapt;
	// Apply high contrast tone mapping operator
	ProgressReporter progress(xRes*yRes, "Tone Mapping");
	for (int y = 0; y < yRes; ++y) {
		Float yc = Float(y) / Float(yRes-1);
		for (int x = 0; x < xRes; ++x) {
			Float xc = Float(x) / Float(xRes-1);
			// Compute adaptation luminance
			Float dwidth = 1.f / Float(max(xRes, yRes));
			Float maxWidth = 32.f / Float(max(xRes, yRes));
			Float width = dwidth, prevWidth = 0.f;
			Float Yadapt;
			Float prevlc = 0.f;
			const Float maxLocalContrast = .5f;
			while (1) {
				// Compute local contrast at $(x,y)$
				Float b0 = pyramid.Lookup(xc, yc, width, 0.f, 0.f, width);
				Float b1 = pyramid.Lookup(xc, yc, 2.f*width, 0.f, 0.f, 2.f*width);
				Float lc = fabsf((b0 - b1) / b0);
				// If maximum contrast is exceeded, compute adaptation luminance
				if (lc > maxLocalContrast) {
					Float t = (lc - prevlc) / maxLocalContrast;
					Float w = Lerp(t, prevWidth, width);
					Yadapt = pyramid.Lookup(xc, yc, w, 0.f, 0.f, w);
					break;
				}
				// Increase search region and prepare to compute contrast again
				prevlc = lc;
				prevWidth = width;
				width += dwidth;
				if (width >= maxWidth) {
					Yadapt = pyramid.Lookup(xc, yc, maxWidth, 0.f, 0.f, maxWidth);
					break;
				}
			}
			// Apply tone mapping based on adaptation luminance
			Float scale = T(Yadapt, CYmin, CYmax, di.maxDisplayY) /
				Yadapt;
			int off = x + y*xRes;
			X(off) *= scale;
			Y(off) *= scale;
			Z(off) *= scale;
		}
		progress.Update(xRes);
	}
	fprintf(stderr, "\n");
}
extern "C" ToneMap *CreateToneMap(const ParamSet &ps) {
	return new HighContrastOp;
}
