// volumegrid.cc*
// Source Code Copyright
/*
 * lrt source code Copyright(c) 1998-2003 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For non-commercial use only.
 * No warranty, express or implied, for this software.
 */
#include "volume.h"
// VolumeGrid Declarations
class VolumeGrid : public DensityRegion {
public:
	// VolumeGrid Methods
	VolumeGrid(const Spectrum &sa, const Spectrum &ss, Float gg,
	 		const Spectrum &emit, const BBox &e, const Transform &v2w,
			const string &filename);
	~VolumeGrid() { delete[] d; }
	BBox WorldBound() const { return WorldToVolume.GetInverse()(extent); }
	bool Intersect(const Ray &r, Float *t0, Float *t1) const {
		Ray ray = WorldToVolume(r);
		return extent.IntersectP(ray, t0, t1);
	}
	Float density(const Point &Pobj) const;
private:
	// VolumeGrid Private Data
	Float *d;
	int nx, ny, nz;
	BBox extent;
};
#define SAMP(x,y,z) (d[(z)*nx*ny + (y)*nx + (x)])
// VolumeGrid Definitions
VolumeGrid::VolumeGrid(const Spectrum &sa, const Spectrum &ss, Float gg,
 		const Spectrum &emit, const BBox &e, const Transform &v2w,
		const string &filename)
	: DensityRegion(sa, ss, gg, emit, v2w) {
	extent = e;
	FILE *f = fopen(filename.c_str(), "r");
	if (!f) {
		fprintf(stderr, "Unable to open volume file %s\n",
			filename.c_str());
		d = new Float[1];
		d[0] = 0;
		nx = ny = nz = 1;
	}
	// Process volume data from file
	fread(&nx, sizeof(int), 1, f);
	fread(&ny, sizeof(int), 1, f);
	fread(&nz, sizeof(int), 1, f);
	d = new Float[nx*ny*nz];
	for (int i = 0; i < nx*ny*nz; ++i) {
		unsigned char c;
		fread(&c, sizeof(unsigned char), 1, f);
		d[i] = c * (1.f/255.f);
	}
	fclose(f);
}
Float VolumeGrid::density(const Point &Pobj) const {
	if (!extent.Inside(Pobj)) return 0;
	// Compute voxel coordinates and offsets for [[Pobj]]
	Float voxx = (Pobj.x - extent.pMin.x) /
		(extent.pMax.x - extent.pMin.x) * (nx-1);
	Float voxy = (Pobj.y - extent.pMin.y) /
		(extent.pMax.y - extent.pMin.y) * (ny-1);
	Float voxz = (Pobj.z - extent.pMin.z) /
		(extent.pMax.z - extent.pMin.z) * (nz-1);
	int vx = Clamp(Floor2Int(voxx), 0, nx - 2);
	int vy = Clamp(Floor2Int(voxy), 0, ny - 2);
	int vz = Clamp(Floor2Int(voxz), 0, nz - 2);
	Float dx = voxx - vx;
	Float dy = voxy - vy;
	Float dz = voxz - vz;
	// Trilinearly interpolate density values to compute local density
	Float d00 = Lerp(dx, SAMP(vx, vy, vz),     SAMP(vx+1, vy, vz));
	Float d10 = Lerp(dx, SAMP(vx, vy+1, vz),   SAMP(vx+1, vy+1, vz));
	Float d01 = Lerp(dx, SAMP(vx, vy, vz+1),   SAMP(vx+1, vy, vz+1));
	Float d11 = Lerp(dx, SAMP(vx, vy+1, vz+1), SAMP(vx+1, vy+1, vz+1));
	Float d0 = Lerp(dy, d00, d10);
	Float d1 = Lerp(dy, d01, d11);
	return Lerp(dz, d0, d1);
}
extern "C" VolumeRegion *CreateVolumeRegion(const Transform &volume2world,
		const ParamSet &params) {
	Spectrum sigma_a = params.FindOneSpectrum("sigma_a", 0.);
	Spectrum sigma_s = params.FindOneSpectrum("sigma_s", 0.);
	Float g = params.FindOneFloat("g", 0.);
	Spectrum Le = params.FindOneSpectrum("Le", 0.);
	Point p0 = params.FindOnePoint("p0", Point(0,0,0));
	Point p1 = params.FindOnePoint("p1", Point(1,1,1));
	string filename = params.FindOneString("filename", "");
	return new VolumeGrid(sigma_a, sigma_s, g, Le, BBox(p0, p1),
		volume2world, filename);
}
